#include "fles.h"
#include "tokensP.h"
#include "newP.h"
#include "matrix.h"
#include <matrix2.h>
#include <sparse.h> 
#include <sparse2.h>

#define NEED_DEFINES 0
#if NEED_DEFINES
#  define NMODES 1
#  define LHS_DAT        "lhs.dat"
#  define RHS_DAT        "rhs.dat"
#  define SOL_DAT        "sol.dat"
#  define POD_DAT        "pod.dat"
#  define RHS_XMGR       "rhs.xmgr"
#  define APPROX_SOL_XMGR  "approx_sol.xmgr"
#  define APPROX_RHS_XMGR  "approx_rhs.xmgr"
#  define SOL_XMGR        "sol.xmgr"
#else
#  define NMODES 1
#  define LHS_DAT         "lhs6.out"
#  define RHS_DAT         "rhs6.out"
#  define SOL_DAT         "sol6.out"
#  define POD_DAT         "phi.out"
#  define RHS_XMGR "rhs-set=6.xmgr"
#  define SOL_XMGR "sol-set=6.xmgr"
#  define APPROX_RHS_XMGR "rhs-set=6-nmodes=1.xmgr"
#  define APPROX_SOL_XMGR "sol-set=6-nmodes=1.xmgr"
#  define APPROX_SOL_ADAPTIVE_XMGR "soladapt-set=6-nmodes=1.xmgr"
#  define SOL_DIFF_XMGR "soldiff-set=6-nmodes=1.xmgr"
#  define SOL_DIFF_ADAPTIVE_XMGR "soldiffadapt-set=6-nmodes=1.xmgr"
#endif

static fles_solver_ctx_t slesctx = 0;
static int tot_of_equations =        5535;
static double nz_per_row = 0.086;
static SPMAT *lhs = NULL;
static VEC *rhs = NULL;
static VEC *sol = NULL;
static VEC *fullsol = NULL;
static VEC *phis[NMODES+1];
static MAT *phi_t_lhs_phi = NULL;
static VEC *phi_t_rhs = NULL;
static VEC *approx_sol = NULL;
static int nmodes = NMODES;

static int
tot_of_equations_ (void *client)
{
  return tot_of_equations;
}

static int
error_notify (fles_error_enum error_num,
              int line,
              char *file,
              char *detail,
              void *client_data)
{
  char buf[1024];
  sprintf (buf, "%s (l.%d), %s", file, line, detail);
  fprintf (stderr, "%s\n", buf);
  abort ();
  return 0;
}

static void
make_fles (int *argc, char ***argv)
{
  slesctx = fles_new_spetsc_solver (NULL, argc, argv, tot_of_equations_, error_notify);
  slesctx->set_option (slesctx, PETSC_OPTS_NZ_PER_ROW, (void *)&nz_per_row);
}
  
static void
check_symmetry ()
{
  int i, j;
  int unsymmetric = 0;
  fprintf (stderr, "Symmetry check: ");
  for (i = 0; i < tot_of_equations; i++)
    for (j = i; j < tot_of_equations; j++) {
      double ad = sp_get_val (lhs, i, j);
      double bd = sp_get_val (lhs, j, i);
      if (fabs (ad - bd) > 1e-9) {
        unsymmetric = 1;
        fprintf (stderr, "Unsymmetry: a[%d][%d] = %g != a[%d][%d] = %g\n",
                 i, j, ad, j, i, bd);
      }
    }
  if (unsymmetric) fprintf (stderr, "Error: Unsymmetric\n");
  else             fprintf (stderr, "Symmetric\n");
}

static void
load_lhs (char *file)
{
  FILE *fp = NULL;
  tokens_parser parser = tokens_new_parser ();

  lhs = sp_resize (lhs, tot_of_equations, tot_of_equations);
  if (lhs == 0) {
    fprintf (stderr, "Error: no memory\n");
    abort ();
  }
  sp_zero (lhs);
  
  if (!tokens_open_file (parser, file)) {
    fprintf (stderr, "Error: no such file? %s\n", file);
    abort ();
  }
  
  fprintf (stderr, "Loading lhs: \"%s\"\n", file);
  slesctx->zero_lhs (slesctx);

  fp = fopen ("lhs.mat", "w");
  while (tokens_next_line (parser)) {
    int ntok = tokens_total_of_tokens (parser);
    int col[1], row[1];
    double val[1];
    const char *rowtok = tokens_token_as_string (parser, 2);
    if (sscanf (rowtok, "%d:", &row[0]) != 1) {
      fprintf (stderr, "Error: got %s\n", rowtok);
      abort ();
    }
    if (row[0] >= tot_of_equations) {
      fprintf (stderr, "Error: wrong number of equations %d\n", tot_of_equations);
      abort ();
    }
    {
      int i=3;
      while (i <= ntok) {
        col[0] = tokens_token_as_int (parser, i); i++;
        val[0] = tokens_token_as_double (parser, i); i++;
        /* printf ("assembling [%d][%d]=%g ", row[0], col[0], val[0]); */
        slesctx->assemble_mat (slesctx, FLES_FALSE, FLES_ADD, 1, row, 1, col, val);
        {
          double v = sp_get_val (lhs, row[0], col[0]) + val[0];
          sp_set_val (lhs, row[0], col[0], v);
          if (v != 0) fprintf (fp, "%d %d %g\n", row[0], col[0], v);
        }
      }
    }
    /* printf ("\n"); */
  }
  slesctx->finish_lhs (slesctx, FLES_FINAL_FINISH);
  check_symmetry ();

  fclose (fp);

  tokens_close_curr_file (parser);
  tokens_delete_parser (parser);
}
  
static void
load_rhs (char *file)
{
  tokens_parser parser = tokens_new_parser ();

  rhs = v_resize (rhs, tot_of_equations);
  if (rhs == 0) {
    fprintf (stderr, "Error: no memory\n");
    abort ();
  }

  if (!tokens_open_file (parser, file)) {
    fprintf (stderr, "Error: no such file? %s\n", file);
    abort ();
  }
  
  fprintf (stderr, "Loading rhs: \"%s\"\n", file);
  slesctx->zero_rhs (slesctx);
  
  {
    int row[1];
    row[0] = 0;
    while (tokens_next_line (parser)) {
      double val[1];
      int ntok = tokens_total_of_tokens (parser);
      if (ntok < 1) {
        fprintf (stderr, "Error: no tokens?\n");
        abort ();
      }        
      if (row[0] >= tot_of_equations) {
        fprintf (stderr, "Error: wrong number of equations %d\n", tot_of_equations);
        abort ();
      }
      val[0] = tokens_token_as_double (parser, 1);
      slesctx->assemble_vect (slesctx, FLES_ADD, 1, row, val);
      rhs->ve[row[0]] += val[0];
      /* printf ("\n"); */
      row[0]++;
    }
  }
  slesctx->finish_rhs (slesctx);
  fprintf (stderr, "Rhs norm: %g\n", v_norm2 (rhs));
  
  tokens_close_curr_file (parser);
  tokens_delete_parser (parser);
}
  
static void
load_sol (char *file)
{
  tokens_parser parser = tokens_new_parser ();

  sol = v_resize (sol, tot_of_equations);
  if (sol == 0) {
    fprintf (stderr, "Error: no memory\n");
    abort ();
  }
  v_zero (sol);

  if (!tokens_open_file (parser, file)) {
    fprintf (stderr, "Error: no such file? %s\n", file);
    abort ();
  }
  
  fprintf (stderr, "Loading sol: \"%s\"\n", file);

  {
    int row[1];
    row[0] = 0;
    while (tokens_next_line (parser)) {
      double val[1];
      int ntok = tokens_total_of_tokens (parser);
      if (ntok < 1) {
        fprintf (stderr, "Error: no tokens?\n");
        abort ();
      }        
      if (row[0] >= tot_of_equations) {
        fprintf (stderr, "Error: wrong number of equations %d\n", tot_of_equations);
        abort ();
      }
      val[0] = tokens_token_as_double (parser, 1);
      sol->ve[row[0]] += val[0];
      /* printf ("\n"); */
      row[0]++;
    }
  }
  
  tokens_close_curr_file (parser);
  tokens_delete_parser (parser);
}

static void
compute_residual_as_loaded ()
{
  double resnorm;
  VEC *res = NULL;
  res = sp_mv_mlt (lhs, sol, res);
  v_sub (res, rhs, res);
  resnorm = v_norm2 (res);
  fprintf (stderr, "Residual norm (as loaded): %g\n", resnorm);
  v_free (res);
}

static void
compute_residual_as_computed_full ()
{
  double resnorm;
  VEC *res = NULL;
  res = sp_mv_mlt (lhs, fullsol, res);
  v_sub (res, rhs, res);
  resnorm = v_norm2 (res);
  fprintf (stderr, "Residual norm (as computed, full): %g\n", resnorm);
  v_free (res);
}

static void
solve ()
{
  fprintf (stderr, "Full solution\n");
  if (!slesctx->solve (slesctx)) {
  }
  fullsol = v_resize (fullsol, tot_of_equations);
  if (fullsol == 0) {
    fprintf (stderr, "Error: no memory\n");
    abort ();
  }
  {
    int row;
    for (row = 0; row < tot_of_equations; row++) {
      fullsol->ve[row] = slesctx->get_from_sol (slesctx, row);
    }
  }
}

static void
read_mode (int mode, tokens_parser parser)
{
  double ev = 0;
  phis[mode] = v_resize (NULL, tot_of_equations);
  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) {
      fprintf (stderr, "Bad data in POD file");
      abort ();
    }
    key = tokens_token_as_string (parser, 1);
    if (tokens_token_and_keyword_equiv (key, "END")) {
      return;
    } else {
      int id;
      double ux, uy, uz;
      if (tokens_total_of_tokens (parser) < 4) {
        fprintf (stderr, "Bad data in POD file");
        abort ();
      }
      id = tokens_token_as_int (parser, 1);
      ux = tokens_token_as_double (parser, 2);
      uy = tokens_token_as_double (parser, 3);
      uz = tokens_token_as_double (parser, 4);
      phis[mode]->ve[((id-1)*3)+0] = ux;
      phis[mode]->ve[((id-1)*3)+1] = uy;
      phis[mode]->ve[((id-1)*3)+2] = uz;
    }
  }
}

static void
skip_till_END (tokens_parser parser)
{
  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) {
      fprintf (stderr, "Bad data in POD file\n");
      abort ();
    }
    key = tokens_token_as_string (parser, 1);
    if (tokens_token_and_keyword_equiv (key, "END")) {
      return;
    } 
  }
}

static void
check_orthonormality ()
{
  fprintf (stderr, "Checking orthonormality\n");
  {
    int i, j;
    int orthonormal = 1;
    double maxdiff = 0;
    for (i = 0; i < nmodes; i++) {
      for (j = 0; j < nmodes; j++) {
        if (i == j) { 
          if (fabs (in_prod (phis[i], phis[j]) - 1) > 1e-6) {
            orthonormal = 0;
            maxdiff = max (maxdiff, fabs (in_prod (phis[i], phis[j]) - 1));
          }
        } else {
          if (fabs (in_prod (phis[i], phis[j])    ) > 1e-6) {
            orthonormal = 0;
            maxdiff = max (maxdiff, fabs (in_prod (phis[i], phis[j])));
          }
        }
      }
    }
    if (orthonormal) fprintf (stderr, "Orthonormal\n");
    else             fprintf (stderr, "NOT Orthonormal; max diff %g\n", maxdiff);
  }
}

static void
load_pod (const char *file)
{
  tokens_parser parser;
  int mode = -1;
  
  parser  = tokens_new_parser ();

  if (!tokens_open_file (parser, (char *) file)) {
    fprintf (stderr, "Could not open POD file \"%s\"", file);
    abort ();
  }

  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) {
      fprintf (stderr, "Bad data in POD file \"%s\"", file);
      abort ();
    }
    key = tokens_token_as_string (parser, 1);
    if        (tokens_token_and_keyword_equiv (key, "MODE")) {
      if (tokens_total_of_tokens (parser) < 2) {
        fprintf (stderr, "Bad data in POD file \"%s\"", file);
        abort ();
      }
      mode = tokens_token_as_int (parser, 2);
      if (mode < NMODES) {
        read_mode (mode, parser);
      } else {
        skip_till_END (parser);
      }
    } else if (tokens_token_and_keyword_equiv (key, "MEAN")) {
      skip_till_END (parser);
    } else {
      fprintf (stderr, "Unknown keyword \"%s\"", key);
      abort ();
    }
  }
  
  tokens_close_curr_file (parser);
  tokens_delete_parser (parser);

  check_orthonormality ();
}

static void
dump_mat (MAT *m, char *name, double eps)
{
  int i, j;
  fprintf (stderr, "***************************************************************\n");
  fprintf (stderr, "*  %s\n", name);
  for (i = 0; i < m->m; i++) {
    fprintf (stderr, " %3d:   ", i);
    for (j = 0; j < m->n; j++) {
      if (fabs (m->me[i][j]) > eps) fprintf (stderr, " %g", m->me[i][j]);
      else                          fprintf (stderr, " 0");
    }
    fprintf (stderr, "\n");
  }
  fprintf (stderr, "***************************************************************\n");
}

static void
construct_transf_lhs ()
{
  int i, j;
  phi_t_lhs_phi = m_resize (phi_t_lhs_phi, nmodes, nmodes);
  fprintf (stderr, "Constructing transformed lhs (%d modes)\n", nmodes);
  m_zero (phi_t_lhs_phi);
  for (i = 0; i < nmodes; i++) {
    for (j = 0; j < nmodes; j++) {
      double v;
      VEC *res = v_get (tot_of_equations);
      res = sp_mv_mlt (lhs, phis[j], res);
      v = in_prod (res, phis[i]);
      phi_t_lhs_phi->me[i][j] = v;
    }
  }
  /*  dump_mat (phi_t_lhs_phi, "phi_t_lhs_phi", 0);  */
}

static void
construct_transf_rhs ()
{
  int i, j;
  phi_t_rhs = v_resize (phi_t_rhs, nmodes);
  fprintf (stderr, "Constructing transformed rhs (%d modes)\n", nmodes);
  v_zero (phi_t_rhs);
  for (i = 0; i < nmodes; i++) {
    phi_t_rhs->ve[i] = in_prod (rhs, phis[i]);
  }
  /* v_dump (stderr, phi_t_rhs); */
}

static void
construct_approx_sol ()
{
  int i, j;
  MAT *Atransf = NULL;
  VEC *btransf = NULL;
  VEC *ampl = v_get (nmodes);
  PERM *pivot = px_get (nmodes);
  fprintf (stderr, "Constructing approximate solution (%d modes)\n", nmodes);
  Atransf = m_copy (phi_t_lhs_phi, Atransf);
  Atransf = LUfactor (Atransf, pivot);
  btransf = v_copy (phi_t_rhs, btransf);
  ampl = LUsolve (Atransf, pivot, btransf, ampl);
  approx_sol = v_get (tot_of_equations);
  if (approx_sol == 0) {
    fprintf (stderr, "Error: no memory\n");
    abort ();
  }
  fprintf (stderr, "Approximate solution\n");
  v_zero (approx_sol);
  for (i = 0; i < tot_of_equations; i++) {
    for (j = 0; j < nmodes; j++) {
      approx_sol->ve[i] += ampl->ve[j] * phis[j]->ve[i];
    }
  }
  {
    VEC *v = v_get (tot_of_equations);
    double a, r;
    v = sp_mv_mlt (lhs, sol, v);
    a = in_prod (v, sol);
    r = in_prod (rhs, sol);
    fprintf (stderr, "Amplitude for the 'exact' solution as a mode: %g\n", a / r);
  }
}

static void
compute_residual_as_approx ()
{
  double resnorm;
  VEC *res = NULL;
  res = sp_mv_mlt (lhs, approx_sol, res);
  v_sub (res, rhs, res);
  resnorm = v_norm2 (res);
  fprintf (stderr, "Residual norm (as approximated by %d modes): %g\n", nmodes, resnorm);
  v_free (res);
}

static void
dump_sol ()
{
  int i;
  FILE *fp = fopen (SOL_XMGR, "w");
  if (!fp) {
    fprintf (stderr, "Failed to open %s\n", SOL_XMGR);
    abort ();
  }
  for (i = 0; i < tot_of_equations; i++) {
    fprintf (fp, "%d %g\n", i, sol->ve[i]);
  }
  fclose (fp);
}

static void
dump_approx_sol ()
{
  int i;
  FILE *fp = fopen (APPROX_SOL_XMGR, "w");
  if (!fp) {
    fprintf (stderr, "Failed to open %s\n", APPROX_SOL_XMGR);
    abort ();
  }
  for (i = 0; i < tot_of_equations; i++) {
    fprintf (fp, "%d %g\n", i, approx_sol->ve[i]);
  }
  fclose (fp);
}

static void
dump_approx_sol_diff ()
{
  int i;
  FILE *fp = fopen (SOL_DIFF_XMGR, "w");
  if (!fp) {
    fprintf (stderr, "Failed to open %s\n", SOL_DIFF_XMGR);
    abort ();
  }
  for (i = 0; i < tot_of_equations; i++) {
    fprintf (fp, "%d %g\n", i, (sol->ve[i] - approx_sol->ve[i]));
  }
  fclose (fp);
}

static void
dump_approx_sol_adaptive ()
{
  int i;
  FILE *fp = fopen (APPROX_SOL_ADAPTIVE_XMGR, "w");
  if (!fp) {
    fprintf (stderr, "Failed to open %s\n", APPROX_SOL_ADAPTIVE_XMGR);
    abort ();
  }
  for (i = 0; i < tot_of_equations; i++) {
    fprintf (fp, "%d %g\n", i, approx_sol->ve[i]);
  }
  fclose (fp);
}

static void
dump_rhs ()
{
  int i;
  FILE *fp = fopen (RHS_XMGR, "w");
  if (!fp) {
    fprintf (stderr, "Failed to open %s\n", RHS_XMGR);
    abort ();
  }
  for (i = 0; i < tot_of_equations; i++) {
    fprintf (fp, "%d %g\n", i, rhs->ve[i]);
  }
  fclose (fp);
}

static void
dump_approx_rhs ()
{
  int i;
  VEC *approx_rhs = NULL;
  FILE *fp = fopen (APPROX_RHS_XMGR, "w");
  if (!fp) {
    fprintf (stderr, "Failed to open %s\n", APPROX_RHS_XMGR);
    abort ();
  }
  approx_rhs = sp_mv_mlt (lhs, approx_sol, approx_rhs);
  for (i = 0; i < tot_of_equations; i++) {
    fprintf (fp, "%d %g\n", i, approx_rhs->ve[i]);
  }
  fclose (fp);
}

static void
dump_approx_sol_diff_adaptive ()
{
  int i;
  FILE *fp = fopen (SOL_DIFF_ADAPTIVE_XMGR, "w");
  if (!fp) {
    fprintf (stderr, "Failed to open %s\n", SOL_DIFF_ADAPTIVE_XMGR);
    abort ();
  }
  for (i = 0; i < tot_of_equations; i++) {
    fprintf (fp, "%d %g\n", i, (sol->ve[i] - approx_sol->ve[i]));
  }
  fclose (fp);
}

static void
solve_with_hint ()
{
  fprintf (stderr, "Full solution with solution hint\n");
  slesctx->set_option (slesctx, PETSC_OPTS_SOL_HINT, (void *)approx_sol->ve);
  if (!slesctx->solve (slesctx)) {
    fprintf (stderr, "Failed in solve");
    abort ();
  }
}

static void
add_rhs_as_mode ()
{
  int j;
  VEC *add_phi = NULL;
  add_phi = v_resize (NULL, tot_of_equations);
  if (add_phi == 0) {
    fprintf (stderr, "Error: no memory\n");
    abort ();
  }
  v_zero (add_phi);
  {
    int i;
    for (i = 0; i < tot_of_equations; i++) add_phi->ve[i] = rhs->ve[i];
  }
  for (j = 0; j < NMODES; j++) {
    double dp = in_prod (phis[j], add_phi);
    add_phi = v_mltadd (add_phi, phis[j], -dp, add_phi);
  }
  {
    double dp = sqrt(in_prod (add_phi, add_phi));
    int i;
    for (i = 0; i < tot_of_equations; i++) add_phi->ve[i] /= dp;
  }
  phis[NMODES] = add_phi;
  nmodes = NMODES+1;
  check_orthonormality ();
}

static void
check_precond ()
{
  MAT *Adense = NULL;
  MAT *Phi = NULL;
  MAT *tmp = NULL;
  MAT *Ainvtilde = NULL;
  MAT *id = NULL;
  MAT *phi_t_lhs_phi_inv = NULL;

  fprintf (stderr, "Checking the approximate inverse of the lhs\n");
  
  Adense = sp_m2dense (lhs, Adense);
  dump_mat (Adense, "Adense", 0);
  
  Phi = m_resize (Phi, tot_of_equations, nmodes);
  {
    int i, j;
    for (i = 0; i < tot_of_equations; i++) {
      for (j = 0; j < nmodes; j++) {
        Phi->me[i][j] = phis[j]->ve[i];
      }
    }
  }
  dump_mat (Phi, "Phi", 0);
  tmp = mtrm_mlt (Phi, Adense, tmp);
  id  = m_mlt (tmp, Phi, id);
  dump_mat (id, "Phi^t * Adense * Phi", 1e-6);
  
  tmp = mtrm_mlt (Phi, Phi, tmp);
  dump_mat (tmp, "Phi^t * Phi", 1e-9);
  tmp = mmtr_mlt (Phi, Phi, tmp);
  dump_mat (tmp, "Phi * Phi^t", 1e-9);

  dump_mat (phi_t_lhs_phi, "phi_t_lhs_phi", 0);
  phi_t_lhs_phi_inv = m_inverse (phi_t_lhs_phi, phi_t_lhs_phi_inv);
  dump_mat (phi_t_lhs_phi_inv, "phi_t_lhs_phi_inv", 0);
  tmp = m_mlt (phi_t_lhs_phi_inv, phi_t_lhs_phi, tmp);
  dump_mat (tmp, "phi_t_lhs_phi_inv * phi_t_lhs_phi", 1e-6);

  tmp = m_mlt (Phi, phi_t_lhs_phi_inv, tmp);
  Ainvtilde = mmtr_mlt (tmp, Phi, Ainvtilde);
  dump_mat (Ainvtilde, "Ainvtilde", 0);

  id = m_mlt (Ainvtilde, Adense, id);
  dump_mat (id, "approx id", 1e-6);

  {
    double adcond, idcond;
    VEC *diag = v_get (Adense->m);
    MAT *factored = m_copy (Adense, NULL);
    factored = QRfactor (factored, diag);
    adcond = QRcondest (factored);
    factored = m_copy (id, factored);
    factored = QRfactor (factored, diag);
    idcond = QRcondest (factored);
    fprintf (stderr, "Condition number of Adense = %g; of Ainvtilde * Adense = %g\n", adcond, idcond);
  }
}

static void
my_solve ()
{
  int k = 0;
  VEC *b = v_get (tot_of_equations);
  VEC *phi = v_get (tot_of_equations);
  VEC *Aphi = v_get (tot_of_equations);
  VEC *xk = v_get (tot_of_equations);
  SPMAT *A = lhs;
  double a = 1;

  v_zero (xk);
  b = v_copy (rhs, b);
  phi = v_copy (b, phi);
  while (k < 100) {
    Aphi = sp_mv_mlt (A, phi, Aphi);
    a    = in_prod (phi, b) / in_prod (phi, Aphi);
    xk   = v_mltadd (xk, phi, -a, xk);
    b    = v_mltadd (b, Aphi, -a, b);
    phi  = v_copy (b, phi);
    sv_mlt ((1/sqrt(in_prod(phi,phi))), phi, phi);
    k = k+1;
    fprintf (stderr, "Iteration %d: ", k);
    {
      int j;
      double n = 0;
      for (j = 0; j < tot_of_equations; j++) {
        n = (sol->ve[j] - xk->ve[j])*(sol->ve[j] - xk->ve[j]);
      }
      fprintf (stderr, "a=%g, ||x-x~||/||x||=%g\n", a, sqrt(n/v_norm2(sol)));
    }
  }
}

static void
my_cg (int zero_guess)
{
  int k = 0;
  VEC *b = rhs;
  VEC *p = v_get (tot_of_equations);
  VEC *w = v_get (tot_of_equations);
  VEC *r = v_get (tot_of_equations);
  VEC *x = v_get (tot_of_equations);
  SPMAT *A = lhs;
  double alpha, beta, rho, rho_prev = 0, bnorm;
  double rtol = 1e-6;
  double atol = 1e-6;
  int monitor = ('t' == 't'? 1 : 0);

  if (atol > 0) { /* atol takes precedence */
     rtol = 0;
  }

  /* initial guess */
  fprintf (stderr, "My CG\n");
  if (zero_guess)   {
    fprintf (stderr, "Zero guess\n");
    v_zero (x);
  } else {
    fprintf (stderr, "Initial hint\n");
    x = v_copy (approx_sol, x);
  }
  p = sp_mv_mlt (A, x, p);
  r = v_mltadd (b, p, -1, r);
  bnorm = sqrt (in_prod (b, b));
  rho = in_prod (r, r);
  k = 0;
  while (!((sqrt(rho) < rtol * bnorm) || (sqrt (rho) < atol))) {
    k++;
    if (k == 1) {
       p = v_copy (r, p);
    } else {
       beta = rho / rho_prev;
       p = v_mltadd (r, p, beta, p);
    }
    w = sp_mv_mlt (A, p, w);
    alpha = rho / in_prod (p, w);
    x = v_mltadd (x, p, alpha, x);
    r = v_mltadd (r, w, -alpha, r);
    rho_prev = rho;
    rho = in_prod (r, r);
    if (monitor) {
      int j;
      double n = 0;
      fprintf (stderr, "Iteration %d: ", k);
      for (j = 0; j < tot_of_equations; j++) {
        n = (sol->ve[j] - x->ve[j])*(sol->ve[j] - x->ve[j]);
      }
      fprintf (stderr, "rho=%g, ||x-x~||/||x||=%g\n", sqrt(rho), sqrt(n/v_norm2(sol)));
    }
  }
  fprintf (stderr, "It took %d iterations to converge to residual norm %g\n", k, sqrt(rho));
}

int
main (int argc, char **argv)
{
  make_fles (&argc, &argv);
  load_lhs (LHS_DAT);
  load_rhs (RHS_DAT);
  load_sol (SOL_DAT);
  compute_residual_as_loaded ();
  solve ();
  compute_residual_as_computed_full ();
  load_pod (POD_DAT);
  construct_transf_lhs ();
  construct_transf_rhs ();
  /* check_precond (); */
  construct_approx_sol ();
  compute_residual_as_approx ();
  dump_sol ();
  dump_approx_sol ();
  dump_approx_sol_diff ();
  dump_rhs ();
  dump_approx_rhs ();
  solve_with_hint ();
  add_rhs_as_mode ();
  construct_transf_lhs ();
  construct_transf_rhs ();
  construct_approx_sol ();
  compute_residual_as_approx ();
  solve_with_hint ();
  dump_approx_sol_adaptive ();
  dump_approx_sol_diff_adaptive ();
  /* my_solve (); */
  my_cg (0);
  my_cg (1);
  return 0;
}


