// -*- Mode: c++ -*-
// $Id: LoopMesh.h,v 1.1.1.1 2001/07/24 02:41:53 pkrysl Exp $
// $Source: /home/cvs/WASP/problems/FE/subdiv/Coke/Loop/LoopMesh.h,v $

#ifndef __LOOP_LOOPMESH_H__
#define __LOOP_LOOPMESH_H__


#include "Loop/LoopTesselator.h"
#include "Loop/Topology.h"
#include "Mesh/MeshTp.h"

#include "Loop/LoopTypes.h"

#include <stack>

/**
 * LoopMesh Class
 */
template <class T>
class LoopMesh : public MeshTp<T> {
  template <class Mesh> friend class Accessor;

public:
  DECLARE_LOOP_TYPES( T )
  
protected:
  /** The subdivision stencil for this Loop subdivision mesh. */
  const Stencil& m_stencil;

protected:
  template <class Accessor>
  void even(TriangleMr* triangle, VertexLabel v, Accessor accessor) const {
#if DEBSUB
    cerr << "even: from " << triangle->vertex(v)->point(triangle->depth()) << " at depth " << (triangle->depth()) << " to ";
#endif
    VertexMr* vertex = triangle->vertex(v);
    accessor.setValid(vertex, triangle->depth() + 1);
    accessor.clear(vertex, triangle->depth() + 1);
    m_stencil.even(triangle, v, accessor, accessor(vertex, triangle->depth() + 1));
#if DEBSUB
    cerr << triangle->vertex(v)->point(triangle->depth() + 1) << " at depth " << (triangle->depth() + 1) << endl;
#endif
  }

  template <class Accessor>
  void odd(TriangleMr* triangle, EdgeLabel e, Accessor accessor) const {
#if DEBSUB
    cerr << "odd....";
#endif
    VertexMr* vertex = triangle->child(T0)->vertex(OppositeVertex(e));
    accessor.setValid(vertex, triangle->depth() + 1);
    accessor.clear(vertex, triangle->depth() + 1);
    m_stencil.odd(triangle, e, accessor, accessor(vertex, triangle->depth() + 1));
#if DEBSUB
    cerr << "\b\b\b\b\b\b\b";
    cerr << "odd:\t" << vertex->point(triangle->depth() + 1) << " at depth " << (triangle->depth() + 1) << endl;
#endif
  }

public:
  /**
   * Constructs an empty mesh with a specified subdivision stencil.
   * @param stencil  subdivision stencil
   */
  LoopMesh(const Stencil& stencil)
   : MeshTp<T>(), m_stencil(stencil) { }
  
  /**
   * Copy constructor.
   */
  LoopMesh(const LoopMesh& instance)
   : MeshTp<T>(instance),
     m_stencil(instance.m_stencil) { }
  
  /**
   * Destructor.
   */
  ~LoopMesh(void) { }

  /**
   * Evaluates the limit surface at a designated vertex in the mesh.
   * @param triangle  triangle containing vertex
   * @param v  label of vertex in triangle
   * @return limit position of vertex
   */
  CVec limit(TriangleMr* triangle, VertexLabel v) {
    VertexMr* vertex = triangle->vertex(v);
    typename VertexMr::Accessor accessor;

    if (vertex->isLimitValid() == false) {
      CVec limit = m_stencil.limit(triangle, v, accessor, m_stencil);
      accessor.setLimit(vertex, limit);
    }

    return (accessor.limit(vertex));
  }

  /**
   * Applies the Loop stencil to calculate the limit normal of a specified 
   * vertex of a given subdivision triangle.
   * @param triangle  subdivision triangle
   * @param v  vertex for which to calculate the limit normal
   * @return vertex limit normal
   */
  template <class TriangleMr>
  CVec normal(TriangleMr* triangle,
       VertexLabel v) const
  {
    int depth = triangle->depth();
    int valence = triangle->valence(v);
    float n = (float)valence;
    CVec du, dv;
    int i;

#if 0
    CVec p1 = triangle->vertex(V1)->point(depth);
    CVec p2 = triangle->vertex(V2)->point(depth);
    CVec p3 = triangle->vertex(V3)->point(depth);
    CVec normal = (p2-p1).cross(p3-1);
    if (triangle->orientation() == TriangleMr::CLOCKWISE)
      normal = -normal;
    normal.normalize();
    return normal;
#endif

    typename TriangleMr::VertexIt vi = triangle->vertexBegin(v);
    typename TriangleMr::VertexIt ve = triangle->vertexEnd(v);

    //cerr << "Point: " << triangle->vertex(v)->point(depth);
    //cerr << "  du = ";
    for (i = 0; vi != ve; i += 1, ++vi) {
      float theta = 2.0f * M_PI * (float)(i - 1) / n;
      VertexMr *mv = (*vi);
      //cerr << "+" << cos(theta) << "* (" << mv->point(depth) << ") ";
      du += cos(theta) * mv->point(depth);
      dv += sin(theta) * mv->point(depth);
    }
    if (i != valence) {
      cerr << "[" << i << " " << valence << "]\n";
      cerr << "point: " << triangle->vertex(v)->point(depth) << endl;
    }
    assert(i == valence);
    //cerr << endl << endl;
    
    CVec normal = du;
    normal = normal.cross(dv);
    normal.normalize();

    //if (triangle->orientation() == TriangleMr::CLOCKWISE)
    //  normal = -normal;

    //cerr << "\n      normal: " << normal;
    //cerr << "du: " << du << " dv: " << dv << endl;

    return (normal);
  }


  /**
   * Subdivides this Loop subdivision mesh uniformly to the specified depth.
   * @param depth  depth to which to uniformly subdivide this mesh
   */
  template <class Accessor>
  void subdivide(int depth, Accessor accessor) {
    TriangleIt ti = triangleBegin();
    TriangleIt te = triangleEnd();
    for (; ti != te; ++ti) {
      Triangle* triangle = *ti;
      ((Mesh*)this)->subdivide(triangle->root(), depth, accessor);
    }
  }

  /**
   * Subdivides this Loop subdivision mesh adaptively
   * @param angle   max angle between two normals
   */
  template <class Accessor>
  void subdivideAdaptively(float angle, int depth, Accessor accessor) {
    TriangleIt ti = triangleBegin();
    TriangleIt te = triangleEnd();
    for (; ti != te; ++ti) {
      Triangle* triangle = *ti;
      ((Mesh*)this)->subdivideAdaptively(triangle->root(), angle, depth, accessor);
    }
  }

  /**
   * Invalidate
   */
  void invalidate() {
    TriangleIt ti = triangleBegin();
    TriangleIt te = triangleEnd();
    for (; ti != te; ++ti) {
      Triangle* triangle = *ti;
      if (triangle->root())
	((Mesh*)this)->invalidate(triangle->root());
    }
  }

  /**
   * Subdivides the specified triangle in the subdivision hierarchy to the
   * specified depth.
   * @param triangle  triangle to subdivide
   * @param depth  depth to which to subdivide triangle
   */
  void invalidate(TriangleMr* triangle) {
    assert(triangle != NULL);

    for (int v = V1; v <= V3; v += 1) {
      VertexMr* vertex = triangle->vertex(VertexLabel(v));
      vertex->setInvalid();
    }

    if (!triangle->isLeaf()) {
      invalidate(triangle->child(T0));
      invalidate(triangle->child(T1));
      invalidate(triangle->child(T2));
      invalidate(triangle->child(T3));
    }
  }

  class PointAccessor {
  public:
    CVec& operator()(VertexMr* vertex, int depth) const {
#if DEBSUB
      if (!vertex->isValid(depth)) {
	cerr << "depth = " << depth << endl;
	cerr << "   point: " << vertex->point(0) << endl;
      }
#endif
      assert(vertex->isValid(depth));
      return (vertex->point(depth));
    }

    void setValid(VertexMr* vertex, int depth) const {
      vertex->setValid(depth);
    }

    void clear(VertexMr* vertex, int depth) const {
      assert(vertex->isValid(depth));
      vertex->point(depth) = CVec(0,0,0);
    }
  };

  void subdivide(int depth) {
    subdivide(depth, PointAccessor());
  }

  void subdivideAdaptively(float angle, int depth) {
    subdivideAdaptively(angle, depth, PointAccessor());
  }

  /**
   * A convenience method to subdivide the specified triangle in the
   * subdivision hierarchy once.
   * @param triangle  triangle to subdivide
   */
  template <class Accessor>
  void subdivide(TriangleMr* triangle, Accessor accessor) {
    ((Mesh*)this)->subdivide(triangle, triangle->depth() + 1, accessor);
  }

  void subdivide(TriangleMr* triangle) {
    subdivide(triangle, PointAccessor());
  }

  /**
   * Subdivides the specified triangle in the subdivision hierarchy to the
   * specified depth.
   * @param triangle  triangle to subdivide
   * @param depth  depth to which to subdivide triangle
   */
  template <class Accessor>
  void subdivide(TriangleMr* triangle, int depth, Accessor accessor) {
    assert(depth <= __MAX_DEPTH);
    assert(triangle->depth() <= depth);

    if (triangle->depth() == depth)
      return;

    if (triangle->isLeaf()) {
      triangle->expand();
    }

    validateOneRing(triangle);

    int below = triangle->depth() + 1;
    for (int v = V1; v <= V3; v += 1) {
      VertexMr* vertex = triangle->vertex(VertexLabel(v));
      assert(vertex->isValid(triangle->depth()));
      if (!vertex->isValid(below)) {
	((Mesh*)this)->even(triangle, VertexLabel(v), accessor);
      }
    }

    TriangleMr* centre = triangle->child(T0);
    for (int e = E1; e <= E3; e += 1) {
      VertexMr* vertex = centre->vertex(OppositeVertex(EdgeLabel(e)));
      if (!vertex->isValid(below)) {
	((Mesh*)this)->odd(triangle, EdgeLabel(e), accessor);
      }
    }

    if (depth > below) {
      for (int t = T0; t <= T3; t += 1) {
	((Mesh*)this)->subdivide(triangle->child(TriangleLabel(t)), depth, accessor);
      }
    }
  }

  /**
   * Subdivides the specified triangle in the subdivision hierarchy to the
   * specified depth.
   * @param triangle  triangle to subdivide
   * @param depth  depth to which to subdivide triangle
   */
  template <class Accessor>
  void subdivideAdaptively(TriangleMr* triangle, float angle, int depth, Accessor accessor) {
    assert(depth <= __MAX_DEPTH);
    assert(triangle->depth() <= depth);

    if (triangle->depth() == depth)
      return;

    if (triangle->isLeaf()) {
      triangle->expand();
    }

    validateOneRing(triangle);

    CVec N0 = normal(triangle, V1);
    CVec N1 = normal(triangle, V2);
    CVec N2 = normal(triangle, V3);
    Real a01 = abs(N0. dot (N1));
    Real a12 = abs(N1. dot (N2));
    Real a23 = abs(N2. dot (N0));

    Real cosangle = cos(angle);

    if (a01 >= cosangle  &&  a12 >= cosangle  &&  a23 >= cosangle)
      return;

    //cerr << "depth = " << triangle->depth() << " a01 = " << a01 << " a12 = " << a12 << " a23 = " << a23 << " cosangle = " << cosangle << endl;

    int below = triangle->depth() + 1;
    for (int v = V1; v <= V3; v += 1) {
      VertexMr* vertex = triangle->vertex(VertexLabel(v));
      assert(vertex->isValid(triangle->depth()));
      if (!vertex->isValid(below)) {
	((Mesh*)this)->even(triangle, VertexLabel(v), accessor);
      }
    }

    TriangleMr* centre = triangle->child(T0);
    for (int e = E1; e <= E3; e += 1) {
      VertexMr* vertex = centre->vertex(OppositeVertex(EdgeLabel(e)));
      if (!vertex->isValid(below)) {
	((Mesh*)this)->odd(triangle, EdgeLabel(e), accessor);
      }
    }

    if (depth > below) {
      for (int t = T0; t <= T3; t += 1) {
	((Mesh*)this)->subdivideAdaptively(triangle->child(TriangleLabel(t)), angle, depth, accessor);
      }
    }
  }

  template <class Accessor>
  void dontsubdivideAdaptively(TriangleMr* triangle, float angle, int depth, Accessor accessor) {
    assert(depth <= __MAX_DEPTH);
    assert(triangle->depth() <= depth);

    if (triangle->depth() == depth)
      return;

    if (triangle->isLeaf()) {
      triangle->expand();
    }

    validateOneRing(triangle);

    int below = triangle->depth() + 1;
    for (int v = V1; v <= V3; v += 1) {
      VertexMr* vertex = triangle->vertex(VertexLabel(v));
      assert(vertex->isValid(triangle->depth()));
      if (!vertex->isValid(below)) {
	((Mesh*)this)->even(triangle, VertexLabel(v), accessor);
      }
    }

    TriangleMr* centre = triangle->child(T0);
    for (int e = E1; e <= E3; e += 1) {
      VertexMr* vertex = centre->vertex(OppositeVertex(EdgeLabel(e)));
      if (!vertex->isValid(below)) {
	((Mesh*)this)->odd(triangle, EdgeLabel(e), accessor);
      }
    }

    if (depth <= below) 
      return;

    CVec N0 = normal(triangle, V1);
    CVec N1 = normal(triangle, V2);
    CVec N2 = normal(triangle, V3);
    Real a01 = abs(N0. dot (N1));
    Real a12 = abs(N1. dot (N2));
    Real a23 = abs(N2. dot (N0));

    if (a01 <= angle  &&  a12 <= angle  &&  a23 <= angle)
      return;

    if (depth > below) {
      for (int t = T0; t <= T3; t += 1) {
	((Mesh*)this)->subdivideAdaptively(triangle->child(TriangleLabel(t)), angle, depth, accessor);
      }
    }
  }


  void subdivide(TriangleMr* triangle, int depth) {
    subdivide(triangle, depth, PointAccessor());
  }

  void subdivideAdaptively(TriangleMr* triangle, float angle, int depth) {
    subdivideAdaptively(triangle, angle, depth, PointAccessor());
  }
  
  /**
   * Ensures that the one-ring neighbours of the specified subdivision triangle
   * exist and are valid.
   * @param triangle  triangle
   */
  void validateOneRing(TriangleMr* triangle) {
#if DEBSUB
    int dep = triangle->depth();
    cerr << "validating one-ring of triangle (" << triangle->vertex(V1)->point(dep) 
	 << "), (" << triangle->vertex(V2)->point(dep) 
	 << "), (" << triangle->vertex(V3)->point(dep) << ")" << endl;
#endif

    if (triangle->isRoot())
      return;

    if (triangle->getValidOneRingFlag())
      return;

    cerr << "<";

    TriangleMr* parent = triangle->parent();
    int depth = triangle->depth();
    Permutation p;
    for (int v = V1; v <= V3; v += 1) {
      // In 3/4 of all cases, this point is incident to an edge of the
      // triangle's parent.
      if (VertexLabel(v) != ParentVertex(triangle->label())) {
	EdgeLabel e = OppositeEdge(VertexLabel(v));
	TriangleMr* adjacent = parent->adjacentTriangle(e, p);
	if (adjacent != 0)
	  ((Mesh*)this)->subdivide(adjacent, depth);
      } else {
	typename TriangleMr::TriangleIt ti = parent->triangleBegin(VertexLabel(v));
	typename TriangleMr::TriangleIt te = parent->triangleEnd(VertexLabel(v));
	while (ti != te) {
	  ((Mesh*)this)->subdivide(*ti, depth);
	  ++ti;
	}
      }
    }

    triangle->setValidOneRingFlag();
    cerr << ">";
  }

public:

  void SanityCheck() {

    TriangleIt ti = triangleBegin();
    TriangleIt te = triangleEnd();


    return;
    while (ti != te) {
      stack<typename Mesh::TriangleMr*> triangles;
      Triangle* control = *ti++;
      triangles.push(control->root());

      while (!triangles.empty()) {
	TriangleMr* triangle = triangles.top();
	triangles.pop();

	if (!triangle->isRoot()) {
	  TriangleLabel t = triangle->label();
	  assert(triangle->parent()->child(t) == triangle);
	}

	for (int e = E1; e <= E3; e++) {
	  Permutation p;
	  TriangleMr* adjacent;
	  adjacent = triangle->adjacentTriangle(EdgeLabel(e), p);
	  assert(adjacent);
	  if (adjacent->depth() < triangle->depth()) continue;
	  assert(adjacent->adjacentTriangle(p.edge(EdgeLabel(e))) == triangle);
	}

	if (!triangle->isLeaf()) {
	  for (int t = T0; t <= T3; t++) {
	    triangles.push(triangle->child(TriangleLabel(t)));
	  }
	}
      }
    }
  }

};

template <class T>
istream& operator>>(istream& is, LoopMesh<T>& mesh)
{
    SMF::readMesh(is, mesh);
    return (is);
}

MESHTYPE(TesselatedMesh, MeshTp, TriangleTp, PointVertex)

template <class T>
ostream& operator<<(ostream& os, LoopMesh<T>& mesh)
{
  DECLARE_LOOP_TYPES( T )

  typedef LoopTesselator<Mesh> Tesselator;

  Tesselator tesselator(mesh);

  tesselator.tesselate();

  TesselatedMesh *tesselation;

  tesselation = tesselator.generateMesh(TesselatedMesh::Types());

  os << *tesselation;
  delete (tesselation);

  return (os);
}

#define LOOPTYPE_INSTANTIATE_TYPES(TYPE_STRUCT, MTYPE, MCLASS, TCLASS, VCLASS, MRTCLASS, MRVCLASS, SCLASS)   \
MESHTYPE_INSTANTIATE_TYPES( TYPE_STRUCT, MTYPE, MCLASS, TCLASS, VCLASS )    \
typedef MRTCLASS<TYPE_STRUCT> MTYPE##_TriangleMr;                           \
typedef MRVCLASS              MTYPE##_VertexMr;                             \
typedef SCLASS                MTYPE##_Stencil;

#define LOOPTYPE_DEFINE_TYPE_STRUCT(TYPE_STRUCT, MTYPE, MCLASS, TCLASS, VCLASS, MRTCLASS, MRVCLASS, SCLASS)   \
MESHTYPE_DEFINE_TYPE_STRUCT( MTYPE##_MeshTypes, MTYPE, MCLASS, TCLASS, VCLASS ) \
struct TYPE_STRUCT : public MTYPE##_MeshTypes {          \
  typedef MTYPE##_TriangleMr          TriangleMr;        \
  typedef MTYPE##_VertexMr            VertexMr;          \
  typedef MTYPE##_Stencil             Stencil;           \
};

#define LOOPTYPE(MTYPE, MCLASS, TCLASS, VCLASS, MRTCLASS, MRVCLASS, SCLASS)                  \
struct MTYPE##_Types;                                                                        \
LOOPTYPE_INSTANTIATE_TYPES(  MTYPE##_Types, MTYPE, MCLASS, TCLASS, VCLASS, MRTCLASS, MRVCLASS, SCLASS) \
LOOPTYPE_DEFINE_TYPE_STRUCT( MTYPE##_Types, MTYPE, MCLASS, TCLASS, VCLASS, MRTCLASS, MRVCLASS, SCLASS) \

#endif  /* __LOOP_LOOPMESH_H__ */
