// -*- Mode: c++ -*-
// $Id: LoopStencil.h,v 1.1.1.1 2001/07/24 02:41:51 pkrysl Exp $
// $Source: /home/cvs/WASP/problems/FE/subdiv/Coke/Loop/LoopStencil.h,v $

#ifndef __LOOP_LOOPSTENCIL_H__
#define __LOOP_LOOPSTENCIL_H__

#include <assert.h>
#include <math.h>
#include "Loop/Permutation.h"
#include "Loop/Stencil.h"
#include "Loop/Topology.h"
#include "Util/cvec.h"

/**
 * LoopStencil Class
 */
class LoopStencil {
public:
  /**
   * Default constructor.
   */
  LoopStencil(void) { }

  /**
   * Copy constructor.
   */
  LoopStencil(const LoopStencil& instance) { }

  /**
   * Destructor.
   */
  ~LoopStencil(void) { }

  /**
   * Assignment operator.
   */
  LoopStencil& operator=(const LoopStencil& instance) {
    return (*this);
  }

  /**
   * Applies the Loop stencil for even vertices to a specified vertex of a
   * given subdivision triangle.  The new position is stored in the
   * subdivision hierarchy and a mutable reference is returned.
   * @param triangle  subdivision triangle
   * @param v  vertex to which to apply subdivision rule
   * @return position of even vertex
   */
  template <class TriangleMr, class Accessor, class Accumulator>
  void even(TriangleMr* triangle, VertexLabel v, Accessor accessor, Accumulator &result) const {
    typename TriangleMr::VertexMr* vertex = triangle->vertex(v);
    int depth = triangle->depth();
    int i;

    // In 9/16 of all cases, this point is incident to an edge of the
    // triangle's parent.
    TriangleLabel t = triangle->label();
    if (!triangle->isRoot() && (t == T0 || v != ParentVertex(t))) {
      EdgeLabel e = OppositeEdge(v);
      VertexLabel vi = CCWVertex(e);
      VertexLabel vj = CWVertex(e);
      Permutation p;

      TriangleMr* parent = triangle->parent();
      TriangleMr* adjacent = parent->adjacentTriangle(e, p)->child(T0);
      TriangleMr* centre = parent->child(T0);

      result += 0.625 * accessor(vertex, depth);
      result += 0.0625 * accessor(parent->vertex(vi), depth);
      result += 0.0625 * accessor(parent->vertex(vj), depth);
      result += 0.0625 * accessor(centre->vertex(vi), depth);
      result += 0.0625 * accessor(centre->vertex(vj), depth);
      result += 0.0625 * accessor(adjacent->vertex(p.vertex(vi)), depth);
      result += 0.0625 * accessor(adjacent->vertex(p.vertex(vj)), depth);
    } else {
      int valence = triangle->valence(v);
      float n = (float)valence;
      float gamma = 3.0f + 2.0f * cos(2.0f * M_PI / n);
      float beta = (0.125f / n) * (5.0f - gamma * gamma / 8.0f);
      float alpha = 1.0f - (n * beta);
      result += alpha * accessor(vertex, depth);

      typename TriangleMr::VertexIt vi = triangle->vertexBegin(v);
      typename TriangleMr::VertexIt ve = triangle->vertexEnd(v);
      for (i = 0; vi != ve; i += 1, ++vi) {
	result += beta * accessor(*vi, depth);
      }

      assert(i == valence);
    }
  }

  /**
   * Applies the Loop stencil to calculate the limit position of a specified 
   * vertex of a given subdivision triangle.
   * @param triangle  subdivision triangle
   * @param v  vertex for which to calculate the limit position
   * @return vertex limit position
   */
  template <class TriangleMr, class Accessor, class Stencil>
  CVec limit(TriangleMr* triangle,
	     VertexLabel v,
	     Accessor accessor,
	     const Stencil& stencil) const
  {
    typename TriangleMr::VertexMr* vertex = triangle->vertex(v);
    TriangleLabel t = triangle->label();
    int depth = triangle->depth();

    int valence = triangle->valence(v);
    float n = (float)valence;
    float gamma = 3.0f + 2.0f * cos(2.0f * M_PI / n);
    float beta = (1.0f / n) * (5.0f - gamma * gamma / 8.0f);
    float omega = 3.0f / beta;
    float rho = 1.0f / (omega + n);
    CVec limit = (omega * rho) * accessor(vertex, depth);

    if (triangle->isRoot()) {
      int i;
      typedef typename TriangleMr::Triangle::Vertex Vertex;

      typename Vertex::VertexIt vi = triangle->triangle()->vertex(v)->vertexBegin();
      typename Vertex::VertexIt ve = triangle->triangle()->vertex(v)->vertexEnd();
      for (i = 0; vi != ve; i += 1, ++vi) {
	limit += rho * accessor((*vi)->vertex(), 0);
      }
      assert(i == valence);
    } else if (t == T0 || v != ParentVertex(t)) {
      TriangleMr* parent = triangle->parent();
      VertexLabel next = NextVertex(v);
      VertexLabel previous = PreviousVertex(v);
      limit += rho * accessor(parent->child(T0)->vertex(next), depth);
      limit += rho * accessor(parent->child(T0)->vertex(previous), depth);
      limit += rho * accessor(parent->vertex(next), depth);
      limit += rho * accessor(parent->vertex(previous), depth);

      TriangleMr* adjacent;
      Permutation p;
      EdgeLabel e = OppositeEdge(v);
      adjacent = triangle->parent()->adjacentTriangle(e, p);
      CVec c1, c2;
      Odd(adjacent, p.edge(NextEdge(e)), accessor, c1, stencil);
      Odd(adjacent, p.edge(PreviousEdge(e)), accessor, c2, stencil);
      limit += rho * c1;
      limit += rho * c2;
    } else {
      int i;
      typename TriangleMr::VertexIt vi = triangle->parent()->vertexBegin(v);
      typename TriangleMr::VertexIt ve = triangle->parent()->vertexEnd(v);
      for (i = 0; vi != ve; i += 1, ++vi) {
	typename TriangleMr::VertexRecord record = *vi;
	EdgeLabel e = record.edge();
	CVec c1;
	Odd(record.triangle(), e, accessor, c1, stencil);
	limit += rho * c1;
      }
      assert(i == valence);
    }
    
    return (limit);
  }

  /**
   * Applies the Loop stencil for odd vertices to a specified edge of a given
   * subdivision triangle.  The new position is stored in the subdivision
   * hierarchy and a mutable reference is returned.
   * @param triangle  subdivision triangle
   * @param e  edge to which to apply subdivision rule
   * @return position of odd vertex
   */
  template <class TriangleMr, class Accessor, class Accumulator>
  void odd(TriangleMr* triangle, EdgeLabel e, Accessor accessor, Accumulator &result) const {
    int depth = triangle->depth();
    VertexLabel v = OppositeVertex(e);
    Permutation p;
    TriangleMr* adjacent = triangle->adjacentTriangle(e, p);

    result += 0.375 * accessor(triangle->vertex(CCWVertex(e)), depth);
    result += 0.375 * accessor(triangle->vertex(CWVertex(e)), depth);
    result += 0.125 * accessor(triangle->vertex(v), depth);
    result += 0.125 * accessor(adjacent->vertex(p.vertex(v)), depth);
  }
};

#endif  /* __LOOP_LOOPSTENCIL_H__ */
