// -*- Mode: c++ -*-
// $Id: TesselatedMesh.h,v 1.1.1.1 2001/07/24 02:41:51 pkrysl Exp $
// $Source: /home/cvs/WASP/problems/FE/subdiv/Coke/Loop/TesselatedMesh.h,v $

#ifndef __LOOP_TESSELATEDMESH_H__
#define __LOOP_TESSELATEDMESH_H__

#include "Loop/Permutation.h"
#include "Mesh/MeshTp.h"
#include "Mesh/PointVertex.h"
#include "Util/cvec.h"

template <class TriangleDm>
class TesselatedVertexTag {
public:
  TesselatedVertexTag(void) { }

  virtual ~TesselatedVertexTag(void) { }

  virtual CVec coords(TriangleDm* domain) = 0;
};

template <class TriangleDm>
class TesselatedInteriorVertexTag : public TesselatedVertexTag<TriangleDm> {
private:
  const CVec m_coords;

  TriangleDm* m_domain;

public:
  TesselatedInteriorVertexTag(TriangleDm* domain, const CVec& coords)
   : TesselatedVertexTag<TriangleDm>(), m_domain(domain), m_coords(coords) { }

  virtual ~TesselatedInteriorVertexTag(void) { }

  virtual CVec coords(TriangleDm* domain) {
    assert(domain == m_domain);
    return (m_coords);
  }
};

template <class TriangleDm>
class TesselatedEdgeVertexTag : public TesselatedVertexTag<TriangleDm> {
private:
  const CVec m_coords;

  TriangleDm* m_domain;

  EdgeLabel m_edge;

public:
  TesselatedEdgeVertexTag(TriangleDm* domain,
			  EdgeLabel edge,
			  const CVec& coords)
   : TesselatedVertexTag<TriangleDm>(),
     m_domain(domain), m_edge(edge), m_coords(coords) { }

  virtual ~TesselatedEdgeVertexTag(void) { }

  virtual CVec coords(TriangleDm* domain) {
    assert(domain == m_domain || domain == m_domain->triangle(m_edge));

    if (domain == m_domain)
      return (m_coords);

    Permutation p = m_domain->permutation(m_edge);
    return CVec(m_coords(p.vertex(V1)),
		m_coords(p.vertex(V2)),
		m_coords(p.vertex(V3)));
  }
};

template <class TriangleDm>
class TesselatedControlVertexTag : public TesselatedVertexTag<TriangleDm> {
private:
  TriangleDm* m_domain;

  VertexLabel m_vertex;

public:
  TesselatedControlVertexTag(TriangleDm* domain,
			     VertexLabel vertex)
   : TesselatedVertexTag<TriangleDm>(),
     m_domain(domain), m_vertex(vertex) { }

  virtual ~TesselatedControlVertexTag(void) { }

  virtual CVec coords(TriangleDm* domain) {
    assert(domain->hasIncidentVertex(m_domain->vertex(m_vertex)));

    CVec coords(0.0f, 0.0f, 0.0f);
    coords(domain->indexOfVertex(m_domain->vertex(m_vertex))) = 1.0f;
    return (coords);
  }
};

template <class Triangle, class Vertex, class TriangleDm, class Secondary>
class TesselatedVertex : public TriangleDm::Vertex {
private:
  TesselatedVertexTag<TriangleDm>* m_tag;

  Secondary m_secondary;

public:
  TesselatedVertex(TesselatedVertexTag<TriangleDm>* tag, Secondary secondary)
   : TriangleDm::Vertex(), m_tag(tag), m_secondary(secondary) { }

  ~TesselatedVertex(void) {
    delete (m_tag);
  }

  CVec coords(TriangleDm* domain) {
    return (m_tag->coords(domain));
  }

  Secondary secondary(void) {
    return (m_secondary);
  }
};

template <class Triangle, class Vertex, class TriangleDm, class Secondary>
class TesselatedTriangle : public TriangleDm::Triangle {
private:
  TriangleDm* m_domain;

  Secondary m_secondary;

public:
  TesselatedTriangle(TriangleDm* domain, Secondary secondary)
   : TriangleDm::Triangle(), m_domain(domain), m_secondary(secondary)
  { }

  ~TesselatedTriangle(void) { }

  TriangleDm* domain(void) const {
    return (m_domain);
  }

  CVec coords(const CVec& coords) {
    assert(coords.x() >= 0.0f && coords.y() >= 0.0f && coords.z() >= 0.0f);
    assert(fabs(coords.x() + coords.y() + coords.z() - 1.0f) < 1.0e-7);

    return (coords.x() * vertex(V1)->coords(m_domain) +
	    coords.y() * vertex(V2)->coords(m_domain) +
	    coords.z() * vertex(V3)->coords(m_domain));
  }

  Secondary secondary(void) {
    return (m_secondary);
  }
};

/**
 * TesselatedMesh Class
 */
template <class T, class Mesh, class Triangle, class Vertex>
class TesselatedMesh : public MeshTp<T> {
public:
  /**
   * Constructs an empty mesh.
   */
  TesselatedMesh(void) : MeshTp<T>() { }

  /**
   * Destructor.
   */
  ~TesselatedMesh(void) { }

  Triangle* triangle(int index) {
    return (m_triangles[index]);
  }
};

#endif  /* __LOOP_TESSELATEDMESH_H__ */
