// -*- Mode: c++ -*-
// $Id: Cloner.h,v 1.1.1.1 2001/07/24 02:41:34 pkrysl Exp $
// $Source: /home/cvs/WASP/problems/FE/subdiv/Coke/Mesh/Cloner.h,v $

#ifndef __MESH_CLONER_H__
#define __MESH_CLONER_H__

#include <hash_map.h>
#include "Mesh/Accessor.h"
#include "Mesh/Labels.h"

#include "Mesh/MeshTypes.h"

/**
 * Cloner Class
 */
template <class T>
class Cloner : public Accessor<typename T::Mesh> {
public:
  DECLARE_MESH_TYPES( T );

  typedef hash_map<Triangle*, Triangle*, Triangle::hash> TriangleMap;
  typedef hash_map<Vertex*, Vertex*, Vertex::hash> VertexMap;

protected:
  Mesh& m_originalMesh;
  TriangleMap m_triangleMap;
  VertexMap m_vertexMap;

protected:
  /**
   * Duplicates the topological components of the original mesh in the cloned
   * mesh.
   */
  void cloneMesh(void) {
    // Clone vertices.
    Mesh::VertexIt vi = m_originalMesh.vertexBegin();
    Mesh::VertexIt ve = m_originalMesh.vertexEnd();
    while (vi != ve) {
      Vertex* originalVertex = *vi++;
      Vertex* vertex = new Vertex(*originalVertex);
      assert(vertex);
      m_vertexMap[originalVertex] = vertex;
      getVertices().push_back(vertex);
    }

    // Clone triangles.
    Mesh::TriangleIt ti = m_originalMesh.triangleBegin();
    Mesh::TriangleIt te = m_originalMesh.triangleEnd();
    while (ti != te) {
      Triangle* originalTriangle = *ti++;
      Triangle* triangle = new Triangle(*originalTriangle);
      assert(triangle);
      m_triangleMap[originalTriangle] = triangle;
      getTriangles().push_back(triangle);

      // Induce same vertices in new triangle.
      triangle->setVertices(m_vertexMap[originalTriangle->vertex(V1)],
			    m_vertexMap[originalTriangle->vertex(V2)],
			    m_vertexMap[originalTriangle->vertex(V3)]);

      // Induce same topological structure in new triangle.
      for (int e = E1; e <= E3; e += 1) {
	Triangle* adjacent = originalTriangle->triangle(EdgeLabel(e));
	if (!adjacent)
	  continue;

	if (m_triangleMap.find(adjacent) != m_triangleMap.end()) {
	  adjacent = m_triangleMap[adjacent];
	  setAdjacentTriangle(triangle, EdgeLabel(e), adjacent);

	  Vertex* vertex = triangle->vertex(VertexLabel(e));
	  EdgeLabel f = adjacent->indexOfEdge(triangle->ccwVertex(vertex),
					      triangle->cwVertex(vertex));
	  setAdjacentTriangle(adjacent, f, triangle);
	}
      }
    }

    // Induce same topological structure in new vertices.
    vi = m_originalMesh.vertexBegin();
    ve = m_originalMesh.vertexEnd();
    while (vi != ve) {
      Vertex* vertex = *vi++;
      Triangle* triangle = vertex->triangle();

      vertex = m_vertexMap[vertex];
      triangle = m_triangleMap[triangle];
      setIncidentTriangle(vertex, triangle);
    }
  }


public:
  /**
   * Constructs a cloning object for the specified mesh and generates a
   * duplicate of the mesh.
   * @param mesh  mesh to clone
   */
  Cloner(Mesh& mesh)
   : Accessor<Mesh>(*(new Mesh(mesh))), m_originalMesh(mesh)
  {
    cloneMesh();
  }


  /**
   * Returns the clone.
   * @return clone generated by this object
   */
  Mesh* getClone(void) {
    return (&mesh());
  }


  /**
   * Returns a map from the triangles in the original mesh to the triangles in
   * the clone.
   * @return map from original triangles to triangle clones
   */
  TriangleMap& getTriangleMap(void) {
    return (m_triangleMap);
  }


  /**
   * Returns a map from the vertices in the original mesh to the vertices in
   * the clone.
   * @return map from original vertices to vertex clones
   */
  VertexMap& getVertexMap(void) {
    return (m_vertexMap);
  }
};

#endif  /* __MESH_CLONER_H__ */
