// -*- Mode: c++ -*-
// $Id: EdgeTp.h,v 1.1.1.1 2001/07/24 02:41:34 pkrysl Exp $
// $Source: /home/cvs/WASP/problems/FE/subdiv/Coke/Mesh/EdgeTp.h,v $

#ifndef __MESH_EDGETP_H__
#define __MESH_EDGETP_H__

#include <function.h>
#include "Mesh/Labels.h"
#include "Util/util.h"

#include "Mesh/MeshTypes.h"

/**
 * EdgeTp Class
 */
template <class T>
class EdgeTp {
public:
  DECLARE_MESH_TYPES( T )

private:
  /** A triangle incident to this edge. */
  Triangle* m_triangle;

  /** The vertices incident to this edge. */
  Vertex* m_vertices[2];

public:
  /**
   * Comparator in lexicographic order.
   */
  struct compare : public binary_function<EdgeTp*, EdgeTp*, bool> {
    bool operator()(const EdgeTp* ei, const EdgeTp* ej) const {
#ifdef _WIN32
      const Vertex* mini = __min(ei->vertex(V1), ei->vertex(V2));
      const Vertex* minj = __min(ej->vertex(V1), ej->vertex(V2));
#else
      const Vertex* mini = min(ei->vertex(V1), ei->vertex(V2));
      const Vertex* minj = min(ej->vertex(V1), ej->vertex(V2));
#endif
      if (mini < minj) {
	return (true);
      } else if (mini == minj) {
#ifdef _WIN32
	const Vertex* maxi = __max(ei->vertex(V1), ei->vertex(V2));
	const Vertex* maxj = __max(ej->vertex(V1), ej->vertex(V2));
#else
	const Vertex* maxi = max(ei->vertex(V1), ei->vertex(V2));
	const Vertex* maxj = max(ej->vertex(V1), ej->vertex(V2));
#endif
	return (maxi < maxj);
      }

      return (false);
    }
  };

public:
  /**
   * Constructs an uninitialized edge.
   */
  EdgeTp(void) : m_triangle(0) {
    m_vertices[V1] = 0;
    m_vertices[V2] = 0;
  }

  /**
   * Copy constructor.
   * @param instance  edge instance to copy
   */
  EdgeTp(const EdgeTp& instance) {
    operator=(instance);
  }

  /**
   * Constructs an edge incident to two specified vertices and a triangle.
   * @param vi  first vertex
   * @param vj  second vertex
   * @param triangle  incident triangle
   */
  EdgeTp(Vertex* vi, Vertex* vj, Triangle* triangle) {
    assert(triangle->hasIncidentVertex(vi) &&
	   triangle->hasIncidentVertex(vj));
    assert(vi != vj);

    m_triangle = triangle;
    m_vertices[V1] = vi;
    m_vertices[V2] = vj;
  }

  /**
   * Constructs an edge incident to two specified vertices.
   * @param vi  first vertex
   * @param vj  second vertex
   */
  EdgeTp(Vertex* vi, Vertex* vj) {
    m_vertices[V1] = vi;
    m_vertices[V2] = vj;

    typename Vertex::TriangleIt ti = vi->triangleBegin();
    typename Vertex::TriangleIt te = vi->triangleEnd();
    for (; ti != te; ++ti) {
      if ((*ti)->hasIncidentVertex(vj)) {
	m_triangle = *ti;
	return;
      }
    }
    die();
  }

  EdgeTp(Triangle* triangle, EdgeLabel e) {
    m_vertices[V1] = triangle->vertex(PreviousVertex(VertexLabel((int)e)));
    m_vertices[V2] = triangle->vertex(NextVertex(VertexLabel((int)e)));
    m_triangle = triangle;
  }

  EdgeTp(Triangle* triangle, Triangle* opposite) {
    int v;

    for (v = V1; v <= V3; v += 1) {
      if (!opposite->hasIncidentVertex(triangle->vertex(VertexLabel(v))))
	break;
      assert(v != V3);
    }

    m_triangle = triangle;
    m_vertices[V1] = triangle->vertex(PreviousVertex(VertexLabel(v)));
    m_vertices[V2] = triangle->vertex(NextVertex(VertexLabel(v)));
  }

  /**
   * Destructor.
   */
  ~EdgeTp(void) { }

  /**
   * Assignment operator.
   * @param instance  edge instance to copy
   */
  EdgeTp& operator=(const EdgeTp& instance) {
    m_triangle = instance.m_triangle;
    m_vertices[V1] = instance.m_vertices[V1];
    m_vertices[V2] = instance.m_vertices[V2];
    return (*this);
  }

  /**
   * Returns the index of the specified vertex incident to this edge.
   * @param vertex  vertex incident to this edge
   * @return vertex index
   */
  int enumerateVertex(const Vertex* vertex) const {
    if (vertex == (Vertex*)m_vertices[V1]) {
      return (V1);
    } else {
      assert(vertex == (Vertex*)m_vertices[V2]);
      return (V2);
    }
  }

  /**
   * Returns true if this edge lies on a boundary.
   * @return true if this edge lies on a boundary
   */
  bool isBoundaryEdge(void) const {
    return (oppositeTriangle() == 0);
  }


  /**
   * Returns the triangle across this edge opposite the reference triangle
   * returned by the triangle method.
   * @return triangle opposite the reference triangle
   */
  Triangle* oppositeTriangle(void) const {
    return (m_triangle->oppositeTriangle(m_vertices[V1], m_vertices[V2]));
  }

  /**
     * Returns a triangle incident to this edge.
     * @return triangle incident to this edge
     */
  Triangle* triangle(void) const {
    return (m_triangle);
  }

  /**
     * Returns the incident vertex associated with the specified index
     * relative to this edge.  Vertices are indexed in construction order.
     * @param index  vertex index
     * @return vertex associated with index
     */
  Vertex* vertex(const int index) const {
    return (m_vertices[index]);
  }
};

#endif  /* __MESH_EDGETP_H__ */
