// -*- Mode: c++ -*-
// $Id: Eraser.h,v 1.1.1.1 2001/07/24 02:41:34 pkrysl Exp $
// $Source: /home/cvs/WASP/problems/FE/subdiv/Coke/Mesh/Eraser.h,v $

#ifndef __MESH_ERASER_H__
#define __MESH_ERASER_H__

#include <hash_set.h>
#include <set.h>
#include "Mesh/Accessor.h"
#include "Mesh/EdgeTp.h"
#include "Mesh/Labels.h"

#include "Mesh/MeshTypes.h"

/**
 * Eraser Class
 */
template <class T>
class Eraser : public Accessor<Mesh> {
public:
  DECLARE_MESH_TYPES( T );

protected:
  typedef set<Edge*, typename Edge::compare> EdgeSet;
  typedef hash_set<Triangle*, typename Triangle::hash> TriangleSet;
  typedef hash_set<Vertex*, typename Vertex::hash> VertexSet;

protected:
  /** The set of edges shared by the triangle set and its complement. */
  EdgeSet _interfaceEdgeSet;

  /** The set of vertices shared by the triangle set and its complement. */
  VertexSet _interfaceVertexSet;

  /** The set of triangles to remove from the mesh. */
  TriangleSet _triangleSet;

  /** The set of vertices in the triangle set. */
  VertexSet _vertexSet;

public:
  /**
   * Constructs an eraser module for a specified mesh.
   * @param mesh  mesh
   */
  Eraser(Mesh& mesh) : Accessor<Mesh>(mesh) { }
  

  /**
   * Removes the selected triangles from the mesh.
   */
  void removeTriangles(void) {
    TriangleSet::iterator ti, te;
    EdgeSet::iterator ei, ee;
    VertexSet::iterator vi, ve;

    // Identify edges on the interface.
    ti = _triangleSet.begin();
    te = _triangleSet.end();
    while (ti != te) {
      Triangle* triangle = *ti++;

      for (int e = E1; e <= E3; ++e) {
	Triangle* opposite = triangle->triangle(EdgeLabel(e));
	if (!opposite)
	  continue;

	if (_triangleSet.find(opposite) == _triangleSet.end()) {
	  Edge* edge = new Edge(triangle, EdgeLabel(e));
	  assert(edge);

	  _interfaceEdgeSet.insert(edge);
	  _interfaceVertexSet.insert(edge->vertex(V1));
	  _interfaceVertexSet.insert(edge->vertex(V2));
	}
      }
    }

    // Verify that the triangle set complement is a 2-manifold.
    vi = _interfaceVertexSet.begin();
    ve = _interfaceVertexSet.end();
    while (vi != ve) {
      Vertex* vertex = *vi++;
      Vertex::TriangleIt vti = vertex->triangleBegin();
      Vertex::TriangleIt vte = vertex->triangleEnd();
      Triangle* triangle;
      do {
	triangle = *vti++;
      } while (_triangleSet.find(triangle) != _triangleSet.end());

      vti = vertex->triangleBegin(triangle);
      vte = vertex->triangleEnd(triangle);
      int components = 0;  // components of complement incident to vertex
      bool inside = false;

      while (vti != vte) {
	triangle = *vti++;
	if (_triangleSet.find(triangle) != _triangleSet.end()) {
	  if (!inside) {
	    inside = true;
	    components++;
	  }
	} else {
	  inside = false;
	}
      }

      if (components != 1) {
	// Triangle set complement is not a 2-manifold.
	die();
      }
    }

    // Make the mesh self-consistent at the interface.
    vi = _interfaceVertexSet.begin();
    ve = _interfaceVertexSet.end();
    while (vi != ve) {
      Vertex* vertex = *vi++;
      Vertex::TriangleIt vti = vertex->triangleBegin();
      Vertex::TriangleIt vte = vertex->triangleEnd();
      Triangle* triangle;
      do {
	triangle = *vti++;
      } while (_triangleSet.find(triangle) == _triangleSet.end());

      vti = vertex->triangleBegin(triangle);
      vte = vertex->triangleEnd(triangle);
      for (;;) {
	triangle = *vti++;
	if (_triangleSet.find(triangle) == _triangleSet.end()) {
	  setIncidentTriangle(vertex, triangle);
	  break;
	}

	if (vti == vte) {
	  die();
	}
      }
    }

    ei = _interfaceEdgeSet.begin();
    ee = _interfaceEdgeSet.end();
    while (ei != ee) {
      Edge* edge = *ei++;
      Triangle* triangle = edge->oppositeTriangle();
      assert(_triangleSet.find(triangle) == _triangleSet.end());

      EdgeLabel e = triangle->indexOfEdge(edge->vertex(V1), edge->vertex(V2));
      setAdjacentTriangle(triangle, e, 0);
      delete (edge);
    }

    // Remove selected region from the mesh.
    ti = _triangleSet.begin();
    te = _triangleSet.end();
    while (ti != te) {
      Triangle* triangle = *ti++;
      Mesh::TriangleIt mti;
      for (mti = mesh().triangleBegin(); *mti != triangle; ++mti);
      getTriangles().erase(mti);
      delete (triangle);
    }

    vi = _vertexSet.begin();
    ve = _vertexSet.end();
    while (vi != ve) {
      Vertex* vertex = *vi++;
      if (_interfaceVertexSet.find(vertex) == _interfaceVertexSet.end()) {
	Mesh::VertexIt vti;
	for (vti = mesh().vertexBegin(); *vti != vertex; ++vti);
	getVertices().erase(vti);
	delete (vertex);
      }
    }
  }


  /**
   * Selects a triangle to be removed from the mesh.
   * @param triangle  triangle to remove from mesh
   */
  void selectTriangle(Triangle* triangle) {
    _triangleSet.insert(triangle);
    for (int v = V1; v <= V3; ++v) {
      _vertexSet.insert(triangle->vertex(VertexLabel(v)));
    }
  }
};

#endif  /* __MESH_ERASER_H__ */
