// -*- Mode: c++ -*-
// $Id: SMF.h,v 1.1.1.1 2001/07/24 02:41:34 pkrysl Exp $
// $Source: /home/cvs/WASP/problems/FE/subdiv/Coke/Mesh/SMF.h,v $

#ifndef __MESH_SMF_H__
#define __MESH_SMF_H__

#include <assert.h>
#include <algo.h>
#include <function.h>
#include <hash_map.h>
#include <iostream.h>
#include "Mesh/Assembler.h"
#include "Mesh/Labels.h"
#include "Util/cvec.h"

#include "Mesh/MeshTypes.h"

/**
 * SMF Class
 */
class SMF {
public:
  /**
   * This convenience method reads a mesh from an input stream which encodes
   * a mesh in SMF format.
   * @param is  input stream
   * @param mesh  mesh to read
   */
  template <class Mesh>
  static void readMesh(istream& is, Mesh& mesh) {
    MeshReader< typename Mesh::Types > reader;
    reader.readMesh(is, mesh);
  }

  /**
   * This convenience method writes a mesh to an output stream in SMF format.
   * @param os  output stream
   * @param mesh  mesh to write
   */
  template <class Mesh>
  static void writeMesh(ostream& os, const Mesh& mesh) {
    MeshWriter<typename Mesh::Types> writer;
    writer.writeMesh(os, mesh);
  }
  
private:
  /**
   * Prevent instantiation of this class.
   */
  SMF(void) { }

public:
  /**
   * SMF::MeshReader Class
   */
  template <class T>
  class MeshReader {
  public:
    DECLARE_MESH_TYPES( T )

  protected:
    /** Mesh assembler. */
    Assembler<T>* m_assembler;
    
    /** Number of triangle fields. */
    int m_triangleCount;
    
    /** Number of vertex fields. */
    int m_vertexCount;

  public:
    /**
     * Default constructor.
     */
    MeshReader(void) { }
    
    /**
     * Destructor.
     */
    ~MeshReader(void) {
      delete (m_assembler);
    }
    
    /**
     * Parses a mesh in SMF format using the specified input stream.
     * @param is  input stream
     * @param mesh  mesh to construct
     */
    void readMesh(istream& is, Mesh& mesh) {
      m_assembler = new Assembler<T>(mesh);
      assert(m_assembler);
      
      readHeader(is);
      readGeometry(is);
      readTopology(is);
      
      m_assembler->assembleMesh();
    }
    
  protected:
    /**
     * Parses the SMF header.
     * @param is  input stream
     */
    void readHeader(istream& is) {
      is >> m_vertexCount >> m_triangleCount;
    }
    
    
    /**
     * Parses the mesh geometry in SMF format.
     * @param is  input stream
     */
    void readGeometry(istream& is) {
      for (int i = 0; i < m_vertexCount; i += 1) {
	CVec point;
	is >> point;
	
	Vertex* vertex = new Vertex();
	assert(vertex);
	vertex->setPoint(point);
	m_assembler->addVertex(vertex);
      }
    }
    
    /**
     * Parses the mesh topology in SMF format.
     * @param is  input stream
     */
    void readTopology(istream& is) {
      for (int i = 0; i < m_triangleCount; i += 1) {
	int v1, v2, v3;
	is >> v1 >> v2 >> v3;
	
	Triangle* triangle = new Triangle();
	assert(triangle);
	triangle->setVertices(m_assembler->getVertex(v1),
			      m_assembler->getVertex(v2),
			      m_assembler->getVertex(v3));
	m_assembler->addTriangle(triangle);
      }
    }
  };
  
  /**
   * SMF::MeshWriter Class
   */
  template <class T>
  class MeshWriter {
  public:
    DECLARE_MESH_TYPES( T )

    typedef typename Mesh::const_VertexIt const_VertexIt;
    typedef typename Mesh::const_TriangleIt const_TriangleIt;

  protected:
    /** Mesh to write. */
    const Mesh* m_mesh;
    
    /** Table of vertex indices. */
    hash_map<Vertex*, int, typename Vertex::hash> m_table;
    
  public:
    /**
     * Default constructor.
     */
    MeshWriter(void) { }
    
    /**
     * Destructor.
     */
    ~MeshWriter(void) { }
    
    /**
     * Writes a mesh in SMF format to a specified output stream.
     * @param  os  output stream
     * @param  mesh  mesh to write
     */
    void writeMesh(ostream& os, const Mesh& mesh) {
	m_mesh = &mesh;
	
	writeHeader(os);
	writeGeometry(os);
	writeTopology(os);
      }
    
  protected:
    /**
     * Writes the SMF header.
     * @param os  output stream
     */
    void writeHeader(ostream& os) {
      os << m_mesh->vertexSize() << ' '
	 << m_mesh->triangleSize() << endl;
    }
    
    /**
     * Writes the mesh geometry in SMF format.
     * @param os  output stream
     */
    void writeGeometry(ostream& os) {
      const_VertexIt vi = m_mesh->vertexBegin();
      const_VertexIt ve = m_mesh->vertexEnd();
      for (int i = 0; vi != ve; i += 1) {
	os << (*vi)->point() << endl;
	m_table[*vi] = i;
	++vi;
      }
    }
    
    /**
     * Writes the mesh topology in SMF format.
     * @param os  output stream
     */
    void writeTopology(ostream& os) {
      const_TriangleIt ti = m_mesh->triangleBegin();
      const_TriangleIt te = m_mesh->triangleEnd();
      
      for (; ti != te; ++ti) {
	os << m_table[(*ti)->vertex(V1)] << ' '
	   << m_table[(*ti)->vertex(V2)] << ' '
	   << m_table[(*ti)->vertex(V3)] << endl;
      }
    }
  };
};

#endif  /* __MESH_SMF_H__ */
