// -*- Mode: c++ -*-
// $Id: TriangleTp.h,v 1.1.1.1 2001/07/24 02:41:34 pkrysl Exp $
// $Source: /home/cvs/WASP/problems/FE/subdiv/Coke/Mesh/TriangleTp.h,v $

#ifndef __MESH_TRIANGLETP_H__
#define __MESH_TRIANGLETP_H__

#include "Mesh/Accessor.h"
#include "Mesh/Labels.h"
#include "Util/util.h"

#include "Mesh/MeshTypes.h"

#include "Mesh/VertexTp.h"

/**
 * TriangleTp Class
 *
 * TriangleTp methods which accept VertexTp<TriangleIf, VertexIf>* are
 * provided for compatibility with VertexTp methods.
 */
template <class T>
class TriangleTp {
  template <class TYPES> friend class Accessor;

public:
  DECLARE_MESH_TYPES( T )

  typedef VertexTp<T> VertexTp;

public:
  /**
   * Hash function for mesh triangles.
   */
  struct hash {
    size_t operator()(Triangle* triangle) const {
      return size_t(triangle);
    }
  };

private:
  /** The triangles adjacent to this triangle across its edges. */
  Triangle* m_triangles[3];

  /** The vertices of this triangle, in counter-clockwise orientation. */
  Vertex* m_vertices[3];

private:
  /**
   * Assignment operator: undefined operation.
   */
  TriangleTp& operator=(const TriangleTp&) {
    die();
    return (*this);
  }

protected:
  void setAdjacentTriangle(const EdgeLabel e, Triangle* triangle) {
    m_triangles[e] = triangle;
  }

  void setIncidentVertex(const VertexLabel v, VertexTp* vertex) {
    m_vertices[v] = vertex;
  }

public:
  /**
   * Constructs an uninitialized triangle.  Initially the triangle does not
   * belong to the topology of a mesh.
   */
  TriangleTp(void) {
    m_triangles[E1] = 0;
    m_triangles[E2] = 0;
    m_triangles[E3] = 0;
    m_vertices[V1] = 0;
    m_vertices[V2] = 0;
    m_vertices[V3] = 0;
  }

  /**
   * Copy constructor.
   */
  TriangleTp(const TriangleTp&) {
    m_triangles[E1] = 0;
    m_triangles[E2] = 0;
    m_triangles[E3] = 0;
    m_vertices[V1] = 0;
    m_vertices[V2] = 0;
    m_vertices[V3] = 0;
  }

  /**
   * Destructor.
   */
  ~TriangleTp(void) { }

  /**
   * Returns the triangle which is the counter-clockwise neighbor of this
   * triangle with respect to the specified common vertex.
   * @param v  vertex incident to this triangle
   * @return counter-clockwise neighbor of this triangle relative to vertex
   */
  Triangle* ccwTriangle(const VertexTp* v) const {
    return (triangle(EdgeLabel(NextVertex(indexOfVertex(v)))));
  }

  /**
   * Returns the vertex which is the counter-clockwise neighbor of a
   * specified vertex with respect to this triangle.
   * @param vertex  v incident to this triangle
   * @return counter-clockwise neighbor of vertex relative to this triangle
   */
  Vertex* ccwVertex(const VertexTp* v) const {
    return (vertex(NextVertex(indexOfVertex(v))));
  }

  /**
   * Returns the triangle which is the clockwise neighbour of this triangle
   * with respect to the specified common vertex.
   * @param v  vertex incident to this triangle
   * @return clockwise neighbor of this triangle relative to vertex
   */
  Triangle* cwTriangle(const VertexTp* v) const {
    return (triangle(EdgeLabel(PreviousVertex(indexOfVertex(v)))));
  }

  /**
   * Returns the vertex which is the clockwise neighbor of a specified
   * vertex with respect to this triangle.
   * @param v  vertex incident to this triangle
   * @return clockwise neighbor of vertex relative to this triangle
   */
  Vertex* cwVertex(const VertexTp* v) const {
    return (vertex(PreviousVertex(indexOfVertex(v))));
  }

  /**
   * Returns the index of the edge of this triangle incident to the specified
   * vertices.
   * given by the specified vertices.
   * @param vi  first vertex
   * @param vj  second vertex
   * @return index of incident edge
   */
  EdgeLabel indexOfEdge(const VertexTp* vi, const VertexTp* vj) const {
    if (vi == (VertexTp*)vertex(V1)) {
      if (vj == (VertexTp*)vertex(V2))
	return (E3);
      else
	return (E2);
    } else if (vi == (VertexTp*)vertex(V2)) {
      if (vj == (VertexTp*)vertex(V1))
	return (E3);
      else
	return (E1);
    } else {
      if (vj == (VertexTp*)vertex(V1))
	return (E2);
      else if (vj == (VertexTp*)vertex(V2))
	return (E1);
      else
	die();
    }

    return (EdgeLabel(0));  // prevent compiler warning
  }

  /**
   * Returns the index of the specified triangle adjacent to this triangle.
   * @param t  triangle adjacent to this triangle
   * @return triangle index
   */
  EdgeLabel indexOfTriangle(const Triangle* t) const {
    if (triangle(E1) == t)
      return (E1);
    else if (triangle(E2) == t)
      return (E2);
    else if (triangle(E3) == t)
      return (E3);
    else
      die();

    return (EdgeLabel(0));  // prevent compiler warning
  }

  /**
   * Returns the index of the specified vertex incident to this triangle.
   * @param v  vertex incident to this triangle
   * @return vertex index
   */
  VertexLabel indexOfVertex(const VertexTp* v) const {
    if (vertex(V1) == v)
      return (V1);
    else if (vertex(V2) == v)
      return (V2);
    else if (vertex(V3) == v)
      return (V3);
    else
      die();

    return (VertexLabel(0));  // prevent compiler warning
  }

  /**
   * Returns true if the specified triangle is adjacent to this triangle.
   * @param t  triangle
   * @return  true if this triangle contains the specified triangle
   */
  bool hasAdjacentTriangle(const Triangle* t) const {
    return (t == (Triangle*)triangle(E1) ||
	    t == (Triangle*)triangle(E2) ||
	    t == (Triangle*)triangle(E3));
  }

  /**
   * Returns true if the specified vertex is incident to this triangle.
   * @param v  vertex
   * @return  true if this triangle contains vertex
   */
  bool hasIncidentVertex(const VertexTp* v) const {
    return (v == (VertexTp*)vertex(V1) ||
	    v == (VertexTp*)vertex(V2) ||
	    v == (VertexTp*)vertex(V3));
  }
  
  /**
   * Returns the adjacent triangle opposite the specified vertex.
   * @param v  vertex incident to this triangle
   * @return triangle opposite vertex relative to this triangle
   */
  Triangle* oppositeTriangle(const VertexTp* v) const {
    return (triangle(EdgeLabel(indexOfVertex(v))));
  }

  /**
   * Returns the adjacent triangle opposite the edge specified by two
   * vertices incident to this triangle.
   * @param vi  first vertex incident to this triangle
   * @param vj  second vertex incident to this triangle
   * @return triangle opposite the edge specified by (vi,vj)
   */
  Triangle* oppositeTriangle(const VertexTp* vi, const VertexTp* vj) const {
    return (triangle(indexOfEdge(vi, vj)));
  }

  /**
   * Returns the vertex incident to this triangle opposite the edge specified
   * by the two other vertices incident to this triangle.
   * @param vi  first vertex incident to this triangle
   * @param vj  second vertex incident to this triangle
   * @return vertex incident to this triangle opposite the edge  specified by
   *         (vi,vj)
   */     
  Vertex* oppositeVertex(const VertexTp* vi, const VertexTp* vj) const {
    return (vertex(VertexLabel(indexOfEdge(vi, vj))));
  }

  Vertex* oppositeVertex(const Triangle* t) const {
    return (vertex(VertexLabel(indexOfTriangle(t))));
  }

  /**
   * Defines the vertices of this triangle, in counter-clockwise orientation.
   * The triangle must be non-degenerate: each vertex must be unique.  This
   * method must be called exactly once for each triangle.
   * @param v1  first vertex
   * @param v2  second vertex
   * @param v3  third vertex
   */
  void setVertices(Vertex* v1, Vertex* v2, Vertex* v3) {
    assert(m_vertices[V1] == 0 && m_vertices[V2] == 0 && m_vertices[V3] == 0);
    assert(v1 != 0 && v2 != 0 && v3 != 0);
    assert(v1 != v2 && v2 != v3 && v3 != v1);

    m_vertices[V1] = v1;
    m_vertices[V2] = v2;
    m_vertices[V3] = v3;
  }

  /**
   * Returns the adjacent triangle associated with the specified index
   * relative to this triangle.  Triangles are ordered counter-clockwise
   * with respect to this triangle.
   * @param index  triangle index
   * @return triangle associated with index
   */
  Triangle* triangle(const EdgeLabel index) const {
    return (m_triangles[index]);
  }

  /**
   * Returns the incident vertex associated with the specified index
   * relative to this triangle.  Vertices are ordered counter-clockwise
   * with respect to this triangle.
   * @param index  vertex index
   * @return vertex associated with index
   */
  Vertex* vertex(const VertexLabel index) const {
    return (m_vertices[index]);
  }

  /* Methods for backwards compatibility */
  bool boundary(const EdgeLabel e) const {
    return (!t(e));
  }

  bool contains(const VertexTp* v) const { 
    return (v == v1() || v == v2() || v == v3());
  }

  bool contains(const Triangle* t) const {
    return (t == t1() || t == t2() || t == t3());
  }

  EdgeLabel edge(const VertexTp *va, const VertexTp *vb) const {
    return EdgeLabel(opposite(va,vb));
  }

  VertexLabel opposite(const VertexTp* va, const VertexTp* vb) const {
    assert(contains(va) && contains(vb));
    return (va == v1() ? (vb == v2() ? V3 : V2) :
	    (va == v2() ? (vb == v3() ? V1 : V3) : (vb == v1() ? V2 : V1)));
  }

  Triangle* t1(void) const { return (m_triangles[E1]); }
  Triangle* t2(void) const { return (m_triangles[E2]); }
  Triangle* t3(void) const { return (m_triangles[E3]); }
  Triangle* t(const EdgeLabel e) const { return (m_triangles[e]); }

  EdgeLabel t(const Triangle* t) const {
    assert(contains(t)); return (t == t1() ? E1 : (t == t2() ? E2 : E3));
  }

  Vertex* v(const VertexLabel index) const { return m_vertices[index]; }
  Vertex* v1(void) const { return (m_vertices[V1]); }
  Vertex* v2(void) const { return (m_vertices[V2]); }
  Vertex* v3(void) const { return (m_vertices[V3]); }

  VertexLabel v(const VertexTp* v) const {
    assert(contains(v)); return (v == v1() ? V1 : (v == v2() ? V2 : V3));
  }

  Vertex* v(const VertexTp *va, const VertexTp *vb) const {
    return (m_vertices[opposite(va,vb)]);
  }

  Triangle* t(const VertexTp *va) const {
    return (m_triangles[v(va)]);
  }

  Triangle* t(const VertexTp *va, const VertexTp *vb) const {
    return (m_triangles[opposite(va, vb)]);
  }
};

#endif  /** __MESH_TRIANGLETP_H__ */
