// -*- Mode: c++ -*-
// $Id: VertexTp.h,v 1.1.1.1 2001/07/24 02:41:34 pkrysl Exp $
// $Source: /home/cvs/WASP/problems/FE/subdiv/Coke/Mesh/VertexTp.h,v $

#ifndef __MESH_VERTEXTP_H__
#define __MESH_VERTEXTP_H__

#include <iterator.h>
#include "Mesh/Accessor.h"
#include "Util/util.h"

#include "Mesh/MeshTypes.h"

/**
 * VertexTp Class
 */
template <class T>
class VertexTp {
  template <class TYPES> friend class Accessor;

public:
  DECLARE_MESH_TYPES( T )

  /**
   * VertexTp::TriangleIt Class
   */
  class TriangleIt : public forward_iterator<Triangle*, int> {
    friend class VertexTp;

  private:
    /** The final triangle in the sequence. */
    Triangle* m_last;

    /** The next triangle in the sequence. */
    Triangle* m_next;

    /** The vertex about which to iterate. */
    const VertexTp* m_vertex;

  private:
    /**
     * Constructs an iterator over the triangles incident to the specified
     * vertex.  The sequence begins with the specified triangle.
     * @param vertex  vertex
     * @param first  first triangle in sequence
     * @param end  true if iterator denotes end of sequence
     */
    TriangleIt(const VertexTp* vertex, Triangle* first, bool end = false)
      : m_vertex(vertex),
	m_last(first),
	m_next(end ? 0 : first) { }

  public:
    /**
     * Default constructor.
     */
    TriangleIt(void) { }

    /**
     * Copy constructor.
     * @param instance  iterator to copy
     */
    TriangleIt(const TriangleIt& instance)
      : m_last(instance.m_last),
	m_next(instance.m_next),
	m_vertex(instance.m_vertex) { }

    /**
     * Destructor.
     */
    ~TriangleIt(void) { }

    /**
     * Copies the specified iterator.
     * @param instance  iterator to copy
     */
    TriangleIt& operator=(const TriangleIt& instance) {
      m_last = instance.m_last;
      m_next = instance.m_next;
      m_vertex = instance.m_vertex;
      return (*this);
    }

    /**
     * Returns true if this iterator is positioned identically to the
     * specified iterator.
     * @param instance  iterator to compare
     * @return true if positions are identical
     */
    bool operator==(const TriangleIt& instance) const {
      return (instance.m_next == m_next &&
	      instance.m_last == m_last &&
	      instance.m_vertex == m_vertex);
    }

#ifdef _WIN32
    bool operator!=(const TriangleIt& instance) const {
      return !(instance == *this);
    }
#endif

    /**
     * Returns the triangle at the current iterator position.
     * @return triangle at current position
     */
    Triangle* operator*(void) const {
      return (m_next);
    }

    /**
     * Advances this iterator to the next triangle in the sequence.
     * @return this iterator
     */
    TriangleIt& operator++(void) {
      Triangle* current = m_next;

      m_next = current->ccwTriangle(m_vertex);
      if (!m_next)
	m_next = m_vertex->triangle();
      if (m_next == m_last)
	m_next = 0;

      return (*this);
    }

    /**
     * Advances this iterator to the next triangle in the sequence.
     * @return copy of this iterator prior to advancement
     */
    TriangleIt operator++(int) {
      TriangleIt iterator = *this;
      ++(*this);
      return (iterator);
    }
  };  /* VertexTp::TriangleIt */

  /**
   * VertexTp::VertexIt Class
   */
  class VertexIt : public forward_iterator<Vertex*, int> {
    friend class VertexTp;

  private:
    /** The final vertex in the sequence. */
    const Vertex* m_last;

    /** The next vertex in the sequence. */
    Vertex* m_next;

    /** The triangle which contains the next vertex in the sequence. */
    const Triangle* m_triangle;

    /** The vertex about which to iterate. */
    const VertexTp* m_vertex;

  private:
    /**
     * Constructs an iterator over the vertices adjacent to the specified
     * vertex.  The sequence begins with the vertex <tt>first</tt>.  A triangle
     * is provided which is incident to both of the vertices, and has the
     * property that the second vertex is counter-clockwise from the first
     * unless these two vertices constitute a boundary edge.
     * @param	vertex  vertex
     * @param	first  first vertex in sequence
     * @param	triangle  triangle incident to first and second
     * @param	end  true if iterator denotes end of sequence
     */
    VertexIt(const VertexTp* vertex,
	     Vertex* first,
	     const Triangle* triangle,
	     bool end = false)
      : m_triangle(triangle),
	m_vertex(vertex),
	m_last(first),
	m_next(end ? 0 : first) { }

  public:
    /**
     * Default constructor.
     */
    VertexIt(void) { }

    /**
     * Copy constructor.
     * @param instance  iterator to copy
     */
    VertexIt(const VertexIt& instance)
      : m_last(instance.m_last),
	m_next(instance.m_next),
	m_triangle(instance.m_triangle),
	m_vertex(instance.m_vertex) { }

    /**
     * Destructor.
     */
    ~VertexIt(void) { }

    /**
     * Copies the specified iterator.
     * @param instance  iterator to copy
     */
    VertexIt& operator=(const VertexIt& instance) {
      m_last = instance.m_last;
      m_next = instance.m_next;
      m_triangle = instance.m_triangle;
      m_vertex = instance.m_vertex;
      return (*this);
    }

    /**
     * Returns true if this iterator is positioned identically to the
     * specified iterator.
     * @param instance  iterator to compare
     * @return true if positions are identical
     */
    bool operator==(const VertexIt& instance) const {
      return (instance.m_next == m_next &&
	      instance.m_last == m_last &&
	      instance.m_vertex == m_vertex);
    }

#ifdef _WIN32
    bool operator!=(const VertexIt& instance) const {
      return (!(instance== *this));
    }
#endif

    /**
     * Returns the vertex at the current iterator position.
     * @return vertex at current position
     */
    Vertex* operator*(void) const {
      return (m_next);
    }

    /**
     * Advances this iterator to the next vertex in the sequence.
     * @return this iterator
     */
    VertexIt& operator++(void) {
      if (!m_triangle) {
	m_triangle = m_vertex->triangle();
	m_next = m_triangle->ccwVertex(m_vertex);
      } else {
	m_next = m_triangle->cwVertex(m_vertex);
	m_triangle = m_triangle->ccwTriangle(m_vertex);
      }

      if (m_next == m_last)
	m_next = 0;

      return (*this);
    }

    /**
     * Advances this iterator to the next vertex in the sequence.
     * @return copy of this iterator prior to advancement
     */
    VertexIt operator++(int) {
      VertexIt iterator = *this;
      ++(*this);
      return (iterator);
    }

    /* Methods for backwards compatibility */
    // NJL: remove this method once edge iterator is implemented
    operator const Triangle*(void) const { 
      return (m_triangle);
    }
  };  /* VertexTp::VertexIt */

  /**
   * Hash function for mesh vertices.
   */
  struct hash {
    size_t operator()(Vertex* vertex) const {
      return size_t(vertex);
    }
  };

private:
  /** 
   * A triangle incident to this vertex.  If this vertex lies on a boundary,
   * triangle is the clockwise-most triangle on the boundary.
   */
  Triangle* m_triangle;

private:
  /**
   * Assignment operator: undefined operation.
   */
  VertexTp& operator=(const VertexTp&) {
    die();
    return (*this);
  }

protected:
  void setIncidentTriangle(Triangle* triangle) {
    assert(triangle->hasIncidentVertex(this));
    m_triangle = triangle;
  }

public:
  /**
   * Constructs an uninitialized vertex.  Initially the vertex does not
   * belong to the topology of a mesh.
   */
  VertexTp(void) : m_triangle(0) { }

  /**
   * Copy constructor.
   */
  VertexTp(const VertexTp&) : m_triangle(0) { }

  /**
   * Destructor.
   */
  ~VertexTp(void) { }

  /**
   * Returns true if this vertex lies on a boundary.
   * @return true if this vertex lies on a boundary
   */
  bool isBoundaryVertex(void) const {
    return (!triangle()->cwTriangle(this));
  }

  /**
   * Returns a triangle incident to this vertex.  If this vertex lies on a
   * boundary, the clockwise-most triangle on the boundary is returned.
   * @return triangle incident to this vertex
   */
  Triangle* triangle(void) const {
    return (m_triangle);
  }

  /**
   * Returns an iterator over the triangles incident to this vertex,
   * positioned at the start of the list.  Triangles are traversed in
   * counter-clockwise order, beginning with the reference triangle
   * returned by the triangle method.
   * @return iterator over triangles incident to this vertex, positioned at
   *	     start of list
   */
  TriangleIt triangleBegin(void) const {
    return TriangleIt(this, triangle());
  }

  /**
   * Returns an iterator over the triangles incident to this vertex,
   * positioned at the specified triangle.  Triangles are traversed in
   * counter-clockwise order.
   * @param first  first triangle in sequence, incident to this vertex
   * @return iterator over triangles incident to this vertex, positioned at
   *	     start of list
   */
  TriangleIt triangleBegin(Triangle* first) const {
    assert(first->contains(this));
    return TriangleIt(this, first);
  }

  /**
   * Returns an iterator over the triangles incident to this vertex,
   * positioned at the end of the list.
   * @return iterator over incident triangles, positioned at end of list
   */
  TriangleIt triangleEnd(void) const {
    return TriangleIt(this, triangle(), true);
  }

  /**
   * Returns an iterator over the triangles incident to this vertex,
   * positioned at the end of the list which originates at the specified
   * triangle.
   * @param first  first triangle in sequence, incident to this vertex
   * @return iterator over incident triangles, positioned at end of list
   */
  TriangleIt triangleEnd(Triangle* first) const {
    assert(first->contains(this));
    return TriangleIt(this, first, true);
  }

  /**
   * Returns the valence of this vertex.
   * @return valence of this vertex
   */
  int valence(void) const {
    int valence = 0;
    Triangle* next = triangle();

    do {
      next = next->ccwTriangle(this);
      valence += 1;
    } while (next != triangle() && next != 0);

    if (isBoundaryVertex())
      valence += 1;

    return (valence);
  }

  /**
   * Returns an iterator over the vertices adjacent to this vertex,
   * positioned at the start of the list.  Vertices are traversed in
   * counter-clockwise order, beginning with the vertex counter-clockwise
   * from this vertex indicent to the reference triangle returned by the
   * triangle method.
   * @return iterator over adjacent vertices, positioned at start of list
   */
  VertexIt vertexBegin(void) const {
    return VertexIt(this, triangle()->ccwVertex(this), triangle());
  }

  /**
   * Returns an iterator over the vertices adjacent to this vertex,
   * positioned at the specified vertex.  Vertices are traversed in
   * counter-clockwise order.
   * @param first  first vertex in sequence, adjacent to this vertex
   * @return iterator over adjacent vertices, positioned at start of list
   */
  VertexIt vertexBegin(Vertex* first) const {
    Triangle* incident = triangle();
    while (true) {
      if (first == incident->ccwVertex(this))
	break;

      Triangle* next = incident->ccwTriangle(this);
      if (next == triangle()) {
	die();
      } else if (!next) {
	if (first == incident->cwVertex(this)) {
	  incident = next;
	  break;
	} else {
	  die();
	}
      } else {
	incident = next;
      }
    }

    return VertexIt(this, first, incident);
  }

  /**
   * Returns an iterator over the vertices adjacent to this vertex,
   * positioned at the end of the list.
   * @return iterator over adjacent vertices, positioned at end of list
   */
  VertexIt vertexEnd(void) const {
    return VertexIt(this, triangle()->ccwVertex(this), 0, true);
  }

  /**
   * Returns an iterator over the vertices adjacent to this vertex,
   * positioned at the end of the list which originates at the specified
   * vertex.
   * @param first  first vertex in sequence, adjacent to this vertex
   * @return iterator over adjacent vertices, positioned at end of list
   */
  VertexIt vertexEnd(Vertex* first) const {
    return VertexIt(this, first, 0, true);
  }
};

#endif  /* __MESH_VERTEXTP_H__ */
