// -*- Mode: c++ -*-
// $Id: Assembler.h,v 1.1.1.1 2001/07/24 02:41:34 pkrysl Exp $
// $Source: /home/cvs/WASP/problems/FE/subdiv/Coke/Mesh/bak/Assembler.h,v $

#ifndef __ASSEMBLER_H__
#define __ASSEMBLER_H__

#include <algo.h>
#include <set.h>
#include <vector.h>
#include "Mesh/Accessor.h"
#include "Mesh/EdgeTp.h"
#include "Mesh/Labels.h"

/**
 * Assembler Class
 */
template <class TriangleIf, class VertexIf>
class Assembler : private Accessor<TriangleIf, VertexIf> {
public:
  typedef vector<Triangle*>::iterator TriangleIt;
  typedef vector<Vertex*>::iterator VertexIt;

private:
  typedef EdgeTp<Triangle, Vertex> Edge;

  vector<Triangle*> m_triangles;
  vector<Vertex*> m_vertices;
  set<Vertex*> m_boundaryVertices;
  bool m_assembled;

public:
  /**
   * Constructs an assembler for a specified mesh.
   * @param mesh  mesh to assemble
   */
  Assembler(Mesh& mesh)
   : Accessor<TriangleIf, VertexIf>(mesh), m_assembled(false) { }

  /**
   * Assembles the added components into the mesh.  The components must
   * define a topological 2-manifold.  If this condition is not satisfied,
   * the method returns false and the original mesh is unchanged.  If
   * satisfied, the method returns true and the new components are
   * integrated into the mesh.
   * @return true if components assembled into mesh
   */
  bool assembleMesh(void) {
    typedef EdgeTp<Triangle, Vertex> Edge;
    set<Edge*, Edge::compare> edges;

    // Construct edges on the boundary vertices from the original mesh.
    set<Vertex*>::iterator bi = m_boundaryVertices.begin();
    set<Vertex*>::iterator be = m_boundaryVertices.end();
    for (; bi != be; ++bi) {
      Vertex* vertex = *bi;
      Vertex* neighbour;
      Triangle* triangle;
      bool hasNeighbour = false;

      // Search for counter-clockwise neighbour.
      triangle = vertex->triangle();
      neighbour = triangle->ccwVertex(vertex);
      if (m_boundaryVertices.count(neighbour) != 0) {
	Edge* edge = new Edge(vertex, neighbour, triangle);
        assert(edge);
	if (edges.count(edge) == 0)
          edges.insert(edge);
	else
          delete (edge);
	hasNeighbour = true;
      }

      // Search for clockwise neighbour.
      // PS: whoa! running through the list to find the last guy??? Bad,
      // PS: bad, bad...
      // NJL: why? existing mesh is a valid 2-manifold so this is safe
      Vertex::TriangleIt ti = vertex->triangleBegin();
      Vertex::TriangleIt te = vertex->triangleEnd();
      for (; ti != te; ++ti)
	triangle = *ti;
      neighbour = triangle->cwVertex(vertex);
      if (m_boundaryVertices.count(neighbour) != 0) {
	Edge* edge = new Edge(vertex, neighbour, triangle);
        assert(edge);
	if (edges.count(edge) == 0)
          edges.insert(edge);
	else
          delete (edge);
	hasNeighbour = true;
      }

      assert(hasNeighbour);
    }

    // Connect the triangles along their common edges.
    // New triangles are connected to triangles in the
    // original mesh along the common boundary.
    // New vertices are connected to their incident
    // triangles.
    TriangleIt ti = m_triangles.begin();
    TriangleIt te = m_triangles.end();
    for (; ti != te; ++ti) {
      Triangle* triangle = *ti;
      int v;

      // Connect triangle to its neighbours.
      for (v = V1; v <= V3; v += 1) {
	Vertex* vi = triangle->vertex(VertexLabel(v));
	Vertex* vj = triangle->vertex(NextVertex(VertexLabel(v)));
	Edge* edge = new Edge(vi, vj, triangle);
	assert(edge);

	if (edges.count(edge) != 0) {
	  Edge* temp = *edges.find(edge);
	  delete (edge);
	  edge = temp;

	  // Verify that neighbour is free on that edge.
	  Triangle* neighbour = edge->triangle();
	  Vertex* opposite = neighbour->oppositeVertex(vi, vj);
	  int index = neighbour->indexOfVertex(opposite);
	  assert(neighbour->triangle(EdgeName(index)) == 0);

	  // Connect triangles along edge.
	  setAdjacentTriangle(neighbour, EdgeName(index), triangle);
	  opposite = triangle->oppositeVertex(vi, vj);
	  index = triangle->indexOfVertex(opposite);
	  setAdjacentTriangle(triangle, EdgeName(index), neighbour);
	} else {
	  edges.insert(edge);
	}
      }

      // Initially connect free vertices to triangle.
      for (v = V1; v <= V3; v += 1) {
	Vertex* vertex = triangle->vertex(VertexLabel(v));
	if (vertex->triangle() == 0) 
	  setIncidentTriangle(vertex, triangle);
      }
    }

    // %EG: I've modified this!
    // Verify that each new vertex has an incident triangle and
    // properly define the incidence relationship according to the
    // clockwise-most triangle convention.
    VertexIt vi = m_vertices.begin(); 
    VertexIt ve = m_vertices.end();
    int ignored = 0;
    for (; vi != ve; ++vi) {
      Vertex* vertex = *vi;
      Triangle* triangle = vertex->triangle();
      if (triangle == 0) {
	//static int count = 0;
	//cerr << "Warning: isolated vertex at " << vertex->point();
	//cerr << " ignored (" << ++count << " vertices ignored)" << endl;
	++ignored;
	m_boundaryVertices.insert(vertex);
	continue;
      }

      // Locate clockwise-most triangle.
      if (triangle->cwTriangle(vertex) != 0) {
	Triangle* current = triangle->cwTriangle(vertex);
	for (;;) {
	  Triangle* next = current->cwTriangle(vertex);
	  if (next == 0) {
	    setIncidentTriangle(vertex, current);
	    break;
	  } else if (next == triangle) {
	    break;
	  } else {
	    current = next;
	  }
	  // PS: uh oh. How will you know that this can't loop forever???
	  // NJL: agreed.  need to rethink for non-oriented meshes.
	}
      }		
    }
    
    if (ignored) {
      cerr << "Mesh Assembler WARNING: " << ignored << " isolated vertices ignored." << endl;
    }

    // Verify that the neighbourhood of each vertex is isomorphic
    // to the open disk or the half-open disk.
    for (ti = m_triangles.begin(); ti != te; ++ti) {
      Triangle* triangle = *ti;
      for (int v = V1; v <= V3; ++v) {
	Vertex* vertex = triangle->vertex(VertexLabel(v));
	Vertex::TriangleIt vti = vertex->triangleBegin();
	Vertex::TriangleIt vte = vertex->triangleEnd();
	bool found = false;
	for (; vti != vte; ++vti) {
	  if (triangle == *vti) {
	    found = true;
	    break;
	  }
	}
	assert(found);
      }
    }

    // The new components define a topological 2-manifold.
    // Integrate with the original mesh.
    // PS: can't you just assign these???
    // NJL: no, we may be adding to a non-empty mesh
    for (ti = m_triangles.begin(); ti != te; ++ti) {
      getTriangles().push_back(*ti);
    }
    for (vi = m_vertices.begin(); vi != ve; ++vi) {
      if (m_boundaryVertices.count(*vi) == 0)
	getVertices().push_back(*vi);
    }

    for_each(edges.begin(), edges.end(), Delete<Edge>());

    return (true);
  }

  /**
   * Receives a triangle to be assembled into the mesh.
   * @param triangle  triangle to add to mesh
   * @return index of triangle in assembly list
   */
  int addTriangle(Triangle* triangle) {
    assert(!m_assembled);
    assert(triangle->triangle(E1) == 0 &&
	   triangle->triangle(E2) == 0 &&
	   triangle->triangle(E3) == 0);

    m_triangles.push_back(triangle);
    return (m_triangles.size() - 1);
  }

  /**
   * Receives a vertex to be assembled into the mesh.
   * @param vertex  vertex to add to mesh
   * @return index of vertex in assembly list
   */
  int addVertex(Vertex* vertex) {
    assert(!m_assembled);

    // Record separately vertices which already belong to the mesh.
    if (vertex->triangle() != 0) {
      assert(vertex->isBoundaryVertex());
      m_boundaryVertices.insert(vertex);
    }

    m_vertices.push_back(vertex);
    return (m_vertices.size() - 1);
  }

  /**
   * Returns the triangle given by the specified index.  Triangles are
   * indexed from zero in the order in which they are added to the mesh.
   * @param index  triangle index
   * @return triangle given by index
   */
  Triangle* getTriangle(int index) {
    return (m_triangles[index]);
  }

  /**
   * Returns the vertex given by the specified index.  Vertices are
   * indexed from zero in the order in which they are added to the mesh.
   * @param index  vertex index
   * @return vertex given by index
   */
  Vertex* getVertex(int index) {
    return (m_vertices[index]);
  }

  /**
   * Returns an iterator over the triangles which were assembled into the
   * mesh, positioned at the start of the list.
   * @return iterator over added triangles, positioned at start of list
   */
  TriangleIt triangleBegin(void) {
    assert(m_assembled);
    return (m_triangles.begin());
  }

  /**
   * Returns an iterator over the triangles which were assembled into the
   * mesh, positioned at the end of the list.
   * @return iterator over added triangles, positioned at end of list
   */
  TriangleIt triangleEnd(void) {
    assert(m_assembled);
    return (m_triangles.end());
  }

  /**
   * Returns an iterator over the vertices which were assembled into the
   * mesh, positioned at the start of the list.
   * @return iterator over added vertices, positioned at start of list
   */
  VertexIt vertexBegin(void) {
    assert(m_assembled);
    return (m_vertices.begin());
  }

  /**
   * Returns an iterator over the vertices which were assembled into the
   * mesh, positioned at the end of the list.
   * @return iterator over added vertices, positioned at end of list
   */
  VertexIt vertexEnd(void) {
    assert(m_assembled);
    return (m_vertices.end());
  }
};

#endif  /* __ASSEMBLER_H__ */
