#ifndef __SIMFORCES_H
#define __SIMFORCES_H

#include "Quadrature/QuadCoeffs.h"

#include "Util/forall.h"
#include "SimTypes.h"

#include "Util/Real.h"
#include "Util/Matrix3.h"
#include "Util/Tensor3.h"

#include "iostream.h"

#define _IEEE 1

#ifndef __GNUC__
#include "nan.h"
#endif

template <class T>
class SimForces {
public:
  DECLARE_SIM_TYPES(T)

  static void Initialize(Mesh &mesh) {

    // Clear the external force on all the vertices

    forall(typename Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      Gext(v) = Zero3Vec;
    endforall;

  }

protected:
  static Real3Vec n[3], m[2];

public:
  static void BeginAreaIntegral(Mesh &mesh) {
    // Clear the internal (static) force on all the vertices

    forall(typename Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      Gstat(v) = Zero3Vec;
      Gext(v) = Zero3Vec;
    endforall;
  }

  static void BeginThicknessIntegral(const GaussPoint *p, Real dA) {
    n[0] = Zero3Vec;
    n[1] = Zero3Vec;
    n[2] = Zero3Vec;

    m[0] = Zero3Vec;
    m[1] = Zero3Vec;
  }

  static void AddVolumeElement(const GaussPoint *p, typename Surface::e2Iter e2Idx, Real dV) {
    CnCn3Tensor S = p->S(e2Idx);
    Real muR = p->muR(e2Idx);

    Real3Vec S0 = Real3Vec(S(0)(0), S(0)(1), S(0)(2));
    Real3Vec S1 = Real3Vec(S(1)(0), S(1)(1), S(1)(2));
    //Real3Vec S2 = Real3Vec(S(2)(0), S(2)(1), S(2)(2));

    Real3Matrix gkovC = MatrixColumns(p->gC(0, e2Idx), 
				      p->gC(1, e2Idx), 
				      p->gC(2, e2Idx));
    
    n[0] += (gkovC * S0) * muR * e2Idx.SimpsonsWeight();
    n[1] += (gkovC * S1) * muR * e2Idx.SimpsonsWeight();
    //n[2] += (gkovC * S2) * muR * e2Idx.SimpsonsWeight();
    
    m[0] += (gkovC * S0) * muR * e2Idx.SimpsonsWeight() * e2Idx.e2(p);
    m[1] += (gkovC * S1) * muR * e2Idx.SimpsonsWeight() * e2Idx.e2(p);
  }

  static void EndThicknessIntegral(const GaussPoint *p, Real simpsonsHeight) {
    n[0] *= simpsonsHeight;
    n[1] *= simpsonsHeight;
    //n[2] *= simpsonsHeight;
    
    m[0] *= simpsonsHeight;
    m[1] *= simpsonsHeight;
  }

  static void BeginVisitSupportVertices(const GaussPoint *p) {
  }

  static void VisitSupportVertex(const GaussPoint *p, Vertex *v, 
				 QuadCoeffs *w, Real dA_vertex) {
    // Compute the contribution of p to the internal force on v
	 
    Real3Vec total = Zero3Vec;
    
    // Membrane
    total += p->da_dx(0,w) * n[0];
    total += p->da_dx(1,w) * n[1];
    //total += p->da_dx(2,w) * n[2] * p->Lambda(); //WARNING, not computed above
    
    // Bending
    total += p->da2de_dx(0,w) * m[0] * p->Lambda();
    total += p->da2de_dx(1,w) * m[1] * p->Lambda();

    total = total * dA_vertex;
    
    Gstat(v) += total;
  }

  static void EndVisitSupportVertices(const GaussPoint *p) {
  }

  static void EndAreaIntegral(Mesh &mesh) {
  }

  static void Update(Mesh &mesh) {
  }

  static Real3Vec Gstat(const Vertex *v) {
    return v->forcesData.m_Gstat;
  }

  static Real3Vec Gext(const Vertex *v) {
    return v->forcesData.m_Gext;
  }

protected:
  static Real3Vec &Gstat(Vertex *v) {
    return v->forcesData.m_Gstat;
  }

  static Real3Vec &Gext(Vertex *v) {
    return v->forcesData.m_Gext;
  }

private:  

  Real3Vec m_Gstat;
  Real3Vec m_Gext;

};


template<class T>
Real3Vec SimForces<T>::n[3];

template<class T>
Real3Vec SimForces<T>::m[2];


#endif


#if 0
      cerr << "gkovR row 0: " << gkovR.row(0) << endl;
      cerr << "gkovR row 1: " << gkovR.row(1) << endl;
      cerr << "gkovR row 2: " << gkovR.row(2) << endl;

      Real n0x = n[0].x();
      Real n1x = n[1].x();
      Real n2x = n[2].x();
      Real n0y = n[0].y();
      Real n1y = n[1].y();
      Real n2y = n[2].y();
      Real n0z = n[0].z();
      Real n1z = n[1].z();
      Real n2z = n[2].z();
#endif

#if 0
	cerr << w->N() << "\t" << w->dN0() << "\t" << w->dN1() << "\t" << w->dN00() << "\t" << w->dN01() << "\t" << w->dN11() << endl;

	Real3Matrix da_dx0 = p->da_dx(0,w);
	Real3Matrix da_dx1 = p->da_dx(1,w);
	Real3Matrix da_dx2 = p->da_dx(2,w);




	Real da0_dx_00 = da_dx0.row(0).x();
	Real da0_dx_01 = da_dx0.row(0).y();
	Real da0_dx_02 = da_dx0.row(0).z();
	Real da0_dx_10 = da_dx0.row(1).x();
	Real da0_dx_11 = da_dx0.row(1).y();
	Real da0_dx_12 = da_dx0.row(1).z();
	Real da0_dx_20 = da_dx0.row(2).x();
	Real da0_dx_21 = da_dx0.row(2).y();
	Real da0_dx_22 = da_dx0.row(2).z();

	Real da1_dx_00 = da_dx1.row(0).x();
	Real da1_dx_01 = da_dx1.row(0).y();
	Real da1_dx_02 = da_dx1.row(0).z();
	Real da1_dx_10 = da_dx1.row(1).x();
	Real da1_dx_11 = da_dx1.row(1).y();
	Real da1_dx_12 = da_dx1.row(1).z();
	Real da1_dx_20 = da_dx1.row(2).x();
	Real da1_dx_21 = da_dx1.row(2).y();
	Real da1_dx_22 = da_dx1.row(2).z();

	Real da2_dx_00 = da_dx2.row(0).x();
	Real da2_dx_01 = da_dx2.row(0).y();
	Real da2_dx_02 = da_dx2.row(0).z();
	Real da2_dx_10 = da_dx2.row(1).x();
	Real da2_dx_11 = da_dx2.row(1).y();
	Real da2_dx_12 = da_dx2.row(1).z();
	Real da2_dx_20 = da_dx2.row(2).x();
	Real da2_dx_21 = da_dx2.row(2).y();
	Real da2_dx_22 = da_dx2.row(2).z();

	Real3Vec total0 = p->da_dx0(0,w) * n[0];
	Real3Vec total1 = p->da_dx1(1,w) * n[1];
	Real3Vec total2 = p->da_dx2(2,w) * n[2] * p->Lambda();



#endif


#if 0
//	if (total2.l2() > 1e-9) {
	  cerr << "xR: " << v->xR() << endl;
	  cerr << "xC: " << v->xC() << endl << endl;
	  cerr << "total0: " << total0 << endl;
	  cerr << "total1: " << total1 << endl;
	  cerr << "total2: " << total2 << endl;
	  cerr << endl;
 	  cerr << "g0R: " << g0R << endl;
	  cerr << "g1R: " << g1R << endl;
	  cerr << "g2R: " << g2R << endl;
	  cerr << endl;
 	  cerr << "g0C: " << g0C << endl;
	  cerr << "g1C: " << g1C << endl;
	  cerr << "g2C: " << g2C << endl;
	  cerr << endl;
	  cerr << "n0: " << n[0] << endl;
	  cerr << "n1: " << n[1] << endl;
	  cerr << "n2: " << n[2] << endl;
	  cerr << endl;
	  cerr << "da0/dx row 0: " << da_dx0.row(0) << endl;
	  cerr << "da0/dx row 1: " << da_dx0.row(1) << endl;
	  cerr << "da0/dx row 2: " << da_dx0.row(2) << endl;
	  cerr << endl;
	  cerr << "da1/dx row 0: " << da_dx1.row(0) << endl;
	  cerr << "da1/dx row 1: " << da_dx1.row(1) << endl;
	  cerr << "da1/dx row 2: " << da_dx1.row(2) << endl;
	  cerr << endl;
	  cerr << "da2/dx row 0: " << da_dx2.row(0) << endl;
	  cerr << "da2/dx row 1: " << da_dx2.row(1) << endl;
	  cerr << "da2/dx row 2: " << da_dx2.row(2) << endl;
	  cerr << endl;
	  cerr << "S row 0: " << S.row(0) << endl;
	  cerr << "S row 1: " << S.row(1) << endl;
	  cerr << "S row 2: " << S.row(2) << endl;
	  assert(0);
//	}


	if (IsNANorINF(total)) { 
	  cerr << "wow\n";
	}
	//cerr << "... " << total << "\t" << p->JacobianR() << " " << p->W << " ";

#endif

