#ifndef __SIMGAUSSPOINT_H
#define __SIMGAUSSPOINT_H

#include "Quadrature/Diamond.h"
#include "Util/Real.h"
#include "SimTypes.h"

// Performance results for 50 iterations of sphere2 pull test:
//
//                      caching
//                     yes   no
//
// optimization:  no    23  116
//               yes    10   44   
//
// (units are user seconds reported by time program)

#define SimGaussPoint_USECACHE


#ifdef SimGaussPoint_USECACHE

#define CACHE_HIT (cache_this == this)
#define CACHE_HIT_e2Idx (cache_this == this  &&  cache_e2Idx == e2Idx)
#define CACHE_HIT_w (cache_this == this  &&  cache_w == w)

#else

#define CACHE_HIT (0)
#define CACHE_HIT_e2Idx (0)
#define CACHE_HIT_w (0)

#endif

template <class T>
class SimGaussPoint : public Diamond<T> {
public:
  DECLARE_SIM_TYPES(T)

  typedef typename Surface::e2Iter e2Iter;

  Material materialData;
  Surface surfaceData;

  SimGaussPoint(Triangle *t, EdgeName e)
    : Diamond<T>(t,e) {
  }

  Real W() {
    if (Boundary())
      return m_W / 2.0;
    else
      return m_W;
  }

  // Surface Properties
  /////////////////////////////

  Real3Vec xC() const {
    return Surface::xC((const GaussPoint*)this);
  }

  Real3Vec x1C() const {
    return Surface::x1C((const GaussPoint*)this);
  }

  Real3Vec x2C() const {
    return Surface::x2C((const GaussPoint*)this);
  }

  Real3Vec xR() const {
    return Surface::xR((const GaussPoint*)this);
  }

  Real3Matrix da_dx(int i, const QuadCoeffs *w) const {
    assert(i>=0 && i<=2);
    if CACHE_HIT_w
      return cache_da_dx[i];
    else
      return da_dxEval(i, w);
  }

  Real3Matrix da_dxEval(int i, const QuadCoeffs *w) const {
    assert(i>=0 && i<=2);
    return Surface::da_dx((const GaussPoint*)this, i, w);
  }

  Real3Matrix da2de_dx(int i, const QuadCoeffs *w) const {
    assert(i>=0 && i<=1);
    if CACHE_HIT_w
      return cache_da2de_dx[i];
    else
      return da2de_dxEval(i, w);
  }

  Real3Matrix da2de_dxEval(int i, const QuadCoeffs *w) const {
    assert(i>=0 && i<=1);
    return Surface::da2de_dx((const GaussPoint*)this, i, w);
  }

  Real3Vec aC(int i) const {
    assert(i>=0 && i<=2);
    if CACHE_HIT
      return cache_aC[i];
    else
      return aCEval(i);
  }

  Real3Vec aCEval(int i) const {
    assert(i>=0 && i<=2);
    return Surface::aC((const GaussPoint*)this, i);
  }

  Real3Vec aR(int i) const {
    assert(i>=0 && i<=2);
    if CACHE_HIT
      return cache_aR[i];
    else
      return aREval(i);
  }

  Real3Vec aREval(int i) const {
    assert(i>=0 && i<=2);
    return Surface::aR((const GaussPoint*)this, i);
  }

  Real3Vec daCde(int i, int j) const {
    assert(i>=0 && i<=2);
    assert(j>=0 && j<=1);
    if CACHE_HIT
      return cache_daCde[i][j];
    else
      return daCdeEval(i,j);
  }

  Real3Vec daCdeEval(int i, int j) const {
    assert(i>=0 && i<=2);
    assert(j>=0 && j<=1);
    return Surface::daCde((const GaussPoint*)this, i, j);
  }

  Real3Vec daRde(int i, int j) const {
    assert(i>=0 && i<=2);
    assert(j>=0 && j<=1);
    if CACHE_HIT
      return cache_daRde[i][j];
    else
      return daRdeEval(i,j);
  }

  Real3Vec daRdeEval(int i, int j) const {
    assert(i>=0 && i<=2);
    assert(j>=0 && j<=1);
    return Surface::daRde((const GaussPoint*)this, i, j);
  }

  Real3Vec gC(int i, e2Iter e2Idx) const {
    assert(i>=0 && i<=2);
    if CACHE_HIT_e2Idx
      return cache_gC[i];
    else
      return gCEval(i, e2Idx);
  }

  Real3Vec gCEval(int i, e2Iter e2Idx) const {
    assert(i>=0 && i<=2);
    return Surface::gC((const GaussPoint*)this, i, e2Idx);
  }

  Real3Vec gR(int i, e2Iter e2Idx) const {
    assert(i>=0 && i<=2);
    if CACHE_HIT_e2Idx
      return cache_gR[i];
    else
      return gREval(i, e2Idx);
  }

  Real3Vec gREval(int i, e2Iter e2Idx) const {
    assert(i>=0 && i<=2);
    return Surface::gR((const GaussPoint*)this, i, e2Idx);
  }

  Real JacobianC() const {
    if CACHE_HIT
      return cache_JacobianC;
    else
      return JacobianCEval();
  }

  Real JacobianCEval() const {
    return Surface::JacobianC((const GaussPoint*)this);
  }

  Real JacobianR() const {
    if CACHE_HIT
      return cache_JacobianR;
    else
      return JacobianREval();
  }

  Real JacobianREval() const {
    return Surface::JacobianR((const GaussPoint*)this);
  }

  Real hC() const {
    return hR() * Lambda();
  }

  Real hR() const {
    if CACHE_HIT
      return cache_hR;
    else
      return hREval();
  }

  Real LayerHeight() const {
    return Surface::LayerHeight((const GaussPoint*)this);
  }

  Real hREval() const {
    return Surface::hR((const GaussPoint*)this);
  }

  Real muR(e2Iter e2Idx) const {
    if CACHE_HIT_e2Idx
      return cache_muR;
    else
      return muREval(e2Idx);
  }

  Real muREval(e2Iter e2Idx) const {
    return Surface::muR((const GaussPoint*)this, e2Idx);
  }

  // Material Properties
  //////////////////////////////

  Real Lambda() const {
    if CACHE_HIT
      return cache_Lambda;
    else
      return LambdaEval();
  }

  Real LambdaEval() const {
    return Material::Lambda((const GaussPoint*)this);
  }

  CnCn3Tensor S(e2Iter e2Idx) const {
    if CACHE_HIT_e2Idx
      return cache_S;
    else
      return SEval(e2Idx);
  }

  CnCn3Tensor SEval(e2Iter e2Idx) const {
    return Material::S((const GaussPoint*)this, e2Idx);
  }

  CvCv3Tensor Strain(e2Iter e2Idx) const {
    if CACHE_HIT_e2Idx
      return cache_Strain;
    else
      return StrainEval(e2Idx);
  }

  CvCv3Tensor StrainEval(e2Iter e2Idx) const {
    return Material::Strain((const GaussPoint*)this, e2Idx);
  }

  CvCv3Tensor CartesianStrain(e2Iter e2Idx) const {
    return Material::CartesianStrain((const GaussPoint*)this, e2Idx);
  }

  CnCn3Tensor CartesianStress(e2Iter e2Idx) const {
    return Material::CartesianStress((const GaussPoint*)this, e2Idx);
  }

  void FillCache() const {

#ifdef SimGaussPoint_USECACHE

    cache_this  = (const GaussPoint*)this;
    cache_e2Idx = Surface::e2End();
    cache_w     = NULL;

    cache_aC[0] = aCEval(0); 
    cache_aC[1] = aCEval(1); 

    cache_JacobianC = JacobianCEval();

    cache_aC[2] = aCEval(2); 

    cache_aR[0] = aREval(0); 
    cache_aR[1] = aREval(1); 

    cache_JacobianR = JacobianREval();

    cache_aR[2] = aREval(2); 

    cache_daRde[0][0] = daRdeEval(0, 0);
    cache_daRde[0][1] = daRdeEval(0, 1);
    cache_daRde[1][0] = daRdeEval(1, 0);
    cache_daRde[1][1] = daRdeEval(1, 1);
    cache_daRde[2][0] = daRdeEval(2, 0);
    cache_daRde[2][1] = daRdeEval(2, 1);

    cache_daCde[0][0] = daCdeEval(0, 0);
    cache_daCde[0][1] = daCdeEval(0, 1);
    cache_daCde[1][0] = daCdeEval(1, 0);
    cache_daCde[1][1] = daCdeEval(1, 1);
    cache_daCde[2][0] = daCdeEval(2, 0);
    cache_daCde[2][1] = daCdeEval(2, 1);

    cache_hR = hREval();

    cache_Lambda = LambdaEval();

#endif

  }

  void FillCache(e2Iter e2Idx) const {

#ifdef SimGaussPoint_USECACHE

    cache_e2Idx = e2Idx;

    cache_gC[0] = gCEval(0, e2Idx); 
    cache_gC[1] = gCEval(1, e2Idx); 
    cache_gC[2] = gCEval(2, e2Idx); 

    cache_gR[0] = gREval(0, e2Idx); 
    cache_gR[1] = gREval(1, e2Idx); 
    cache_gR[2] = gREval(2, e2Idx); 

    cache_muR = muREval(e2Idx);

    cache_Strain = StrainEval(e2Idx);

    cache_S = SEval(e2Idx);

#endif

  }

  void FillCache(const QuadCoeffs *w) const {    

#ifdef SimGaussPoint_USECACHE

    cache_da_dx[0] = da_dxEval(0, w);
    cache_da_dx[1] = da_dxEval(1, w);
    cache_da_dx[2] = da_dxEval(2, w);

    cache_da2de_dx[0] = da2de_dxEval(0, w);
    cache_da2de_dx[1] = da2de_dxEval(1, w);

    cache_w = w;

#endif

  }

private:

  static const Real m_W;

  // Cache status

  static const SimGaussPoint *cache_this;
  static e2Iter cache_e2Idx;
  static const QuadCoeffs *cache_w;

  // Membrane-dependent properties (e0 and e1)

  // Surface

  static Real3Vec cache_aC[3], cache_aR[3];

  static Real cache_JacobianC, cache_JacobianR;

  static Real cache_hR;

  static Real3Vec cache_daRde[3][2];
  static Real3Vec cache_daCde[3][2];

  // Material

  static Real cache_Lambda;


  // Membrane and offset-dependent properties (e0, e1, e2)

  // Surface

  static Real3Vec cache_gC[3], cache_gR[3];

  static Real cache_muR;

  // Material

  static CnCn3Tensor cache_S;

  static CvCv3Tensor cache_Strain;


  // Membrane and Support-Vertex-dependent properties (e0, e1, w)

  static Real3Matrix cache_da_dx[3];

  static Real3Matrix cache_da2de_dx[2];
};

template <class T>
const Real SimGaussPoint<T>::m_W = sqrt(3.0)/6.0;


// Cache status

template <class T> const SimGaussPoint<T> *SimGaussPoint<T>::cache_this;
template <class T> SimGaussPoint<T>::e2Iter SimGaussPoint<T>::cache_e2Idx;
template <class T> const QuadCoeffs *SimGaussPoint<T>::cache_w;

// Membrane-dependent properties (e0 and e1)

// Surface

template <class T> Real3Vec SimGaussPoint<T>::cache_aC[3];
template <class T> Real3Vec SimGaussPoint<T>::cache_aR[3];

template <class T> Real SimGaussPoint<T>::cache_JacobianC;
template <class T> Real SimGaussPoint<T>::cache_JacobianR;

template <class T> Real SimGaussPoint<T>::cache_hR;

template <class T> Real3Vec SimGaussPoint<T>::cache_daRde[3][2];
template <class T> Real3Vec SimGaussPoint<T>::cache_daCde[3][2];

// Material

template <class T> Real SimGaussPoint<T>::cache_Lambda;


// Membrane and offset-dependent properties (e0, e1, e2)

// Surface

template <class T> Real3Vec SimGaussPoint<T>::cache_gC[3];
template <class T> Real3Vec SimGaussPoint<T>::cache_gR[3];

template <class T> Real SimGaussPoint<T>::cache_muR;

// Material

template <class T> CnCn3Tensor SimGaussPoint<T>::cache_S;

template <class T> CvCv3Tensor SimGaussPoint<T>::cache_Strain;


// Membrane and Support-Vertex-dependent properties (e0, e1, w)

template <class T> Real3Matrix SimGaussPoint<T>::cache_da_dx[3];

template <class T> Real3Matrix SimGaussPoint<T>::cache_da2de_dx[2];

#endif






