#ifndef __SIMGEOMETRY_H
#define __SIMGEOMETRY_H

#include "Util/forall.h"
#include "Util/Real.h"
#include "SimTypes.h"

#include "fstream.h"

template <class T>
class SimGeometry {
public:
  DECLARE_SIM_TYPES(T)

  static void Initialize(Mesh &mesh) {
    forall(typename Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      SetxC (v, xR(v));
      Setx1C(v, Zero3Vec);
      Setx2C(v, Zero3Vec);

      SetxCprev (v, xR(v));
      Setx1Cprev(v, Zero3Vec);
      Setx2Cprev(v, Zero3Vec);
    endforall;
  }

  static void SkewCurrent(Mesh &mesh, unsigned dependentCoord, 
			  unsigned independentCoord, Real factor) {
    assert(dependentCoord < 3 && independentCoord < 3);
    forall(typename Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      Real3Vec x = v->xC();
      x(dependentCoord) += factor * v->point()(independentCoord);
      v->SetxC(x);
      v->SetxCprev(x);
    endforall;
  }

  static void ScaleReference(Mesh &mesh, Real factor) {
    forall(typename Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      v->SetxR( factor * v->xR() );
    endforall;
  }

  static void ScaleReference(Mesh &mesh, Real fX, Real fY, Real fZ) {
    forall(typename Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      Real3Vec x = v->xR();
      x.x() *= fX;
      x.y() *= fY;
      x.z() *= fZ;
      v->SetxR( x );
    endforall;
  }

  static void ScaleCurrent(Mesh &mesh, Real factor) {
    forall(typename Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      SetxCprev(v, xCprev(v)*factor);
      SetxC(    v, xCprev(v));
    endforall;
  }

  static void ScaleCurrent(Mesh &mesh, Real fX, Real fY, Real fZ) {
    forall(typename Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      Real3Vec x = xCprev(v);
      x.x() *= fX;
      x.y() *= fY;
      x.z() *= fZ;
      SetxCprev(v, x);
      SetxC(    v, x);
    endforall;
  }

  static void TransformReference(Mesh &mesh, const Real3Matrix &S) {
    forall(typename Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      v->SetxR( S * v->xR() );
    endforall;
  }

  static void PullCurrent(Mesh &mesh, Real fraction) {
    Vertex *a, *b;
    Real dist = 0.0;
    a = b = NULL;
    forall(typename Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      if (a==NULL) a = v;
      Real d = (xR(v) - xR(a)).l2();
      if (d > dist) {
	dist = d;
	b = v;
      }
    endforall;
    assert(a && b && dist > 0.0);
    SetxCprev(a, xCprev(a) + fraction * (xR(a) - xR(b)));
    SetxCprev(b, xCprev(b) + fraction * (xR(b) - xR(a)));
    SetxC(a, xCprev(a));
    SetxC(b, xCprev(b));
  }

  static void Update(Mesh &mesh) {
  }


  // Current Coordinates

  static Real3Vec DispC(const Vertex *v) {
    return v->geometryData.m_dispC[0];
  }

  static void SetDispC(Vertex *v, Real3Vec dispC) {
    v->geometryData.m_dispC[0] = dispC;
  }


  static Real3Vec xC(const Vertex *v) {
    return xR(v) + v->geometryData.m_dispC[0];
  }

  static void SetxC(Vertex *v, Real3Vec xC) {
    v->geometryData.m_dispC[0] = xC - xR(v);
  }


  static Real3Vec x1C(const Vertex *v) {
    return v->geometryData.m_dispC[1];
  }

  static void Setx1C(Vertex *v, Real3Vec x1C) {
    v->geometryData.m_dispC[1] = x1C;
  }


  static Real3Vec x2C(const Vertex *v) {
    return v->geometryData.m_dispC[2];
  }

  static void Setx2C(Vertex *v, Real3Vec x2C) {
    v->geometryData.m_dispC[2] = x2C;
  }

  
  // Previous Coordinates (at last time step)
  
  static Real3Vec DispCprev(const Vertex *v) {
    return v->geometryData.m_dispCprev[0];
  }

  static void SetDispCprev(Vertex *v, Real3Vec dispCprev) {
    v->geometryData.m_dispCprev[0] = dispCprev;
  }


  static Real3Vec xCprev(const Vertex *v) {
    return xR(v) + v->geometryData.m_dispCprev[0];
  }

  static void SetxCprev(Vertex *v, Real3Vec xCprev) {
    v->geometryData.m_dispCprev[0] = xCprev - xR(v);
  }


  static Real3Vec x1Cprev(const Vertex *v) {
    return v->geometryData.m_dispCprev[1];
  }

  static void Setx1Cprev(Vertex *v, Real3Vec x1Cprev) {
    v->geometryData.m_dispCprev[1] = x1Cprev;
  }


  static Real3Vec x2Cprev(const Vertex *v) {
    return v->geometryData.m_dispCprev[2];
  }

  static void Setx2Cprev(Vertex *v, Real3Vec x2Cprev) {
    v->geometryData.m_dispCprev[2] = x2Cprev;
  }


  // Next Coordinates (proposed new coordinates)

  static Real3Vec DispCnext(const Vertex *v) {
    return v->geometryData.m_dispCnext;
  }

  static void SetDispCnext(Vertex *v, Real3Vec dispCnext) {
    v->geometryData.m_dispCnext = dispCnext;
  }


  static Real3Vec xCnext(const Vertex *v) {
    return xR(v) + v->geometryData.m_dispCnext;
  }

  static void SetxCnext(Vertex *v, Real3Vec xCnext) {
    v->geometryData.m_dispCnext = xCnext - xR(v);
  }



  // Temporary Coordinate Holders

  static Real3Vec DispCtemp(const Vertex *v, int tempIdx) {
    return v->geometryData.m_dispCtemp[tempIdx][0];
  }

  static void SetDispCtemp(Vertex *v, int tempIdx, Real3Vec dispCtemp) {
    v->geometryData.m_dispCtemp[tempIdx][0] = dispCtemp;
  }


  static Real3Vec xCtemp(const Vertex *v, int tempIdx) {
    return xR(v) + v->geometryData.m_dispCtemp[tempIdx][0];
  }

  static void SetxCtemp(Vertex *v, int tempIdx, Real3Vec xCtemp) {
    v->geometryData.m_dispCtemp[tempIdx][0] = xCtemp - xR(v);
  }


  static Real3Vec x1Ctemp(const Vertex *v, int tempIdx) {
    return v->geometryData.m_dispCtemp[tempIdx][1];
  }

  static void Setx1Ctemp(Vertex *v, int tempIdx, Real3Vec x1Ctemp) {
    v->geometryData.m_dispCtemp[tempIdx][1] = x1Ctemp;
  }


  static Real3Vec x2Ctemp(const Vertex *v, int tempIdx) {
    return v->geometryData.m_dispCtemp[tempIdx][2];
  }

  static void Setx2Ctemp(Vertex *v, int tempIdx, Real3Vec x2Ctemp) {
    v->geometryData.m_dispCtemp[tempIdx][2] = x2Ctemp;
  }



  // Reference Coordinates

  static Real3Vec xR(const Vertex *v) {
      return v->point();
  }

  static void SetxR(Vertex *v, Real3Vec xR) {
    v->point() = xR;
  }

  // Acceleration due to contact

  static Real3Vec x2Contact(const Vertex *v) {
    return v->geometryData.m_x2Contact;
  }

  static void Setx2Contact(Vertex *v, Real3Vec x2Contact) {
    v->geometryData.m_x2Contact = x2Contact;
  }


  // Load and save geometry state

  static void SaveState(Vertex *v, ofstream &fstate) {
    Real3Vec data;

    data = xR(v);
    fstate.write((char*)(&data), sizeof(data));

    data = DispC(v);
    fstate.write((char*)(&data), sizeof(data));
    
    data = x1C(v);
    fstate.write((char*)(&data), sizeof(data));
      
    data = x2C(v);
    fstate.write((char*)(&data), sizeof(data));

    data = DispCprev(v);
    fstate.write((char*)(&data), sizeof(data));
    
    data = x1Cprev(v);
    fstate.write((char*)(&data), sizeof(data));
      
    data = x2Cprev(v);
    fstate.write((char*)(&data), sizeof(data));
  }

  static void SaveStateASCII(Vertex *v, ostream &fstate) {
    fstate << xR(v) << endl;
    fstate << DispC(v) << endl;
    fstate << x1C(v) << endl;
    fstate << x2C(v) << endl;
    fstate << DispCprev(v) << endl;
    fstate << x1Cprev(v) << endl;
    fstate << x2Cprev(v) << endl;
    fstate << endl;
  }

  static void LoadState(Vertex *v, ifstream &fstate) {
    Real3Vec data;

    fstate.read((char*)(&data), sizeof(data));
    SetxR(v, data);

    fstate.read((char*)(&data), sizeof(data));
    SetDispC(v, data);

    fstate.read((char*)(&data), sizeof(data));
    Setx1C(v, data);

    fstate.read((char*)(&data), sizeof(data));
    Setx2C(v, data);

    fstate.read((char*)(&data), sizeof(data));
    SetDispCprev(v, data);

    fstate.read((char*)(&data), sizeof(data));
    Setx1Cprev(v, data);

    fstate.read((char*)(&data), sizeof(data));
    Setx2Cprev(v, data);
  }

private:

  Real3Vec m_dispC[3];

  Real3Vec m_dispCtemp[4][3];

  Real3Vec m_dispCprev[3];

  Real3Vec m_dispCnext;

  Real3Vec m_x2Contact;

};

#endif




