#ifndef __SIMSURFACE_H
#define __SIMSURFACE_H

#include "SimTypes.h"

#include "Util/util.h"
#include "Util/forall.h"

#include "Maple/da3_dx.c"
#include "Maple/defgradf1.c"
#include "Maple/ader.c"

template <class T>
class SimSurface {
public:
  DECLARE_SIM_TYPES(T)

  class e2Iter {

  public:

    enum StartState {START, MIDDLE, END, INVALID = -99};

    e2Iter (StartState state = INVALID) {
      switch (state) {
      case START: 
	m_idx = 0;
	break;

      case MIDDLE:
	m_idx = Parameters::thicknessSamples/2;
	assert(Parameters::thicknessSamples % 2 == 1);
	break;

      case END:
	m_idx = Parameters::thicknessSamples;
	break;

      case INVALID:
	m_idx = Parameters::thicknessSamples + 100;
	break;
      }
    }

    Real e2(const GaussPoint *p) {
      assert(m_idx >= 0 && m_idx < Parameters::thicknessSamples);
      return p->surfaceData.m_e2[m_idx];
    }

    e2Iter operator ++() {
      ++m_idx;
      return *this;
    }

    e2Iter operator --() {
      --m_idx;
      return *this;
    }

    int idx() {
      assert(m_idx >= 0 && m_idx < Parameters::thicknessSamples);
      return m_idx;
    }

    bool operator == (e2Iter b) {
      return m_idx == b.m_idx;
    }

    bool operator != (e2Iter b) {
      return m_idx != b.m_idx;
    }

    e2Iter &operator = (e2Iter b) {
      m_idx = b.m_idx;
      return *this;
    }

    Real SimpsonsWeight() {
      assert(m_idx >= 0 && m_idx < Parameters::thicknessSamples);
      return simpsonsWeight[m_idx];
    }

  private:
    
    int m_idx;
  };

  static void Initialize(Mesh &mesh) {
    // Uniform thickness assumption
    // Initialize thickness at every gauss point

    const int sampleCount = Parameters::thicknessSamples;
    assert(sampleCount > 2   &&  (sampleCount % 2) == 1);
    simpsonsWeight = new Real[sampleCount];
    bool two = true;
    for (int i = 0; i < sampleCount; ++i) {
      simpsonsWeight[i] = two ? 2.0 : 4.0;
      two = !two;
    }
    simpsonsWeight[0]             = 1.0;
    simpsonsWeight[sampleCount-1] = 1.0;

    forall(typename Mesh::GaussPointIt, p, mesh.gaussPointBegin(), mesh.gaussPointEnd())
      p->surfaceData.Init( p );
    endforall;
  }

  static void RecomputeReference(Mesh &mesh) {
    forall(typename Mesh::GaussPointIt, p, mesh.gaussPointBegin(), mesh.gaussPointEnd())
      p->surfaceData.ComputeReference( p );
    endforall;
  }

  static void Update(Mesh &mesh) {
  }

  static e2Iter e2Begin() {
    return e2Iter(e2Iter::START);
  }

  static e2Iter e2Middle() {
    return e2Iter(e2Iter::MIDDLE);
  }

  static e2Iter e2End() {
    return e2Iter(e2Iter::END);
  }

  static Real3Vec xC(const GaussPoint *p) {
    return p->EvalN(ptr_fun(&Geometry::xC), Real3Vec());
  }

  static Real3Vec x1C(const GaussPoint *p) {
    return p->EvalN(ptr_fun(&Geometry::x1C), Real3Vec());
  }

  static Real3Vec x2C(const GaussPoint *p) {
    return p->EvalN(ptr_fun(&Geometry::x2C), Real3Vec());
  }

  static Real3Vec xREval(const GaussPoint *p) {
    return p->EvalN(ptr_fun(&Geometry::xR), Real3Vec());
  }

  static Real3Vec xR(const GaussPoint *p) {
    return p->surfaceData.m_xR;
  }

  static Real3Vec aC(const GaussPoint *p, int idx) {
    assert(idx >= 0  &&  idx <= 2  &&  p);
    switch (idx) { 
    case 0: return p->Evald0(ptr_fun(&Geometry::xC), Real3Vec());
    case 1: return p->Evald1(ptr_fun(&Geometry::xC), Real3Vec());
    case 2: return p->aC(0). cross (p->aC(1)) / p->JacobianC();
    }
    die();
    return Zero3Vec;
  }

  static Real3Vec aREval(const GaussPoint *p, int idx) {
    assert(idx >= 0  &&  idx <= 2  &&  p);
    switch (idx) { 
    case 0: return p->Evald0(ptr_fun(&Geometry::xR), Real3Vec());
    case 1: return p->Evald1(ptr_fun(&Geometry::xR), Real3Vec());
    case 2: return p->aR(0) . cross(p->aR(1)) / p->JacobianR();
    }
    die();
    return Zero3Vec;
  }

  static Real3Vec aR(const GaussPoint *p, int idx) {
    assert(idx >= 0  &&  idx <= 2  &&  p);
    return p->surfaceData.m_aR[idx];
  }

  // Compute da0R w.r.t. theta[j]
  static Real3Vec a0Rder(const GaussPoint *p, int i) {
    assert(i >= 0  &&  i <= 1  &&  p);
    switch (i) { 
    case 0: return p->Evald00(ptr_fun(&Geometry::xR), Real3Vec());
    case 1: return p->Evald01(ptr_fun(&Geometry::xR), Real3Vec());
    }
    die();
    return Zero3Vec;
  }

  // Compute da1R w.r.t. theta[j]
  static Real3Vec a1Rder(const GaussPoint *p, int i) {
    assert(i >= 0  &&  i <= 1  &&  p);
    switch (i) { 
    case 0: return p->Evald01(ptr_fun(&Geometry::xR), Real3Vec());
    case 1: return p->Evald11(ptr_fun(&Geometry::xR), Real3Vec());
    }
    die();
    return Zero3Vec;
  }

  // Compute da2R w.r.t. theta[j]
  static Real3Vec a2Rder(const GaussPoint *p, int i) {
    assert(i >= 0  &&  i <= 1  &&  p);
    Real3Vec a0  = p->aR(0);
    Real3Vec a1  = p->aR(1);
    Real3Vec a2  = p->aR(2);
    Real3Vec a00 = p->daRde(0,0);
    Real3Vec a01 = p->daRde(0,1);
    Real3Vec a11 = p->daRde(1,1);
    Real a3der[3];
    switch (i) { 
    case 0: {
      sa3der0(a3der, 
	      a0.x(), a0.y(), a0.z(),
	      a1.x(), a1.y(), a1.z(),
	      a2.x(), a2.y(), a2.z(),
	      a00.x(), a00.y(), a00.z(),
	      a11.x(), a11.y(), a11.z(),
	      a01.x(), a01.y(), a01.z(),
	      a01.x(), a01.y(), a01.z());
      return Real3Vec(a3der);
    }
    case 1: 
      sa3der1(a3der, 
	      a0.x(), a0.y(), a0.z(),
	      a1.x(), a1.y(), a1.z(),
	      a2.x(), a2.y(), a2.z(),
	      a00.x(), a00.y(), a00.z(),
	      a11.x(), a11.y(), a11.z(),
	      a01.x(), a01.y(), a01.z(),
	      a01.x(), a01.y(), a01.z());
      return Real3Vec(a3der);
    }
    die();
    return Zero3Vec;
  }

  // Compute daR[i] w.r.t. theta[j]
  static Real3Vec daRdeEval(const GaussPoint *p, int i, int j) {
    assert(i >= 0  &&  i <= 2  &&  p);
    assert(j >= 0  &&  j <= 1  &&  p);
    switch (i) { 
    case 0: return a0Rder(p, j);
    case 1: return a1Rder(p, j);
    case 2: return a2Rder(p, j);
    }
    die();
    return Zero3Vec;
  }

  // Compute daR[i] w.r.t. theta[j]
  static Real3Vec daRde(const GaussPoint *p, int i, int j) {
    assert(i >= 0  &&  i <= 2  &&  p);
    assert(j >= 0  &&  j <= 1  &&  p);
    return p->surfaceData.m_daRde[i][j];
  }

  // Compute da0C w.r.t. theta[j]
  static Real3Vec a0Cder(const GaussPoint *p, int i) {
    assert(i >= 0  &&  i <= 1  &&  p);
    switch (i) { 
    case 0: return p->Evald00(ptr_fun(&Geometry::xC), Real3Vec());
    case 1: return p->Evald01(ptr_fun(&Geometry::xC), Real3Vec());
    }
    die();
    return Zero3Vec;
  }

  // Compute da1C w.r.t. theta[j]
  static Real3Vec a1Cder(const GaussPoint *p, int i) {
    assert(i >= 0  &&  i <= 1  &&  p);
    switch (i) { 
    case 0: return p->Evald01(ptr_fun(&Geometry::xC), Real3Vec());
    case 1: return p->Evald11(ptr_fun(&Geometry::xC), Real3Vec());
    }
    die();
    return Zero3Vec;
  }

  // Compute da2C w.r.t. theta[j]
  static Real3Vec a2Cder(const GaussPoint *p, int i) {
    assert(i >= 0  &&  i <= 1  &&  p);
    Real3Vec a0  = p->aC(0);
    Real3Vec a1  = p->aC(1);
    Real3Vec a2  = p->aC(2);
    Real3Vec a00 = p->daCde(0,0);
    Real3Vec a01 = p->daCde(0,1);
    Real3Vec a11 = p->daCde(1,1);
    Real a3der[3];
    switch (i) { 
    case 0: {
      sa3der0(a3der, 
	      a0.x(), a0.y(), a0.z(),
	      a1.x(), a1.y(), a1.z(),
	      a2.x(), a2.y(), a2.z(),
	      a00.x(), a00.y(), a00.z(),
	      a11.x(), a11.y(), a11.z(),
	      a01.x(), a01.y(), a01.z(),
	      a01.x(), a01.y(), a01.z());
      return Real3Vec(a3der);
    }
    case 1: 
      sa3der1(a3der, 
	      a0.x(), a0.y(), a0.z(),
	      a1.x(), a1.y(), a1.z(),
	      a2.x(), a2.y(), a2.z(),
	      a00.x(), a00.y(), a00.z(),
	      a11.x(), a11.y(), a11.z(),
	      a01.x(), a01.y(), a01.z(),
	      a01.x(), a01.y(), a01.z());
      return Real3Vec(a3der);
    }
    die();
    return Zero3Vec;
  }

  // Compute daC[i] w.r.t. theta[j]
  static Real3Vec daCde(const GaussPoint *p, int i, int j) {
    assert(i >= 0  &&  i <= 2  &&  p);
    switch (i) { 
    case 0: return a0Cder(p, j);
    case 1: return a1Cder(p, j);
    case 2: return a2Cder(p, j);
    }
    die();
    return Zero3Vec;
  }

  // da[idx] w.r.t. x
  static Real3Matrix da_dx(const GaussPoint *p, int idx, const QuadCoeffs *w) {
    assert(idx >= 0  &&  idx <= 2  &&  p);
    switch (idx) { 
    case 0: {
      return MatrixDiagonal(w->dN0(), w->dN0(), w->dN0());
    }
    case 1: {
      return MatrixDiagonal(w->dN1(), w->dN1(), w->dN1());
    }
    case 2: {
      Real akovc[3][3], da3_x[3], da3_y[3], da3_z[3];

      akovc[0][0] = p->aC(0).x();
      akovc[0][1] = p->aC(0).y();
      akovc[0][2] = p->aC(0).z();
      akovc[1][0] = p->aC(1).x();
      akovc[1][1] = p->aC(1).y();
      akovc[1][2] = p->aC(1).z();
      
      defgradf0secx (da3_x, akovc, w->dN0(), w->dN1());
      defgradf0secy (da3_y, akovc, w->dN0(), w->dN1());
      defgradf0secz (da3_z, akovc, w->dN0(), w->dN1());

      return MatrixRows(Real3Vec(da3_x), 
			Real3Vec(da3_y), 
			Real3Vec(da3_z));
      }
    }
    die();
    return Zero3Matrix;
  }

  // (da2 w.r.t. theta[i]) w.r.t. x
  static Real3Matrix da2de_dx(const GaussPoint *p, int i, const QuadCoeffs *w) {
    assert(i >= 0  &&  i <= 1  &&  p  && w);

    Real3Vec a0  = p->aR(0);
    Real3Vec a1  = p->aR(1);
    Real3Vec a00 = p->daRde(0,0);
    Real3Vec a01 = p->daRde(0,1);
    Real3Vec a11 = p->daRde(1,1);

    switch (i) { 
    case 0: {
      Real da20derx[3];
      Real da20dery[3];
      Real da20derz[3];
      defgrad1fx(da20derx, 
		 a0.x(), a0.y(), a0.z(),
		 a1.x(), a1.y(), a1.z(),
		 a00.x(), a00.y(), a00.z(),
		 a11.x(), a11.y(), a11.z(),
		 a01.x(), a01.y(), a01.z(),
		 a01.x(), a01.y(), a01.z(),
		 w->dN0(), w->dN1(), w->dN00(), 
		 w->dN01(), w->dN11());
      defgrad1fy(da20dery, 
		 a0.x(), a0.y(), a0.z(),
		 a1.x(), a1.y(), a1.z(),
		 a00.x(), a00.y(), a00.z(),
		 a11.x(), a11.y(), a11.z(),
		 a01.x(), a01.y(), a01.z(),
		 a01.x(), a01.y(), a01.z(),
		 w->dN0(), w->dN1(), w->dN00(), 
		 w->dN01(), w->dN11());
      defgrad1fz(da20derz, 
		 a0.x(), a0.y(), a0.z(),
		 a1.x(), a1.y(), a1.z(),
		 a00.x(), a00.y(), a00.z(),
		 a11.x(), a11.y(), a11.z(),
		 a01.x(), a01.y(), a01.z(),
		 a01.x(), a01.y(), a01.z(),
		 w->dN0(), w->dN1(), w->dN00(), 
		 w->dN01(), w->dN11());
      return MatrixRows(Real3Vec(da20derx), 
			Real3Vec(da20dery), 
			Real3Vec(da20derz));
    }
    case 1: {
      Real da21derx[3];
      Real da21dery[3];
      Real da21derz[3];
      defgrad2fx(da21derx, 
		 a0.x(), a0.y(), a0.z(),
		 a1.x(), a1.y(), a1.z(),
		 a00.x(), a00.y(), a00.z(),
		 a11.x(), a11.y(), a11.z(),
		 a01.x(), a01.y(), a01.z(),
		 a01.x(), a01.y(), a01.z(),
		 w->dN0(), w->dN1(), w->dN00(), 
		 w->dN01(), w->dN11());
      defgrad2fy(da21dery, 
		 a0.x(), a0.y(), a0.z(),
		 a1.x(), a1.y(), a1.z(),
		 a00.x(), a00.y(), a00.z(),
		 a11.x(), a11.y(), a11.z(),
		 a01.x(), a01.y(), a01.z(),
		 a01.x(), a01.y(), a01.z(),
		 w->dN0(), w->dN1(), w->dN00(), 
		 w->dN01(), w->dN11());
      defgrad2fz(da21derz, 
		 a0.x(), a0.y(), a0.z(),
		 a1.x(), a1.y(), a1.z(),
		 a00.x(), a00.y(), a00.z(),
		 a11.x(), a11.y(), a11.z(),
		 a01.x(), a01.y(), a01.z(),
		 a01.x(), a01.y(), a01.z(),
		 w->dN0(), w->dN1(), w->dN00(), 
		 w->dN01(), w->dN11());
      return MatrixRows(Real3Vec(da21derx), 
			Real3Vec(da21dery), 
			Real3Vec(da21derz));
    }
    }
    die();
    return Zero3Matrix;
  }  

  static Real JacobianC(const GaussPoint *p) {
    assert(p);
    return (p->aC(0). cross (p->aC(1))). l2();
  }

  static Real JacobianREval(const GaussPoint *p) {
    assert(p);
    return (p->aR(0). cross (p->aR(1))). l2();
  }

  static Real JacobianR(const GaussPoint *p) {
    assert(p);
    return p->surfaceData.m_JacobianR;
  }

     
      //  gkov[i][j] = akov[i][j]+thet3*stretch*a3d[i][j];

  static Real3Vec gC(const GaussPoint *p, int idx, e2Iter e2Idx) {
    assert(idx >= 0  &&  idx <= 2  &&  p);
    switch (idx) {
    case 0: return p->aC(0) + e2Idx.e2(p) * p->Lambda() * daCde(p,2,0);
    case 1: return p->aC(1) + e2Idx.e2(p) * p->Lambda() * daCde(p,2,1);
    case 2: return p->aC(2);
    }
    die();
    return Zero3Vec;
  }
  
  static Real3Vec gREval(const GaussPoint *p, int idx, e2Iter e2Idx) {
    assert(idx >= 0  &&  idx <= 2  &&  p);
    switch (idx) { 
    case 0: return p->aR(0) + e2Idx.e2(p) * daRde(p,2,0);
    case 1: return p->aR(1) + e2Idx.e2(p) * daRde(p,2,1);
    case 2: return p->aR(2);
    }
    die();
    return Zero3Vec;
  }

  static Real3Vec gR(const GaussPoint *p, int idx, e2Iter e2Idx) {
    assert(idx >= 0  &&  idx <= 2  &&  p);
    return p->surfaceData.m_gR[idx][e2Idx.idx()];
  }

  static Real hR(const GaussPoint *p) {
    return p->surfaceData.m_hR;
  }

  static Real LayerHeight(const GaussPoint *p) {
    return p->surfaceData.m_layerHeight;
  }

  static Real muREval(const GaussPoint *p, e2Iter e2Idx) {
    return ((p->gR(0,e2Idx). cross (p->gR(1,e2Idx))). dot (p->gR(2,e2Idx)))
         / ((p->aR(0).    cross (p->aR(1))).    dot (p->aR(2)));
  }

  static Real muR(const GaussPoint *p, e2Iter e2Idx) {
    return p->surfaceData.m_muR[e2Idx.idx()];
  }

protected:
  void Init(GaussPoint *p) {
    int sampleCount = Parameters::thicknessSamples;

    m_e2    = new Real    [sampleCount];
    m_muR   = new Real    [sampleCount];
    m_gR[0] = new Real3Vec[sampleCount];
    m_gR[1] = new Real3Vec[sampleCount];
    m_gR[2] = new Real3Vec[sampleCount];

    assert(m_e2 && m_muR && m_gR[0] && m_gR[1] && m_gR[2]);

    p->surfaceData.ComputeReference( p );
  }

  void ComputeReference(GaussPoint *p) {
    int sampleCount = Parameters::thicknessSamples;

    assert(m_e2 && m_muR && m_gR[0] && m_gR[1] && m_gR[2]);

    m_xR = xREval(p);

    m_aR[0] = aREval(p, 0);
    m_aR[1] = aREval(p, 1);

    m_JacobianR = JacobianREval(p);

    m_aR[2] = aREval(p, 2);

    m_daRde[0][0] = daRdeEval(p, 0, 0);
    m_daRde[0][1] = daRdeEval(p, 0, 1);
    m_daRde[1][0] = daRdeEval(p, 1, 0);
    m_daRde[1][1] = daRdeEval(p, 1, 1);
    m_daRde[2][0] = daRdeEval(p, 2, 0);
    m_daRde[2][1] = daRdeEval(p, 2, 1);

    m_hR = Parameters::h;

    m_layerHeight = m_hR / ((Real) sampleCount - 1.0);

    for (e2Iter e2Idx = e2Begin();
	 e2Idx != e2End();
	 ++e2Idx) {

      int i = e2Idx.idx();

      Real e2 = i * m_layerHeight  -  m_hR / 2.0;

      m_e2[i] = e2;
      
      m_gR[0][i] = gREval(p, 0, e2Idx);
      m_gR[1][i] = gREval(p, 1, e2Idx);
      m_gR[2][i] = gREval(p, 2, e2Idx);

      m_muR[i] = muREval(p, e2Idx);
    }
  }


protected:

  // Reference properties that do not depend on e2;

  Real3Vec m_xR;
  Real3Vec m_aR[3];				    
  Real m_JacobianR;
  Real3Vec m_daRde[3][2];
  Real m_hR;

  // Reference properties used for thickness integration;

  Real m_layerHeight;
  Real *m_e2;
  Real *m_muR;
  Real3Vec *m_gR[3];

  // Static data members

  static Real *simpsonsWeight;
};

template <class T>
Real *SimSurface<T>::simpsonsWeight;

#endif


