#ifndef __MATRIX3_H
#define __MATRIX3_H

#include "Vector3.h"
#include <iostream.h>

#include "Util/Real.h"
#include "Util/Invert.h"

template <class Scalar>
class Matrix3 {
public:
  typedef Vector3<Scalar> Row;
  typedef Vector3<Scalar> Column;

  static const Matrix3 Zero;

  Matrix3(void) {
    m_r[0] = m_r[1] = m_r[2] = Row::Zero;
  }

  Matrix3(const Matrix3 &t) {
    m_r[0] = t.m_r[0];
    m_r[1] = t.m_r[1];
    m_r[2] = t.m_r[2];
  }

  Matrix3(const Scalar m[3][3]) {
    int i, j;
    for (i = 0; i < 3; ++i)
      for (j = 0; j < 3; ++j)
	row(i)(j) = m[i][j];
  }
  
  Row &operator()(int i) {
    return row(i);
  }

  Row operator()(int i) const {
    return row(i);
  }

  Matrix3 operator *(const Matrix3 B) const { 
    Matrix3 M;
    for (int i=0; i<3; ++i)
      for (int j=0; j<3; ++j)
	M(i)(j) = row(i). dot (B.col(j));
    return M;
  }

  Column operator *(const Column &v) const {
    return Column(row(0). dot (v),
		  row(1). dot (v),
		  row(2). dot (v));
  }

  Row &row(int idx) {
    assert(idx >= 0   &&  idx <= 2);
    return m_r[idx];
  }

  Row row(int idx) const {
    assert(idx >= 0   &&  idx <= 2);
    return m_r[idx];
  }

  Column col(int idx) const {
    assert(idx >= 0   &&  idx <= 2);
    return Column(row(0)(idx), row(1)(idx), row(2)(idx));
  }

  Matrix3 transpose() const {
    Matrix3 M;
    M.row(0) = Row(row(0).x(), row(1).x(), row(2).x());
    M.row(1) = Row(row(0).y(), row(1).y(), row(2).y());
    M.row(2) = Row(row(0).z(), row(1).z(), row(2).z());
    return M;
  }

  Scalar Trace() const {
    return row(0)(0) + row(1)(1) + row(2)(2);
  }

  Matrix3 inverse() const {
    Scalar m[3][3];
    int i, j;

    for (i = 0; i < 3; ++i)
      for (j = 0; j < 3; ++j)
	m[i][j] = row(i)(j);

    MapleInvert3by3(m);

    return Matrix3(m);
  }

private:
  Row m_r[3];
};

template <class Scalar>
const Matrix3<Scalar> Matrix3<Scalar>::Zero = Matrix3<Scalar>();

template <class Scalar>
inline
Matrix3<Scalar>
MatrixDiagonal(const Scalar d0,
	       const Scalar d1,
	       const Scalar d2) {
  Matrix3<Scalar> M;
  M(0)(0) = d0;
  M(1)(1) = d1;
  M(2)(2) = d2;
  return M;
}

template <class Scalar>
inline
Matrix3<Scalar>
MatrixRows(const Vector3<Scalar> row0,
	   const Vector3<Scalar> row1,
	   const Vector3<Scalar> row2) {
  Matrix3<Scalar> M;
  M(0) = row0;
  M(1) = row1;
  M(2) = row2;
  return M;
}

template <class Scalar>
inline
Matrix3<Scalar>
MatrixColumns(const Vector3<Scalar> col0,
	      const Vector3<Scalar> col1,
	      const Vector3<Scalar> col2) {
  Matrix3<Scalar> M;
  M(0) = col0;
  M(1) = col1;
  M(2) = col2;
  return M.transpose();
}

#ifdef linux
#define ostream_withassign _IO_ostream_withassign
#endif

template <class Scalar>
ostream_withassign &operator <<(ostream_withassign &o, Matrix3<Scalar> T) {
  int i, j;
  for (i=0; i<3; ++i) {
    for (j=0; j<3; ++j)
      o << T(i)(j) << " ";
    o << endl;
  }
  return o;
}

typedef Matrix3<Real> Real3Matrix;

extern const Real3Matrix Zero3Matrix;


#endif
