#include "Wconfig.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wloads.h"
#include "Wdispbc.h"
#include "Wivcond.h"
#include "Wmats.h"
#include "Wcontact.h"
#include "Wvset.h"
#include "Wpvm.h"
#include "Wmpi.h"
#include "Woptions.h"
#include "Wcdi.h"
#include "Wnewmarki.h"
#include "Wpod.h"


static void
update_ext_load_work(W_domain_t *d);
void
assemble_mass (W_domain_t *d);

#define VL(E_nd, W_hat) VECT_LOOP(1, d->E_nd, W_hat)

#define theELEM an_elem
#define REG_ELEM_LOOP(dom, to_do)                                           \
   if (dom->elems_reg_for_integrate != NULL) {                              \
     W_elem_t *an_elem;                                                     \
                                                                            \
     loop_through_list(dom->elems_reg_for_integrate, an_elem, W_elem_t *) { \
       to_do;                                                               \
     }                                                                      \
   }


#define DICT_LOOP_START(dict, vartype, var)                                     \
  {                                                                             \
    vartype *var = NULL;                                                        \
    dict_iterator d_iter_ = dict_get_iterator ();                               \
    dict_reset_iterator (d_iter_);                                              \
    while ((var = (vartype *) dict_iter_objs (dict, d_iter_)) != NULL) {
#define DICT_LOOP_END                           \
    }                                           \
    dict_del_iterator (d_iter_);                \
  }

#define NODE_LOOP_START(domain, var) DICT_LOOP_START(domain->nodes, W_node_t, var)
#define NODE_LOOP_END                DICT_LOOP_END

#define ELEM_LOOP_START(domain, var) DICT_LOOP_START(domain->elements, W_elem_t, var)
#define ELEM_LOOP_END                DICT_LOOP_END

#define LOAD_LOOP_START(domain, var) DICT_LOOP_START(domain->loads, W_load_t, var)
#define LOAD_LOOP_END                DICT_LOOP_END

#define MATERIAL_LOOP_START(domain, var) DICT_LOOP_START(domain->materials, W_material_t, var)
#define MATERIAL_LOOP_END                DICT_LOOP_END

#define DISP_BC_LOOP_START(domain, var) DICT_LOOP_START(domain->disp_bcs, W_disp_bc_t, var)
#define DISP_BC_LOOP_END                DICT_LOOP_END

#define CONTACT_LOOP_START(domain, var) DICT_LOOP_START(domain->cont_surfs, W_contact_surf_t, var)
#define CONTACT_LOOP_END                DICT_LOOP_END

#define IV_COND_LOOP_START(domain, var) DICT_LOOP_START(domain->iv_conds, struct W_initial_velocity_cond_t, var)
#define IV_COND_LOOP_END                DICT_LOOP_END



/*
   Get a node by index.
 */
W_node_t * /* node */
WD_get_node_by_idx (W_domain_t *d, /* domain */
                     W_index_t idx /* index */
                     ) 
{
  return dict_find_obj_with_int_name (d->nodes, idx);
}

 
/*
   Add a node to the domain.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_node (W_domain_t *d, /* add a node to the domain */
             W_index_t  idx, /* index */
             W_node_t   *node /* node */
             )
{
  W_node_t *n = dict_find_obj_with_int_name (d->nodes, idx);
  if (n) {
    sprintf (W_buf (), "Redefining node %d", (int) idx);
    REPORT_FATAL_ERROR (d->errh, W_buf ());
    return FALSE;
  } else {
    dict_add_obj_int_name (d->nodes, node, idx);
    return TRUE;
  }
} 

/*
   Iterate nodes.
   Supply curr_node = NULL, if starting iteration.
 */
W_node_t * /* node or NULL if there are no more */
WD_get_next_node (W_domain_t *d, /* domain */
                  W_node_t *curr /* node or NULL if starting */
                  )
{
  static dict_iterator diter = NULL;
  if (!diter) diter = dict_get_iterator ();
  if (!curr) {
    dict_reset_iterator (diter);
  }
  return dict_iter_objs (d->nodes, diter);
}

/*
   Iterate elements.
   Supply curr_elem = NULL, if starting iteration.
 */
W_elem_t * /* element or NULL if there are no more */
WD_get_next_element (W_domain_t *d, /* domain */
                    W_elem_t *curr /* element or NULL if starting */
                    )
{
  static dict_iterator diter = NULL;
  if (!diter) diter = dict_get_iterator ();
  if (!curr) {
    dict_reset_iterator (diter);
  }
  return dict_iter_objs (d->elements, diter);
}


/*
   Add a new element to the domain.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_element (W_domain_t     *d, /* domain */
                W_index_t      idx, /* index */
                W_elem_t       *elem /* element */
                )
{
  W_elem_t *e = dict_find_obj_with_int_name (d->elements, idx);
  if (e) {
    sprintf (W_buf (), "Redefining element %d", (int) idx);
    REPORT_FATAL_ERROR (d->errh, W_buf ());
    return FALSE;
  } else {
    dict_add_obj_int_name (d->elements, elem, idx);
    return TRUE;
  }
}

/*
   Get an element by index.
 */
W_elem_t * /* element or NULL if there is no such element index */
WD_get_elem_by_idx (W_domain_t *d, /* domain */
                    W_index_t idx /* index */
                    ) 
{
  return dict_find_obj_with_int_name (d->elements, idx);
}

  
/*
   Get a material by index.
 */
W_material_t * /* material or NULL if there is no such material index */
WD_get_material_by_idx (W_domain_t *d, /* domain */
                        W_index_t idx /* index */
                        )
{
  return dict_find_obj_with_int_name (d->materials, idx);
}


/*
   Add material to the domain.
 */
W_bool_t/* TRUE if successful; FALSE otherwise */
WD_add_material (W_domain_t     *d, /* domain */
                W_index_t      idx, /* index */
                W_material_t   *material /* material */
                )
{
  W_material_t *n = dict_find_obj_with_int_name (d->materials, idx);
  if (n) {
    REPORT_ERROR(d->errh, "Redefining material");
    return FALSE;
  } else {
    dict_add_obj_int_name (d->materials, material, idx);
    return TRUE;
  }
}
 

/*
   Add a time function.
 */
W_bool_t/* TRUE if successful; FALSE otherwise */
WD_add_time_func (W_domain_t      *d, /* domain */
                  W_index_t       idx, /* index */
                  W_time_func_t   *time_func /* time function */
                  )
{
  W_time_func_t *tf = dict_find_obj_with_int_name (d->time_funcs, idx);
  if (tf) {
    REPORT_ERROR(d->errh, "Redefining time function");
    return FALSE;
  } else {
    dict_add_obj_int_name (d->time_funcs, time_func, idx);
    return TRUE;
  }
}
 

/*
   Get time function by index.
 */
W_time_func_t * /* time function or NULL if there is no such index */
WD_get_time_func_by_idx (W_domain_t *d, /* domain */
                          W_index_t idx /* index */
                          ) 
{
  return dict_find_obj_with_int_name (d->time_funcs, idx);
}

 
/*
   Add external load.
 */
W_bool_t/* TRUE if successful; FALSE otherwise */
WD_add_load (W_domain_t *d, /* domain */
             W_load_t   *load /* load */
             )
{
  static int idx = 1;
  dict_add_obj_int_name (d->loads, load, idx++);
  return TRUE;
}

 
/*
   Add a displacement BC to the domain.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_disp_bc (W_domain_t *d, /* add a node to the domain */
                W_index_t  idx, /* index */
                W_disp_bc_t   *disp_bc /* node */
                )
{
  W_disp_bc_t *bc = dict_find_obj_with_int_name (d->disp_bcs, idx);
  if (bc) {
    REPORT_ERROR(d->errh, "Redefining displacement BC");
    return FALSE;
  } else {
    dict_add_obj_int_name (d->disp_bcs, disp_bc, idx);
    return TRUE;
  }
} 

/*
   Iterate displacement boundary conditions.
   Supply curr_node = NULL, if starting iteration.
 */
W_disp_bc_t * /* node or NULL if there are no more */
WD_get_next_disp_bc (W_domain_t *d, /* domain */
                     W_disp_bc_t *curr /* node or NULL if starting */
                     )
{
  static dict_iterator diter = NULL;
  if (!diter) diter = dict_get_iterator ();
  if (!curr) {
    dict_reset_iterator (diter);
  }
  return dict_iter_objs (d->disp_bcs, diter);
}
 

 
/*
   Add a displacement BC to the domain.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_iv_cond (W_domain_t *d, /* add a node to the domain */
                W_index_t  idx, /* index */
                W_initial_velocity_cond_p iv_cond /* node */
                )
{
  W_initial_velocity_cond_p ivc = dict_find_obj_with_int_name (d->iv_conds, idx);
  if (ivc) {
    REPORT_ERROR(d->errh, "Redefining");
    return FALSE;
  } else {
    dict_add_obj_int_name (d->iv_conds, iv_cond, idx);
    return TRUE;
  }
} 

/*
   Iterate displacement boundary conditions.
   Supply curr_node = NULL, if starting iteration.
 */
W_disp_bc_t * /* node or NULL if there are no more */
WD_get_next_iv_cond (W_domain_t *d, /* domain */
                     W_initial_velocity_cond_p curr /* node or NULL if starting */
                     )
{
  static dict_iterator diter = NULL;
  if (!diter) diter = dict_get_iterator ();
  if (!curr) {
    dict_reset_iterator (diter);
  }
  return dict_iter_objs (d->iv_conds, diter);
}
 


/*
   Add a contact surface to the domain.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_add_cont_surf (W_domain_t *d, /* domain */
                  W_index_t  idx, /* index */
                  W_contact_surf_t   *conts /* contact surface */
                  )
{
  W_contact_surf_t *cs = dict_find_obj_with_int_name (d->cont_surfs, idx);
  if (cs) {
    REPORT_ERROR(d->errh, "Redefining contact surface");
    return FALSE;
  } else {
    dict_add_obj_int_name (d->cont_surfs, conts, idx);
    return TRUE;
  }
} 

/*
   Iterate displacement boundary conditions.
   Supply curr = NULL, if starting iteration.
 */
W_contact_surf_t * /* node or NULL if there are no more */
WD_get_next_cont_surf (W_domain_t *d, /* domain */
                     W_contact_surf_t  *curr /*  NULL if starting */
                     )
{
  static dict_iterator diter = NULL;
  if (!diter) diter = dict_get_iterator ();
  if (!curr) {
    dict_reset_iterator (diter);
  }
  return dict_iter_objs (d->cont_surfs, diter);
}
 
/*
   Get contact surface by index.
 */
W_contact_surf_t * /* time function or NULL if there is no such index */
WD_get_cont_surf_by_idx (W_domain_t *d, /* domain */
                          W_index_t idx /* index */
                          ) 
{
  return dict_find_obj_with_int_name (d->cont_surfs, idx);
}


static struct {
  char          reader_tag;
  reader_proc   reader;
} readers[] = {
  { 'E',  W_element_reader         },
  { 'N',  W_node_reader            }, 
  { 'L',  W_load_reader            },
  { 'M',  W_material_reader        },
  { 'T',  W_time_func_reader       },
  { 'D',  W_disp_bc_reader         },
  { 'C',  W_contact_reader         },
  { 'O',  W_options_reader         },
#if defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER)
  { 'P',  W_pvm_run_reader         },
#endif
#if defined (MPI_WASP) && MPI_WASP
  { 'P',  W_mpi_run_reader         },
#endif
  { 'H',  W_history_clients_reader },
  { 'S',  W_vset_table_reader      },
  { 'I',  W_init_cond_reader       },
  { 'R',  W_pod_reader             },
  { '\0', W_read_and_discard_block }
};

static reader_proc current_reader = NULL;
static int current_reader_tag_num = 0;

static reader_proc 
get_reader_proc (const char *input_line);


static W_bool_t
read_problem_data (W_domain_t *d)
{
  char buffer[MAXFILENAMECHAR];

  while ((current_reader = get_reader_proc (W_get_input_line (d))) != NULL){
    if (!(*current_reader)(d)) {
      sprintf(buffer, "Failed in data reader: reader tag = %c\n",
              readers[current_reader_tag_num].reader_tag);
      REPORT_FATAL_ERROR(d->errh, buffer);
      return FALSE;
    }
  }
  return TRUE;
}

static int
get_tag_num (const char *input_line)
{
  char first_non_blank = '\0';
  int i, n;
  for (i = 0, n = strlen (input_line); i < n; i++) {
    if (!isspace (input_line[i])) {
      first_non_blank = input_line[i];
      break;
    }
  }
  i = 0;
  while (readers[i].reader_tag != '\0') {
    if (readers[i].reader_tag == first_non_blank) {
      current_reader_tag_num = i;
      return i;
    }
    i++;
  }
  /* at this point we're at the last entry in the table of readers;
     that means no reader has been defined for this block, warn the user
     and skip it (by returning index of W_read_and_discard_block())
   */
  sprintf (W_buf (), "Skipping unrecognized block \"%s\"", input_line);
  REPORT_WARNING (NULL, W_buf ());
  return i;
}



static reader_proc 
get_reader_proc (const char *input_line)
{
  if (input_line == NULL)
    return NULL;
  else 
    return readers[get_tag_num(input_line)].reader;
}



/*
   Create a new domain by reading an input file.
 */
W_domain_t * /* domain */
WD_new_domain (struct W_options_t *o)
{
  W_domain_t *d;
  const char *input_line;
  
  d = w_new (W_domain_t);
  CHECK_MALLOC (d);

  d->options = o;
  d->errh    = WEH_new_handler (WOPT_OPT (d->options, error_log_name));
  d->logh    = W_new_log_handler (WOPT_OPT (d->options, log_name));
  
  d->parser  = tokens_new_parser ();

  if (!tokens_open_file (d->parser, WOPT_OPT (d->options, in_file_name))) {
    REPORT_FATAL_ERROR (d->errh, "Could not open input file");
  }
  
  if (WOPT_OPT (d->options, verbose)) {
    W_update_log (d->logh, "Input file:");
    sprintf (W_buf (), " %s", WOPT_OPT (d->options, in_file_name));
    W_update_log (d->logh, W_buf ());
  }

  if (WOPT_OPT (d->options, expeval_init) != NULL) {
    int stat;
    double f;
    stat = expeval_eval (WOPT_OPT (d->options, expeval_init), &f);
    if (stat != EXPEVAL_E_OK) {
      sprintf (W_buf (), "%s: \"%s\"", expeval_err_msgs[stat],
               WOPT_OPT (d->options, expeval_init));
      REPORT_FATAL_ERROR (d->errh, W_buf ());
    }
  }
  
  input_line = W_get_input_line (d);
  if (input_line != NULL) 
    DUPLICATE_STRING (d->comment, input_line);
  if (WOPT_OPT (d->options, verbose)) {
    W_update_log (d->logh, "Problem description:");
    sprintf (W_buf (), " %s", d->comment);
    W_update_log (d->logh, W_buf ());
  }

  input_line = W_get_input_line (d);
  if (input_line == NULL) /* Just skip this line;
                             the total used to be used in older versions,
                             but one line still has to be supplied at this point
                             (one could use 'ignore' or supply a zero)
                          */ 
      BAD_INPUT ();
  

  d->cd_integrator      = NULL;
  d->newmark_integrator = NULL;

  d->avg_nodal_mass = 0;
  d->total_mass     = 0;
  
  d->history_writer = NULL;
  if (!WOPT_OPT (d->options, no_output))
    d->history_writer = WHR_new_history_writer (d);

  d->nodes = dict_create_dictionary ();
  CHECK_MALLOC (d->nodes);

  d->elements = dict_create_dictionary ();
  CHECK_MALLOC (d->elements);

  d->elems_reg_for_integrate = NULL; /* No elements require to */
                                     /* be called when the domain */
                                     /* is integrated in time */
  
  d->materials = dict_create_dictionary ();
  CHECK_MALLOC (d->materials);
  
  d->time_funcs = dict_create_dictionary ();
  CHECK_MALLOC (d->time_funcs);

  d->loads = dict_create_dictionary ();
  CHECK_MALLOC (d->loads);

  d->disp_bcs = dict_create_dictionary ();
  CHECK_MALLOC (d->disp_bcs);

  d->iv_conds = dict_create_dictionary ();
  CHECK_MALLOC (d->iv_conds);

  d->cont_surfs = dict_create_dictionary ();
  CHECK_MALLOC (d->cont_surfs);

  d->ext_load_work   = 0;

#if defined (MPI_WASP) && MPI_WASP
  d->mpi_process.domain = d;
  d->mpi_process.rank   = -1;
  d->mpi_process.size   = 0;
#endif

  d->pod = NULL;

  if (!read_problem_data (d)) {
    REPORT_FATAL_ERROR (d->errh, "Data not read");
  }

  tokens_close_curr_file (d->parser);
  
  d->mass_damping_coeff      = WOPT_OPT (d->options, rayleigh_mass_damping_coeff);
  d->curr_mass_damping_coeff = d->mass_damping_coeff;
  d->mass_damping_time_func  = WOPT_OPT (d->options, mass_damping_time_func);

#if defined (MPI_WASP) && MPI_WASP
  if (!WOPT_OPT (d->options, serial)) W_mpi_build_comm_buffs (&d->mpi_process);
#endif

  d->response_observer_func = NULL;

#if TIMED_WASP
  d->effl_timer = W_timer_new ();
  d->solv_timer = W_timer_new ();
  d->ucfg_timer = W_timer_new ();
  d->fina_timer = W_timer_new ();
#if defined (MPI_WASP) && MPI_WASP
  d->comm_timer = W_timer_new ();
#endif
#endif

  d->check_interrupt = NULL;
  
  return d;
}


void
W_domain_set_response_observer_func (W_domain_t *d,
                                     W_response_observer_func response_observer_func)
{
  d->response_observer_func = response_observer_func;
}

/*
   Register element for integration.
   Needs to be done only if the element has internal degrees of freedom.
   If an element registers, it needs to define the functions related
   to the central difference algorithm.
 */
void
WD_register_elem_for_integration (W_domain_t *d, /* domain */
                                  W_elem_t *e /* element */
                                  )
{
  if (d->elems_reg_for_integrate == NULL)
    d->elems_reg_for_integrate = make_list();

  add_to_tail(d->elems_reg_for_integrate, e);
}



static void
make_pod_transf (W_domain_t *d)
{
  int i, j, k;
  MAT *Phit_M_Phi = NULL, *Phi = NULL, *Phit_M_Phi_inv = NULL;
  int M, ndof, dofpn;
  double *Phi_row;
  double *massmat, nmass;
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif
  double mean;
  
  M = W_pod_get_nmodes (d->pod);
  Phit_M_Phi = m_get (M, M); CHECK_MALLOC (Phit_M_Phi);
  ndof = WD_tot_of_equations (d);
  Phi_row = w_new_array (M, double); CHECK_MALLOC (Phi_row);
  massmat = w_new_array (ndof, double); CHECK_MALLOC (massmat);

  Phi = m_get (ndof, M); CHECK_MALLOC (Phi);

  FOR (k, ndof) {
    massmat[k] = 0;
    W_pod_dof_modes (d->pod, k, Phi_row, &mean);
    FOR (i, M) { Phi->me[k][i] = Phi_row[i]; }
  }

  NODE_LOOP_START (d, n) {
    nmass = WN_nodal_mass (n);
    dofpn = WN_eqn_numbers_and_constraints (n, ens, constrained);
    FOR (k, dofpn) { massmat[ens[k]] = nmass; }
  } NODE_LOOP_END;

  FOR (i, M) {
    FOR (j, M) {
      Phit_M_Phi->me[i][j] = 0;
      FOR (k, ndof) {
        Phit_M_Phi->me[i][j] += massmat[k] * Phi->me[k][i] * Phi->me[k][j];
      }
    }
  }

  Phit_M_Phi_inv = m_inverse (Phit_M_Phi, Phit_M_Phi_inv = NULL);
  
  d->Phit_M_Phi_inv_Phit
    = mmtr_mlt (Phit_M_Phi_inv, Phi, d->Phit_M_Phi_inv_Phit = NULL);

  d->pod_a = w_new_array (M, double); CHECK_MALLOC (d->pod_a);

  m_free (Phit_M_Phi_inv);
  m_free (Phit_M_Phi);
  w_free (massmat);
  m_free (Phi);
}

/*
  Solve for the primary unknowns (accelerations) in terms of the POD
  modes.
 */
void
WD_solve_for_pod (W_domain_t *d /* domain */)
{
  int M, k, i, j, dofpn;
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif
  W_vect_t accel;
  double *a;
  double mean, f[3], *Phi_row;
  W_bool_t apply[] = { TRUE, TRUE, TRUE, TRUE, TRUE, TRUE };
#if TIMED_WASP
  W_timer_restart (d->solv_timer);
#endif


  M = W_pod_get_nmodes (d->pod);

  Phi_row = w_new_array (M, double); CHECK_MALLOC (Phi_row);
  
  if (d->Phit_M_Phi_inv_Phit == NULL) make_pod_transf (d);

/* POD unknowns */
  a = d->pod_a;
  FOR (k, M) {a[k] = 0;}
  
  NODE_LOOP_START (d, n) {
    WN_get_force (n, &f[0], &f[1], &f[2]);
    dofpn = WN_eqn_numbers_and_constraints (n, ens, constrained);
    FOR (i, M) {
      FOR (k, dofpn) {
        j = ens[k];
        a[i] += d->Phit_M_Phi_inv_Phit->me[i][j] * f[k];
      }
    }
  } NODE_LOOP_END;

  NODE_LOOP_START (d, n) {
    dofpn = WN_eqn_numbers_and_constraints (n, ens, constrained);

    W_pod_dof_modes (d->pod, ens[0], Phi_row, &mean);
    accel.x = 0;
    FOR (i, M) { accel.x += Phi_row[i] * a[i]; }

    W_pod_dof_modes (d->pod, ens[1], Phi_row, &mean);
    accel.y = 0;
    FOR (i, M) { accel.y += Phi_row[i] * a[i]; }

    W_pod_dof_modes (d->pod, ens[2], Phi_row, &mean);
    accel.z = 0;
    FOR (i, M) { accel.z += Phi_row[i] * a[i]; }

    WN_set_force (n, &accel, apply); /* force stores the new accelerations
                                      computed in the regular solve
                                     */
  } NODE_LOOP_END;

  DISP_BC_LOOP_START (d, bc) {
    WBC_enforce (bc);
  } DISP_BC_LOOP_END;

  w_free (Phi_row);
  
#if TIMED_WASP
  W_timer_accumulate (d->solv_timer);
#endif
}


/*
   Integrate the equations of motion for the domain.
 */
void
WD_integrate (W_domain_t *d /* integrate */
              )
{
  W_float_t start_time = 0;
  double real_time, usr_time, systm_time;
  W_postfile_times_list_t OL;
  W_bool_t should_flush
#ifdef X_XFACE
    = (WOPT_OPT (d->options, X_xface)                      /* flush for X interface */
       || WOPT_OPT (d->options, time_between_outputs) == 0 /* for forced output each step */
       );
#else
  = 0;
#endif

  OL = WOPT_OPT (d->options, postfile_times_list);
  
#if defined (MPI_WASP) && MPI_WASP
  /* Wait until all processes have been started up and arrive at the barrier */
  if (!options.serial) MPI_Barrier (MPI_COMM_WORLD); 
#endif
  
  if (d->cd_integrator == NULL) {
    if (d->pod && W_pod_is_in_use_mode (d->pod)) {
      if (WOPT_OPT (d->options, integrator_detail) == WARBURTON_CD_VARIANT) {
        d->cd_integrator =
          W_new_cd_Warburton_integrator
          (d, 0.0, WD_compute_dt(d),
           (W_cdi_calc_eff_loads_proc)WD_calc_eff_loads,
           (W_cdi_solve_for_proc)WD_solve_for_pod,
           (W_cdi_update_config_proc)WD_update_config_Warburton_variant,
           (W_cdi_finalize_step_proc)WD_finalize_step);
      } else if (WOPT_OPT (d->options, integrator_detail) == NEWMARK_CD_VARIANT ||
                 WOPT_OPT (d->options, integrator_detail) == 0 /* default */
                 ) {
        d->cd_integrator =
          W_new_cd_Newmark_integrator
          (d, 0.0, WD_compute_dt(d),
           (W_cdi_calc_eff_loads_proc)WD_calc_eff_loads,
           (W_cdi_solve_for_proc)WD_solve_for_pod,
           (W_cdi_update_config_proc)WD_update_config_Newmark_variant,
           (W_cdi_finalize_step_proc)WD_finalize_step);
      } else {
        REPORT_FATAL_ERROR(d->errh, "Unknown central difference integrator");
      }
    } else {
      if (WOPT_OPT (d->options, integrator_detail) == WARBURTON_CD_VARIANT) {
        d->cd_integrator =
          W_new_cd_Warburton_integrator
          (d, 0.0, WD_compute_dt(d),
           (W_cdi_calc_eff_loads_proc)WD_calc_eff_loads,
           (W_cdi_solve_for_proc)WD_solve_for,
           (W_cdi_update_config_proc)WD_update_config_Warburton_variant,
           (W_cdi_finalize_step_proc)WD_finalize_step);
      } else if (WOPT_OPT (d->options, integrator_detail) == NEWMARK_CD_VARIANT ||
                 WOPT_OPT (d->options, integrator_detail) == 0 /* default */
                 ) {
        d->cd_integrator =
          W_new_cd_Newmark_integrator
          (d, 0.0, WD_compute_dt(d),
           (W_cdi_calc_eff_loads_proc)WD_calc_eff_loads,
           (W_cdi_solve_for_proc)WD_solve_for,
           (W_cdi_update_config_proc)WD_update_config_Newmark_variant,
           (W_cdi_finalize_step_proc)WD_finalize_step);
      } else {
        REPORT_FATAL_ERROR(d->errh, "Unknown central difference integrator");
      }
    }
#   if !defined(PVM_WASP_MASTER)
    assemble_mass (d);
    WD_apply_initial_conditions (d);
#   endif
  }

  WD_flush_history(d);

  start_time = WD_time(d);
  if (WOPT_OPT (d->options, running_patch_test))
    WOPT_SET_OPT (d->options, target_time, 1.9 * W_cd_integrator_dt(d->cd_integrator));

 advance:
  do {
    if (d->check_interrupt) d->check_interrupt (d);
    if (WD_time(d) - start_time >= WOPT_OPT (d->options, time_between_outputs)) {
      start_time = WD_time(d); 
                                          
      if (!WOPT_OPT (d->options, no_output)) {        
        if (WOPT_OPT (d->options, verbose)) {
          W_get_times (&real_time, &usr_time, &systm_time);
          sprintf(W_buf(), "Time=%g (r=%g; u=%g; s=%g)",
                  WD_time(d), real_time, usr_time, systm_time);
	  W_update_log(d->logh, W_buf());
	}
        WHR_update_history(d->history_writer, should_flush);
      }
#ifdef X_XFACE
      if (WOPT_OPT (d->options, X_xface)) {
        WD_build_grep (d, FALSE);
        sprintf(W_buf(), "Time = %g", WD_time(d));
        xiface_show_graphics (WOPT_OPT (d->options, interactive_during_outputs),
                              W_buf(), WD_time (d));
      }
#endif
    }
    if (   (!WOPT_OPT (d->options, no_output))
        && (   (OL.ntimes > 0) && (OL.curr < OL.ntimes))) {
      if ((OL.ntimes > 0) && (WD_time(d) >= OL.times[OL.curr])) {
        WD_write_postfile(d, OL.times[OL.curr]);
        OL.curr++;
      }
    } else if (WOPT_OPT (d->options, write_postfile_file_now)) {
      WD_write_postfile(d, WD_time(d));
      WOPT_SET_OPT (d->options, write_postfile_file_now, FALSE);
    }
    if (d->response_observer_func) (d->response_observer_func) (d, FALSE);
    W_advance_cd_integrator(d->cd_integrator);
  } while (WD_time(d) < WOPT_OPT (d->options, target_time));
  
#ifdef X_XFACE
  if (WOPT_OPT (d->options, X_xface)) {
    W_get_times(&real_time, &usr_time, &systm_time);
    sprintf(W_buf(), "Time reached = %g (r=%g; u=%g; s=%g)",
            WD_time(d), real_time, usr_time, systm_time);
    xiface_loop(1, W_buf());
    if (WD_time (d) < WOPT_OPT (d->options, target_time)) goto advance;
  }
#endif
  
  if (d->response_observer_func) (d->response_observer_func) (d, TRUE);

  WD_clean_up_before_exit(d);
  
  if (WOPT_OPT (d->options, verbose)) {
    sprintf(W_buf(), "Reached target time = %g", WD_time(d));
    W_update_log(d->logh, W_buf());
    W_get_times(&real_time, &usr_time, &systm_time);
    sprintf(W_buf(), "   Used time [sec]: real %g, user %g, sys %g",
            real_time, usr_time, systm_time);
    W_update_log(d->logh, W_buf());
  }

#ifdef NEED_STAT_MALLOC
  W_stat_malloc_statistics(stderr);
#endif
}



/*
   Get the current time. 
 */
W_float_t /* time */
WD_time (W_domain_t  *d /* domain */
         )
{
  assert(d != NULL);
  if      (d->cd_integrator)
    return W_cd_integrator_time (d->cd_integrator);
  else if (d->newmark_integrator)
    return W_newmark_integrator_time (d->newmark_integrator);
  else
    return 0;
}


/*
   Get the current time step. 
 */
W_float_t /* time */
WD_time_step (W_domain_t  *d /* domain */
              )
{
  if      (d->cd_integrator)
    return W_cd_integrator_dt (d->cd_integrator);
  else if (d->newmark_integrator)
    return W_newmark_integrator_dt (d->newmark_integrator);
  else
    return 0;
}




#if !(defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER)) /* serial code */
/*
   Compute the time step.
 */
W_float_t /* time step */
WD_compute_dt (W_domain_t *d /* domain */)
{
  double dt = FLT_MAX;

  if (WOPT_OPT (d->options, time_step) == 0) {
    ELEM_LOOP_START (d, e) {
      dt = min(dt, WE_suggested_time_step(e));
   } ELEM_LOOP_END;
    dt *= WOPT_OPT (d->options, time_step_reduction);
  } else
    dt = WOPT_OPT (d->options, time_step);

#if defined (MPI_WASP) && MPI_WASP
  {
    double sdt = dt;
    MPI_Allreduce (&sdt, &dt, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
  }
#endif
  
  if (WOPT_OPT (d->options, verbose)) {
    sprintf(W_buf(), "Time step = %g", (double)dt);
    W_update_log(d->logh, W_buf());
  }
  return dt;
}

#endif /* serial code for WD_compute_dt */


#if !(defined(PVM_WASP_WORKER) || defined(PVM_WASP_MASTER))
/* serial or MPI code */
/*
   Compute the effective nodal loads.
 */
void
WD_calc_eff_loads (W_domain_t *d /* domain */)
{
#if TIMED_WASP
  W_timer_restart (d->effl_timer);
#endif
 
  d->curr_mass_damping_coeff = d->mass_damping_coeff;
  if (d->mass_damping_coeff != 0) {
    if (d->mass_damping_time_func > 0) {
      W_time_func_t *tf
        = WD_get_time_func_by_idx (d, d->mass_damping_time_func);
      d->curr_mass_damping_coeff *= WTF_func_val (tf, WD_time(d));
    }
  }

  NODE_LOOP_START (d, n) {
    WN_prep_force_assembly (n);
  } NODE_LOOP_END;

  LOAD_LOOP_START (d, l) {
    WL_assemble_load (l, d);
  } LOAD_LOOP_END;
  
  CONTACT_LOOP_START (d, c) {
    WC_search (c);
  } CONTACT_LOOP_END;

  REG_ELEM_LOOP(d, WE_calc_eff_loads(theELEM));

  ELEM_LOOP_START (d, e) {
    WE_assemble_restoring_forces (e, W_MAT_STATE_UPDATE_FINAL);
  } ELEM_LOOP_END;
  
#if W_ROT_DOFS
  NODE_LOOP_START (d, n) {
    WN_assemble_ang_vel_dep_load (n);
  } NODE_LOOP_END;
#endif

  if (d->mass_damping_coeff != 0) {
    W_float_t dt = W_cd_integrator_dt(d->cd_integrator);
    NODE_LOOP_START (d, n) {
      WN_assemble_damping (n, d->mass_damping_coeff, dt);
    } NODE_LOOP_END;
  }

#if TIMED_WASP
  W_timer_accumulate (d->effl_timer);
#endif

#if defined (MPI_WASP) && MPI_WASP && MPI_WASP_NODE_CUT
  if (!WOPT_OPT (d->options, serial)) W_mpi_exchange_forces (d);
#endif
}
#endif /* serial code for WD_calc_eff_loads */

#if !(defined(PVM_WASP_MASTER))
 /* serial or worker's code, or MPI code */
/*
   Solve for the primary position unknowns.
 */
void
WD_solve_for(W_domain_t *d /* domain */)
{
#if TIMED_WASP
  W_timer_restart (d->solv_timer);
#endif
 
  NODE_LOOP_START (d, n) {
    WN_solve_for (n);
  } NODE_LOOP_END;

  REG_ELEM_LOOP(d, WE_solve_for (theELEM));

  DISP_BC_LOOP_START (d, bc) {
    WBC_enforce (bc);
  } DISP_BC_LOOP_END;

#if TIMED_WASP
  W_timer_accumulate (d->solv_timer);
#endif

#if defined (MPI_WASP) && MPI_WASP && MPI_WASP_ELEM_CUT
  if (!WOPT_OPT (d->options, serial)) W_mpi_exchange_accels (d);
#endif
}
#endif /* serial code for WD_solve_for */


void
assemble_mass (W_domain_t *d /* domain */)
{
  int nnodes;
  double nmass;

  d->avg_nodal_mass = 0;

  NODE_LOOP_START (d, n) {
    WN_prep_mass_assembly (n);
  } NODE_LOOP_END;
  
  ELEM_LOOP_START (d, e) {
    WE_assemble_mass(e);
  } ELEM_LOOP_END;

#if W_ROT_DOFS
  NODE_LOOP_START (d, n) {
    WN_resolve_rot_inertia (n);
  } NODE_LOOP_END;
#endif

#if defined (MPI_WASP) && MPI_WASP_NODE_CUT
  if (!WOPT_OPT (d->options, serial)) W_mpi_exchange_masses (d);
#endif
   
  {
    double rj, pmr = 0, nmr = 0, omega_bar = 0.25;
    nnodes = 0;
    NODE_LOOP_START (d, n) {
      W_coords_3d_t *x = &n->ref_coords;
      rj = sqrt (x->x*x->x);
      pmr += n->mass * v_bar * (rj*rj*rj*rj);
      nmr += n->mass * (rj*rj);
      nmass = WN_nodal_mass (n);
      d->avg_nodal_mass += nmass;
      if (nmass <= 0 &&
          (n->can_move_in_dir[0] ||
           n->can_move_in_dir[1] ||
           n->can_move_in_dir[2])) {
        W_bool_t free[] = {FALSE,FALSE,FALSE,FALSE,FALSE,FALSE};
        sprintf (W_buf (), "Zero or negative mass associated with node %d", (int)n->id);
        REPORT_WARNING (d->errh, W_buf ());
        WN_fix_node (n, free);
      }
      nnodes++;
    } NODE_LOOP_END;
    printf ("nmr=%g, pmr=%g, omega=%g\n", nmr, pmr, pmr/nmr);
  }

  d->avg_nodal_mass = d->avg_nodal_mass / nnodes;
  d->total_mass     = WD_total_mass (d);
}


#if !defined(PVM_WASP_MASTER) /* serial or worker's code */
/*
   Update the configuration for the Warburton's variant of the
   central differences.
 */
void
WD_update_config_Warburton_variant (W_domain_t *d /* domain */)
{
   W_float_t dt = W_cd_integrator_dt(d->cd_integrator);
#if TIMED_WASP
   W_timer_restart (d->ucfg_timer);
#endif
   
   NODE_LOOP_START (d, n) {
     WN_update_config_Warburton_variant (n, d, dt);
   } NODE_LOOP_END;
   REG_ELEM_LOOP(d, WE_update_config_Warburton_variant(theELEM));
   update_ext_load_work(d);
   
#if TIMED_WASP
   W_timer_accumulate (d->ucfg_timer);
#endif
}
#endif /* serial or worker's code */

#if !defined(PVM_WASP_MASTER) /* serial or worker's code */
/*
   Update the configuration for the Newmark's variant of the
   central differences.
 */
void
WD_update_config_Newmark_variant (W_domain_t *d /* domain */)
{
  W_float_t dt = W_cd_integrator_dt(d->cd_integrator);
#if TIMED_WASP
  W_timer_restart (d->ucfg_timer);
#endif
  
  NODE_LOOP_START (d, n) {
    WN_update_config_Newmark_variant (n, d, dt);
  } NODE_LOOP_END;
  REG_ELEM_LOOP(d, WE_update_config_Newmark_variant(theELEM));
  update_ext_load_work(d);
   
#if TIMED_WASP
   W_timer_accumulate (d->ucfg_timer);
#endif
}
#endif /* serial or worker's code */

 
#ifdef X_XFACE

static int 
redraw_view (NODE ptr, NODE elem)
{
  EVFastRedraw ((EView *)elem);
  return 1;
}
/*
   Build graphic representation of domain.
 */
void
WD_build_grep (W_domain_t *d /* domain */,
               W_bool_t immed_draw
               )
{
  ELEM_LOOP_START (d, e) {
    WE_modify_elem_grep (e, WOPT_OPT (d->options, displ_scale), immed_draw);
  } ELEM_LOOP_END;
  NODE_LOOP_START (d, n) {
    WN_modify_node_grep (n, WOPT_OPT (d->options, displ_scale),
                         WOPT_OPT (d->options, draw_numbers), immed_draw);
  } NODE_LOOP_END;
  LOAD_LOOP_START (d, l) {
    WL_modify_load_grep (l, d, immed_draw);
  } LOAD_LOOP_END;

  if (!immed_draw)
    EMDispatchToDependentViews (ESIModel (), redraw_view, NULL);
}

#endif

/*
   Compute the total mass of the domain.
 */
W_float_t /* total mass */
WD_total_mass (W_domain_t *d /* domain */)
{
  assert(d != NULL);
  
  if (d->total_mass == 0) {
    ELEM_LOOP_START (d, e) {
      d->total_mass += WE_elem_total_mass(e);
    } ELEM_LOOP_END;
  } 
  return d->total_mass;
}



/*
   Compute the average mass of the nodes.
 */
W_float_t /* total mass */
WD_avg_nodal_mass (W_domain_t *d /* domain */)
{
  return d->avg_nodal_mass;
}






/*
   Compute current kinetic energy.
 */
W_float_t /* energy */
WD_current_kinetic_energy (W_domain_t *d /* domain */)
{
  W_float_t e = 0;
  
  NODE_LOOP_START (d, n) {
    e += WN_node_kinetic_energy (n);
  } NODE_LOOP_END;
  REG_ELEM_LOOP(d, e += WE_kinetic_energy(theELEM));
  return e;
}



/*
   Compute current work of deformation.
 */
W_float_t /* value of work of deformation */
WD_current_deformation_energy (W_domain_t *d /* domain */)
{
  W_float_t en = 0;

  ELEM_LOOP_START (d, e) {
    en += WE_deformation_energy(e);
  } ELEM_LOOP_END;
  return en;
}


static void
update_ext_load_work(W_domain_t *d /* domain */)
{
  W_float_t dew = 0;

  LOAD_LOOP_START (d, l) {
    dew += WL_incr_ext_load_work (l, d);
  } LOAD_LOOP_END;
  d->ext_load_work += dew;
}


/*
   Compute value of current work of external loads.
 */
W_float_t
WD_current_ext_load_work (W_domain_t *d /* domain */) 
{
  return (d->ext_load_work);
}

static void
pod_proj_coeffs (W_domain_t *d, int how_many, double *a,
                 void (*accessor) (W_node_t *n, double *x, double *y, double *z) 
                 )
{
  W_pod_p pod = d->pod;
  int j, k, M = W_pod_get_nmodes (pod);
  double mean, *modes;
  W_node_t *n;
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif
  double fr[3];

  modes = w_new_array (M, double); CHECK_MALLOC (modes);
  
  FOR (k, how_many) { a[k] = 0; }

  NODE_LOOP_START (d, n) {
    (*accessor) (n, &fr[0], &fr[1], &fr[2]);
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    FOR (j, 3) {
      W_pod_dof_modes (pod, ens[j], modes, &mean);
      FOR (k, how_many) { a[k] += fr[j] * modes[k]; }
    }
  } NODE_LOOP_END;

  w_free (modes);
}

static void
record_pod_mode_participation (W_domain_t *d)
{
  W_pod_p pod = d->pod;
  int j, M = W_pod_get_nmodes (pod);
  char name[128];
  double *coeffs, sum;

  coeffs = w_new_array (M, double); CHECK_MALLOC (coeffs);
  /* project onto modes */
  pod_proj_coeffs (d, M, coeffs, WN_get_displ);
  /* normalize coefficients */
  sum = 0; FOR (j, M) { sum += coeffs[j]*coeffs[j]; }; sum = sqrt (sum);
  /* save history */
  FOR (j, M) {
    sprintf (name, "nDpart-m%d", j);
    WHR_store (d->history_writer, name, WD_time (d), coeffs[j] / sum);
  }
  /* project onto modes */
  pod_proj_coeffs (d, M, coeffs, WN_get_velocity);
  /* normalize coefficients */
  sum = 0; FOR (j, M) { sum += coeffs[j]*coeffs[j]; }; sum = sqrt (sum);
  /* save history */
  FOR (j, M) {
    sprintf (name, "nVpart-m%d", j);
    WHR_store (d->history_writer, name, WD_time (d), coeffs[j] / sum);
  }
  w_free (coeffs);
}

/*
   Update the history records.
 */
void
WD_update_history (W_domain_t *d /* domain */
                   )
{
  W_float_t ke, de, ew;
  double time = WD_time (d);

  ke = WD_current_kinetic_energy(d);
  de = WD_current_deformation_energy(d);
  ew = WD_current_ext_load_work(d);
      
  WHR_store (d->history_writer, "KinetE", time, ke);
  WHR_store (d->history_writer, "DeformE", time, de);
  WHR_store (d->history_writer, "ExtLWork", time, ew);
  WHR_store (d->history_writer, "TotalE", time, (ke + de));

  if (d->pod && W_pod_is_in_use_mode (d->pod))
    record_pod_mode_participation (d);
}


static void
proj_iv_coeffs (W_domain_t *d, double *a)
{
  W_pod_p pod = d->pod;
  int M = W_pod_get_nmodes (pod);
  int j, k;
  double iv[3], mean, *dof_modes;
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else 
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif

  dof_modes = w_new_array (M, double); CHECK_MALLOC (dof_modes);
  FOR (k, M) { a[k] = 0; }
  NODE_LOOP_START (d, n) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    iv[0] = n->velocity_t.x; iv[1] = n->velocity_t.y; iv[2] = n->velocity_t.z; 
    FOR (j, 3) {
      W_pod_dof_modes (pod, ens[j], dof_modes, &mean);
      FOR (k, M) {
        a[k] += iv[j] * dof_modes[k];
      }
    }
  } NODE_LOOP_END;
  w_free (dof_modes);
}

static void
proj_initial_velocity (W_domain_t *d)
{
  double *a;
  W_pod_p pod = d->pod;
  int j, k, M = W_pod_get_nmodes (pod);
  double iv[3], mean, *dof_modes;
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else 
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif

  dof_modes = w_new_array (M, double); CHECK_MALLOC (dof_modes);
  a         = w_new_array (M, double); CHECK_MALLOC (a);

  proj_iv_coeffs (d, a);

  NODE_LOOP_START (d, n) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    FOR (j, 3) {
      W_pod_dof_modes (pod, ens[j], dof_modes, &mean);
      iv[j] = 0;
      FOR (k, M) {
        iv[j] += a[k] * dof_modes[k];
      }
    }
    n->velocity_t.x = iv[0]; n->velocity_t.y = iv[1]; n->velocity_t.z = iv[2];
  } NODE_LOOP_END;

  w_free (a);
  w_free (dof_modes);
}

/*
   Apply initial conditions to the entities constituting domain.
 */
void
WD_apply_initial_conditions (W_domain_t *d /* domain */)
{
  W_vect_t rc, v;
  W_float_t dt = 0;

  if (!(d->cd_integrator || d->newmark_integrator)) REPORT_ERROR(d->errh, "no integrator");

  dt = WD_time_step (d);

  NODE_LOOP_START (d, n) {
    WN_get_ref_coords (n, &rc.x, &rc.y, &rc.z);
    WD_initial_velocity (d, &rc, &v);
    WN_setup_init_conds (n, &v, dt);
  } NODE_LOOP_END;

  if (d->pod && W_pod_is_in_use_mode (d->pod)) proj_initial_velocity (d);

  ELEM_LOOP_START (d, e) {
    WE_apply_initial_conditions (e);
  } ELEM_LOOP_END;

  DISP_BC_LOOP_START (d, bc) {
    WBC_init_disp_bc (bc);
  } DISP_BC_LOOP_END; 
}


/*
   Compute initial velocity at a given point.
 */
void
WD_initial_velocity (W_domain_t *d /* domain */,
                     W_loc_t *at, /* location */
                     W_loc_t *v_initial /* returns the velocity */
                     )
{
  W_velocity_3d_t zerov = {0,0,0};
  *v_initial = zerov;

  IV_COND_LOOP_START (d, ivc) {
    if (WIVC_eval (ivc, at, v_initial))
      return;
  } IV_COND_LOOP_END;
}

 



static void
write_hexvue_file (W_domain_t *d /* domain */, ckit_iobuf_t IOB)
{
  char buf[132];
  int n_output_values = 0;
  
  W_get_date_and_time (buf);
  ckit_io_put_line (IOB, "! HEXVUE file (C) 1995 Petr Krysl.");
  sprintf (W_buf (), "!     Generated by WASP on %s.", buf);
  ckit_io_put_line (IOB, W_buf ());
  sprintf (W_buf (), "%s; Time: %g (file %s, %s)",
          d->comment, WD_time(d), WOPT_OPT (d->options, in_file_name), buf);
  ckit_io_put_line (IOB, W_buf ());
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_DISPL_MASK)
    n_output_values += 3;
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_STRESS_MASK)
    n_output_values += 6;
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_EQPLS_MASK)
    n_output_values += 1;
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_VON_MISES_MASK)
    n_output_values += 1;
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_MATERIAL_MASK)
    n_output_values += 1;
  
  sprintf (W_buf (), "%d ! # of values per vertex", n_output_values);
  ckit_io_put_line (IOB, W_buf ());

  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_DISPL_MASK) {
    ckit_io_put_line (IOB, "\"Displacement U\""); 
    ckit_io_put_line (IOB, "\"Displacement V\""); 
    ckit_io_put_line (IOB, "\"Displacement W\"");
  }
  
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_STRESS_MASK) {
    if        (WOPT_OPT (d->options, output_stress_type) == W_OUTPUT_STRESS_CAUCHY) {
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_11\""); 
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_12\""); 
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_13\"");
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_22\""); 
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_23\""); 
      ckit_io_put_line (IOB, "\"Cauchy stress sigma_33\"");
    } else if (WOPT_OPT (d->options, output_stress_type) == W_OUTPUT_STRESS_2ND_PK) {
      ckit_io_put_line (IOB, "\"2nd PK stress S_11\""); 
      ckit_io_put_line (IOB, "\"2nd PK stress S_12\""); 
      ckit_io_put_line (IOB, "\"2nd PK stress S_13\"");
      ckit_io_put_line (IOB, "\"2nd PK stress S_22\""); 
      ckit_io_put_line (IOB, "\"2nd PK stress S_23\""); 
      ckit_io_put_line (IOB, "\"2nd PK stress S_33\"");
    }
  }

  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_EQPLS_MASK) {
    ckit_io_put_line (IOB, "\"Equivalent plastic strain\"");
  }
  
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_VON_MISES_MASK) {
    ckit_io_put_line (IOB, "\"Von Mises stress\"");
  }
 
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_MATERIAL_MASK) {
    ckit_io_put_line (IOB, "\"Material ID\"");
  }

  ckit_io_put_line (IOB, "! END OF HEADER");
  
  ELEM_LOOP_START (d, e) {
    WE_write_postproc_file (e, IOB,
                            HEXVUE_POSTPROC_FILE,
                            WOPT_OPT (d->options, hexvue_format), NULL);
  } ELEM_LOOP_END;
  ckit_io_put_line (IOB, "END ! HEXVUE file.");
}


static void
write_ve3d_file (W_domain_t *d /* domain */, ckit_iobuf_t IOB)
{
  char buf[132];
  
  W_get_date_and_time(buf);
  
  ckit_io_put_line (IOB, "! VE3D file (C) 1996 Petr Krysl.");
  sprintf (W_buf (), "!    Generated by WASP (%s).\n", buf);
  ckit_io_put_line (IOB, W_buf ());
  sprintf (W_buf (), "BODY \"%s\"\n\n", WOPT_OPT (d->options, in_file_name));
  ckit_io_put_line (IOB, W_buf ());

#if 0                           /* RAW */
  ELEM_LOOP_START (d, e) {
    if (WE_elem_elcode (e) == ELCODE_EFG_SUPERELEM) {
      sprintf (W_buf (), "REGION \"%s\"",
               WE_elem_name(e);
               ckit_io_put_line (IOB, W_buf ());
               WE_write_postproc_file (e, IOB,
                                       VE3D_POSTPROC_FILE, "ascii", NULL);
               ckit_io_put_line (IOB, "END_REGION");
             }
    }
  } ELEM_LOOP_END;
#endif

  sprintf (W_buf (), "END_BODY \"%s\"\n\n", WOPT_OPT (d->options, in_file_name));
  ckit_io_put_line (IOB, W_buf ());
}


static void
write_tecplot_file (W_domain_t *d /* domain */, ckit_iobuf_t IOB)
{
  char buf[132];
  W_index_t count;
  long int I_equals_offset;

  W_get_date_and_time(buf);
  sprintf(W_buf (), "TITLE=\"%s (%s): %s\"\n",
          WOPT_OPT (d->options, in_file_name), buf, d->comment);
  ckit_io_put_line (IOB, W_buf ());
  ckit_io_put_line (IOB, "VARIABLES=\"X\", \"Y\", \"Z\""); 
  
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_DISPL_MASK)
    ckit_io_put_line (IOB, ", \"U\", \"V\", \"W\"");
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_STRESS_MASK)
    ckit_io_put_line (IOB, ", \"S11\", \"S12\", \"S13\", \"S22\", \"S23\", \"S33\"");
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_EQPLS_MASK)
    ckit_io_put_line (IOB, ", \"EQUIV_PL_STRAIN\"");
  if (WOPT_OPT (d->options, postfile_output_spec) & W_POSTFILE_VON_MISES_MASK)
    ckit_io_put_line (IOB, ", \"VON_MISES_STRESS\"");
  ckit_io_put_line (IOB, "");

  ckit_io_put_line (IOB, "ZONE I=");
  I_equals_offset = ckit_io_get_file_pos (IOB);
  ckit_io_put_line (IOB, "                F=POINT");

  count = 1;
  NODE_LOOP_START (d, n) {
    WN_write_tecplot_file (n, IOB, WOPT_OPT (d->options, postfile_output_spec), &count);
  } NODE_LOOP_END;
  ELEM_LOOP_START (d, e) {
    WE_write_postproc_file (e, IOB, TECPLOT_I_POSTPROC_FILE, "ascii", &count);
  } ELEM_LOOP_END;
  ckit_io_set_file_pos (IOB, I_equals_offset);
  sprintf(W_buf (), "%ld,", count-1);
  ckit_io_put_line_no_nl (IOB, W_buf ());
}


/*
   Write a post-processing file.
   
 */
void
WD_write_named_postfile (W_domain_t *d /* domain */,
                         char *out, 
                         int type
                   )
{
  FILE *stream;
  ckit_iobuf_t IOB;

  stream = ckit_io_open_stream (out, "w");
  if (stream == NULL) REPORT_ERROR (d->errh, (char *)ckit_io_error_explanation ());
  IOB = ckit_io_new_buffer (stream);
  if (IOB == NULL) REPORT_ERROR (d->errh, (char *)ckit_io_error_explanation ());

  switch (type) {
  case HEXVUE_POSTPROC_FILE:
    write_hexvue_file(d, IOB);
    break;
  case VE3D_POSTPROC_FILE:
    write_ve3d_file(d, IOB);
    break;
  case TECPLOT_I_POSTPROC_FILE:
    write_tecplot_file(d, IOB);
    break;
  default:
    break;
  }
  ckit_io_delete_buffer (IOB);
  fclose (stream);
}

void
WD_write_postfile (W_domain_t *d /* domain */,
                   double label_time
                   )
{
  FILE *stream;
  ckit_iobuf_t IOB;
  char *out = W_postfile_name (label_time,
                               WOPT_OPT (d->options, in_file_basename),
                               WOPT_OPT (d->options, postfile_type),
                               WOPT_OPT (d->options, postfile_number_seq));
  int type = WOPT_OPT (d->options, postfile_type);

  WD_write_named_postfile (d, out, type);
}





/*
   Print info for a domain.
 */
void
WD_print_info (W_domain_t *d /* domain */,
               FILE *stream /* print to stream */
               )
{
  fprintf(stream, "Domain\n");
  fprintf(stream, "End Domain\n");
}


/*
   Print info for a given element.
 */
void
WD_print_elem_info(W_domain_t *d /* domain */,
                   W_index_t elid, /* element index */
                   W_info_spec_t spec, /* specification */
                   FILE *stream /* output stream */
                   )
{
  W_elem_t *el;

  el = WD_get_elem_by_idx(d, elid);
  if (el == NULL) {
    fprintf(stream, "No such element: %ld\n", elid);
  } else {
    WE_print_info(el, spec, stream, 0);
  }
}

/*
   Print info for a given node.
 */
void
WD_print_node_info (W_domain_t *d /* domain */,
                   W_index_t nid, /* node index */
                   W_info_spec_t spec, /* specification */
                   FILE *stream) /* output stream */
{
  W_node_t *n;

  n = WD_get_node_by_idx(d, nid);
  if (n == NULL) {
    fprintf(stream, "No such node: %ld\n", nid);
  } else {
    fprintf(stream, "Node %ld\n", nid);
    WN_print_info(n, d, spec, stream, 0);
  }
}


/*
   Clean up before the exit from the program.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WD_clean_up_before_exit (W_domain_t *d /* domain */)
{ 
  WD_flush_history(d);
  return TRUE;
}


/*
   Flush history buffers.
 */
void
WD_flush_history (W_domain_t *d /* domain */)
{
  /* Flush history records to disk to clean up */
  if (!WOPT_OPT (d->options, no_output))
    WHR_update_history (d->history_writer, TRUE);
}


/*
   Finalize the current time step.
 This function can: (i) set a new time step to use, (ii) give a chance 
 to registered elements to do some action (e.g. propagate a crack). 
 */
W_float_t /* new time step */
WD_finalize_step (W_domain_t *d /* domain */)
{
  W_float_t dt = W_cd_integrator_dt(d->cd_integrator);
  W_float_t newdt = dt;
  W_float_t sdt;
#if TIMED_WASP
  W_timer_restart (d->fina_timer);
#endif
   
  REG_ELEM_LOOP(d, sdt = WE_finalize_step(theELEM));

#if TIMED_WASP
  W_timer_accumulate (d->fina_timer);
#endif


  if (d->pod) W_pod_accumulate (d->pod);

  return newdt; /* We are ignoring the suggestions from the elements */
                /* for the time being.  Needs to be implemented. */
}


/*
   Probe some field.
 */
W_bool_t
WD_probe (W_domain_t *d, /* domain */
          void *options, /* options */
          W_loc_t *loc, /* at location */
          unsigned long spec, /* specification */
          W_probe_package_t *result /* returns result */
	 )
{
  ELEM_LOOP_START (d, e) {
    if (WE_probe (e, options, loc, spec, result)) {
      return TRUE;
    }
  } ELEM_LOOP_END;
  return FALSE;
}

/*
   Get mass damping coefficient.
  */
double /* coefficient */
WD_mass_damping_coeff (W_domain_t *d /* domain */ 
                       )
{
  return d->curr_mass_damping_coeff;
}



/*
   Get ratio of current and the input mass damping coefficient.
 */
double /* ratio */
WD_mass_damping_coeff_time_dep (W_domain_t *d /* domain */) 
{
  return (d->mass_damping_coeff != 0
          ? d->curr_mass_damping_coeff / d->mass_damping_coeff
          : 0);
}

typedef struct fef_ref_t {
  W_elem_t *e;
  int       nface;
  int       nn;
  W_node_t *n[4];
  int       ref;
}              fef_ref_t;


static LIST_OF (fef_ref_t *)
find_frl (W_node_t *n, dict_dictionary nfrd)
{
  return dict_find_obj_with_handle_name (nfrd, n);
}


static W_node_t *
min_id_fvn (W_node_t *fvn[4], int nfvn)
{
  int mini = 0, minid = INT_MAX, i;

  FOR (i, nfvn) {
    if (fvn[i]->id < minid) {
      minid = fvn[i]->id;
      mini  = i;
    }
  }
  return fvn[mini];
}

static fef_ref_t *
find_fr (LIST frl, int nfvn, W_node_t *fvn[4])
{
  int had = 0;
  int i, j;
  fef_ref_t *fr;
  
  loop_through_list (frl, fr, fef_ref_t *) {
    if (fr->nn == nfvn) {
      had = 0;
      FOR (i, nfvn) {
        FOR (j, nfvn) {
          if (fvn[i] == fr->n[j]) {
            had++;
            break;
          }
        }
      }
      if (had == nfvn) return fr;
    }
  }
  return NULL;
}

/*
   Get list of contact faces.
 */
static LIST_OF (fef_ref_t *)
get_list_of_bdry_faces (W_domain_t *d)
{
  dict_dictionary nfrld;
  int i, j, nf, nfvn;
  LIST_OF (fef_ref_t *) frefl;
  W_node_t *fvn[4], *n;
  LIST_OF (fef_ref_t *) frl, tmpl;
  fef_ref_t *fr;

  /* accumulate FE faces */
  
  nfrld = dict_create_dictionary ();
  
  tmpl = make_list ();
  
  ELEM_LOOP_START (d, e) {
    nf = WE_num_of_faces (e);
    FOR (i, nf) {
      nfvn = WE_get_face_vertex_nodes (e, i, fvn);
      n    = min_id_fvn (fvn, nfvn);
      frl  = find_frl (n, nfrld);
      if (!frl) {
        frl = make_list ();
        dict_add_obj_handle_name (nfrld, frl, n);
      }
      fr = find_fr (frl, nfvn, fvn);
      if (!fr) {
        fr = w_new (fef_ref_t); CHECK_MALLOC (fr);
        fr->e     = e;
        fr->nface = i;
        fr->nn    = nfvn;
        FOR (j, fr->nn) { fr->n[j] = fvn[j]; }
        fr->ref   = 0;
        add_to_tail (frl, fr);
        add_to_tail (tmpl, fr);
      } 
      fr->ref++;
    }
  } ELEM_LOOP_END;

  /* create face list by inspecting the accumulated normals */
  
  frefl = make_list ();
  fr     = NULL;
  while ((fr = (fef_ref_t *) get_list_next (tmpl, fr))) {
    if (fr->ref == 1) {
      add_to_tail (frefl, fr);
    } 
  }

  /* cleanup */
  
  free_list (tmpl, NOT_NODES);
  {
    dict_iterator diter;
    diter = dict_get_iterator ();
    dict_reset_iterator (diter);
    while ((frl = (LIST) dict_iter_objs (nfrld, diter))) {
      free_list (frl, NOT_NODES);
    }
    dict_del_iterator (diter);
  }
  dict_destroy_dictionary (nfrld);

  return frefl;
}

/*
   Get list of contact faces.
 */
LIST_OF (WCF_contact_face_t *)
WD_get_list_of_contact_faces (W_domain_t *d)
{
  LIST_OF (WCF_contact_face_t *) cfaces;
  LIST_OF (fef_ref_t *) frl;
  int i, nf;
  fef_ref_t *fr;
#define MAX_CF_PER_FEF 16
  WCF_contact_face_t *cf[MAX_CF_PER_FEF];

  /* find boundary faces */
  
  frl = get_list_of_bdry_faces (d);
  
  /* create contact face list  */
  
  cfaces = make_list ();
  fr     = NULL;
  while ((fr = (fef_ref_t *) get_list_next (frl, fr))) {
    if (fr->ref == 1) {
      nf = WE_get_contact_faces (fr->e, fr->nface, cf, MAX_CF_PER_FEF);
      if (nf > MAX_CF_PER_FEF)
        REPORT_FATAL_ERROR (d->errh, "too many contact faces");
      FOR (i, nf) { add_to_tail (cfaces, cf[i]); }
    } 
  }

  /* cleanup */
  
  free_list (frl, AND_NODES);

  return cfaces;
}



void 
WD_assemble_eff_stiff_newmark (W_domain_t *d)
{
  double beta, gamma, ic00, ic01, dt = WD_time_step (d);
    
  d->curr_mass_damping_coeff = d->mass_damping_coeff;
  if (d->mass_damping_coeff != 0) {
    if (d->mass_damping_time_func > 0) {
      W_time_func_t *tf
        = WD_get_time_func_by_idx (d, d->mass_damping_time_func);
      d->curr_mass_damping_coeff *= WTF_func_val (tf, WD_time(d));
    }
  }

  W_newmark_get_consts (d->newmark_integrator, &beta, &gamma);
  ic00 = 1.0 / dt / dt / beta;
  ic01 = gamma / dt / beta;
  
  ELEM_LOOP_START (d, e) {
    WE_assemble_eff_stiffness (e, 1.0, ic00 + d->curr_mass_damping_coeff * ic01,
                               WOPT_OPT (d->options, mass_type));
  } ELEM_LOOP_END;
}

void
WD_assemble_residual_forces_newmark (W_domain_t *d,
                                     W_mat_state_update_form_enum update_form,
                                     double *extf_norm, double *inertf_norm)
{
  double f[3];
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else 
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif
  int iens[3];
  
  d->curr_mass_damping_coeff = d->mass_damping_coeff;
  if (d->mass_damping_coeff != 0) {
    if (d->mass_damping_time_func > 0) {
      W_time_func_t *tf
        = WD_get_time_func_by_idx (d, d->mass_damping_time_func);
      d->curr_mass_damping_coeff *= WTF_func_val (tf, WD_time(d));
    }
  }

  *extf_norm   = 0;
  *inertf_norm = 0;

  NODE_LOOP_START (d, n) {
    WN_prep_force_assembly (n);
  } NODE_LOOP_END;

  LOAD_LOOP_START (d, l) {
    WL_assemble_load (l, d);
  } LOAD_LOOP_END;

  NODE_LOOP_START (d, n) {
    *extf_norm += WN_extf_norm_squared (n);
    *inertf_norm += WN_inertf_norm_squared (n);
  } NODE_LOOP_END;
  *extf_norm   = sqrt (*extf_norm);
  *inertf_norm = sqrt (*inertf_norm);
  
  ELEM_LOOP_START (d, e) {
    WE_assemble_residual_forces (e, update_form, WOPT_OPT (d->options, mass_type));
  } ELEM_LOOP_END;

  if (d->mass_damping_coeff != 0) {
    W_float_t dt = WD_time_step (d);
    NODE_LOOP_START (d, n) {
      WN_assemble_damping (n, d->mass_damping_coeff, dt);
    } NODE_LOOP_END;
  }

  NODE_LOOP_START (d, n) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    WN_get_force (n, &f[0], &f[1], &f[2]);
    iens[0] = ens[0]; iens[1] = ens[1]; iens[2] = ens[2];
    W_newmark_assemble_rhs (d->newmark_integrator, FLES_ADD, 3, iens, f);
  } NODE_LOOP_END;
}

void
WD_pred_conf_newmark (W_domain_t *d)
{
  double beta, gamma;
  W_vect_t disp;
  W_bool_t apply[] = {TRUE, TRUE, TRUE};
  double dt = WD_time_step (d);

  W_newmark_get_consts (d->newmark_integrator, &beta, &gamma);

  NODE_LOOP_START (d, n) {
    WN_get_displ (n, &disp.x, &disp.y, &disp.z);
    WN_set_prev_displ (n, &disp, apply); /* restart value */
    WN_predict_config_Newmark_implicit (n, d, dt, beta, gamma);
  } NODE_LOOP_END;

  DISP_BC_LOOP_START (d, bc) {
    WBC_enforce (bc);
  } DISP_BC_LOOP_END;
}

void
WD_updt_conf_newmark (W_domain_t *d)
{
  double beta, gamma, ic00, ic01, dt = WD_time_step (d);
  W_vect_t du;
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif
  int j;
  
  W_newmark_get_consts (d->newmark_integrator, &beta, &gamma);
  ic00 = 1.0 / dt / dt / beta;
  ic01 = gamma / dt / beta;

  NODE_LOOP_START (d, n) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    du.x = W_newmark_get_sol (d->newmark_integrator, ens[0]);
    du.y = W_newmark_get_sol (d->newmark_integrator, ens[1]);
    du.z = W_newmark_get_sol (d->newmark_integrator, ens[2]);
    WN_update_config_Newmark_implicit (n, d, &du, ic00, ic01);
  } NODE_LOOP_END;
#ifdef X_XFACE                  /* RAW for debugging */
  if (WOPT_OPT (d->options, X_xface) && WOPT_OPT (d->options, interactive_during_outputs)) {
    WD_build_grep (d, FALSE);
    xiface_show_graphics (1, "iteration", WD_time (d));
  }
#endif
}


static void
compute_omega (W_domain_t *d)
{
  double r2, sum_m_r_v, sum_m_r2;
  double x, y, z, vx, vy, vz, omega;
  
  sum_m_r_v = sum_m_r2 = 0;
  NODE_LOOP_START (d, n) {
    WN_get_coords (n, &x, &y, &z);
    WN_get_velocity (n, &vx, &vy, &vz);
    r2 = x*x + y*y + z*z;
    sum_m_r2 += n->mass * r2;
    sum_m_r_v += n->mass * (x * vy - y * vx);
  } NODE_LOOP_END;

  printf ("%g %g\n", WD_time (d), sum_m_r_v / sum_m_r2);
}


W_float_t
WD_finalize_newmark (W_domain_t *d)
{
  if (d->pod) W_pod_accumulate (d->pod);
  update_ext_load_work (d);
  compute_omega (d);
  return WD_time_step (d);
}


int
WD_tot_of_equations (W_domain_t *d)
{
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif
  W_index_t maxeqn;

  maxeqn = 0;
  NODE_LOOP_START (d, n) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    maxeqn = max (maxeqn, ens[0]);
    maxeqn = max (maxeqn, ens[1]);
    maxeqn = max (maxeqn, ens[2]);
  } NODE_LOOP_END;
  return (maxeqn + 1);
}

/*
   Integrate the equations of motion for the domain.  Use implicit (Newmark)
   integration.
 */
void
WD_integrate_implicit (W_domain_t *d /* integrate */
                       )
{
  W_float_t start_time = 0;
  double real_time, usr_time, systm_time;
  W_postfile_times_list_t OL;
  W_bool_t should_flush
#ifdef X_XFACE
    = (WOPT_OPT (d->options, X_xface)                      /* flush for X interface */
       || WOPT_OPT (d->options, time_between_outputs) == 0 /* for forced output each step */
       );
#else
  = 0;
#endif

  OL = WOPT_OPT (d->options, postfile_times_list);
  
  if (d->newmark_integrator == NULL) {
    
    d->newmark_integrator =
      W_newmark_new_integrator (
                                d, 0.0, WD_compute_dt (d),
                                0, 0, /* Newmark constants take default values */
                                (fles_tot_of_equations_callback)
                                WD_tot_of_equations,
                                (W_newmark_assemble_eff_stiff_proc)
                                WD_assemble_eff_stiff_newmark,
                                (W_newmark_residual_force_proc)
                                WD_assemble_residual_forces_newmark,
                                (W_newmark_predict_config_proc)
                                WD_pred_conf_newmark,
                                (W_newmark_update_config_proc)
                                WD_updt_conf_newmark,
                                (W_newmark_finalize_step_proc)
                                WD_finalize_newmark,
                                WOPT_OPT (d->options, newmarki_options)
                                );
    CHECK_MALLOC (d->newmark_integrator);
  }
  assemble_mass (d);
  WD_apply_initial_conditions (d);

  WD_flush_history(d);

  start_time = WD_time (d);

 advance:
  do {
    if (d->check_interrupt) d->check_interrupt (d);
    if (WD_time(d) - start_time >= WOPT_OPT (d->options, time_between_outputs)) {
      start_time = WD_time(d); 
                                          
      if (!WOPT_OPT (d->options, no_output)) {        
        if (WOPT_OPT (d->options, verbose)) {
          W_get_times (&real_time, &usr_time, &systm_time);
          sprintf(W_buf(), "Time=%g (r=%g; u=%g; s=%g)",
                  WD_time(d), real_time, usr_time, systm_time);
	  W_update_log(d->logh, W_buf());
	}
        WHR_update_history(d->history_writer, should_flush);
      }
#ifdef X_XFACE
      if (WOPT_OPT (d->options, X_xface)) {
        WD_build_grep (d, FALSE);
        sprintf(W_buf(), "Time = %g", WD_time(d));
        xiface_show_graphics (WOPT_OPT (d->options, interactive_during_outputs),
                              W_buf(), WD_time (d));
      }
#endif
    }
    if (   (!WOPT_OPT (d->options, no_output))
        && (   (OL.ntimes > 0) && (OL.curr < OL.ntimes))) {
      if ((OL.ntimes > 0) && (WD_time(d) >= OL.times[OL.curr])) {
        WD_write_postfile(d, OL.times[OL.curr]);
        OL.curr++;
      }
    } else if (WOPT_OPT (d->options, write_postfile_file_now)) {
      WD_write_postfile(d, WD_time(d));
      WOPT_SET_OPT (d->options, write_postfile_file_now, FALSE);
    }
    if (d->response_observer_func) (d->response_observer_func) (d, FALSE);
    W_newmark_advance_integrator (d->newmark_integrator);
  } while (WD_time (d) < WOPT_OPT (d->options, target_time));
  
#ifdef X_XFACE
  if (WOPT_OPT (d->options, X_xface)) {
    W_get_times(&real_time, &usr_time, &systm_time);
    sprintf(W_buf(), "Time reached = %g (r=%g; u=%g; s=%g)",
            WD_time(d), real_time, usr_time, systm_time);
    xiface_loop(1, W_buf());
    if (WD_time (d) < WOPT_OPT (d->options, target_time)) goto advance;
  }
#endif

  if (d->response_observer_func) (d->response_observer_func) (d, TRUE);

  WD_clean_up_before_exit(d);

  if (WOPT_OPT (d->options, verbose)) {
    sprintf(W_buf(), "Reached target time = %g", WD_time(d));
    W_update_log(d->logh, W_buf());
    W_get_times(&real_time, &usr_time, &systm_time);
    sprintf(W_buf(), "   Used time [sec]: real %g, user %g, sys %g",
            real_time, usr_time, systm_time);
    W_update_log(d->logh, W_buf());
  }

#ifdef NEED_STAT_MALLOC
  W_stat_malloc_statistics(stderr);
#endif
#undef OL
}


void
WD_set_interrupt_func (W_domain_t *d, W_check_interrupt_func ifunc)
{
  d->check_interrupt = ifunc;
}
