#include "Wconfig.h"
#include "Wxf.h"
#include "Wnewmarki.h"
#include "Woptions.h"
#include "Wpod.h"
#include "Whistry.h"
#include "Wutil.h"

#define DBGPRINT(stuff) 

#define SOLV_SPETSC    0
#define SOLV_DMESCHACH 1
#define SOLV_SMESCHACH 2

#define PODP W_pod_p 

typedef struct history_client_data_t {
  double                            residual_norm_0;
  double                            residual_norm;
  double                            displ_incr_norm_0;
  double                            displ_incr_norm;
  double                            pod_full_residual_norm_0;
  double                            pod_full_residual_norm;
  double                            pod_full_residual_max_term;
  double                            pod_full_residual_proj_error;
  double                            extf_norm_0;
  double                            inertf_norm_0;
  double                            extf_norm;
  double                            inertf_norm;
  int                               niter;
}              history_client_data_t;


typedef struct pod_transf_t {
  W_bool_t                          adaptive;
  int                               nitadapt;
  int                               nmodes; 
  double                           *dof_modes; 
  double                           *coeff;
  int                              *idcs; 
  MAT                              *Phi; 
  MAT                              *lhs; 
  MAT                              *aux; 
  double                           *ser_lhs; 
  int                               full_residual_size; 
  double                           *full_residual; 
  double                           *full_residual_proj; 
  double                            full_residual_norm_0;
  double                            full_residual_norm;
  double                            full_residual_max_term;
  double                            full_residual_proj_error;
  double                            adaptive_residual_norm;
} pod_transf_t;


typedef struct W_newmark_integrator_t {
  W_domain_t                       *domain;
  W_float_t                         t;
  W_float_t                         dt;
  int                               iter;
  int                               max_iter;
  fles_solver_ctx_t                 solver_ctx;
  double                            gamma;
  double                            beta;
  fles_tot_of_equations_callback    tot_of_equations_callback;
  W_newmark_finalize_step_proc      finalize_step;
  W_newmark_assemble_eff_stiff_proc assemble_eff_stiff;
  W_newmark_residual_force_proc     residual_force;
  W_newmark_update_config_proc      update_config;
  W_newmark_predict_config_proc     predict_config;
  double                            extf_norm_0;
  double                            inertf_norm_0;
  double                            extf_norm;
  double                            inertf_norm;
  double                            residual_norm_0;
  double                            residual_norm;
  double                            displ_incr_norm_0;
  double                            displ_incr_norm;
  double                            conv_ratio;
  W_bool_t                          full_newton;
  int                               recomp_keff_after;
  int                               modnewton_after;
  W_bool_t                          use_pod;
  pod_transf_t                      podt;
  W_timer_t                        *ass_rhs_timer; 
  W_timer_t                        *ass_lhs_timer; 
  W_timer_t                        *solve_timer; 
  W_timer_t                        *updt_timer;
  W_timer_t                        *pred_timer;
  W_bool_t                          report_iter;
  W_bool_t                          pass_on_max;
  history_client_data_t             hist_buf;
}              W_newmark_integrator_t;

static void
resize_for_pod (W_newmark_integrator_t *i)
{
  int j;

  i->podt.nmodes = W_pod_get_nmodes (i->domain->pod);

  i->podt.dof_modes = w_new_array (i->podt.nmodes, double);
  CHECK_MALLOC (i->podt.dof_modes);

  i->podt.coeff = w_new_array (i->podt.nmodes, double);
  CHECK_MALLOC (i->podt.coeff);

  i->podt.idcs = w_new_array (i->podt.nmodes, int);
  CHECK_MALLOC (i->podt.idcs);
  FOR (j, i->podt.nmodes) { i->podt.idcs[j] = j; }

  catch(E_MEM,
        { 
          i->podt.Phi = m_get (3, i->podt.nmodes);
        }, {
          REPORT_FATAL_ERROR (i->domain->errh, "Not enough memory");
        });
  catch(E_MEM,
        {
          i->podt.lhs = m_get (i->podt.nmodes, i->podt.nmodes);
        }, {
          REPORT_FATAL_ERROR (i->domain->errh, "Not enough memory");
        });
  catch(E_MEM,
        {
          i->podt.aux = m_get (i->podt.nmodes, i->podt.nmodes);
        }, {
          REPORT_FATAL_ERROR (i->domain->errh, "Not enough memory");
        });

  i->podt.ser_lhs = w_new_array (i->podt.nmodes*i->podt.nmodes, double);
  CHECK_MALLOC (i->podt.ser_lhs);
  i->podt.full_residual_size = i->tot_of_equations_callback (i->domain);
  i->podt.full_residual = w_new_array (i->podt.full_residual_size, double);
  CHECK_MALLOC (i->podt.full_residual);
  i->podt.full_residual_proj = w_new_array (i->podt.full_residual_size, double);
  CHECK_MALLOC (i->podt.full_residual_proj);
}



static void
dump_mat (MAT *m, char *file)
{
  FILE *fp;

  fp = fopen (file, "w");
  if (fp == NULL) {
    return;
  }
  m_foutput (fp, m);
  fclose (fp);
}

static int
error_notify (fles_error_enum error_num,
              int line,
              char *file,
              char *detail,
              void *client)
{
  char buf[1024];
  W_domain_t *d = (W_domain_t *)client;
  W_newmark_integrator_t *i = d->newmark_integrator;
  DBGPRINT (if (error_num == FLES_GENERAL_ERROR) {
    i->solver_ctx->dump_lhs (i->solver_ctx, "lhs");
  });
  sprintf (buf, "%s (l.%d), %s", file, line, detail);
  REPORT_FATAL_ERROR (i->domain->errh, buf);
  return 0;
}


static int
pod_tot_of_equations_callback (W_domain_t *d)
{
  W_newmark_integrator_t *i = d->newmark_integrator;
  return i->podt.nmodes;
}

static void
save_hist (W_newmark_integrator_t *i)
{
  i->hist_buf.niter                        = i->iter;
  i->hist_buf.residual_norm_0              = i->residual_norm_0;
  i->hist_buf.residual_norm                = i->residual_norm;
  i->hist_buf.displ_incr_norm_0            = i->displ_incr_norm_0;
  i->hist_buf.displ_incr_norm              = i->displ_incr_norm;
  i->hist_buf.pod_full_residual_max_term   = i->podt.full_residual_max_term;
  i->hist_buf.pod_full_residual_norm_0     = i->podt.full_residual_norm_0;
  i->hist_buf.pod_full_residual_norm       = i->podt.full_residual_norm;
  i->hist_buf.pod_full_residual_proj_error = i->podt.full_residual_proj_error;
  i->hist_buf.inertf_norm_0                = i->inertf_norm_0;
  i->hist_buf.extf_norm_0                  = i->extf_norm_0;
  i->hist_buf.inertf_norm                  = i->inertf_norm;
  i->hist_buf.extf_norm                    = i->extf_norm;
}


/*
   New Newmark integrator.
 */
W_newmark_integrator_t * /* integrator */
W_newmark_new_integrator (
     void                   *domain, /* domain */
     W_float_t               t0, /* initial time */
     W_float_t               dt, /* time step */
     W_float_t               beta,
     W_float_t               gamma,
     fles_tot_of_equations_callback
                             tot_of_equations_callback,
     W_newmark_assemble_eff_stiff_proc
                             assemble_eff_stiff, /* procedure to assemble
                                                    effective stiffness */
     W_newmark_residual_force_proc
                             residual_force, /* procedure to assemble
                                                    residual force */
     W_newmark_predict_config_proc
                             predict_config, /* procedure to predict configuration */
     W_newmark_update_config_proc
                             update_config, /* procedure to update configuration */
     W_newmark_finalize_step_proc
                             finalize_step, /* procedure to finalize step */
     char *optstring
     )
{  
  W_newmark_integrator_t *i;
  int max_iter = 12;
  double conv_ratio = 0.0001;
  int use_pod = FALSE;
  int adaptive_pod = FALSE;
  int nitadapt = 1;
  fles_tot_of_equations_callback ftec = tot_of_equations_callback;
  int solver = SOLV_DMESCHACH;
  double nz_per_row = 0.01;
  W_bool_t full_newton = TRUE;
  int recomp_keff_after = 1;
  int modnewton_after = 2;
  W_bool_t report_iter = FALSE;
  W_bool_t pass_on_max = FALSE;
  
  /* Create integrator */
  i = w_new  (W_newmark_integrator_t);
  CHECK_MALLOC(i);

  /* Domain */
  i->domain = domain;
  
  /* Select time step dt */
  i->t  = t0;
  i->dt = dt;
  if (gamma <= 0) i->gamma = 0.5;
  else            i->gamma = gamma;
  if (beta <= 0)  i->beta  = 0.25 * (0.5 + i->gamma) * (0.5 + i->gamma);
  else            i->beta  = beta;
  i->tot_of_equations_callback = tot_of_equations_callback;
  i->iter = 0;
  i->residual_norm_0 = 0;
  i->residual_norm = 0;
  i->displ_incr_norm_0 = 0;
  i->displ_incr_norm = 0;
  i->extf_norm = i->extf_norm_0 = 0;
  i->inertf_norm = i->inertf_norm_0 = 0;
  
  if (optstring) {
    int pos;
    tokens_parser parser = tokens_new_parser ();
    if (tokens_parse_line (parser, optstring)) {
      if ((pos = tokens_have_keyword (parser, 1, "POD")) > 0) {
        use_pod = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "ADAPTIVE_POD")) > 0) {
        adaptive_pod = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "NITADAPT")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          nitadapt = tokens_token_as_int (parser, pos+1);
        }
      }
       if ((pos = tokens_have_keyword (parser, 1, "MODNEWTON")) > 0) {
        full_newton = FALSE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "REPORT_ITER")) > 0) {
        report_iter = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "PASS_ON_MAX")) > 0) {
        pass_on_max = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "MAX_ITER")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          max_iter = tokens_token_as_int (parser, pos+1);
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "RECOMP_KEFF_AFTER")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          recomp_keff_after = tokens_token_as_int (parser, pos+1);
        }
        if (recomp_keff_after < 1) recomp_keff_after = 1;
      }
      if ((pos = tokens_have_keyword (parser, 1, "MODNEWTON_AFTER")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          modnewton_after = tokens_token_as_int (parser, pos+1);
        }
        if (modnewton_after < 1) modnewton_after = 1;
      }
      if ((pos = tokens_have_keyword (parser, 1, "GAMMA")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          i->gamma = tokens_token_as_double (parser, pos+1);
          i->beta  = 0.25 * (0.5 + i->gamma) * (0.5 + i->gamma);
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "CONV_RATIO")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          conv_ratio = tokens_token_as_double (parser, pos+1);
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "NZ_PER_ROW")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          nz_per_row = tokens_token_as_double (parser, pos+1);
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "SOLVER")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          const char *solvname;
          solvname = tokens_token_as_string (parser, pos+1);
          if        (tokens_token_and_keyword_equiv (solvname, "SMESCHACH")) {
            solver = SOLV_SMESCHACH;
          } else if (tokens_token_and_keyword_equiv (solvname, "DMESCHACH")) {
            solver = SOLV_DMESCHACH;
          } else if (tokens_token_and_keyword_equiv (solvname, "SPETSC")) {
            solver = SOLV_SPETSC;
          }
        }
      }
    }
    tokens_delete_parser (parser);
  }

  i->max_iter    = max_iter;
  i->conv_ratio  = conv_ratio;
  i->full_newton = full_newton;
  i->recomp_keff_after = recomp_keff_after;
  i->modnewton_after = modnewton_after;
  i->report_iter = report_iter;
  i->pass_on_max = pass_on_max;

  /* Initialize procedures */
  i->assemble_eff_stiff  = assemble_eff_stiff;
  i->update_config       = update_config;
  i->predict_config      = predict_config;
  i->residual_force      = residual_force;
  i->finalize_step       = finalize_step;

  i->use_pod = use_pod;
  i->podt.adaptive                 = adaptive_pod;
  i->podt.nitadapt                 = nitadapt;
  i->podt.nmodes                   = 0;
  i->podt.dof_modes                = NULL;
  i->podt.coeff                    = NULL;
  i->podt.idcs                     = NULL;
  i->podt.Phi                      = NULL;
  i->podt.lhs                      = NULL;
  i->podt.aux                      = NULL;
  i->podt.full_residual_norm_0     = 0;
  i->podt.full_residual_norm       = 0;
  i->podt.full_residual_proj_error = 0;

  /* Initialize solver context */
  if (i->use_pod) {
    resize_for_pod (i);
    if (i->podt.nmodes <= 0) REPORT_FATAL_ERROR (i->domain->errh, "Wrong number of equations");
    ftec = (fles_tot_of_equations_callback) pod_tot_of_equations_callback;
    if (WOPT_OPT (i->domain->options, verbose)) {
      sprintf (W_buf (), "Newmark: Using POD; tot_of_equations = %d", i->podt.nmodes);
      W_update_log (i->domain->logh, W_buf ());
    }
  }

  i->solver_ctx = NULL;
  switch (solver) {
  case SOLV_SMESCHACH:
    i->solver_ctx = fles_new_sparse_meschach_solver (i->domain, ftec, error_notify);
    {
      int method = FLES_SMESCHACH_SPCG;
      i->solver_ctx->set_option (i->solver_ctx, FLES_SMESCHACH_METHOD, (void *)method);
    } 
    break;
#if SOLVER_PETSC
  case SOLV_SPETSC:
    i->solver_ctx = fles_new_spetsc_solver (i->domain,
                                            WOPT_OPT (i->domain->options, argcp),
                                            WOPT_OPT (i->domain->options, argvp),
                                            ftec, error_notify);
    {
      i->solver_ctx->set_option (i->solver_ctx, PETSC_OPTS_NZ_PER_ROW, (void *)&nz_per_row);
    }
    break;
#endif
  case SOLV_DMESCHACH:
  default:
    i->solver_ctx = fles_new_dense_meschach_solver (i->domain, ftec, error_notify);
    break;
  }
  if (i->solver_ctx == NULL) REPORT_FATAL_ERROR (i->domain->errh, "No solver");

  i->ass_rhs_timer = W_timer_new ();
  i->ass_lhs_timer = W_timer_new ();
  i->solve_timer   = W_timer_new ();
  i->updt_timer    = W_timer_new ();
  i->pred_timer    = W_timer_new ();

  save_hist (i);

  return i;
}

static void
get_sol_vect (W_newmark_integrator_t *i)
{
  int j;
  double *coeff = i->podt.coeff;

  if (i->podt.nmodes <= 0) resize_for_pod (i);

  FOR (j, i->podt.nmodes) {
    coeff[j] = i->solver_ctx->get_from_sol (i->solver_ctx, j);
  }
}

static void
solve_for_Delta_u (W_newmark_integrator_t *i)
{
  W_timer_restart (i->solve_timer);
  DBGPRINT (
  i->solver_ctx->dump_lhs (i->solver_ctx, "lhs");
  i->solver_ctx->dump_rhs (i->solver_ctx, "rhs");
  );

  if (!i->solver_ctx->solve (i->solver_ctx))
    REPORT_FATAL_ERROR (i->domain->errh, "Failed in solve_for_Delta_u()");

  DBGPRINT (
  i->solver_ctx->dump_sol (i->solver_ctx, "sol");
  );

  i->displ_incr_norm = i->solver_ctx->sol_norm (i->solver_ctx, FLES_NORM_2);
  
  if (i->use_pod) get_sol_vect (i);
  W_timer_accumulate (i->solve_timer);
}

static double
vdot (double *a, double *b, int n)
{
  int i;
  double r = 0;
  FOR (i, n) { r += a[i] * b[i]; }
  return r;
}

                                /* subtract c = a - b */
static void
vsub (double *c, double *a, double *b, int n)
{
  int i;
  FOR (i, n) { c[i] = a[i] - b[i]; }
}

static void
pod_zero_full_residual (W_newmark_integrator_t *i)
{
  int j, n = i->podt.full_residual_size;
  double *fr = i->podt.full_residual;
  
  if (!fr) REPORT_FATAL_ERROR (i->domain->errh, "Not allocated");
  FOR (j, n) { fr[j] = 0; }
}

static double
pod_full_residual_norm (W_newmark_integrator_t *i)
{
  double *fr = i->podt.full_residual;
  
  if (!fr) REPORT_FATAL_ERROR (i->domain->errh, "Not allocated");
  return sqrt (vdot (fr, fr, i->podt.full_residual_size));
}

static double
pod_full_residual_max_term (W_newmark_integrator_t *i)
{
  double *fr = i->podt.full_residual;
  int j, n = i->podt.full_residual_size;
  double maxterm = 0;
  
  if (!fr) REPORT_FATAL_ERROR (i->domain->errh, "Not allocated");
  FOR (j, n) {
    maxterm = max (fabs (fr[j]), maxterm);
  }
  return maxterm;
}

static void
pod_full_residual_proj_coeffs (W_newmark_integrator_t *i, int how_many, double *a)
{
  PODP pod = i->domain->pod;
  int j, k, n = i->podt.full_residual_size, M = i->podt.nmodes;
  double *fr = i->podt.full_residual;
  double mean, *dof_modes = i->podt.dof_modes;
  
  FOR (k, M) { a[k] = 0; }
  FOR (j, n) { 
    W_pod_dof_modes (pod, j, dof_modes, &mean);
    FOR (k, how_many) { a[k] += fr[j] * dof_modes[k]; }
  }
}

static void
pod_full_residual_project (W_newmark_integrator_t *i, int how_many)
{
  PODP pod = i->domain->pod;
  int j, k, n = i->podt.full_residual_size, M = i->podt.nmodes;
  double *fr = i->podt.full_residual;
  double *pfr = i->podt.full_residual_proj;
  double mean, *a, *dof_modes = i->podt.dof_modes;
#define DEBUG_PROJ 0
  
  if (!fr) REPORT_FATAL_ERROR (i->domain->errh, "Not allocated");
  if (!pfr) {
    pfr = i->podt.full_residual_proj = w_new_array (n, double);
    CHECK_MALLOC (pfr);
  }
  a = w_new_array (M, double); CHECK_MALLOC (a);

  pod_full_residual_proj_coeffs (i, how_many, a);

#if DEBUG_PROJ /* RAW just for debugging */
  FOR (k, M) { printf ("a[%d] = %g\n", k, a[k]); }
#endif

  FOR (j, n) {
    pfr[j] = 0;
    W_pod_dof_modes (pod, j, dof_modes, &mean);
    FOR (k, M) {
      pfr[j] += a[k] * dof_modes[k];
    }
  }

  w_free (a);
}


static double
pod_full_residual_proj_error (W_newmark_integrator_t *i)
{
  double result = 0, frnorm, renorm, *fr, *pfr;
  int n =  i->podt.full_residual_size;
  
  pod_full_residual_project (i, i->podt.nmodes);

  fr  = i->podt.full_residual;
  pfr = i->podt.full_residual_proj;

  frnorm = vdot (fr, fr, n);

  /* difference between the full residual and its projection */
  vsub (pfr, fr, pfr, n);

  renorm = vdot (pfr, pfr, n);

  return sqrt (renorm / frnorm);
}

static int
converged (W_newmark_integrator_t *i)
{
  i->residual_norm = i->solver_ctx->rhs_norm (i->solver_ctx, FLES_NORM_2);
  if (i->use_pod) {
    if (i->podt.adaptive)
      i->residual_norm = i->podt.adaptive_residual_norm;
    i->podt.full_residual_norm = pod_full_residual_norm (i);
    i->podt.full_residual_max_term = pod_full_residual_max_term (i);
    if (i->iter == 1)
      i->podt.full_residual_proj_error = pod_full_residual_proj_error (i);
  }
  if (i->iter == 1) {
    i->residual_norm_0           = i->residual_norm;
    i->displ_incr_norm_0         = i->displ_incr_norm;
    i->podt.full_residual_norm_0 = i->podt.full_residual_norm;
  }
  
  if (i->report_iter) {
    double rrat = (i->residual_norm_0 > 0
                   ? i->residual_norm / i->residual_norm_0
                   : (i->residual_norm == 0
                      ? 1
                      : FLT_MAX
                      ));
    double drat = (i->displ_incr_norm_0 > 0
                   ? i->displ_incr_norm / i->displ_incr_norm_0
                   : (i->displ_incr_norm == 0
                      ? 1
                      : FLT_MAX
                      ));
    double frrat = (i->podt.full_residual_norm_0 > 0
                    ? i->podt.full_residual_norm / i->podt.full_residual_norm_0
                    : 1
                    );

    if (i->use_pod) {
      sprintf (W_buf (), "   %d: |R|/|R0|=%g, |fR|/|fR0|=%g, |fR|/(|Ma|+|P|)=%g",
               i->iter, rrat,
               frrat,
               i->podt.full_residual_norm / (i->extf_norm + i->inertf_norm));
    } else {
      sprintf (W_buf (), "   %d: |R|/|R0|=%g",
               i->iter, rrat);
    }
    W_update_log (i->domain->logh, W_buf ());
  }
  
  if (i->iter >= i->max_iter) {
    if (i->pass_on_max) return TRUE;
    else { REPORT_FATAL_ERROR (i->domain->errh, "Divergent in equilibrium iteration"); return FALSE; }
  } else {
    return ((i->residual_norm == 0 && i->residual_norm_0 == 0) ||
            i->residual_norm / i->residual_norm_0 <= i->conv_ratio
            );
  }
}

static void
assemble_k_eff (W_newmark_integrator_t *i)
{
  int first_n = i->modnewton_after;
  int fn1 = first_n - 1;
  if (i->full_newton ||
      (                         /* modified Newton-Raphson */
       i->iter < fn1 ||           /* first time around */
       ((i->iter >= fn1) &&         /* first n-iterations with full Newton */
        (i->iter  - fn1) / i->recomp_keff_after * i->recomp_keff_after
        == (i->iter - fn1))
       )
      ) {
    W_timer_restart (i->ass_lhs_timer);
    i->solver_ctx->zero_lhs (i->solver_ctx);
    i->assemble_eff_stiff (i->domain);
    i->solver_ctx->finish_lhs (i->solver_ctx, FLES_FINAL_FINISH);
    W_timer_accumulate (i->ass_lhs_timer);
  }
}

static void
update_config (W_newmark_integrator_t *i)
{
  W_timer_restart (i->updt_timer);
  i->update_config (i->domain);
  W_timer_accumulate (i->updt_timer);
}

static void
predict_config (W_newmark_integrator_t *i)
{
  W_timer_restart (i->pred_timer);
  i->predict_config (i->domain);
  W_timer_accumulate (i->pred_timer);
}

static void
pod_init_advance (W_newmark_integrator_t *i)
{
  PODP pod = i->domain->pod;
  
  if (i->podt.adaptive)
    W_pod_revert_to_orig_modes (pod);
  W_pod_rotate (pod); 
}

static void
pod_start_residual_assembly (W_newmark_integrator_t *i)
{
  PODP pod = i->domain->pod;
  
  pod_zero_full_residual (i);
#if 0
  if (i->podt.adaptive && i->iter < i->podt.nitadapt)
    W_pod_revert_to_orig_modes (pod);
#endif
}

static void
pod_finish_residual_assembly (W_newmark_integrator_t *i)
{
  PODP pod = i->domain->pod;
  
  if (i->podt.adaptive && i->iter < i->podt.nitadapt) {
    int n =  i->podt.full_residual_size;
    double *fr, *pfr;
    pod_full_residual_project (i, i->podt.nmodes - 1);
    fr  = i->podt.full_residual;
    pfr = i->podt.full_residual_proj;
    vsub (pfr, fr, pfr, n);
    W_pod_include_mode (pod, i->podt.nmodes - 1, pfr);
    W_pod_orthonormalize (pod);
  }
  pod_full_residual_proj_coeffs (i, i->podt.nmodes, i->podt.coeff);
  if (i->podt.adaptive) {
    i->podt.adaptive_residual_norm
      = sqrt (vdot (i->podt.coeff, i->podt.coeff, i->podt.nmodes - 1));
  }
  i->solver_ctx->assemble_vect (i->solver_ctx, FLES_ADD,
                                i->podt.nmodes, i->podt.idcs, i->podt.coeff);
}

static void
assemble_R (W_newmark_integrator_t *i, W_mat_state_update_form_enum update_form)
{
  W_timer_restart (i->ass_rhs_timer);

  i->solver_ctx->zero_rhs (i->solver_ctx);

  if (i->use_pod) pod_start_residual_assembly (i);
  i->residual_force (i->domain, update_form, &i->extf_norm, &i->inertf_norm);
  if (i->use_pod) pod_finish_residual_assembly (i);

  if (i->iter < 1) {
    i->extf_norm_0 = i->extf_norm;
    i->inertf_norm_0 = i->inertf_norm;
  }
  i->solver_ctx->finish_rhs (i->solver_ctx);
  W_timer_accumulate (i->ass_rhs_timer);
}

static void
finalize_step (W_newmark_integrator_t *i)
{
  if (i->finalize_step)
    i->finalize_step (i->domain);
}

/*
   Advance the integrator in time by a single time step.
 */
void
W_newmark_advance_integrator (W_newmark_integrator_t *i /* integrator */)
{
  i->iter = 0;
  i->t += i->dt;

  if (i->use_pod) pod_init_advance (i);
  predict_config (i);
  do {
    assemble_R (i, W_MAT_STATE_UPDATE_TEMP);
    assemble_k_eff (i);
    solve_for_Delta_u (i);
    update_config (i);
    i->iter++;
  } while (!converged (i));
  assemble_R (i, W_MAT_STATE_UPDATE_FINAL);

  save_hist (i);

  finalize_step (i);
}



/*
   Get the current time from the integrator.
 */
W_float_t /* time */
W_newmark_integrator_time (W_newmark_integrator_t *i/* integrator */)
{
  return i->t;
}



/*
   Get the time step from the integrator.
 */
W_float_t
W_newmark_integrator_dt (W_newmark_integrator_t *i/* integrator */)
{
  return i->dt;
}


static void
build_Phi (W_newmark_integrator_t *i,
           int nr, int *ri
           )
{
  int j, k;
  double mean;
  PODP pod = i->domain->pod;

  if (i->podt.nmodes <= 0) resize_for_pod (i);
  
  i->podt.Phi = m_resize (i->podt.Phi, nr, i->podt.nmodes);

  FOR (j, nr) {
    W_pod_dof_modes (pod, ri[j], i->podt.dof_modes, &mean);
    FOR (k, i->podt.nmodes) { i->podt.Phi->me[j][k] = i->podt.dof_modes[k]; }
  }
}


static void
serialize (MAT *m, double *v, int n)
{
  int i, j, k;
  k = 0;
  FOR (i, n) {
    FOR (j, n) {
      v[k] = m->me[i][j];
      k++;
    }
  }
}

static void
deserialize (double *v, MAT *m, int n)
{
  int i, j, k;
  k = 0;
  FOR (i, n) {
    FOR (j, n) {
      m->me[i][j] = v[k];
      k++;
    }
  }
}

static void
assemble_lhs_pod (W_newmark_integrator_t *i,
                  int nr, int *ri,
                  double *values
                  )
{
  int j, k, l;

  build_Phi (i, nr, ri);

  m_resize (i->podt.lhs, nr, nr);
  deserialize (values, i->podt.lhs, nr);

  mtrm_mlt (i->podt.Phi, i->podt.lhs, i->podt.aux);

  m_mlt (i->podt.aux, i->podt.Phi, i->podt.lhs);

  serialize (i->podt.lhs, i->podt.ser_lhs, i->podt.nmodes);

  i->solver_ctx->assemble_mat (i->solver_ctx,
                               FLES_FALSE, FLES_ADD,
                               i->podt.nmodes, i->podt.idcs, i->podt.nmodes, i->podt.idcs,
                               i->podt.ser_lhs);
}


static void
assemble_lhs_full (W_newmark_integrator_t *i,
                   fles_update_mode_enum update_mode,
                   int nr, int *ri,
                   double *values, W_bool_t *constrained
                   )
{
  int j, k, l;
  W_bool_t need_to_rewrite = FALSE; /* do I need to modify the element matrix
                                       because of some constraints?
                                     */

  FOR (j, nr) { if (constrained[j]) { need_to_rewrite = TRUE; break; } }

  if (need_to_rewrite) {
    l = 0;
    FOR (j, nr) {
      FOR (k, nr) {
        if (j != k && (constrained[j] || constrained[k])) {
          values[l] = 0;
        }
        l++;
      }
    }
  }
  
  i->solver_ctx->assemble_mat (i->solver_ctx,
                               FLES_FALSE, update_mode, nr, ri, nr, ri, values);
}

/*
  Assemble submatrix of the LHS.
 */
void
W_newmark_assemble_lhs (W_newmark_integrator_t *i,
                        fles_update_mode_enum update_mode,
                        int nr, int *ri,
                        double *values,
                        W_bool_t *constrained
                        )
{
  if (i->use_pod) {
    assemble_lhs_pod (i, nr, ri, values);
  } else {
    assemble_lhs_full (i, update_mode, nr, ri, values, constrained);
  }
}

static void
assemble_rhs_pod (W_newmark_integrator_t *i,
                        int nr, int *ri,
                        double *values
                        )
{
  int k;

  FOR (k, nr) { /* assemble the full residual */
    i->podt.full_residual[ri[k]] += values[k];
  }
}

/*
  Assemble subvector of the RHS.
 */
void
W_newmark_assemble_rhs (W_newmark_integrator_t *i,
                        fles_update_mode_enum update_mode,
                        int nr, int *ri,
                        double *values
                        )
{
  if (!i->use_pod) 
    i->solver_ctx->assemble_vect (i->solver_ctx, update_mode, nr, ri, values);
  else {
    assemble_rhs_pod (i, nr, ri, values);
  }
}

void
W_newmark_get_consts (W_newmark_integrator_t *i, double *beta, double *gamma
                      )
{
  *beta = i->beta;
  *gamma = i->gamma;
}

static double 
get_pod_sol (W_newmark_integrator_t *i, W_index_t j)
{
  int k, M;
  double dofv = 0;
  PODP pod = i->domain->pod;
  pod_transf_t *podt = &i->podt;
  double mean, *dof_modes = podt->dof_modes, *coeff = podt->coeff;

  if (i->podt.nmodes <= 0) resize_for_pod (i);
  M = podt->nmodes;
  
  W_pod_dof_modes (pod, j, dof_modes, &mean);
  dofv = 0;
  FOR (k, M) { dofv += dof_modes[k] * coeff[k]; }
  return dofv;
}

W_float_t 
W_newmark_get_sol (W_newmark_integrator_t *i, W_index_t j)
{
  if (!i->use_pod)
    return i->solver_ctx->get_from_sol (i->solver_ctx, j);
  else
    return get_pod_sol (i, j);
}

void
W_newmark_stat (W_newmark_integrator_t *i,
                double *ass_rhs_time,
                double *ass_lhs_time, 
                double *solve_time, 
                double *pred_time,
                double *updt_time
                )
{
  *ass_rhs_time = W_timer_get_time (i->ass_rhs_timer);
  *ass_lhs_time = W_timer_get_time (i->ass_lhs_timer);
  *solve_time   = W_timer_get_time (i->solve_timer);
  *pred_time    = W_timer_get_time (i->pred_timer);
  *updt_time    = W_timer_get_time (i->updt_timer);
}

void
W_newmark_constrain (W_newmark_integrator_t *i, int row)
{
  double one = 1;
  if (!i->use_pod) {
    i->solver_ctx->finish_lhs (i->solver_ctx, FLES_FINAL_FINISH);
    i->solver_ctx->zero_mat_row (i->solver_ctx, row);
    i->solver_ctx->assemble_mat (i->solver_ctx,
                                 FLES_FALSE, FLES_ADD,
                                 1, &row, 1, &row, &one);
  }
}


/*
 Update history.
 Update history by storing the current values of integrator data.
 */
void
W_newmark_update_history (W_newmark_integrator_t *i,
                          W_history_writer_t *hw
                          )
{
  history_client_data_t *cd = &i->hist_buf;
  double time = i->t;
  WHR_store (hw, "nmki-R0", time, cd->residual_norm_0);
  WHR_store (hw, "nmki-R", time, cd->residual_norm);
  WHR_store (hw, "nmki-d0", time, cd->displ_incr_norm_0);
  WHR_store (hw, "nmki-d", time, cd->displ_incr_norm);
  WHR_store (hw, "nmki-fR0", time, cd->pod_full_residual_norm_0);
  WHR_store (hw, "nmki-fR", time, cd->pod_full_residual_norm);
  WHR_store (hw, "nmki-fRmaxt", time, cd->pod_full_residual_max_term);
  WHR_store (hw, "nmki-fRproje", time, cd->pod_full_residual_proj_error);
  WHR_store (hw, "nmki-niter", time, cd->niter);
  WHR_store (hw, "nmki-extf", time, cd->extf_norm);
  WHR_store (hw, "nmki-inertf", time, cd->inertf_norm);
  if (cd->pod_full_residual_norm_0 > 0) {
    WHR_store (hw, "nmki-fR:fR0", time,
               cd->pod_full_residual_norm / cd->pod_full_residual_norm_0);
    WHR_store (hw, "nmki-fR:IP0", time,
               cd->pod_full_residual_norm / (cd->extf_norm_0 + cd->inertf_norm_0));
  }
  if ((cd->extf_norm + cd->inertf_norm) > 0 &&
      i->podt.full_residual_size > 0)
    WHR_store (hw, "nmki-fR:NIP", time,
               cd->pod_full_residual_norm / (cd->extf_norm + cd->inertf_norm) /
               i->podt.full_residual_size);
} 

