#include "Wconfig.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Woptions.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wloads.h"
#include "Wdispbc.h"
#include "Wivcond.h"
#include "Wmats.h"
#include "Wcontact.h"
#include "Wvset.h"
#include "Wpvm.h"
#include "Wmpi.h"
#include "Wcdi.h"
#include "Wnewmarki.h"
#include "Wpod.h"
#include "Wmat3dutil.h"

#define REPLACE FALSE
#define ADD     TRUE


typedef void (*compute_modes_func) (W_pod_p pod);

typedef enum W_pod_method_t { POD_DIRECT = 0, POD_SNAPSHOTS = 1 } W_pod_method_t;

typedef struct W_pod_collect_t {
  double              dt;     /* time interval between data collections */
  double              from_t; /* collect data from this time ... */
  double              up_to_t;/* ... up to this time */
  double              last_t; /* last time the data was collected */
  W_index_t           toteqn; /* dimension of the below array */
  W_pod_method_t      method; /* collection method */
  LIST_OF (double *)  ul;            /* list of arrays; one array per snapshot */
  double             *umean;
  MAT                *evecs;         /* eigen-modes of R */
  VEC                *evals;         /* eigenvalues of R */
  double              cutoff_ev;
  compute_modes_func  compute_modes;
}              W_pod_collect_t;


typedef struct mode_mat_t {
  struct W_pod_t     *pod;
  int                 nm;
  double            **dof_modes;
}              mode_mat_t;

typedef struct W_pod_use_t {
  int                 max_nm;
  int                 dof_modes_dim;
  int                 curr_m;
  LIST_OF (int)       modenl;
  dict_dictionary     node_transf_matd;
  mode_mat_t         *mm;
  mode_mat_t         *mm_orig;
}              W_pod_use_t;

typedef struct W_pod_t {
  W_domain_t         *domain;        /* domain */
  W_pod_collect_t    *collect;
  W_pod_use_t        *use;
}              W_pod_t;


static mode_mat_t *
mm_new (W_pod_t *pod)
{
  W_pod_use_t *u = pod->use;
  mode_mat_t *mm;
  int i, j;
  
  mm = w_new (mode_mat_t); CHECK_MALLOC (mm);
  mm->pod = pod;
  mm->nm  = u->max_nm;
  mm->dof_modes = w_new_array (u->dof_modes_dim, double *);
  CHECK_MALLOC (mm->dof_modes);
  FOR (i, u->dof_modes_dim) {
    mm->dof_modes[i] = w_new_array (u->max_nm + 1, double); 
    CHECK_MALLOC (mm->dof_modes[i]);
    FOR (j, u->max_nm+1) { mm->dof_modes[i][j] = 0; }
  }

  return mm;
}

static void
mm_init (mode_mat_t *mm)
{
  W_pod_use_t *u = mm->pod->use;
  int i, j, k;
  W_node_t *n;
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif
  MAT *ntm;
  
  n = NULL;
  while (n = WD_get_next_node (mm->pod->domain, n)) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    ntm = dict_find_obj_with_int_name (mm->pod->use->node_transf_matd, n->id);
    if (!ntm) {
      REPORT_FATAL_ERROR (mm->pod->domain->errh,
                          "Nodal transformation matrix not found");
    }
    FOR (k, 3) {
      i = ens[k];
      FOR (j, u->max_nm+1) { mm->dof_modes[i][j] = ntm->me[k][j]; }
    }
  }
  mm->nm = u->max_nm;
}

static void
mm_rotate (mode_mat_t *mm, double t[3][3])
{
  W_pod_use_t *u = mm->pod->use;
  int i, j, k;
  W_node_t *n;
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif
  double v[3], tv[3];
  double nm1 = u->max_nm+1;
  
  n = NULL;
  while (n = WD_get_next_node (mm->pod->domain, n)) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    FOR (j, nm1) { 
      FOR (k, 3) {
        i = ens[k];
        v[k] = mm->dof_modes[i][j];
      }
      MAT3DUTIL_MULT_A_3X3_V_3 (tv, t, v);
      FOR (k, 3) {
        i = ens[k];
        mm->dof_modes[i][j] = tv[k];
      }
    }
  }
}

static void
mm_copy (mode_mat_t *mm_dest, mode_mat_t *mm_src)
{
  int i, j;
  int n, m;
  n = mm_dest->pod->use->dof_modes_dim;
  m = mm_dest->pod->use->max_nm + 1;
  FOR (i, n) {
    FOR (j, m) { mm_dest->dof_modes[i][j] = mm_src->dof_modes[i][j]; }
  }
}

static void
mm_dump (mode_mat_t *mm, FILE *fp)
{
  int i, j;
  int n, m;
  n = mm->pod->use->dof_modes_dim;
  m = mm->pod->use->max_nm + 1;
  FOR (j, m) {
    fprintf (fp, "mode %d\n", j);
    FOR (i, n) {
      fprintf (fp, "%d: %g\n", i, mm->dof_modes[i][j]);
    }
  }
}

static void
mm_set (mode_mat_t *mm, int which, double *v)
{
  int i, n;
  double **dof_modes = mm->dof_modes;
  n = mm->pod->use->dof_modes_dim;
  FOR (i, n) {
    dof_modes[i][which] = v[i];
  }
}

static double
mm_dot (mode_mat_t *mm, int m0, int m1)
{
  int i, n;
  double d = 0;
  double **dof_modes = mm->dof_modes;
  n = mm->pod->use->dof_modes_dim;
  FOR (i, n) {
    d += dof_modes[i][m0] * dof_modes[i][m1];
  }
  return d;
}

static void
mm_subtract (mode_mat_t *mm, int which, double d, int from)
{
  int i, n;
  double **dof_modes = mm->dof_modes;
  n = mm->pod->use->dof_modes_dim;
  FOR (i, n) {
    dof_modes[i][from] -= d * dof_modes[i][which];
  }
}

static void
mm_normalize (mode_mat_t *mm, int which)
{
  int i, n;
  double d = 0;
  double **dof_modes = mm->dof_modes;
  n = mm->pod->use->dof_modes_dim;
  d = mm_dot (mm, which, which);
  d = sqrt (d);
  FOR (i, n) {
    dof_modes[i][which] /= d;
  }
}

static void
mm_orthonormalize (mode_mat_t *mm)
{
  int i, j, k, nm = mm->nm;
  double d;
  j = 0;
  mm_normalize (mm, 0);
  for (i = 1; i < nm; i++) {
    for (k = i; k < nm; k++) {
      d = mm_dot (mm, k, j);
      mm_subtract (mm, j, d, k);
    }
    j++;
    mm_normalize (mm, j);
  }
}

static double
dot (double *a, double *b, int M)
{
  int k;
  double accum = 0;
  FOR (k, M) { accum += a[k] * b[k]; }
  return accum;
}

static int
transfmat (W_domain_t *d, double t[3][3])
{
  W_node_t *n;
  double x, y, z, r, ca, sa;

  n = WD_get_node_by_idx (d, 1); /* node #1 */
  if (!n) REPORT_FATAL_ERROR (d->errh, "Node not found");
  WN_get_coords (n, &x, &y, &z);
  r = sqrt (x*x + y*y + z*z); 
  ca = x / r;
  sa = y / r;
  t[0][0] = ca;
  t[0][1] = sa;
  t[0][2] = 0;
  t[1][0] = -sa;
  t[1][1] = ca;
  t[1][2] = 0;
  t[2][0] = 0;
  t[2][1] = 0;
  t[2][2] = 1;
  return 1;
}

static void
accumulate (W_pod_p pod)
{
  W_node_t *n;
  double x[3], X[3], tx[3], d[3];
  int i;
  W_index_t toteqn = WD_tot_of_equations (pod->domain);
  double *v;
  W_pod_collect_t *c = pod->collect;
  double t[3][3];

  if (c->toteqn == 0) c->toteqn = toteqn;
  if (c->toteqn != toteqn) REPORT_FATAL_ERROR (pod->domain->errh, "Change in the number of nodes");

  if (c->ul == NULL) {
    c->ul = make_list (); CHECK_MALLOC (c->ul);
  }

  transfmat (pod->domain, t);
  /* accumulate array at this snapshot */
  v = w_new_array (c->toteqn, double); CHECK_MALLOC (v);
  add_to_tail (c->ul, v);
  i = 0;
  n = NULL;
  while (n = WD_get_next_node (pod->domain, n)) {
    if (i+3 > c->toteqn) REPORT_FATAL_ERROR (pod->domain->errh, "Wrong numbering");
    WN_get_coords (n, &x[0], &x[1], &x[2]);
    WN_get_ref_coords (n, &X[0], &X[1], &X[2]);
    MAT3DUTIL_MULT_A_3X3_V_3 (tx, t, x);
    d[0] = tx[0] - X[0];
    d[1] = tx[1] - X[1];
    d[2] = tx[2] - X[2];
    v[i] = d[0]; i++;
    v[i] = d[1]; i++;
    v[i] = d[2]; i++;
  }
}


void
W_pod_accumulate (W_pod_p pod)
{
  double t = WD_time (pod->domain);

  if (pod->collect != NULL) { 
    if (t >= pod->collect->from_t && t <= pod->collect->up_to_t &&
        t >= pod->collect->last_t + pod->collect->dt 
        ) {
      accumulate (pod);
      pod->collect->last_t = t;
    }
  }
}


typedef struct pod_mode_data_t {
  int     mode;
  double  eval;
} pod_mode_data_t;

static int
compare_modes (pod_mode_data_t *pm0, pod_mode_data_t *pm1)
{
  return (pm0->eval >= pm1->eval);
}

static void
subtract_mean (W_pod_p pod)
{
  int i;
  double one_over_M;
  double *u;
  W_pod_collect_t *c = pod->collect;
  int toteqn = c->toteqn;
  int M = list_length (c->ul);

  one_over_M = 1. / M;

  c->umean = w_new_array (toteqn, double); CHECK_MALLOC (c->umean);
  FOR (i, toteqn) { c->umean[i] = 0; }

  loop_through_list (c->ul, u, double *) {
    FOR (i, toteqn) { c->umean[i] += one_over_M * u[i]; }
  }

  loop_through_list (c->ul, u, double *) {
    FOR (i, toteqn) { u[i] -= c->umean[i]; }
  }
}

static void
make_mode_MoS (W_pod_collect_t *c, int mode, double *u)
{
  int i, j, M = c->evals->dim;
  double aj, *uj;
  double sum = 0;

  FOR (i, c->toteqn) { u[i] = 0; }
  
  FOR (j, M) {
    uj = (double *) get_list_pos (c->ul, j+1);
    if (!uj) REPORT_FATAL_ERROR (NULL, "No data");
    aj = c->evecs->me[j][mode];
    FOR (i, c->toteqn) {
      u[i] += aj * uj[i];
    }
  }
  
  sum = 0;
  FOR (i, c->toteqn) {
    sum += u[i] * u[i];
  }
  sum = sqrt (sum);
  if (sum <= 0 || isnan (sum)) sum = 1;
  FOR (i, c->toteqn) {
    u[i] /= sum;
  }
}

#if 0
#define CHECK_ORTHONORMAL(i,j,d)                                        \
  if ((i == j) && (fabs (d - 1) > 1e-6)) {                              \
    max_d = max (max_d, fabs (d - 1));                                  \
    all_orthonormal = FALSE;                                            \
    sprintf (W_buf (), "Mode %d not normalized: %g", i, d);             \
    REPORT_WARNING (pod->domain->errh, W_buf ());                       \
  }                                                                     \
  if ((i != j) && (fabs (d) > 1e-6)) {                                  \
    max_d = max (max_d, fabs (d));                                      \
    all_orthonormal = FALSE;                                            \
    sprintf (W_buf (), "Modes %d and %d not orthonormal: %g", i, j, d); \
    REPORT_WARNING (pod->domain->errh, W_buf ());                       \
  }
#else
#define CHECK_ORTHONORMAL(i,j,d)                                        \
  if ((i == j) && (fabs (d - 1) > 1e-6)) {                              \
    max_d = max (max_d, fabs (d - 1));                                  \
    all_orthonormal = FALSE;                                            \
  }                                                                     \
  if ((i != j) && (fabs (d) > 1e-6)) {                                  \
    max_d = max (max_d, fabs (d));                                      \
    all_orthonormal = FALSE;                                            \
  }
#endif

static void
check_orthonorm_c (W_pod_p pod)
{
  W_pod_collect_t *c = pod->collect;
  int i, j, k, M = c->evals->dim;
  MAT *evecs = c->evecs;
  double d;
  W_bool_t all_orthonormal = TRUE;
  double max_d = 0;
  
  FOR (i, M) {
    FOR (j, M) {
      if (c->evals->ve[i] > c->cutoff_ev && c->evals->ve[j] > c->cutoff_ev) {
        d = 0;
        FOR (k, c->toteqn) {
          d += evecs->me[k][i] * evecs->me[k][j];
        }
        CHECK_ORTHONORMAL (i, j, d);
      }
    }
  }
  if (!all_orthonormal) {
    sprintf (W_buf (), "Some modes not orthonormal; max mismatch %g", max_d);
    REPORT_WARNING (pod->domain->errh, W_buf ());
  }
}

static void
order_modes_MoS (W_pod_p pod)
{
  LIST ml = make_list ();
  pod_mode_data_t *pm;
  int i, j;
  MAT *new_evecs;
  W_pod_collect_t *c = pod->collect;
  int M = c->evals->dim;
  double *u;

  FOR (i, M) {
    pm = w_new (pod_mode_data_t); CHECK_MALLOC (pm);
    pm->mode = i;
    pm->eval = c->evals->ve[i];
    add_to_tail (ml, pm);
  }

  sort_list_func (ml, compare_modes);

  W_update_log (pod->domain->logh, "Method of Snaphots POD");
  new_evecs = m_get (c->toteqn, M);
  m_zero (new_evecs);
  u = w_new_array (c->toteqn, double); CHECK_MALLOC (u);
  i = 0;
  loop_through_list (ml, pm, pod_mode_data_t *) {
    sprintf (W_buf (), "   Mode %d, eigenvalue %g", i, pm->eval);
    W_update_log (pod->domain->logh, W_buf ());
    make_mode_MoS (c, pm->mode, u);
    FOR (j, c->toteqn) { /* set eigenvector */
      new_evecs->me[j][i] = u[j];
    }
    c->evals->ve[i] = pm->eval; /* eigenvalue */
    i++;
#   if 0 /* RAW check of orthonormality */
    {
      int k, m, n;
      double d;
      d = 0;

      FOR (m, i) {
        FOR (n, i) {
          d = 0;
          FOR (k, c->toteqn) {
            d += new_evecs->me[k][m] * new_evecs->me[k][n];
          }
          if (m == n && fabs (d - 1) > 1e-6 ||
              m != n && fabs (d) > 1e-6)
            printf ("m=%d, n=%d, d=%g\n", m, n, d);
        }
      }
    }
#   endif
  }
  w_free (u);

  m_free (c->evecs);
  c->evecs = new_evecs;

  free_list (ml, AND_NODES);
}


static void
compute_modes_MoS (W_pod_p pod)
{
  int i, j, n;
  double one_over_M;
  double *ui, *uj;
  W_pod_collect_t *c = pod->collect;
  int M = list_length (c->ul);
  MAT *C;

  if (M == 0) { REPORT_FATAL_ERROR (pod->domain->errh, "No collected data"); }

  one_over_M = 1.0 / (double) M;
  
  C = m_get (M, M);
  m_zero (C);

  n = c->toteqn;
  FOR (i, M) {
    ui = (double *) get_list_pos (c->ul, i+1);
    if (ui == NULL) REPORT_FATAL_ERROR (pod->domain->errh, "No such data");
    FOR (j, M) {
      uj = (double *) get_list_pos (c->ul, j+1);
      if (uj == NULL) REPORT_FATAL_ERROR (pod->domain->errh, "No such data");
      C->me[i][j] += one_over_M * dot (ui, uj, n);
    }
  }

  c->evecs = m_resize (c->evecs, M, M); CHECK_MALLOC (c->evecs);
  c->evals = v_resize (c->evals, M); CHECK_MALLOC (c->evals);
#if 0
  {
    MAT *V;
    V = m_get (M, M);
    m_ident (V);
    m_ident (c->evecs);
    svd (C, V, c->evecs, c->evals);
    m_transp (c->evecs, c->evecs);
    m_free (V);
  }
#else
  symmeig (C, c->evecs, c->evals);
#endif
  m_free (C);

  order_modes_MoS (pod);
}




static void
order_modes_DM (W_pod_p pod)
{
  LIST ml = make_list ();
  pod_mode_data_t *pm;
  int i, j;
  W_pod_collect_t *c = pod->collect;
  int M = c->evals->dim;
  MAT *new_evecs;

  FOR (i, M) {
    pm = w_new (pod_mode_data_t); CHECK_MALLOC (pm);
    pm->mode = i;
    pm->eval = c->evals->ve[i];
    add_to_tail (ml, pm);
  }

  sort_list_func (ml, compare_modes);

  W_update_log (pod->domain->logh, "Direct Method of POD");
  new_evecs = m_get (c->evecs->m, c->evecs->m);
  i = 0;
  loop_through_list (ml, pm, pod_mode_data_t *) {
    sprintf (W_buf (), "   Mode %d, eigenvalue %g", i, pm->eval);
    W_update_log (pod->domain->logh, W_buf ());
    FOR (j, M) {
      new_evecs->me[j][i] = c->evecs->me[j][pm->mode];
    }
    c->evals->ve[i] = pm->eval;
    i++;
  }

  m_free (c->evecs);
  c->evecs = new_evecs;

  free_list (ml, AND_NODES);
}



static void
compute_modes_DM (W_pod_p pod)
{
  int i, j, k;
  double one_over_M;
  double *uk;
  W_pod_collect_t *c = pod->collect;
  int toteqn = c->toteqn;
  int M = list_length (c->ul);
  MAT *C;

  if (M == 0) { REPORT_FATAL_ERROR (pod->domain->errh, "No collected data"); }

  one_over_M = 1.0 / (double) M;
  
  C = m_get (toteqn, toteqn);
  m_zero (C);
  
  FOR (k, M) {
    uk = (double *) get_list_pos (c->ul, k+1);
    if (uk == NULL) REPORT_FATAL_ERROR (pod->domain->errh, "No such data");
    FOR (i, toteqn) {
      FOR (j, toteqn) {
        C->me[i][j] += one_over_M * uk[i] * uk[j];
      }
    }
  }

  c->evecs = m_resize (c->evecs, toteqn, toteqn); CHECK_MALLOC (c->evecs);
  c->evals = v_resize (c->evals, toteqn); CHECK_MALLOC (c->evals);
  symmeig (C, c->evecs, c->evals);
  m_free (C);

  FOR (i, toteqn) { /* normalize */
    double sum = 0;
    FOR (j, toteqn) {
      sum += c->evecs->me[j][i]*c->evecs->me[j][i];
    }
    sum = sqrt (sum);
    if (sum <= 0 || isnan (sum)) sum = 1;
    FOR (j, toteqn) {
      c->evecs->me[j][i] /= sum;
    }      
  }

  order_modes_DM (pod);
}


void
W_pod_compute_modes (W_pod_p pod)
{
  if (pod->collect != NULL) {
    subtract_mean (pod);
    pod->collect->compute_modes (pod);
    check_orthonorm_c (pod);
  }
}


static void
set_mean (W_pod_p pod, W_bool_t add, double scale)
{
  W_node_t *n;
  W_vect_t d;
  W_bool_t apply[3] = {TRUE, TRUE, TRUE};
  int i;
  W_pod_collect_t *c = pod->collect;

  if (c->umean == NULL){REPORT_WARNING (pod->domain->errh, "No mean");return;}

  i = 0;
  n = NULL;
  while (n = WD_get_next_node (pod->domain, n)) {
    VECMAC_SCALAR (d, =, 0);
    if (add) {
      WN_get_displ (n, &d.x, &d.y, &d.z);
      d.x  += scale * c->umean[i+0];
      d.y  += scale * c->umean[i+1];
      d.z  += scale * c->umean[i+2];
    } else {
      d.x  = scale * c->umean[i+0];
      d.y  = scale * c->umean[i+1];
      d.z  = scale * c->umean[i+2];
    }
    i += 3;
    WN_set_displ (n, &d, apply);
  }
}


static void
set_mode (W_pod_p pod, int mode, W_bool_t add, double scale)
{
  W_node_t *n;
  W_vect_t d;
  W_bool_t apply[3] = {TRUE, TRUE, TRUE};
  int i, j;
  W_pod_collect_t *c = pod->collect;

  if (c->evecs == NULL){
    REPORT_WARNING (pod->domain->errh, "No component modes matrix"); return;
  }
  if (mode >= c->evals->dim)  {
    REPORT_WARNING (pod->domain->errh, "No such mode"); return;
  }
  if (mode < 0) mode = 0;

  i = 0;
  n = NULL;
  while (n = WD_get_next_node (pod->domain, n)) {
    if (i+3 > c->toteqn) REPORT_FATAL_ERROR (pod->domain->errh, "Wrong numbering");
    if (add) {
      WN_get_displ (n, &d.x, &d.y, &d.z);
      d.x += scale * c->evecs->me[i+0][mode];
      d.y += scale * c->evecs->me[i+1][mode];
      d.z += scale * c->evecs->me[i+2][mode];
    } else {
      d.x = scale * c->evecs->me[i+0][mode];
      d.y = scale * c->evecs->me[i+1][mode];
      d.z = scale * c->evecs->me[i+2][mode];
    }
    WN_set_displ (n, &d, apply);
    i += 3;
  }
}


#ifdef X_XFACE
void
W_pod_show_mode (W_pod_p pod, int mode)
{
  double displ_scale = 1;
  if (pod->collect == NULL) { REPORT_WARNING (pod->domain->errh, "Not in collect mode"); return; }
  displ_scale = WOPT_OPT (pod->domain->options, displ_scale);
  set_mean (pod, REPLACE, 0);
  set_mode (pod, mode, ADD, displ_scale);
  WD_build_grep (pod->domain, TRUE);
}
#endif

double
W_pod_eigenvalue (W_pod_p pod, int mode)
{
  if (pod->collect == NULL) {
    REPORT_WARNING (pod->domain->errh, "Not in collect mode"); return 0;
  }
  if (pod->collect->evecs == NULL) {
    REPORT_WARNING (pod->domain->errh, "No component modes"); return 0;
  }
  if (mode >= pod->collect->evals->dim)  {
    REPORT_WARNING (pod->domain->errh, "No such mode"); return 0;
  }
  if (mode < 0) mode = 0;
  return pod->collect->evals->ve[mode];
}

int
W_pod_tot_modes (W_pod_p pod)
{
  if (pod->collect == NULL) {
    REPORT_WARNING (pod->domain->errh, "Not in collect mode"); return 0;
  }
  if (pod->collect->evecs == NULL) {
    REPORT_WARNING (pod->domain->errh, "No component modes matrix"); return 0;
  }
  return pod->collect->evals->dim;
}

static double
ev_sum (W_pod_p pod)
{
  double sum = 0;
  int i;
  FOR (i, pod->collect->evals->dim) { sum += pod->collect->evals->ve[i]; }
  return sum;
}

int
W_pod_write_mean (W_pod_p pod, FILE *fp)
{
  W_node_t *n;
  W_vect_t d;
  W_pod_collect_t *c = pod->collect;

  if (pod->collect == NULL) { REPORT_WARNING (pod->domain->errh, "Not in collect mode"); return FALSE; }
  if (c->umean == NULL)     { REPORT_WARNING (pod->domain->errh, "No mean data"); return FALSE; }

  set_mean (pod, REPLACE, 1);

  fprintf (fp, "!\n");
  fprintf (fp, "MEAN\n");
  fprintf (fp, "! node_id u_x u_y u_z\n");
  n = NULL;
  while (n = WD_get_next_node (pod->domain, n)) {
    WN_get_displ (n, &d.x, &d.y, &d.z);
    fprintf (fp, "%d %g %g %g\n", (int) n->id, d.x, d.y, d.z);
  }
  fprintf (fp, "END MEAN\n");
  fprintf (fp, "!\n");
  return TRUE;
}


int
W_pod_write_mode (W_pod_p pod, int mode, FILE *fp)
{
  W_node_t *n;
  W_vect_t d;
  W_pod_collect_t *c = pod->collect;
  double evsum = ev_sum (pod);

  if (pod->collect == NULL) {
    REPORT_WARNING (pod->domain->errh, "Not in collect mode"); return FALSE;
  }
  if (c->evecs == NULL) {
    REPORT_WARNING (pod->domain->errh, "No modes"); return FALSE;
  }
  if (mode >= c->evals->dim)  {
    REPORT_WARNING (pod->domain->errh, "No such mode"); return FALSE;
  }
  if (mode < 0) mode = 0;

  set_mode (pod, mode, REPLACE, 1);

  fprintf (fp, "!\n");
  fprintf (fp, "MODE %d ", mode);
  fprintf (fp, "EIGENVALUE %g (%g%%)\n", W_pod_eigenvalue (pod, mode),
           W_pod_eigenvalue (pod, mode) / evsum * 100);
  fprintf (fp, "! node_id u_x u_y u_z\n");
  n = NULL;
  while (n = WD_get_next_node (pod->domain, n)) {
    WN_get_displ (n, &d.x, &d.y, &d.z);
    fprintf (fp, "%d %g %g %g\n", (int) n->id, d.x, d.y, d.z);
  }
  fprintf (fp, "END MODE %d\n", mode);
  fprintf (fp, "!\n");
  return TRUE;
}

static MAT *
new_ntm (W_pod_use_t *u, W_index_t id)
{
  MAT *ntm;
  ntm = m_get (3, u->max_nm+1); CHECK_MALLOC (ntm);
  dict_add_obj_int_name (u->node_transf_matd, ntm, id);
  m_zero (ntm);
  return ntm;
}

static void
read_mean (W_pod_use_t *u, tokens_parser parser, W_domain_t *d)
{
  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) { BAD_INPUT (); }
    key = tokens_token_as_string (parser, 1);
    if (tokens_token_and_keyword_equiv (key, "END")) {
      return;
    } else {
      int id;
      double ux, uy, uz;
      MAT *ntm;
      
      if (tokens_total_of_tokens (parser) < 4) { BAD_INPUT (); }
      id = tokens_token_as_int (parser, 1);
      ux = tokens_token_as_double (parser, 2);
      uy = tokens_token_as_double (parser, 3);
      uz = tokens_token_as_double (parser, 4);
      /* store the mode vector in the node matrix */
      ntm = dict_find_obj_with_int_name (u->node_transf_matd, id);
      if (!ntm) ntm = new_ntm (u, id);
      ntm->me[0][u->max_nm] = ux;
      ntm->me[1][u->max_nm] = uy;
      ntm->me[2][u->max_nm] = uz;
    }
  }
}

static void
read_mode (W_pod_use_t *u, tokens_parser parser, W_domain_t *d)
{
  double ev = 0;
  int mode = u->curr_m;
  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) { BAD_INPUT (); }
    key = tokens_token_as_string (parser, 1);
    if (tokens_token_and_keyword_equiv (key, "END")) {
      u->curr_m++;
      return;
    } else {
      int id;
      double ux, uy, uz;
      MAT *ntm;
      
      if (tokens_total_of_tokens (parser) < 4) { BAD_INPUT (); }
      id = tokens_token_as_int (parser, 1);
      ux = tokens_token_as_double (parser, 2);
      uy = tokens_token_as_double (parser, 3);
      uz = tokens_token_as_double (parser, 4);
      /* store the mode vector in the node matrix */
      ntm = dict_find_obj_with_int_name (u->node_transf_matd, id);
      if (!ntm) ntm = new_ntm (u, id);
      ntm->me[0][mode] = ux;
      ntm->me[1][mode] = uy;
      ntm->me[2][mode] = uz;
    }
  }
}

static void
skip_till_END (tokens_parser parser, W_domain_t *d)
{
  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) { BAD_INPUT (); }
    key = tokens_token_as_string (parser, 1);
    if (tokens_token_and_keyword_equiv (key, "END")) {
      return;
    } 
  }
}

static int
to_be_loaded (W_pod_p pod, int mode)
{
  W_pod_use_t *u = pod->use;
  if (u->modenl) {
    return in_list (u->modenl, (void *)mode);
  } else {
    return (mode >= 0 && mode < u->max_nm);
  }
}

static void
restore_modes (W_pod_p pod, const char *file)
{
  W_pod_use_t *u = pod->use;
  tokens_parser parser;
  int mode = -1;
  W_domain_t *d = pod->domain;
  
  if (u->max_nm <= 0) REPORT_FATAL_ERROR (pod->domain->errh, "Wrong number of modes");

  parser  = tokens_new_parser ();

  if (!tokens_open_file (parser, (char *) file))
    REPORT_FATAL_ERROR (pod->domain->errh, "Could not open POD file");

  u->node_transf_matd   = dict_create_dictionary ();
  CHECK_MALLOC (u->node_transf_matd);
  u->dof_modes_dim = WD_tot_of_equations (pod->domain);
  u->mm = u->mm_orig = NULL;
  
  u->curr_m = 0;

  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) { BAD_INPUT (); }
    key = tokens_token_as_string (parser, 1);
    if        (tokens_token_and_keyword_equiv (key, "MODE")) {
      if (tokens_total_of_tokens (parser) < 2) { BAD_INPUT (); }
      mode = tokens_token_as_int (parser, 2);
      if (to_be_loaded (pod, mode)) {
        read_mode (u, parser, pod->domain);
      } else {
        skip_till_END (parser, pod->domain);
      }
    } else if (tokens_token_and_keyword_equiv (key, "MEAN")) {
      read_mean (u, parser, pod->domain);
    } else {
      REPORT_FATAL_ERROR (pod->domain->errh, "Unknown keyword");
    }
  }
  
  tokens_close_curr_file (parser);
  tokens_delete_parser (parser);
}

static void
check_orthonorm_u (W_pod_p pod)
{
  W_pod_use_t *u = pod->use;
  W_index_t toteqn = WD_tot_of_equations (pod->domain);
  int i, j, k;
  int M = u->max_nm;
  double d;
  W_bool_t all_orthonormal = TRUE;
  double max_d = 0;

  FOR (i, M) {
    FOR (j, M) {
      d = 0;
      FOR (k, toteqn) {
        d += u->mm->dof_modes[k][i] * u->mm->dof_modes[k][j];
      }
      CHECK_ORTHONORMAL (i, j, d);
    }
  }
  if (!all_orthonormal) {
    sprintf (W_buf (), "Some modes not orthonormal; max mismatch %g", max_d);
    REPORT_WARNING (pod->domain->errh, W_buf ());
  }
}

static void
build_dof_modes (W_pod_p pod)
{
  pod->use->mm_orig = mm_new (pod);
  mm_init (pod->use->mm_orig);
  pod->use->mm = mm_new (pod);
  mm_copy (pod->use->mm, pod->use->mm_orig);
  check_orthonorm_u (pod);
}


int
W_pod_get_nmodes (W_pod_p pod)
{
  if (!pod->use) REPORT_FATAL_ERROR (pod->domain->errh, "Not in POD-use mode");
  if (!pod->use->mm)
    return pod->use->max_nm;
  else
    return pod->use->mm->nm;
}

void
W_pod_dof_modes (W_pod_p pod, W_index_t dof, double *v, double *mean)
{
  W_pod_use_t *u = pod->use;
  double **dof_modes;
  int nm;
  int i;
  if (!u) REPORT_FATAL_ERROR (pod->domain->errh, "Not in POD-use mode");
  if (u->mm == NULL) build_dof_modes (pod);
  dof_modes = u->mm->dof_modes;
  nm = u->mm->nm;
  FOR (i, nm) { v[i] = dof_modes[dof][i]; }
  *mean = dof_modes[dof][nm];
}

static void
skip_input (W_domain_t *d)
{
  const char *input_line;
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      REPORT_FATAL_ERROR (d->errh, "Missing END");
    TOKENS_PARSE(input_line);
    if (TOKENS_HAVE_KEYWORD ("END")) {
      break;
    }
  }
}

static W_bool_t use_pod = FALSE;
static W_bool_t collect_pod = FALSE;
static int pod_nm = 0;
static int pod_max_nm = 0;
static LIST_OF (int) modenl = NULL;
static double pod_dt = 0;
static double pod_up_to_t = FLT_MAX;
static double pod_from_t = 0;
static W_pod_method_t method = POD_DIRECT;
static double pod_cutoff_ev = -1; /* no constraint */

static void
parse_command_line_options (W_domain_t *d)
{
  if (WOPT_OPT (d->options, pod_options)) {
    int pos;
    tokens_parser parser = tokens_new_parser ();
    if (tokens_parse_line (parser, WOPT_OPT (d->options, pod_options))) {
      if ((pos = tokens_have_keyword (parser, 1, "USE")) > 0) {
        use_pod = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "COLLECT")) > 0) {
        collect_pod = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "METHOD")) > 0) {
        if (tokens_have_keyword (parser, pos, "DIRECT"))
          method = POD_DIRECT;
        else
          method = POD_SNAPSHOTS;
      }
      if ((pos = tokens_have_keyword (parser, 1, "CUTOFF_EV")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          pod_cutoff_ev = tokens_token_as_double (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "CUTOFF_EV option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "NM")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          pod_nm = tokens_token_as_int (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "NM option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "MAX_NM")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          pod_nm = tokens_token_as_int (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "MAX_NM option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "NMODES")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          pod_nm = pod_max_nm = tokens_token_as_int (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "NMODES option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "MODENL")) > 0) {
        int i = 1, m;
        modenl = make_list (); CHECK_MALLOC (modenl);
        while (tokens_total_of_tokens (parser) >= pos+i &&
               tokens_type_of_token (parser, pos+i) == TOKENS_NUMBER) {
          m = tokens_token_as_int (parser, pos+i);
          add_to_tail (modenl, (void *) m);
          i++;
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "DT")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          pod_dt = tokens_token_as_double (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "DT option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "UP_TO_T")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          pod_up_to_t = tokens_token_as_double (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "UP_TO_T option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "FROM_T")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          pod_from_t = tokens_token_as_double (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "FROM_T option requires a number");
        }
      }
     }
    tokens_delete_parser (parser);
  }
}

/*
  Reduction reader.
  Model reduction through POD (KLE), ie. Proper Orthogonal Decomposition
  (Karhunen-Loeve Expansion).
 */

W_bool_t 
W_pod_reader (W_domain_t *d)
{
  const char *input_line;
  const char *s;
  double inflate = 0;

  parse_command_line_options (d);

  if (!(use_pod || collect_pod)) /* disregard POD */
    return W_read_and_discard_block (d);

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      TOKENS_PARSE(input_line);
      if        (TOKENS_HAVE_KEYWORD ("COLLECT")) {
        if (collect_pod) {
          d->pod = w_new (W_pod_t); CHECK_MALLOC (d->pod);
          d->pod->domain = d;
          d->pod->use    = NULL;
          d->pod->collect = w_new (W_pod_collect_t); CHECK_MALLOC (d->pod->collect);
          d->pod->collect->last_t    = 0;
          d->pod->collect->dt        = pod_dt;
          d->pod->collect->up_to_t   = pod_up_to_t;
          d->pod->collect->from_t    = pod_from_t;
          d->pod->collect->toteqn    = 0;
          d->pod->collect->cutoff_ev = pod_cutoff_ev; 
          d->pod->collect->evecs     = NULL;
          d->pod->collect->evals     = NULL;
          d->pod->collect->method    = method;
          while ((input_line = W_get_input_line(d)) != NULL) {
            TOKENS_PARSE(input_line);
            if        (TOKENS_HAVE_KEYWORD ("DT")) {
              if (d->pod->collect->dt <= 0) /* command line arguments override input file */
                d->pod->collect->dt = TOKEN_DBL (2);
            } else if (TOKENS_HAVE_KEYWORD ("SNAPSHOTS")) {
              method = POD_SNAPSHOTS;
            } else if (TOKENS_HAVE_KEYWORD ("DIRECT")) {
              method = POD_DIRECT;
            } else if (TOKENS_HAVE_KEYWORD ("END")) {
              break;
            }
          }
          d->pod->collect->ul            = NULL;
          d->pod->collect->umean         = NULL;
          switch (d->pod->collect->method) {
          case POD_DIRECT:
            d->pod->collect->compute_modes = compute_modes_DM;
            break;
          case POD_SNAPSHOTS:
          default:
            d->pod->collect->compute_modes = compute_modes_MoS;
          }
        } else {
          skip_input (d);
        }
      } else if (TOKENS_HAVE_KEYWORD ("USE")) {
        if (use_pod) {
          d->pod = w_new (W_pod_t); CHECK_MALLOC (d->pod);
          d->pod->domain = d;
          d->pod->collect = NULL;
          d->pod->use     = w_new (W_pod_use_t); CHECK_MALLOC (d->pod->use);
          d->pod->use->modenl = NULL;
          if (modenl) {
            d->pod->use->max_nm = list_length (modenl);
            d->pod->use->modenl = copy_list (modenl, NOT_NODES);
          } else {
            if (pod_max_nm <= 0) pod_max_nm = pod_nm;
            d->pod->use->max_nm = pod_max_nm;
          }
          d->pod->use->curr_m = 0;
          while ((input_line = W_get_input_line(d)) != NULL) {
            char *file = NULL;
            TOKENS_PARSE(input_line);
            if        (TOKENS_HAVE_KEYWORD ("FILE")) {
              const char *s = TOKEN_STR (2);
              NEWP_DUPL_STRING (file, s,
                                { REPORT_FATAL_ERROR (d->errh, "Failed in string copy"); });
            } else if (TOKENS_HAVE_KEYWORD ("NMODES")) {
              if (d->pod->use->max_nm == 0) /* only if not given on command line*/
                d->pod->use->max_nm = TOKEN_INT (2);
            } else if (TOKENS_HAVE_KEYWORD ("END")) {
              if (file == NULL) {
                sprintf (W_buf (), "%s.pod", WOPT_OPT (d->options, in_file_basename));
                NEWP_DUPL_STRING (file, W_buf (),
                                  { REPORT_FATAL_ERROR (d->errh, "Failed in string copy"); });
              }
              restore_modes (d->pod, file);
              w_free (file);
              break;
            }
          }
        } else {
          skip_input (d);
        }
      } else {
        REPORT_FATAL_ERROR (d->errh, "Unknown keyword");
      }
    }
  }
  return FALSE;
}

int
W_pod_is_in_collect_mode (W_pod_p pod)
{
  return (pod->collect != NULL);
}

int
W_pod_is_in_use_mode (W_pod_p pod)
{
  return (pod->use != NULL);
}

int
W_pod_save_collected_data (W_pod_p pod, FILE *fp) 
{
  W_pod_collect_t *c = pod->collect;
  int toteqn = c->toteqn;
  double *v;
  int sample, j;

  fprintf (fp, "!     \n");
  fprintf (fp, "! POD sampled data, %d nodes (3 displacements each), %d samples\n",
           dict_count (pod->domain->nodes), list_length (c->ul));
  fprintf (fp, "!   x-, y- and z-displacement given on separate lines for each node in its turn\n");
  fprintf (fp, "!     \n");
  sample = 0;
  loop_through_list (c->ul, v, double *) {
    fprintf (fp, "! start sample %d\n", sample);
    FOR (j, toteqn) {
      fprintf (fp, "%g\n", v[j]);
    }
    fprintf (fp, "! end sample %d\n", sample);
    sample++;
  }
  return TRUE;
}


static void
write_header (W_pod_p pod, FILE *fp)
{
  W_pod_collect_t *c = pod->collect;
  fprintf (fp, "! POD modes\n");
  fprintf (fp, "!   Computed %d modes\n", c->evals->dim);
  fprintf (fp, "!   Collected %d samples\n", list_length (c->ul));
  fprintf (fp, "!   From time %g\n", c->from_t);
  fprintf (fp, "!   Up to time %g\n", c->up_to_t);
  switch (c->method) {
  case POD_DIRECT:
    fprintf (fp, "!   Used the direct method\n");
    break;
  case POD_SNAPSHOTS:
    fprintf (fp, "!   Used the snapshots method\n");
    break;
  default:
    fprintf (fp, "!   Used the unknown method\n");
    break;
  }
  if (c->cutoff_ev < 0)
    fprintf (fp, "!   Writing out all computed modes\n");
  else
    fprintf (fp, "!   Writing out modes with eigenvalue above %g\n", c->cutoff_ev);
  fprintf (fp, "! \n");
}


int
W_pod_write_pod (W_pod_p pod, FILE *fp)
{
  int mode;
  write_header (pod, fp);
  FOR (mode, W_pod_tot_modes (pod)) {
    if (W_pod_eigenvalue (pod, mode) > pod->collect->cutoff_ev) 
      W_pod_write_mode (pod, mode, fp);
  }
  W_pod_write_mean (pod, fp);
  return TRUE;
}

int
W_pod_revert_to_orig_modes (W_pod_p pod)
{
  W_pod_use_t *u = pod->use;
  if (!u) REPORT_FATAL_ERROR (pod->domain->errh, "Not in POD-use mode");
  if (u->mm == NULL) return FALSE;
  mm_copy (u->mm, u->mm_orig);
  return TRUE;
}

int
W_pod_orthonormalize (W_pod_p pod)
{
  mm_orthonormalize (pod->use->mm);
  return TRUE;
}

int
W_pod_include_mode (W_pod_p pod, int to_replace, double *data)
{
  mm_set (pod->use->mm, to_replace, data);
  return TRUE;
}

static double
proj_coeff (double *u, int udim, int mode, mode_mat_t *mm)
{
  int j;
  double result = 0;
  FOR (j, udim) {
    result += u[j] * mm->dof_modes[j][mode];
  }
  return result;
}

static void
adjust_configuration (W_pod_p pod)
{
  int j, k;
  double *ubar, *projubar, *vbar, *projvbar, *abar, *projabar, *mcoeff;
  W_node_t *n;
  mode_mat_t *mm = pod->use->mm;
  int nm = mm->nm - 2; /* disregard the last two RBM modes */
  int toteqn = pod->use->dof_modes_dim;
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif
  double r_1, v_1[3], a_1[3], x[3], X[3], rX[3], tx[3];
  double t[3][3], invt[3][3];
  
  ubar      = w_new_array (toteqn, double); CHECK_MALLOC (ubar);
  projubar  = w_new_array (toteqn, double); CHECK_MALLOC (projubar);
  vbar      = w_new_array (toteqn, double); CHECK_MALLOC (vbar);
  projvbar  = w_new_array (toteqn, double); CHECK_MALLOC (projvbar);
  abar      = w_new_array (toteqn, double); CHECK_MALLOC (abar);
  projabar  = w_new_array (toteqn, double); CHECK_MALLOC (projabar);
  mcoeff    = w_new_array (nm, double); CHECK_MALLOC (mcoeff);
  
  transfmat (pod->domain, t);
  MAT3DUTIL_INVERT_3X3 (1, t, invt);

  /* compute coords the rotated c.s. */
  n = NULL;
  while (n = WD_get_next_node (pod->domain, n)) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    
    WN_get_ref_coords (n, &X[0], &X[1], &X[2]);
    MAT3DUTIL_MULT_A_3X3_V_3 (rX, invt, X);
    WN_get_coords (n, &x[0], &x[1], &x[2]);
    FOR (k, 3) { x[k] -= rX[k]; } /* difference between x and rX */
    MAT3DUTIL_MULT_A_3X3_V_3 (tx, t, x); /* transform into barred c.s. */
    FOR (k, 3) { ubar[ens[k]] = tx[k]; }
    
    WN_get_velocity (n, &x[0], &x[1], &x[2]);
    MAT3DUTIL_MULT_A_3X3_V_3 (tx, t, x);
    FOR (k, 3) { vbar[ens[k]] = tx[k]; }
    
    WN_get_acceleration (n, &x[0], &x[1], &x[2]);
    MAT3DUTIL_MULT_A_3X3_V_3 (tx, t, x);
    FOR (k, 3) { abar[ens[k]] = tx[k]; }
  }
  
  /* project into the current pod basis */
  FOR (k, nm) { mcoeff[k] = proj_coeff (ubar, toteqn, k, mm); }
  FOR (j, toteqn) {
    projubar[j] = 0;
    FOR (k, nm) {
      projubar[j] += mcoeff[k] * mm->dof_modes[j][k];
    }
    /* printf ("%d: ubar = %g, projubar = %g\n", j, ubar[j], projubar[j]); */
  }
  FOR (k, nm) { mcoeff[k] = proj_coeff (vbar, toteqn, k, mm); }
  FOR (j, toteqn) {
    projvbar[j] = 0;
    FOR (k, nm) {
      projvbar[j] += mcoeff[k] * mm->dof_modes[j][k];
    }
  }
  FOR (k, nm) { mcoeff[k] = proj_coeff (abar, toteqn, k, mm); }
  FOR (j, toteqn) {
    projabar[j] = 0;
    FOR (k, nm) {
      projabar[j] += mcoeff[k] * mm->dof_modes[j][k];
    }
  }
  
  /* transform into the global basis and adjust the nodal configuration */
  n = WD_get_node_by_idx (pod->domain, 1); /* node #1 */
  if (!n) REPORT_FATAL_ERROR (pod->domain->errh, "Node not found");
  WN_get_ref_coords (n, &x[0], &x[1], &x[2]);
  r_1 = sqrt (x[0]*x[0] + x[1]*x[1] * x[2]*x[2]);
  WN_get_velocity (n, &x[0], &x[1], &x[2]);
  MAT3DUTIL_MULT_A_3X3_V_3 (v_1, t, x);
  WN_get_acceleration (n, &x[0], &x[1], &x[2]);
  MAT3DUTIL_MULT_A_3X3_V_3 (a_1, t, x);
  
  n = NULL;
  while (n = WD_get_next_node (pod->domain, n)) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    WN_get_ref_coords (n, &X[0], &X[1], &X[2]);
    
    FOR (k, 3) { tx[k] = projubar[ens[k]]; }
    MAT3DUTIL_MULT_A_3X3_V_3 (x, invt, tx); /* express in the global c.s. */
    MAT3DUTIL_MULT_A_3X3_V_3 (rX, invt, X);
    printf ("%d %g %g vs. ", n->id, n->displ_t.x, n->displ_t.y);
    n->displ_t.x = x[0] + rX[0] - n->ref_coords.x;
    n->displ_t.y = x[1] + rX[1] - n->ref_coords.y;
    n->displ_t.z = x[2] + rX[2] - n->ref_coords.z;
    printf (" %g %g\n", n->displ_t.x, n->displ_t.y);

    FOR (k, 3) { tx[k] = projvbar[ens[k]]; }
    MAT3DUTIL_MULT_A_3X3_V_3 (x, invt, tx);
    n->velocity_t.x = x[0];
    n->velocity_t.y = x[1];
    n->velocity_t.z = x[2];
    
    FOR (k, 3) { tx[k] = projabar[ens[k]]; }
    MAT3DUTIL_MULT_A_3X3_V_3 (x, invt, tx);
    n->acceleration_t.x = x[0];
    n->acceleration_t.y = x[1];
    n->acceleration_t.z = x[2];
  }

  w_free (ubar); w_free (projubar);
  w_free (vbar); w_free (projvbar);
  w_free (abar); w_free (projabar);
  w_free (mcoeff);
}

static void
insert_rigid_body_motions (W_pod_p pod)
{
  W_node_t *n;
  W_pod_use_t *u = pod->use;
  int i, j;
  double x, y, z, r_1;
  double *xmode, *ymode;

  xmode = w_new_array (u->dof_modes_dim, double); CHECK_MALLOC (xmode);
  ymode = w_new_array (u->dof_modes_dim, double); CHECK_MALLOC (ymode);
  n = WD_get_node_by_idx (pod->domain, 1); /* node #1 == constrained to lie on the circle */
  if (!n) REPORT_FATAL_ERROR (pod->domain->errh, "Node not found");
  WN_get_ref_coords (n, &x, &y, &z);
  r_1 = sqrt (x*x + y*y + z*z);

  /* compute the rigid body modes; we rely on the reference configuration being along
     the x-axis
   */
  i = 0;
  n = NULL;
  while (n = WD_get_next_node (pod->domain, n)) {
    WN_get_ref_coords (n, &x, &y, &z);
    xmode[i+0] = x / r_1;
    xmode[i+1] = 0;
    xmode[i+2] = 0;
    ymode[i+0] = 0;
    ymode[i+1] = x / r_1;
    ymode[i+2] = 0;
    i += 3;
  }

  W_pod_include_mode (pod, u->mm->nm-2, xmode);
  W_pod_include_mode (pod, u->mm->nm-1, ymode);

  w_free (xmode);
  w_free (ymode);
}


int
W_pod_rotate (W_pod_p pod)
{
  double t[3][3], invt[3][3];
  transfmat (pod->domain, t);
  MAT3DUTIL_INVERT_3X3 (1, t, invt);
  W_pod_revert_to_orig_modes (pod);
  insert_rigid_body_motions (pod);
  W_pod_orthonormalize (pod);
  mm_rotate (pod->use->mm, invt);
#if 0
  mm_dump (pod->use->mm_orig, stderr);
  mm_dump (pod->use->mm, stderr);
#endif
  adjust_configuration (pod);
  return TRUE;
}
