#include "fles.h"
#include "tokensP.h"
#include "newP.h"
#include "matrix.h"
#include <matrix2.h>
#include <sparse.h> 
#include <sparse2.h>

#define NEED_DEFINES 0
#if NEED_DEFINES
#  define NMODES 1
#  define LHS_DAT        "lhs.dat"
#  define RHS_DAT        "rhs.dat"
#  define SOL_DAT        "sol.dat"
#  define POD_DAT        "pod.dat"
#  define RHS_XMGR       "rhs.xmgr"
#  define APPROX_SOL_XMGR  "approx_sol.xmgr"
#  define APPROX_RHS_XMGR  "approx_rhs.xmgr"
#  define FULL_SOL_XMGR   "sol.xmgr"
#else
#  define NMODES $NMODES
#  define LHS_DAT "$LHS_DAT"
#  define RHS_DAT "$RHS_DAT"
#  define SOL_DAT "$SOL_DAT" 
#  define POD_DAT "$POD_DAT" 
#  define RHS_XMGR "$RHS_XMGR" 
#  define APPROX_SOL_XMGR "$APPROX_SOL_XMGR" 
#  define APPROX_RHS_XMGR "$APPROX_RHS_XMGR"
#  define FULL_SOL_XMGR "$FULL_SOL_XMGR" 
#endif

static fles_solver_ctx_t slesctx = 0;
static int tot_of_equations = $TOT_OF_EQUATIONS;
static double nz_per_row = 0.14;
static SPMAT *lhs = NULL;
static VEC *rhs = NULL;
static VEC *sol = NULL;
static VEC *fullsol = NULL;
static VEC *modes[NMODES];
static MAT *phi_t_lhs_phi = NULL;
static VEC *phi_t_rhs = NULL;
static VEC *approx_sol = NULL;

static int
tot_of_equations_ (void *client)
{
  return tot_of_equations;
}

static int
error_notify (fles_error_enum error_num,
              int line,
              char *file,
              char *detail,
              void *client_data)
{
  char buf[1024];
  sprintf (buf, "%s (l.%d), %s", file, line, detail);
  fprintf (stderr, "%s\n", buf);
  abort ();
  return 0;
}

static void
make_fles (int *argc, char ***argv)
{
  slesctx = fles_new_spetsc_solver (NULL, argc, argv, tot_of_equations_, error_notify);
  slesctx->set_option (slesctx, PETSC_OPTS_NZ_PER_ROW, (void *)&nz_per_row);
}
  
static void
check_symmetry ()
{
  int i, j;
  int unsymmetric = 0;
  fprintf (stderr, "Symmetry check: ");
  for (i = 0; i < tot_of_equations; i++)
    for (j = i; j < tot_of_equations; j++) {
      double ad = sp_get_val (lhs, i, j);
      double bd = sp_get_val (lhs, j, i);
      if (fabs (ad - bd) > 1e-9) {
        unsymmetric = 1;
        fprintf (stderr, "Unsymmetry: a[%d][%d] = %g != a[%d][%d] = %g\n",
                 i, j, ad, j, i, bd);
      }
    }
  if (unsymmetric) fprintf (stderr, "Error: Unsymmetric\n");
  else             fprintf (stderr, "Symmetric\n");
}

static void
load_lhs (char *file)
{
  tokens_parser parser = tokens_new_parser ();

  lhs = sp_resize (lhs, tot_of_equations, tot_of_equations);
  if (lhs == 0) {
    fprintf (stderr, "Error: no memory\n");
    abort ();
  }
  sp_zero (lhs);
  
  if (!tokens_open_file (parser, file)) {
    fprintf (stderr, "Error: no such file? %s\n", file);
    abort ();
  }
  
  fprintf (stderr, "Loading lhs\n");
  slesctx->zero_lhs (slesctx);
  
  while (tokens_next_line (parser)) {
    int ntok = tokens_total_of_tokens (parser);
    int col[1], row[1];
    double val[1];
    const char *rowtok = tokens_token_as_string (parser, 2);
    if (sscanf (rowtok, "%d:", &row[0]) != 1) {
      fprintf (stderr, "Error: got %s\n", rowtok);
      abort ();
    }
    if (row[0] >= tot_of_equations) {
      fprintf (stderr, "Error: wrong number of equations %d\n", tot_of_equations);
      abort ();
    }
    {
      int i=3;
      while (i <= ntok) {
        col[0] = tokens_token_as_int (parser, i); i++;
        val[0] = tokens_token_as_double (parser, i); i++;
        /* printf ("assembling [%d][%d]=%g ", row[0], col[0], val[0]); */
        slesctx->assemble_mat (slesctx, FLES_FALSE, FLES_ADD, 1, row, 1, col, val);
        {
          double v = sp_get_val (lhs, row[0], col[0]) + val[0];
          sp_set_val (lhs, row[0], col[0], v);
        }
      }
    }
    /* printf ("\n"); */
  }
  slesctx->finish_lhs (slesctx, FLES_FINAL_FINISH);
  check_symmetry ();

  tokens_close_curr_file (parser);
  tokens_delete_parser (parser);
}
  
static void
load_rhs (char *file)
{
  tokens_parser parser = tokens_new_parser ();

  rhs = v_resize (rhs, tot_of_equations);
  if (rhs == 0) {
    fprintf (stderr, "Error: no memory\n");
    abort ();
  }

  if (!tokens_open_file (parser, file)) {
    fprintf (stderr, "Error: no such file? %s\n", file);
    abort ();
  }
  
  fprintf (stderr, "Loading rhs\n");
  slesctx->zero_rhs (slesctx);
  
  {
    int row[1];
    row[0] = 0;
    while (tokens_next_line (parser)) {
      double val[1];
      int ntok = tokens_total_of_tokens (parser);
      if (ntok < 1) {
        fprintf (stderr, "Error: no tokens?\n");
        abort ();
      }        
      if (row[0] >= tot_of_equations) {
        fprintf (stderr, "Error: wrong number of equations %d\n", tot_of_equations);
        abort ();
      }
      val[0] = tokens_token_as_double (parser, 1);
      slesctx->assemble_vect (slesctx, FLES_ADD, 1, row, val);
      rhs->ve[row[0]] += val[0];
      /* printf ("\n"); */
      row[0]++;
    }
  }
  slesctx->finish_rhs (slesctx);
  fprintf (stderr, "Rhs norm: %g\n", v_norm2 (rhs));
  
  tokens_close_curr_file (parser);
  tokens_delete_parser (parser);
}
  
static void
load_sol (char *file)
{
  tokens_parser parser = tokens_new_parser ();

  sol = v_resize (sol, tot_of_equations);
  if (sol == 0) {
    fprintf (stderr, "Error: no memory\n");
    abort ();
  }
  v_zero (sol);

  if (!tokens_open_file (parser, file)) {
    fprintf (stderr, "Error: no such file? %s\n", file);
    abort ();
  }
  
  fprintf (stderr, "Loading sol\n");

  {
    int row[1];
    row[0] = 0;
    while (tokens_next_line (parser)) {
      double val[1];
      int ntok = tokens_total_of_tokens (parser);
      if (ntok < 1) {
        fprintf (stderr, "Error: no tokens?\n");
        abort ();
      }        
      if (row[0] >= tot_of_equations) {
        fprintf (stderr, "Error: wrong number of equations %d\n", tot_of_equations);
        abort ();
      }
      val[0] = tokens_token_as_double (parser, 1);
      sol->ve[row[0]] += val[0];
      /* printf ("\n"); */
      row[0]++;
    }
  }
  
  tokens_close_curr_file (parser);
  tokens_delete_parser (parser);
}

static void
compute_residual_as_loaded ()
{
  double resnorm;
  VEC *res = NULL;
  res = sp_mv_mlt (lhs, sol, res);
  v_sub (res, rhs, res);
  resnorm = v_norm2 (res);
  fprintf (stderr, "Residual norm (as loaded): %g\n", resnorm);
  v_free (res);
}

static void
compute_residual_as_computed_full ()
{
  double resnorm;
  VEC *res = NULL;
  res = sp_mv_mlt (lhs, fullsol, res);
  v_sub (res, rhs, res);
  resnorm = v_norm2 (res);
  fprintf (stderr, "Residual norm (as computed, full): %g\n", resnorm);
  v_free (res);
}

static void
full_solve ()
{
  if (!slesctx->solve (slesctx)) {
  }
  fullsol = v_resize (fullsol, tot_of_equations);
  if (fullsol == 0) {
    fprintf (stderr, "Error: no memory\n");
    abort ();
  }
  {
    int row;
    for (row = 0; row < tot_of_equations; row++) {
      fullsol->ve[row] = slesctx->get_from_sol (slesctx, row);
    }
  }
}

static void
read_mode (int mode, tokens_parser parser)
{
  double ev = 0;
  modes[mode] = v_resize (NULL, tot_of_equations);
  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) {
      fprintf (stderr, "Bad data in POD file");
      abort ();
    }
    key = tokens_token_as_string (parser, 1);
    if (tokens_token_and_keyword_equiv (key, "END")) {
      return;
    } else {
      int id;
      double ux, uy, uz;
      if (tokens_total_of_tokens (parser) < 4) {
        fprintf (stderr, "Bad data in POD file");
        abort ();
      }
      id = tokens_token_as_int (parser, 1);
      ux = tokens_token_as_double (parser, 2);
      uy = tokens_token_as_double (parser, 3);
      uz = tokens_token_as_double (parser, 4);
      modes[mode]->ve[((id-1)*3)+0] = ux;
      modes[mode]->ve[((id-1)*3)+1] = uy;
      modes[mode]->ve[((id-1)*3)+2] = uz;
    }
  }
}

static void
skip_till_END (tokens_parser parser)
{
  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) {
      fprintf (stderr, "Bad data in POD file\n");
      abort ();
    }
    key = tokens_token_as_string (parser, 1);
    if (tokens_token_and_keyword_equiv (key, "END")) {
      return;
    } 
  }
}

static void
load_pod (const char *file)
{
  tokens_parser parser;
  int mode = -1;
  
  parser  = tokens_new_parser ();

  if (!tokens_open_file (parser, (char *) file)) {
    fprintf (stderr, "Could not open POD file \"%s\"", file);
    abort ();
  }

  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) {
      fprintf (stderr, "Bad data in POD file \"%s\"", file);
      abort ();
    }
    key = tokens_token_as_string (parser, 1);
    if        (tokens_token_and_keyword_equiv (key, "MODE")) {
      if (tokens_total_of_tokens (parser) < 2) {
        fprintf (stderr, "Bad data in POD file \"%s\"", file);
        abort ();
      }
      mode = tokens_token_as_int (parser, 2);
      if (mode < NMODES) {
        read_mode (mode, parser);
      } else {
        skip_till_END (parser);
      }
    } else if (tokens_token_and_keyword_equiv (key, "MEAN")) {
      skip_till_END (parser);
    } else {
      fprintf (stderr, "Unknown keyword \"%s\"", key);
      abort ();
    }
  }
  
  tokens_close_curr_file (parser);
  tokens_delete_parser (parser);

  fprintf (stderr, "Checking orthonormality\n");
  {
    int i, j;
    int orthonormal = 1;
    double maxdiff = 0;
    for (i = 0; i < NMODES; i++) {
      for (j = 0; j < NMODES; j++) {
        if (i == j) { 
           if (fabs (in_prod (modes[i], modes[j]) - 1) > 1e-6) {
              orthonormal = 0;
              maxdiff = max (maxdiff, fabs (in_prod (modes[i], modes[j]) - 1));
           }
        } else {
           if (fabs (in_prod (modes[i], modes[j])    ) > 1e-6) {
              orthonormal = 0;
              maxdiff = max (maxdiff, fabs (in_prod (modes[i], modes[j])));
           }
        }
      }
    }
    if (orthonormal) fprintf (stderr, "Orthonormal\n");
    else             fprintf (stderr, "NOT Orthonormal; max diff %g\n", maxdiff);
  }
}

static void
construct_transf_lhs ()
{
  int i, j;
  phi_t_lhs_phi = m_get (NMODES, NMODES);
  m_zero (phi_t_lhs_phi);
  for (i = 0; i < NMODES; i++) {
    for (j = 0; j < NMODES; j++) {
      double v;
      VEC *res = v_get (tot_of_equations);
      res = sp_mv_mlt (lhs, modes[j], res);
      v = in_prod (res, modes[i]);
      phi_t_lhs_phi->me[i][j] = v;
    }
  }
  /*  m_dump (stderr, phi_t_lhs_phi); */
}

static void
construct_transf_rhs ()
{
  int i, j;
  phi_t_rhs = v_get (NMODES);
  v_zero (phi_t_rhs);
  for (i = 0; i < NMODES; i++) {
    phi_t_rhs->ve[i] = in_prod (rhs, modes[i]);
  }
  /* v_dump (stderr, phi_t_rhs); */
}

static void
construct_approx_sol ()
{
  int i, j;
  VEC *ampl = v_get (NMODES);
  PERM *pivot = px_get (NMODES);
  phi_t_lhs_phi = LUfactor (phi_t_lhs_phi, pivot);
  ampl = LUsolve (phi_t_lhs_phi, pivot, phi_t_rhs, ampl);
  approx_sol = v_get (tot_of_equations);
  if (approx_sol == 0) {
    fprintf (stderr, "Error: no memory\n");
    abort ();
  }
  fprintf (stderr, "Approximate solution\n");
  v_zero (approx_sol);
  for (i = 0; i < tot_of_equations; i++) {
    for (j = 0; j < NMODES; j++) {
      approx_sol->ve[i] += ampl->ve[j] * modes[j]->ve[i];
    }
  }
  {
    VEC *v = v_get (tot_of_equations);
    double a, r;
    v = sp_mv_mlt (lhs, sol, v);
    a = in_prod (v, sol);
    r = in_prod (rhs, sol);
    fprintf (stderr, "Amplitude for the 'exact' solution as a mode: %g\n", a / r);
  }
}

static void
compute_residual_as_approx ()
{
  double resnorm;
  VEC *res = NULL;
  res = sp_mv_mlt (lhs, approx_sol, res);
  v_sub (res, rhs, res);
  resnorm = v_norm2 (res);
  fprintf (stderr, "Residual norm (as approximated by %d modes): %g\n", NMODES, resnorm);
  v_free (res);
}

static void
dump_full_sol ()
{
  int i;
  FILE *fp = fopen (FULL_SOL_XMGR, "w");
  if (!fp) {
    fprintf (stderr, "Failed to open %s\n", FULL_SOL_XMGR);
    abort ();
  }
  for (i = 0; i < tot_of_equations; i++) {
    fprintf (fp, "%d %g\n", i, sol->ve[i]);
  }
  fclose (fp);
}

static void
dump_approx_sol ()
{
  int i;
  FILE *fp = fopen (APPROX_SOL_XMGR, "w");
  if (!fp) {
    fprintf (stderr, "Failed to open %s\n", APPROX_SOL_XMGR);
    abort ();
  }
  for (i = 0; i < tot_of_equations; i++) {
    fprintf (fp, "%d %g\n", i, approx_sol->ve[i]);
  }
  fclose (fp);
}

static void
dump_rhs ()
{
  int i;
  FILE *fp = fopen (RHS_XMGR, "w");
  if (!fp) {
    fprintf (stderr, "Failed to open %s\n", RHS_XMGR);
    abort ();
  }
  for (i = 0; i < tot_of_equations; i++) {
    fprintf (fp, "%d %g\n", i, rhs->ve[i]);
  }
  fclose (fp);
}

static void
dump_approx_rhs ()
{
  int i;
  VEC *approx_rhs = NULL;
  FILE *fp = fopen (APPROX_RHS_XMGR, "w");
  if (!fp) {
    fprintf (stderr, "Failed to open %s\n", APPROX_RHS_XMGR);
    abort ();
  }
  approx_rhs = sp_mv_mlt (lhs, approx_sol, approx_rhs);
  for (i = 0; i < tot_of_equations; i++) {
    fprintf (fp, "%d %g\n", i, approx_rhs->ve[i]);
  }
  fclose (fp);
}

static void
dumps ()
{
  dump_full_sol ();
  dump_approx_sol ();
  dump_rhs ();
  dump_approx_rhs ();
}

static void
full_solve_with_hint ()
{
  slesctx->set_option (slesctx, PETSC_OPTS_SOL_HINT, (void *)approx_sol->ve);
  if (!slesctx->solve (slesctx)) {
    fprintf (stderr, "Failed in solve");
    abort ();
  }
}

int
main (int argc, char **argv)
{
  make_fles (&argc, &argv);
  load_lhs (LHS_DAT);
  load_rhs (RHS_DAT);
  load_sol (SOL_DAT);
  compute_residual_as_loaded ();
  full_solve ();
  compute_residual_as_computed_full ();
  load_pod (POD_DAT);
  construct_transf_lhs ();
  construct_transf_rhs ();
  construct_approx_sol ();
  compute_residual_as_approx ();
  dumps ();
  full_solve_with_hint ();
  return 0;
}
