#include "triangulation.h"
#include "pirat.h"
#include "g3d.h"
#include "tokensP.h"

static const int client = 13;
#define CLIENT ((void *)&client)

static void
init_assoc_data (triangulation_t t, int nmodes)
{
  triangulation_node_p n;
  double *modes;
  int i;
  dict_iterator di = dict_get_iterator ();
  
  dict_reset_iterator (di);
  while (n = tri_next_node (t, di)) {
    modes = NEWP_OBJ_ARRAY (3*nmodes, double);
    for (i = 0; i < 3*nmodes; i++) modes[i] = 0;
    /* fprintf (stderr, "Associating %p with %d\n", modes, tri_node_get_id (n));  */
    tri_node_set_assoc_data (n, CLIENT, modes);
    if (tri_node_get_assoc_data (n, CLIENT) != modes) {
      fprintf (stderr, "Mismatch\n");
      exit (1);
    }
  }

  dict_del_iterator (di);
}

static triangulation_t 
load_triangulation (char *file, int nmodes)
{
  FILE *stream;
  triangulation_t t;
  
  stream = fopen (file, "r");
  if (stream == NULL) {
    fprintf (stderr, "Could not open %s\n", file);
    exit (1);
  }

  t = tri_from_afro_file (stream);
  if (t == NULL) {
    fprintf (stderr, "Error #%d\n", tri_status ());
    exit (1);
  }
  fclose (stream);

  init_assoc_data (t, nmodes);
  
  return t;
}  

static void
load_pod_file (triangulation_t t, char *pod_file, int nmodes)
{
  const char *s;
  int mode = 0;
  int reading_mode = FALSE;
  tokens_parser parser = tokens_new_parser ();

  if (!tokens_open_file (parser, pod_file)) {
    fprintf (stderr, "File %s could not be opened\n", pod_file);
    exit (1);
  }

  while (tokens_next_line (parser)) {
    s = tokens_token_as_string (parser, 1);
    if        (tokens_token_and_keyword_equiv (s, "MODE")) {
      /*      fprintf (stderr, "Loading mode %d\n", mode); */
      reading_mode = TRUE;
    } else if (tokens_token_and_keyword_equiv (s, "END")) {
      reading_mode = FALSE;
      mode++;
      if (mode == nmodes) break;
    } else if (reading_mode) {
      double x, y, z;
      double *modes;
      triangulation_node_p n;
      int id;
      if (tokens_total_of_tokens (parser) != 4) {
        fprintf (stderr, "Wrong data\n");
        exit (1);
      }
      if ((tokens_type_of_token (parser, 1) == TOKENS_NUMBER) &&
          (tokens_type_of_token (parser, 2) == TOKENS_NUMBER) &&
          (tokens_type_of_token (parser, 3) == TOKENS_NUMBER) &&
          (tokens_type_of_token (parser, 4) == TOKENS_NUMBER)
          ) {
        id = tokens_token_as_int (parser, 1);
        x  = tokens_token_as_double (parser, 2);
        y  = tokens_token_as_double (parser, 3);
        z  = tokens_token_as_double (parser, 4);
        n  = tri_find_node (t, id);
        if (!n) {
          fprintf (stderr, "No such node: %d\n", id);
          exit (1);
        }
        modes = tri_node_get_assoc_data (n, CLIENT);
        if (!modes) {
          fprintf (stderr, "NULL from tri_node_get_assoc_data() (l.%d) for %d?\n", __LINE__, id);
          exit (1);
        }
        modes[3*mode+0] = x;
        modes[3*mode+1] = y;
        modes[3*mode+2] = z;
#if 0
        fprintf (stderr, "Found %p for %d\n", modes, id);
        fprintf (stderr, "Setting Found %p for %d\n", modes, id);
        { int k; for (k = 0; k < 3*nmodes; k++) { fprintf (stderr, "%g ", modes[k]); } }
        fprintf (stderr, "\n");
#endif
      }
    }
  }

  tokens_close_curr_file (parser);
}
 
#define SHRINK_TRIANGLE_DATA(theSHRINK, fstV, scndV, thrdV) \
  {                                                         \
     double theCV;                                          \
                                                            \
     theCV   = (fstV); theCV += (scndV); theCV += (thrdV);  \
     theCV   *= 0.3333333333333333333333333333333333333333; \
     (fstV)  = theCV + (theSHRINK) * ((fstV)  - theCV);     \
     (scndV) = theCV + (theSHRINK) * ((scndV) - theCV);     \
     (thrdV) = theCV + (theSHRINK) * ((thrdV) - theCV);     \
   }

static triangulation_triangle_p
find_enclosing_tri (pirat_front_end_t *pirat, double x, double y, double max_dist)
{
  triangulation_node_p n;
  int ntries = 0;
  double tri_inflate = 1;
  G3D_box_t range;
  G3D_loc_t pt;
  pt.x = x; pt.y = y; pt.z = 0;
  G3D_INIT_BOX (range, pt);
  G3D_INFLATE_BOX (range, max_dist);
  if (!pirat_search_range (pirat, &range, 100)) {
    fprintf (stderr, "Failed in pirat_search_range()\n");
    exit (1);
  }
 new_try:
  n = pirat_first_in_range (pirat);
  while (n) {
    double ax[3], ay[3];
    triangulation_triangle_p tri = 0;
    while (tri = tri_node_next_incid_triangle (n, tri)) {
      int i;
      triangulation_node_p an[3];
      tri_triangle_get_nodes (tri, an);
      for (i = 0; i < 3; i++) {
        tri_node_get_xy (an[i], &ax[i], &ay[i]);
      }
      SHRINK_TRIANGLE_DATA (tri_inflate, ax[0], ax[1], ax[2]);
      SHRINK_TRIANGLE_DATA (tri_inflate, ay[0], ay[1], ay[2]);
      if (G3D_is_inside_tri2d (x, y, ax[0], ay[0], ax[1], ay[1], ax[2], ay[2])) {
        return tri;
      }
    }
    n = pirat_next_in_range (pirat);
  }
  if (ntries > 8) {
    fprintf (stderr, "Too many tries and still no triangle\n");
    exit (1);
  } else {
    tri_inflate += 0.01;
    goto new_try;
  }
}

static void
interpolate (triangulation_node_p n, int nmodes, pirat_front_end_t *pirat, double max_dist)
{
  double x, y;
  int num_entries, j;
  double *modes;
  int i;
  triangulation_triangle_p tri = NULL;

  modes = tri_node_get_assoc_data (n, CLIENT);
  if (!modes) {
    fprintf (stderr, "NULL from tri_node_get_assoc_data? (l.%d)\n", __LINE__);
    exit (1);
  }

  tri_node_get_xy (n, &x, &y);
  tri = find_enclosing_tri (pirat, x, y, max_dist);
  if (!tri) {
    fprintf (stderr, "No enclosing triangle\n");
    exit (1);
  }
  
  {
    triangulation_node_p an[3];
    double *amodes[3], ax[3], ay[3];
    tri_triangle_get_nodes (tri, an);
    for (i = 0; i < 3; i++) {
      tri_node_get_xy (an[i], &ax[i], &ay[i]);
      amodes[i] = tri_node_get_assoc_data (an[i], CLIENT);
      if (amodes[i] == 0) {
        fprintf (stderr, "No modes\n");
        exit (1);
      }
    }
    for (i = 0; i < 3*nmodes; i++) {
      double v = G3D_interpolate_on_tri2d (x, y, ax[0], ay[0], ax[1], ay[1], ax[2], ay[2],
                                           amodes[0][i], amodes[1][i], amodes[2][i]);
      modes[i] = v;
    }
  }
}

static void
subtract (triangulation_t t, int which, double d, int from, int nmodes)
{
  triangulation_node_p n;
  double *modes;
  dict_iterator di = dict_get_iterator ();
  dict_reset_iterator (di);
  while (n = tri_next_node (t, di)) {
    int j;
    modes = tri_node_get_assoc_data (n, CLIENT);
    for (j = 0; j < 3; j++) { modes[3*from+j] -= d * modes[3*which+j]; }
  }
  dict_del_iterator (di);
}

static double
dot (triangulation_t t, int f1, int f2, int nmodes)
{
  double d = 0;
  triangulation_node_p n;
  double *modes;
  dict_iterator di = dict_get_iterator ();
  dict_reset_iterator (di);
  while (n = tri_next_node (t, di)) {
    int j;
    modes = tri_node_get_assoc_data (n, CLIENT);
    for (j = 0; j < 3; j++) { d += modes[3*f1+j] * modes[3*f2+j]; }
  }
  dict_del_iterator (di);
  return d;
}

static void
normalize (triangulation_t t, int mode, int nmodes)
{
  triangulation_node_p n;
  double *modes, sum;
  dict_iterator di = dict_get_iterator ();
  sum = dot (t, mode, mode, nmodes);
  sum = sqrt (sum);
  dict_reset_iterator (di);
  while (n = tri_next_node (t, di)) {
    int j;
    modes = tri_node_get_assoc_data (n, CLIENT);
    for (j = 0; j < 3; j++) { modes[3*mode+j] /= sum; }
  }
  dict_del_iterator (di);
}

static void
gram_schmidt (triangulation_t t, int nmodes)
{
  int i, j, k;
  double d;
  j = 0;
  normalize (t, 0, nmodes);
  for (i = 1; i < nmodes; i++) {
    for (k = i; k < nmodes; k++) {
      d = dot (t, k, j, nmodes);
      subtract (t, j, d, k, nmodes);
    }
    j++;
    normalize (t, j, nmodes);
  }
}

static void
transfer_pod (triangulation_t srctri, triangulation_t desttri, int nmodes)
{
  double max_edge_length = tri_max_edge_length (srctri);
  pirat_front_end_t *pirat = pirat_new_pirat (20, 1e-3);

  { /* add source nodes to the search structure */
    triangulation_node_p n;
    double max_edge_length;
    dict_iterator di = dict_get_iterator ();
    dict_reset_iterator (di);
    while (n = tri_next_node (srctri, di)) {
      double x, y;
      G3D_loc_t loc; 
      tri_node_get_xy (n, &loc.x, &loc.y); loc.z = 0;
      pirat_add_pt (pirat, n, &loc);
    }
    dict_del_iterator (di);
  }    

  { /* interpolate */
    triangulation_node_p n;
    dict_iterator di = dict_get_iterator ();
    
    dict_reset_iterator (di);
    while (n = tri_next_node (desttri, di)) {
      interpolate (n, nmodes, pirat, max_edge_length);
    }
    
    dict_del_iterator (di);
  }

#if 1
  gram_schmidt (desttri, nmodes); 
#endif
}

static int
count_nodes (triangulation_t t)
{
  triangulation_node_p n;
  int count = 0;
  dict_iterator di = dict_get_iterator ();
  dict_reset_iterator (di);
  while (n = tri_next_node (t, di)) {
    count++;
  }
  dict_del_iterator (di);
  return count;
}
  

static void
write_pod (triangulation_t desttri, int nmodes)
{
  triangulation_node_p n;
  double *modes;
  int mode, i, j;
  int nnodes = count_nodes (desttri);
  
  for (mode = 0; mode < nmodes; mode++) {
    fprintf (stdout, "MODE %d EIGENVALUE 0.0 (0%)\n", mode);
    for (i = 0; i < nnodes; i++) {
      n = tri_find_node (desttri, i+1);
      if (!n) {
        fprintf (stderr, "No such node: %d\n", i+1);
        exit (1);
      }
      if (tri_node_get_id (n) != i+1) {
        fprintf (stderr, "No such node: %d\n", i+1);
        exit (1);
      }
      modes = tri_node_get_assoc_data (n, CLIENT);
      fprintf (stdout, "%d %g %g %g\n", tri_node_get_id (n), modes[3*mode+0], modes[3*mode+1], modes[3*mode+2]);
    }
    fprintf (stdout, "END MODE %d\n", mode);
  }
}

static void
write_mode_comp_on_triangulation (triangulation_t t, int nmodes, char *file, int mode, int comp, double scale, char *color)
{
  int i;
  triangulation_triangle_p tri;
  dict_iterator di = dict_get_iterator ();
  FILE *fp = fopen (file, "w");
  
  fprintf (fp, "# ELIXIR file (C) 1994 Petr Krysl; version 1\n");

  dict_reset_iterator (di);
  while (tri = tri_next_triangle (t, di)) {
    triangulation_node_p an[3];
    double *amodes[3], ax[3], ay[3];
    fprintf (fp, "989685046 28 7 0 0 0\n");
    tri_triangle_get_nodes (tri, an);
    for (i = 0; i < 3; i++) {
      tri_node_get_xy (an[i], &ax[i], &ay[i]);
      amodes[i] = tri_node_get_assoc_data (an[i], CLIENT);
      fprintf (fp, "%g %g %g ", ax[i], ay[i], scale*amodes[i][3*mode+comp]);
    }
    fprintf (fp, "1 0 0 1\n%s\nblack\n", color);
  }

  fprintf (fp, "989685028 0 6 0 0 0\n1\n0\n0\nyellow\n0\n0\nfixed\n0\n""0\n0\n2\n0.900000\n0.523599\n1\n1\n31\n0\n0\nblack\n0\n0\n");

  dict_del_iterator (di);
  fclose (fp);
}

static void
graphic_debug (triangulation_t srctri, triangulation_t desttri, int nmodes)
{
  double scale = 100;
  int mode, comp;
  char buf[512];
  for (mode = 0; mode < nmodes; mode++) {
    for (comp = 0; comp < 2; comp++) {
      sprintf (buf, "src-%d-%d.egf", mode, comp);
      write_mode_comp_on_triangulation (srctri, nmodes, buf, mode, comp, scale, "yellow");
      sprintf (buf, "dest-%d-%d.egf", mode, comp);
      write_mode_comp_on_triangulation (desttri, nmodes, buf, mode, comp, scale, "red");
    }
  }
}

int 
main (int argc, char **argv)
{
  triangulation_t srctri;
  triangulation_t desttri;
  int nmodes;

  if (argc < 5) {
    fprintf (stderr, "Expected arguments: nmodes srctri desttri podfile\n");
    exit (1);
  }

  /* number of modes */
  nmodes = atoi (argv[1]);
  if (nmodes < 1) {
    fprintf (stderr, "Suspect nmodes: %d\n", nmodes);
    exit (1);
  }

  /* load dest triangulation */
  srctri = load_triangulation (argv[2], nmodes);

  /* load src triangulation */
  desttri = load_triangulation (argv[3], nmodes);
    
  /* load pod file */
  load_pod_file (srctri, argv[4], nmodes);

  /* transfer pod data */
  transfer_pod (srctri, desttri, nmodes);

  /* write out */
  write_pod (desttri, nmodes);

  /*  graphic_debug (srctri, desttri, nmodes); */
}
