#ifndef __CONTROLTRIANGLE_H__
#define __CONTROLTRIANGLE_H__

#include "Loop/MultiresTriangle.h"
#include "Loop/Permutation.h"
#include "Mesh/TriangleTp.h"

template <class TriangleIf, class VertexIf>
class Accessor;

/**
 * ControlTriangle Class
 */
template <class TriangleIf, class VertexIf, class VertexMr>
class ControlTriangle : public TriangleTp<TriangleIf, VertexIf> {
  friend class Accessor<TriangleIf, VertexIf>;

public:
  typedef MultiresTriangle<TriangleIf, VertexIf, VertexMr> TriangleMr;

private:
  /**
   * The encoding of the permutation mapping for the triangles adjacent to
   * this triangle.
   */
  short m_perm1:4, m_perm2:4, m_perm3:4;

  /** The root of the subdivision hierarchy at this triangle. */
  TriangleMr* m_root;

public:
  /**
   * Constructs an uninitialized triangle.  Initially the triangle does not
   * belong to the topology of a mesh.
   */
  ControlTriangle(void)
   : TriangleTp<TriangleIf, VertexIf>(),
     m_perm1(0), m_perm2(0), m_perm3(0),
     m_root(0) { }
  
  /**
   * Destructor.
   */
  ~ControlTriangle(void) {
    if (m_root)
      delete (m_root);
  }
  
  /**
   * Returns the permutation mapping for the adjacent triangle corresponding
   * to the specified edge.
   * @param e  triangle edge
   * @return permutation mapping for adjacent triangle across edge
   */
  Permutation permutation(const EdgeLabel e) const {
    switch (e) {
    case E1:
      return Permutation(m_perm1);
    case E2:
      return Permutation(m_perm2);
    case E3:
      return Permutation(m_perm3);
    }

    assert(false);
    return Permutation();
  }

  /**
   * Returns an immutable reference to the root of the subdivision hierarchy
   * at this triangle.
   * @return immutable reference to root of subdivision hierarchy at this
   *         triangle
   */
  TriangleMr* root(void) const {
    return (m_root);
  }

  /**
   * Defines the vertices of this triangle, in counter-clockwise orientation.
   * The triangle must be non-degenerate: each vertex must be unique.  This
   * method must be called exactly once for each triangle.
   * @param v1  first vertex
   * @param v2  second vertex
   * @param v3  third vertex
   */
  void setVertices(VertexIf* v1, VertexIf* v2, VertexIf* v3) {
    TriangleTp<TriangleIf, VertexIf>::setVertices(v1, v2, v3);

    m_root = new TriangleMr((TriangleIf*)this);
    assert(m_root);
  }

protected:
  void setAdjacentTriangle(const EdgeLabel e, TriangleIf* triangle) {
    TriangleTp<TriangleIf, VertexIf>::setAdjacentTriangle(e, triangle);

    switch (e) {
    case E1:
      m_perm1 = Permutation::encode(this, e);
      break;
    case E2:
      m_perm2 = Permutation::encode(this, e);
      break;
    case E3:
      m_perm3 = Permutation::encode(this, e);
      break;
    }
  }

private:
  /**
   * Copy constructor: undefined operation.
   */
  ControlTriangle(const ControlTriangle& instance) {
    assert(false);
  }

  /**
   * Assignment operator: undefined operation.
   */
  ControlTriangle& operator=(const ControlTriangle& instance) {
    assert(false);
    return (*this);
  }
};


#endif  /* __CONTROLTRIANGLE_H__ */
