#ifndef __LOOPMESH_H__
#define __LOOPMESH_H__

#include "Loop/LoopReader.h"
#include "Loop/MultiresTriangle.h"
#include "Loop/LoopTesselator.h"
#include "Mesh/MeshTp.h"

/**
 * LoopMesh Class
 */
template <class TriangleIf,
          class VertexIf,
          class VertexMr,
          class Stencil>
class LoopMesh : public MeshTp<TriangleIf, VertexIf> {
public:
  typedef MultiresTriangle<TriangleIf, VertexIf, VertexMr> TriangleMr;
  
private:
  /** The subdivision stencil for this Loop subdivision mesh. */
  Stencil& m_stencil;

public:
  /**
   * Constructs an empty mesh with a specified subdivision stencil.
   * @param stencil  subdivision stencil
   */
  LoopMesh(Stencil& stencil)
   : MeshTp<TriangleIf, VertexIf>(), m_stencil(stencil) { }
  
  /**
   * Destructor.
   */
  ~LoopMesh(void) { }
  
  /**
   * Ensures that the one-ring neighbours of the specified subdivision triangle
   * exist and are valid.
   * @param triangle  triangle
   */
  void generateOneRing(TriangleMr* triangle) {
    if (triangle->isRoot())
      return;

    TriangleMr* parent = triangle->parent();
    Permutation p;
    for (int v = V1; v <= V3; v += 1) {
      // In 3/4 of all cases this point is incident to an edge of the
      // triangle's parent.
      if (VertexLabel(v) != ParentVertex(triangle->label())) {
	EdgeLabel e = OppositeEdge(VertexLabel(v));
	TriangleMr* adjacent = parent->adjacentTriangle(e, p);
	if (adjacent != 0)
	  subdivide(adjacent, triangle->depth());
      } else {
	VertexLabel pivot = VertexLabel(v);
	VertexLabel outer = CWVertex(pivot);
	TriangleMr::Orientation direction = TriangleMr::COUNTERCLOCKWISE;
	TriangleMr* next = parent;

	do {
	  subdivide(next, triangle->depth());
	  EdgeLabel e = OppositeEdge(outer);
	  next = next->adjacentTriangle(e, p);
	  if (next == 0) {
	    if (direction == TriangleMr::CLOCKWISE) {
	      break;
	    } else {
	      pivot = VertexLabel(v);
	      outer = CCWVertex(pivot);
	      direction = TriangleMr::CLOCKWISE;
	      next = parent;
	      continue;
	    }
	  }

	  pivot = p.vertex(pivot);
	  outer = OppositeVertex(pivot, p.vertex(outer));
	} while (next != parent);
      }
    }
  }

  /**
   * Evaluates the limit surface at a designated vertex in the mesh.
   * @param triangle  triangle containing vertex
   * @param v  label of vertex in triangle
   * @return limit position of vertex
   */
  CVec limit(TriangleMr* triangle, VertexLabel v) {
    // NJL: Guarantee existence of one-ring neighbours here.
    return (m_stencil.limit(triangle, v));
  }

  /**
   * Subdivides this Loop subdivision mesh uniformly to the specified depth.
   * @param depth  depth to which to uniformly subdivide this mesh
   */
  void subdivide(int depth) {
    TriangleIt ti = triangleBegin();
    TriangleIt te = triangleEnd();
    for (; ti != te; ++ti) {
      TriangleIf* triangle = *ti;
      subdivide(triangle->root(), depth);
    }
  }

  /**
   * Subdivides the specified triangle in the subdivision hierarchy to the
   * specified depth.
   * @param triangle  triangle to subdivide
   * @param depth  depth to which to subdivide triangle
   */
  void subdivide(TriangleMr* triangle, int depth) {
    assert(depth < __MAX_DEPTH);
    assert(triangle->depth() <= depth);

    int below = triangle->depth() + 1;
    if (below > depth) {
      return;
    } else if (!triangle->isLeaf()) {
      // Determine if triangle vertices need to be recomputed.
      TriangleMr* centre = triangle->child(T0);
      if (triangle->vertex(V1)->isValid(below) &&
	  triangle->vertex(V2)->isValid(below) &&
	  triangle->vertex(V3)->isValid(below) &&
	  centre->vertex(V1)->isValid(below) &&
	  centre->vertex(V2)->isValid(below) &&
	  centre->vertex(V3)->isValid(below))
      {
	if (depth > below) {
	  for (int t = T0; t <= T3; t += 1) {
	    subdivide(triangle->child(TriangleLabel(t)), depth);
	  }
	}
	return;
      }
    } else if (below <= depth) {
      generateOneRing(triangle);
      triangle->expand();
    }

    for (int v = V1; v <= V3; v += 1) {
      if (!triangle->vertex(VertexLabel(v))->isValid(below))
	m_stencil.even(triangle, VertexLabel(v));
    }

    TriangleMr* centre = triangle->child(T0);
    for (int e = E1; e <= E3; e += 1) {
      if (!centre->vertex(OppositeVertex(EdgeLabel(e)))->isValid(below))
	m_stencil.odd(triangle, EdgeLabel(e));
    }

    if (depth > below) {
      for (int t = T0; t <= T3; t += 1) {
	subdivide(triangle->child(TriangleLabel(t)), depth);
      }
    }
  }
};

template <class TriangleIf, class VertexIf, class VertexMr, class Stencil>
istream& operator>>(istream& is,
		    LoopMesh<TriangleIf, VertexIf, VertexMr, Stencil>& mesh)
{
  LoopReader< LoopMesh<TriangleIf, VertexIf, VertexMr, Stencil> > reader;
  reader.readMesh(is, mesh);
  return (is);
}

template <class TriangleIf, class VertexIf, class VertexMr, class Stencil>
ostream& operator<<(ostream& os,
		    LoopMesh<TriangleIf, VertexIf, VertexMr, Stencil>& mesh)
{
    LoopTesselator< LoopMesh<TriangleIf, VertexIf, VertexMr, Stencil> > t;
    os << (*t(mesh));
    return (os);
}

#define LOOPTYPE(LTYPE, LCLASS, TCLASS, VCLASS, MCLASS, SCLASS)               \
    class LTYPE##_Triangle;                                                   \
    class LTYPE##_Vertex;                                                     \
                                                                              \
    class LTYPE##_Triangle : public TCLASS<LTYPE##_Triangle,                  \
                                           LTYPE##_Vertex,                    \
                                           MCLASS>                            \
    {                                                                         \
    public:                                                                   \
      LTYPE##_Triangle(void)                                                  \
       : TCLASS<LTYPE##_Triangle, LTYPE##_Vertex, MCLASS>() { }               \
      ~LTYPE##_Triangle(void) { }                                             \
    };                                                                        \
                                                                              \
    class LTYPE##_Vertex : public VCLASS<LTYPE##_Triangle,                    \
                                         LTYPE##_Vertex,                      \
                                         MCLASS>                              \
    {                                                                         \
    public:                                                                   \
      LTYPE##_Vertex(void)                                                    \
       : VCLASS<LTYPE##_Triangle, LTYPE##_Vertex, MCLASS>() { }               \
      ~LTYPE##_Vertex(void) { }                                               \
    };                                                                        \
                                                                              \
    typedef LCLASS<LTYPE##_Triangle, LTYPE##_Vertex, MCLASS, SCLASS> LTYPE;   \

#endif  /* __LOOPMESH_H__ */
