#ifndef __LOOPSTENCIL_H__
#define __LOOPSTENCIL_H__

#include <assert.h>
#include "Loop/Topology.h"
#include "Util/cvec.h"

/**
 * LoopStencil Class
 */
class LoopStencil {
public:
  /**
   * Default constructor.
   */
  LoopStencil(void) { }

  /**
   * Copy constructor.
   */
  LoopStencil(const LoopStencil& instance) { }

  /**
   * Destructor.
   */
  ~LoopStencil(void) { }

  /**
   * Assignment operator.
   */
  LoopStencil& operator=(const LoopStencil& instance) {
    return (*this);
  }

  /**
   * Applies the Loop stencil for even vertices to a specified vertex of a
   * given subdivision triangle.  The new position is stored in the
   * subdivision hierarchy and a mutable reference is returned.
   * @param triangle  subdivision triangle
   * @param v  vertex to which to apply subdivision rule
   * @return mutable reference to computed position
   */
  template <class TriangleMr>
  CVec& even(TriangleMr* triangle, VertexLabel v) {
    TriangleMr::Vertex* vertex = triangle->vertex(v);
    int depth = triangle->depth();
    CVec even;

    // In 9/16 of all cases, this point is incident to an edge of the
    // triangle's parent.
    TriangleLabel t = triangle->label();
    if (!triangle->isRoot() && (t == T0 || v != ParentVertex(t))) {
      EdgeLabel e = OppositeEdge(v);
      VertexLabel vi = CCWVertex(e);
      VertexLabel vj = CWVertex(e);
      Permutation p;

      TriangleMr* parent = triangle->parent();
      TriangleMr* adjacent = parent->adjacentTriangle(e, p)->child(T0);
      TriangleMr* centre = parent->child(T0);

      even = 0.625f * vertex->point(depth) + 0.0625f *
	     (parent->vertex(vi)->point(depth) +
	      parent->vertex(vj)->point(depth) +
	      centre->vertex(vi)->point(depth) +
	      centre->vertex(vj)->point(depth) +
	      adjacent->vertex(p.vertex(vi))->point(depth) +
	      adjacent->vertex(p.vertex(vj))->point(depth));
    } else {
      int valence = triangle->valence(v);
      float n = (float)valence;
      float gamma = 3.0f + 2.0f * cos(2.0f * M_PI / n);
      float beta = (0.125f / n) * (5.0f - gamma * gamma / 8.0f);
      float alpha = 1.0f - (n * beta);
      even = alpha * vertex->point(depth);

      TriangleMr::VertexIt vi = triangle->vertexBegin(v);
      TriangleMr::VertexIt ve = triangle->vertexEnd(v);
      for (int i = 0; vi != ve; i += 1, ++vi) {
	even += beta * (*vi)->point(depth);
      }

      assert(i == valence);
    }

    vertex->point(depth + 1) = even;
    vertex->setValid(depth + 1);
    return (vertex->point(depth + 1));
  }

  /**
   * Applies the Loop stencil to calculate the limit position of a specified 
   * vertex of a given subdivision triangle.
   * @param triangle  subdivision triangle
   * @param v  vertex for which to calculate the limit position
   * @return  limit position of vertex
   */
  template <class TriangleMr>
  CVec limit(TriangleMr* triangle, VertexLabel v) {
    TriangleMr::Vertex* vertex = triangle->vertex(v);
    int depth = triangle->depth();

    int valence = triangle->valence(v);
    float n = (float)valence;
    float gamma = 3.0f + 2.0f * cos(2.0f * M_PI / n);
    float beta = (1.0f / n) * (5.0f - gamma * gamma / 8.0f);
    float omega = 3.0f / beta;
    float rho = 1.0f / (omega + n);
    CVec limit = (omega * rho) * vertex->point(depth);

    TriangleMr::VertexIt vi = triangle->vertexBegin(v);
    TriangleMr::VertexIt ve = triangle->vertexEnd(v);
    for (int i = 0; vi != ve; i += 1, ++vi) {
      limit += rho * (*vi)->point(depth);
    }
    
    assert(i == valence);
    return (limit);
  }

  /**
   * Applies the Loop stencil for odd vertices to a specified edge of a given
   * subdivision triangle.  The new position is stored in the subdivision
   * hierarchy and a mutable reference is returned.
   * @param triangle  subdivision triangle
   * @param e  edge to which to apply subdivision rule
   * @return mutable reference to computed position
   */
  template <class TriangleMr>
  CVec& odd(TriangleMr* triangle, EdgeLabel e) {
    int depth = triangle->depth();
    VertexLabel v = OppositeVertex(e);
    Permutation p;
    TriangleMr* adjacent = triangle->adjacentTriangle(e, p);

    CVec odd = 0.375f * (triangle->vertex(CCWVertex(e))->point(depth) +
			 triangle->vertex(CWVertex(e))->point(depth)) +
               0.125f * (triangle->vertex(v)->point(depth) +
			 adjacent->vertex(p.vertex(v))->point(depth));

    TriangleMr::Vertex* vertex = triangle->child(T0)->vertex(v);
    vertex->point(depth + 1) = odd;
    vertex->setValid(depth + 1);
    return (vertex->point(depth + 1));
  }
};

#endif  /* __LOOPSTENCIL_H__ */
