#ifndef __LOOPTESSELATOR_H__
#define __LOOPTESSELATOR_H__

#include <hash_map.h>
#include <stack.h>
#include "Loop/TesselatedMesh.h"
#include "Loop/Topology.h"
#include "Mesh/MeshTp.h"
#include "Mesh/PointVertex.h"
#include "Mesh/TriangleTp.h"

/**
 * LoopTesselator Class
 */
template <class Loop>
class LoopTesselator {
private:
  class Tesselator_Triangle;
  class Tesselator_Vertex;

  class Tesselator_Triangle
   : public TesselatedTriangle<Tesselator_Triangle,
                               Tesselator_Vertex,
                               Loop::Triangle>
  {
  public:
    Tesselator_Triangle(Loop::Triangle* domain)
     : TesselatedTriangle<Tesselator_Triangle,
                          Tesselator_Vertex,
                          Loop::Triangle>(domain) { }
    ~Tesselator_Triangle(void) { }
  };

  class Tesselator_Vertex
   : public TesselatedVertex<Tesselator_Triangle,
                             Tesselator_Vertex,
                             Loop::Triangle>
  {
  public:
    Tesselator_Vertex(TesselatedVertexTag<Loop::Triangle>* tag)
     : TesselatedVertex<Tesselator_Triangle,
                        Tesselator_Vertex,
                        Loop::Triangle>(tag) { }
    ~Tesselator_Vertex(void) { }
  };

public:
  typedef TesselatedMesh<Tesselator_Triangle, Tesselator_Vertex> Mesh;
  typedef Mesh::Triangle Triangle;
  typedef Mesh::Vertex Vertex;
  typedef Assembler<Triangle, Vertex> Assembler;
  typedef Loop::TriangleMr TriangleMr;
  typedef TriangleMr::Orientation Orientation;

public:
  LoopTesselator(bool limitFlag = true) : m_limitFlag(limitFlag) { }

  ~LoopTesselator(void) { }

  Mesh* operator()(Loop& loopMesh) {
    return (tesselate(loopMesh));
  }

  Mesh* tesselate(Loop& loopMesh, int depth = __MAX_DEPTH) {
    assert(depth >= 0);
    Mesh* mesh = new Mesh();
    assert(mesh);
    m_assembler = new Assembler(*mesh);
    assert(m_assembler);
    m_loop = &loopMesh;

    Loop::const_TriangleIt ti = m_loop->triangleBegin();
    Loop::const_TriangleIt te = m_loop->triangleEnd();
    for (; ti != te; ++ti) {
      TriangleMr* triangle = (*ti)->root();
      m_stack.push(triangle);

      while (!m_stack.empty()) {
	triangle = m_stack.top();
	m_stack.pop();

	if (triangle->isLeaf() || triangle->depth() == depth) {
	  tesselate(triangle, depth);
	} else if (triangle->depth() < depth) {
	  for (int t = T0; t <= T3; t += 1) {
	    m_stack.push(triangle->child(TriangleLabel(t)));
	  }
	}
      }
    }

    m_assembler->assembleMesh();
    delete (m_assembler);
    m_loop = 0;
    m_table.clear();

    return (mesh);
  }

private:
  Assembler* m_assembler;

  Loop* m_loop;

  stack<TriangleMr*> m_stack;

  hash_map<TriangleMr::Vertex*, int, TriangleMr::Vertex::hash> m_table;

  bool m_limitFlag;

  void tesselate(TriangleMr* triangle, int depth) {
    Loop::Triangle* domain = triangle->triangle();
    Orientation orientation = triangle->orientation();
    Vertex* vertices[6];

    for (int v = V1; v <= V3; v += 1) {
      vertices[v] = vertex(triangle, VertexLabel(v));
    }

    int mask = 0;
    Permutation p;
    for (int e = E1; e <= E3; e += 1) {
      TriangleMr* adjacent = triangle->adjacentTriangle(EdgeLabel(e), p);
      if (!adjacent || adjacent->isLeaf() || adjacent->depth() == depth)
	continue;

      VertexLabel o = OppositeVertex(EdgeLabel(e));
      vertices[e + 3] = vertex(adjacent->child(T0), p.vertex(o));
      mask |= 1 << e;
    }

    switch (mask) {
    case 0:
      addTriangle(domain, vertices[0], vertices[1], vertices[2], orientation);
      break;
      
    case 1:
      addTriangle(domain, vertices[0], vertices[3], vertices[2], orientation);
      addTriangle(domain, vertices[0], vertices[1], vertices[3], orientation);
      break;
      
    case 2:
      addTriangle(domain, vertices[0], vertices[1], vertices[4], orientation);
      addTriangle(domain, vertices[1], vertices[2], vertices[4], orientation);
      break;
      
    case 3:
      addTriangle(domain, vertices[0], vertices[1], vertices[3], orientation);
      addTriangle(domain, vertices[0], vertices[3], vertices[4], orientation);
      addTriangle(domain, vertices[2], vertices[4], vertices[3], orientation);
      break;
      
    case 4:
      addTriangle(domain, vertices[0], vertices[5], vertices[2], orientation);
      addTriangle(domain, vertices[1], vertices[2], vertices[5], orientation);
      break;
      
    case 5:
      addTriangle(domain, vertices[0], vertices[5], vertices[2], orientation);
      addTriangle(domain, vertices[1], vertices[3], vertices[5], orientation);
      addTriangle(domain, vertices[2], vertices[5], vertices[3], orientation);
      break;
      
    case 6:
      addTriangle(domain, vertices[0], vertices[5], vertices[4], orientation);
      addTriangle(domain, vertices[1], vertices[2], vertices[4], orientation);
      addTriangle(domain, vertices[1], vertices[4], vertices[5], orientation);
      break;
      
    case 7:
      addTriangle(domain, vertices[0], vertices[5], vertices[4], orientation);
      addTriangle(domain, vertices[1], vertices[3], vertices[5], orientation);
      addTriangle(domain, vertices[2], vertices[4], vertices[3], orientation);
      addTriangle(domain, vertices[3], vertices[4], vertices[5], orientation);
      break;
    }
  }

  void addTriangle(Loop::Triangle* domain,
		   Vertex* v1,
		   Vertex* v2,
		   Vertex* v3,
		   int orientation)
  {
    Triangle* triangle = new Triangle(domain);
    assert(triangle);

    if (Orientation(orientation) == TriangleMr::COUNTERCLOCKWISE)
      triangle->setVertices(v1, v2, v3);
    else
      triangle->setVertices(v1, v3, v2);

    m_assembler->addTriangle(triangle);
  }

  Vertex* vertex(TriangleMr* triangle, VertexLabel v) {
    TriangleMr::Vertex* vertex = triangle->vertex(v);
    if (m_table.find(vertex) != m_table.end())
      return (m_assembler->getVertex(m_table[vertex]));

    Loop::Triangle* domain = triangle->triangle();
    TesselatedVertexTag<Loop::Triangle>* tag;
    if (triangle->vertex(v)->isControlVertex()) {
      tag = new TesselatedControlVertexTag<Loop::Triangle>(domain, v);
    } else {
      CVec coords(0.0f, 0.0f, 0.0f);
      coords(v) = 1.0f;

      TriangleMr* next = triangle;
      while (!next->isRoot()) {
	switch (next->label()) {
	case T0:
	  coords = coords.x() * CVec(0.0f, 0.5f, 0.5f) +
	           coords.y() * CVec(0.5f, 0.0f, 0.5f) +
	           coords.z() * CVec(0.5f, 0.5f, 0.0f);
	  break;

	case T1:
	  coords = coords.x() * CVec(1.0f, 0.0f, 0.0f) +
	           coords.y() * CVec(0.5f, 0.0f, 0.5f) +
	           coords.z() * CVec(0.5f, 0.5f, 0.0f);
	  break;

	case T2:
	  coords = coords.x() * CVec(0.0f, 0.5f, 0.5f) +
	           coords.y() * CVec(0.0f, 1.0f, 0.0f) +
	           coords.z() * CVec(0.5f, 0.5f, 0.0f);
	  break;

	case T3:
	  coords = coords.x() * CVec(0.0f, 0.5f, 0.5f) +
	           coords.y() * CVec(0.5f, 0.0f, 0.5f) +
	           coords.z() * CVec(0.0f, 0.0f, 1.0f);
	  break;
	}

	next = next->parent();
      }

      if (coords.x() == 0.0f)
	tag = new TesselatedEdgeVertexTag<Loop::Triangle>(domain, E1, coords);
      else if (coords.y() == 0.0f)
	tag = new TesselatedEdgeVertexTag<Loop::Triangle>(domain, E2, coords);
      else if (coords.z() == 0.0f)
	tag = new TesselatedEdgeVertexTag<Loop::Triangle>(domain, E3, coords);
      else
	tag = new TesselatedInteriorVertexTag<Loop::Triangle>(domain, coords);
    }
    assert(tag);

    Vertex* newVertex = new Vertex(tag);
    assert(newVertex);
    if (m_limitFlag == true)
      newVertex->point() = m_loop->limit(triangle, v);
    else
      newVertex->point() = vertex->point(triangle->depth());

    int index = m_assembler->addVertex(newVertex);
    m_table[vertex] = index;
    return (newVertex);				
  }
};

#endif  /* __TESSELATOR_H__ */
