#ifndef __MULTIRESTRIANGLE_H__
#define __MULTIRESTRIANGLE_H__

#include "Loop/ControlVertex.h"
#include "Loop/Permutation.h"
#include "Loop/Topology.h"

/**
 * MultiresTriangle Class
 */
template <class TriangleIf, class VertexIf, class VertexMr>
class MultiresTriangle {
public:
  enum Orientation { CLOCKWISE = 0, COUNTERCLOCKWISE = 1 };
  typedef TriangleIf Triangle;
  typedef VertexMr Vertex;

public:
  /**
   * MultiresTriangle::TriangleIt Class
   */
  class TriangleIt : public forward_iterator<MultiresTriangle*, int> {
    friend class MultiresTriangle;
    friend class OneRingIterator;

  private:
    Orientation m_direction;

    VertexLabel m_first;

    MultiresTriangle* m_last;

    MultiresTriangle* m_next;

    VertexLabel m_pivot;

    VertexLabel m_vertex;

  private:
    TriangleIt(VertexLabel vertex,
	       MultiresTriangle* triangle,
	       bool end = false)
     : m_direction(COUNTERCLOCKWISE),
       m_first(vertex),
       m_last(triangle),
       m_next(end ? 0 : triangle),
       m_pivot(vertex),
       m_vertex(PreviousVertex(vertex)) { }

  public:
    TriangleIt(void) { }

    TriangleIt(const TriangleIt& instance)
     : m_direction(instance.m_direction),
       m_first(instance.m_first),
       m_last(instance.m_last),
       m_next(instance.m_next),
       m_pivot(instance.m_pivot),
       m_vertex(instance.m_vertex) { }

    TriangleIt& operator=(const TriangleIt& instance) {
      m_direction = instance.m_direction;
      m_first = instance.m_first;
      m_last = instance.m_last;
      m_next = instance.m_next;
      m_pivot = instance.m_pivot;
      m_vertex = instance.m_vertex;
      return (*this);
    }

    bool operator==(const TriangleIt& instance) const {
      return (instance.m_next == m_next &&
	      instance.m_first == m_first &&
	      instance.m_last == m_last);
    }

#ifdef _WIN32
    bool operator!=(const TriangleIt& instance) const {
      return !(instance == *this);
    }
#endif

    MultiresTriangle* operator*(void) const {
      return (m_next);
    }

    TriangleIt& operator++(void) {
      Permutation p;
      EdgeLabel e = IncidentEdge(m_vertex, m_pivot);
      m_next = m_next->adjacentTriangle(e, p);

      if (!m_next) {
	if (m_direction == COUNTERCLOCKWISE) {
	  e = IncidentEdge(NextVertex(m_first), m_first);
	  m_next = m_last->adjacentTriangle(e, p);
	  if (m_next) {
	    m_pivot = p.vertex(m_first);
	    m_vertex = OppositeVertex(p.edge(e));
	    m_direction = CLOCKWISE;
	  }
	} else {
	  m_next = 0;
	}
      } else if (m_next == m_last) {
	m_next = 0;
      } else {
	m_pivot = p.vertex(m_pivot);
	m_vertex = OppositeVertex(p.vertex(m_vertex), m_pivot);
      }

      return (*this);
    }	

    TriangleIt operator++(int) {
      TriangleIt iterator = *this;
      ++(*this);
      return (iterator);
    }
  };  /* MultiresVertex::TriangleIt */

  /**
   * MultiresTriangle::VertexIt Class
   */
  class VertexIt : public forward_iterator<MultiresTriangle*, int> {
    friend class MultiresTriangle;

  private:
    Orientation m_direction;

    VertexLabel m_first;

    MultiresTriangle* m_last;

    VertexMr* m_next;

    VertexLabel m_pivot;

    MultiresTriangle* m_triangle;

    VertexLabel m_vertex;

  private:
    VertexIt(VertexLabel vertex,
	     MultiresTriangle* triangle,
	     bool end = false)
     : m_direction(COUNTERCLOCKWISE),
       m_first(vertex),
       m_last(triangle),
       m_next(end ? 0 : triangle->vertex(PreviousVertex(vertex))),
       m_pivot(vertex),
       m_triangle(triangle),
       m_vertex(PreviousVertex(vertex))
    {
      if (!end) {
	m_next = triangle->vertex(m_vertex);
	if (!m_next) {
	  m_vertex = NextVertex(m_pivot);
	  m_next = triangle->vertex(m_vertex);
	  m_direction = CLOCKWISE;
	}
      }
    }

  public:
    VertexIt(void) { }

    VertexIt(const VertexIt& instance)
     : m_direction(instance.m_direction),
       m_first(instance.m_first),
       m_last(instance.m_last),
       m_next(instance.m_next),
       m_pivot(instance.m_pivot),
       m_triangle(instance.m_triangle),
       m_vertex(instance.m_vertex) { }

    VertexIt& operator=(const VertexIt& instance) {
      m_direction = instance.m_direction;
      m_first = instance.m_first;
      m_last = instance.m_last;
      m_next = instance.m_next;
      m_pivot = instance.m_pivot;
      m_vertex = instance.m_vertex;
      return (*this);
    }

    bool operator==(const VertexIt& instance) const {
      return (instance.m_next == m_next &&
	      instance.m_first == m_first &&
	      instance.m_last == m_last);
    }

#ifdef _WIN32
    bool operator!=(const VertexIt& instance) const {
      return !(instance == *this);
    }
#endif

    Vertex* operator*(void) const {
      return (m_next);
    }

    VertexIt& operator++(void) {
      Permutation p;
      EdgeLabel e = IncidentEdge(m_vertex, m_pivot);
      m_triangle = m_triangle->adjacentTriangle(e, p);
	
      if (!m_triangle) {
	if (m_direction == COUNTERCLOCKWISE) {
	  m_triangle = m_last;
	  m_pivot = m_first;
	  m_vertex = NextVertex(m_pivot);
	  m_next = m_triangle->vertex(m_vertex);
	  m_direction = CLOCKWISE;
	} else {
	  m_next = 0;
	}
      } else if (m_triangle == m_last) {
	m_next = 0;
      } else {
	m_pivot = p.vertex(m_pivot);
	m_vertex = OppositeVertex(p.vertex(m_vertex), m_pivot);
	m_next = m_triangle->vertex(m_vertex);
      }

      return (*this);
    }	

    VertexIt operator++(int) {
      VertexIt iterator = *this;
      ++(*this);
      return (iterator);
    }
  };  /* MultiresVertex::VertexIt */

  /**
   * MultiresTriangle::OneRingIt Class
   */
  class OneRingIt : public forward_iterator<MultiresTriangle*, int> {
    friend class MultiresTriangle;

  private:
    bool m_end;

    MultiresTriangle* m_exclude;

    TriangleIt m_te;

    TriangleIt m_ti;

    MultiresTriangle* m_triangle;

    VertexLabel m_vertex;

  private:
    OneRingIt(MultiresTriangle* triangle, bool end = false)
     : m_end(end), m_triangle(triangle)
    {
      if (end)
	return;

      for (int v = V1; v <= V3; v += 1) {
	m_vertex = VertexLabel(v);

	EdgeLabel e = OppositeEdge(NextVertex(m_vertex));
	m_exclude = m_triangle->adjacentTriangle(e);
	m_ti = TriangleIt(m_vertex, m_triangle);
	m_te = TriangleIt(m_vertex, m_triangle, true);
	++m_ti;

	while (m_ti != m_te) {
	  if (*m_ti != m_exclude)
	    return;
	  ++m_ti;
	}
      }
    
      m_end = true;
    }

  public:
    OneRingIt(void) { }

    OneRingIt(const OneRingIt& instance)
     : m_end(instance.m_end),
       m_exclude(instance.m_exclude),
       m_te(instance.m_te),
       m_ti(instance.m_ti),
       m_triangle(instance.m_triangle),
       m_vertex(instance.m_vertex) { }

    OneRingIt& operator=(const OneRingIt& instance) {
      m_end = instance.m_end;
      m_exclude = instance.m_exclude;
      m_te = instance.m_te;
      m_ti = instance.m_ti;
      m_triangle = instance.m_triangle;
      m_vertex = instance.m_vertex;
    }

    bool operator==(const OneRingIt& instance) const {
      return (instance.m_triangle == m_triangle &&
	      instance.m_end == m_end &&
	      (m_end || instance.m_ti == m_ti));
    }

#ifdef _WIN32
    bool operator!=(const OneRingIt& instance) const {
      return !(instance == *this);
    }
#endif

    MultiresTriangle* operator*(void) const {
      return (*m_ti);
    }
    
    OneRingIt& operator++(void) {
      for (;;) {
	++m_ti;
	if (m_ti == m_te)
	  break;
	else if (*m_ti != m_exclude)
	  return (*this);
      }
      
      for (;;) {
	m_vertex = NextVertex(m_vertex);
	if (m_vertex == V1) {
	  m_end = true;
	  break;
	}

	EdgeLabel e = OppositeEdge(NextVertex(m_vertex));
	m_exclude = m_triangle->adjacentTriangle(e);
	m_ti = TriangleIt(m_vertex, m_triangle);
	m_te = TriangleIt(m_vertex, m_triangle, true);
	++m_ti;

	while (m_ti != m_te) {
	  if (*m_ti != m_exclude)
	    return (*this);
	  ++m_ti;
	}
      }

      return (*this);
    }

    OneRingIt operator++(int) {
      OneRingIt iterator = *this;
      ++(*this);
      return (iterator);
    }
  };  /* MultiresTriangle::OneRingIt */

public:
  static Orientation oppositeOrientation(Orientation orientation) {
    return (Orientation(orientation ^ 1));
  }

public:
  /**
   * Constructs a root triangle in a subdivision hierarchy.
   * @param triangle  corresponding triangle in control net
   */
  MultiresTriangle(Triangle* triangle) {
    m_children[T0] = 0;
    m_children[T1] = 0;
    m_children[T2] = 0;
    m_children[T3] = 0;
    m_depth = 0;
    m_label = T0;
    m_orientation = COUNTERCLOCKWISE;
    m_parent = 0;
    m_triangle = triangle;

    typedef ControlVertex<TriangleIf, VertexIf, VertexMr> ControlVertex;
    for (int v = V1; v <= V3; v += 1) {
      ControlVertex* vertex = (ControlVertex*)triangle->vertex(VertexLabel(v));
      m_vertices[v] = vertex->vertex();
    }
  }

  /**
   * Constructs a leaf triangle within a subdivision hierarchy.
   * @param parent  parent of this triangle in subdivision hierarchy
   * @param label  label of this triangle with respect to parent
   * @param vertex1  first multiresolution vertex for this triangle
   * @param vertex2  second multiresolution vertex for this triangle
   * @param vertex3  third multiresolution vertex for this triangle
   */
  MultiresTriangle(MultiresTriangle* parent,
		   TriangleLabel label,
		   Vertex* vertex1,
		   Vertex* vertex2,
		   Vertex* vertex3)
  {
    m_children[T0] = 0;
    m_children[T1] = 0;
    m_children[T2] = 0;
    m_children[T3] = 0;
    m_depth = parent->depth() + 1;
    m_label = label;
    m_parent = parent;
    m_vertices[V1] = vertex1;
    m_vertices[V2] = vertex2;
    m_vertices[V3] = vertex3;
    m_triangle = parent->triangle();

    switch (label) {
    case T0:
      m_orientation = parent->orientation();
      break;

    case T1:
    case T2:
    case T3:
      m_orientation = oppositeOrientation(parent->orientation());
      break;
    }
  }

  /**
   * Destructor.
   */
  ~MultiresTriangle(void) {
    if (!isLeaf) {
      for (int v = V1; v <= V3; v += 1) {
	MultiresTriangle* triangle;

	triangle = adjacentTriangle(OppositeEdge(VertexLabel(v)));
	if (!triangle) {
	  delete (child(T0)->vertex(VertexLabel(v)));
	}
      }

      for (int t = T0; t <= T3; t += 1) {
	delete (m_children[t]);
      }
    }

    if (m_parent) {
      m_parent->m_children[m_label] = 0;
    }
  }

  MultiresTriangle* adjacentTriangle(EdgeLabel e) {
    Permutation p;
    return (adjacentTriangle(e, p));
  }

  MultiresTriangle* adjacentTriangle(EdgeLabel e, Permutation& p) {
    int stack = 0;  // stack supports maximum depth 16
    int height = 0;
    MultiresTriangle* triangle = this;
    p.identity();

    // Ascend hierarchy.
    for (;;) {
      TriangleLabel t = triangle->label();
      if (t == T0) {
	if (!triangle->isRoot()) {
	  triangle = triangle->parent()->child(TriangleLabel(e + 1));  // FIX
	} else {
	  Triangle* top = triangle->triangle()->triangle(e);
	  if (top == 0)
	    return (0);

	  p = triangle->triangle()->permutation(e);
	  triangle = top->root();
	}

	break;
      } else if (t == TriangleLabel(e + 1)) {  // FIX
	triangle = triangle->parent()->child(T0);
	break;
      } else {
	stack = (stack << 2) | t;
	height += 1;
	triangle = triangle->parent();
	e = ParentEdge(t, e);
      }
    }

    // Descend hierarchy.
    while (height > 0) {
      if (triangle->isLeaf())
	return (0);  // return null if not equally subdivided

      TriangleLabel t = p.triangle(TriangleLabel(stack & 0x3));
      triangle = triangle->child(t);
      stack >>= 2;
      height -= 1;
    }

    return (triangle);
  }

  void expand(void) {
    if (!isLeaf())
      return;

    Permutation p;
    Vertex* midpoints[3];
    for (int v = V1; v <= V3; v += 1) {
      EdgeLabel e = OppositeEdge(VertexLabel(v));
      MultiresTriangle* adjacent = adjacentTriangle(e, p);
      if (adjacent != 0 && !adjacent->isLeaf()) {
	midpoints[v] = adjacent->child(T0)->vertex(p.vertex(VertexLabel(v)));
      } else {
	midpoints[v] = new Vertex(depth() + 1);
	assert(midpoints[v]);
      }
    }

    m_children[T0] = new MultiresTriangle(this, T0,
					  midpoints[V1],
					  midpoints[V2],
					  midpoints[V3]);
    assert(m_children[T0]);
    m_children[T1] = new MultiresTriangle(this, T1,
					  vertex(V1),
					  midpoints[V2],
					  midpoints[V3]);
    assert(m_children[T1]);
    m_children[T2] = new MultiresTriangle(this, T2,
					  midpoints[V1],
					  vertex(V2),
					  midpoints[V3]);
    assert(m_children[T2]);
    m_children[T3] = new MultiresTriangle(this, T3,
					  midpoints[V1],
					  midpoints[V2],
					  vertex(V3));
    assert(m_children[T3]);
  }

  MultiresTriangle* child(TriangleLabel t) const {
    return (m_children[t]);
  }

  int depth(void) const {
    return (m_depth);
  }

  bool isLeaf(void) const {
    return (m_children[T0] == 0);
  }

  bool isRoot(void) const {
    return (m_parent == 0);
  }

  TriangleLabel label(void) const {
    return (m_label);
  }

  OneRingIt oneRingBegin(void) {
    return OneRingIt(this);
  }

  OneRingIt oneRingEnd(void) {
    return OneRingIt(this, true);
  }

  Orientation orientation(void) const {
    return (Orientation(m_orientation));
  }

  MultiresTriangle* parent(void) const {
    return (m_parent);
  }

  Triangle* triangle(void) const {
    return (m_triangle);
  }

  TriangleIt triangleBegin(VertexLabel v) {
    return TriangleIt(v, this);
  }

  TriangleIt triangleEnd(VertexLabel v) {
    return TriangleIt(v, this, true);
  }

  /**
   * Returns the valence of a vertex on the interior of the subdivision mesh.
   * @param v  interior vertex
   * @return valence of vertex
   */
  int valence(VertexLabel v) {
    if (vertex(v)->isControlVertex())
      return (triangle()->vertex(v)->valence());
    else
      return (6);
  }

  Vertex* vertex(VertexLabel v) const {
    return (m_vertices[v]);
  }

  VertexIt vertexBegin(VertexLabel v) {
    return VertexIt(v, this);
  }

  VertexIt vertexEnd(VertexLabel v) {
    return VertexIt(v, this, true);
  }
  
private:
  MultiresTriangle* m_children[4];
  
  int m_depth;
  
  TriangleLabel m_label;
  
  unsigned int m_orientation:1;

  MultiresTriangle* m_parent;
  
  Vertex* m_vertices[3];

  Triangle* m_triangle;


  MultiresTriangle(const MultiresTriangle& instance) {
    assert(false);
  }

  MultiresTriangle& operator=(const MultiresTriangle& instance) {
    assert(false);
    return (*this);
  }
};

#endif  /* __MULTIRESTRIANGLE_H__ */
