#ifndef __MULTIRESVERTEX_H__
#define __MULTIRESVERTEX_H__

#include <assert.h>
#include "Loop/FeatureTag.h"
#include "Loop/Topology.h"
#include "Util/cvec.h"
#include "Util/util.h"

/**
 * MultiresVertex Class
 */
class MultiresVertex {
public:
  /**
   * Hash function for multiresolution vertices.
   */
  struct hash {
    size_t operator()(MultiresVertex* point) const {
      return size_t(point);
    }
  };

protected:
  /**
   * Identifier for invalid multiresolution vertices.
   */
  static const int INVALID;

protected:
  /** The depth in the subdivision hierarchy at which this vertex exists. */
  int m_depth;

  /** The array of positions for this vertex. */
  CVec* m_point;

  /** The feature tag for this vertex. */
  FeatureTag* m_tag;

  /** The maximum depth at which the positions for this vertex are valid. */
  int m_validDepth;

public:
  /**
   * Constructs a multiresolution vertex at the specified depth in the
   * subdivision hierarchy.
   * @param depth  depth at which to construct this vertex
   */
  MultiresVertex(int depth) : m_depth(depth), m_tag(0), m_validDepth(INVALID) {
    m_point = new CVec[__MAX_DEPTH - depth];
    assert(m_point);
  }

  /**
   * Destructor.
   */
  ~MultiresVertex(void) {
    delete[] (m_point);
    if (m_tag)
      delete (m_tag);
  }

  /**
   * Returns the depth in the subdivision hierarchy at which this vertex
   * exists.
   * @return depth of this vertex in the subdivision hierarchy
   */
  int depth(void) {
    return (m_depth);
  }

  /**
   * Returns true if this vertex corresponds to a vertex in the control net
   * of the subdivision surface.
   * @return  true if this vertex corresponds to a control vertex
   */
  bool isControlVertex(void) {
    return (m_depth == 0);
  }

  /**
   * Returns true if the vertex position at the specified depth is valid.
   * @param depth  depth in subdivision hierarchy
   * @return  true if vertex position is valid at depth
   */
  bool isValid(int depth) {
    return (depth <= m_validDepth);
  }

  /**
   * Returns an immutable reference to the position of this vertex at the
   * specified depth.
   * @param depth  depth in subdivision hierarchy
   * @return immutable reference to vertex position at depth
   */
  const CVec& point(int depth) const {
    assert(depth >= m_depth && depth < __MAX_DEPTH);
    return (m_point[depth - m_depth]);
  }

  /**
   * Returns a mutable reference to the position of this vertex at the
   * specified depth.
   * @param depth  depth in subdivision hierarchy
   * @return mutable reference to vertex position at depth
   */
  CVec& point(int depth) {
    assert(depth >= m_depth && depth < __MAX_DEPTH);
    return (m_point[depth - m_depth]);
  }

  /**
   * Invalidates the position of this vertex is invalid for all levels of the
   * subdivision hierarchy.
   */
  void setInvalid(void) {
    m_validDepth = INVALID;
  }

  /**
   * Specifies that the position of this vertex at the given depth is valid.
   * @param depth  depth in subdivision hierarchy
   */
  void setValid(int depth) {
    if (depth == m_depth - 1) {
      setInvalid();
    } else {
      assert(depth == m_depth || (depth > m_depth && isValid(depth - 1)));
      m_validDepth = depth;
    }
  }

  /**
   * Returns the feature tag for this vertex.
   * @return feature tag for this vertex
   */
  FeatureTag* tag(void) const {
    return (m_tag);
  }

  /**
   * Returns a mutable reference to the feature tag for this vertex.
   * @return mutable reference to the feature tag for this vertex.
   */
  FeatureTag*& tag(void) {
    return (m_tag);
  }

protected:
  /**
   * Copy constructor: undefined operation.
   */
  MultiresVertex(const MultiresVertex& instance) {
    die();
  }

  /**
   * Assignment operator: undefined operation.
   */
  MultiresVertex& operator=(const MultiresVertex& instance) {
    die();
    return (*this);
  }
};


#endif  /* __MULTIRESVERTEX_H__ */
