#ifndef __TESSELATEDMESH_H__
#define __TESSELATEDMESH_H__

#include "Loop/Permutation.h"
#include "Mesh/MeshTp.h"
#include "Mesh/PointVertex.h"
#include "Util/cvec.h"

template <class TriangleDm>
class TesselatedVertexTag {
public:
  TesselatedVertexTag(void) { }

  virtual ~TesselatedVertexTag(void) { }

  virtual CVec coords(TriangleDm* domain) = 0;
};

template <class TriangleDm>
class TesselatedInteriorVertexTag : public TesselatedVertexTag<TriangleDm> {
private:
  const CVec m_coords;

  TriangleDm* m_domain;

public:
  TesselatedInteriorVertexTag(TriangleDm* domain, const CVec& coords)
   : TesselatedVertexTag<TriangleDm>(), m_domain(domain), m_coords(coords) { }

  virtual ~TesselatedInteriorVertexTag(void) { }

  virtual CVec coords(TriangleDm* domain) {
    assert(domain == m_domain);
    return (m_coords);
  }
};

template <class TriangleDm>
class TesselatedEdgeVertexTag : public TesselatedVertexTag<TriangleDm> {
private:
  const CVec m_coords;

  TriangleDm* m_domain;

  EdgeLabel m_edge;

public:
  TesselatedEdgeVertexTag(TriangleDm* domain,
			  EdgeLabel edge,
			  const CVec& coords)
   : TesselatedVertexTag<TriangleDm>(),
     m_domain(domain), m_edge(edge), m_coords(coords) { }

  virtual ~TesselatedEdgeVertexTag(void) { }

  virtual CVec coords(TriangleDm* domain) {
    assert(domain == m_domain || domain == m_domain->triangle(m_edge));

    if (domain == m_domain)
      return (m_coords);

    Permutation p = m_domain->permutation(m_edge);
    return CVec(m_coords(p.vertex(V1)),
		m_coords(p.vertex(V2)),
		m_coords(p.vertex(V3)));
  }
};

template <class TriangleDm>
class TesselatedControlVertexTag : public TesselatedVertexTag<TriangleDm> {
private:
  TriangleDm* m_domain;

  VertexLabel m_vertex;

public:
  TesselatedControlVertexTag(TriangleDm* domain,
			     VertexLabel vertex)
   : TesselatedVertexTag<TriangleDm>(),
     m_domain(domain), m_vertex(vertex) { }

  virtual ~TesselatedControlVertexTag(void) { }

  virtual CVec coords(TriangleDm* domain) {
    assert(domain->hasIncidentVertex(m_domain->vertex(m_vertex)));

    CVec coords(0.0f, 0.0f, 0.0f);
    coords(domain->indexOfVertex(m_domain->vertex(m_vertex))) = 1.0f;
    return (coords);
  }
};

template <class TriangleIf, class VertexIf, class TriangleDm>
class TesselatedVertex : public PointVertex<TriangleIf, VertexIf> {
private:
  TesselatedVertexTag<TriangleDm>* m_tag;

public:
  TesselatedVertex(TesselatedVertexTag<TriangleDm>* tag)
   : PointVertex<TriangleIf, VertexIf>(), m_tag(tag) { }

  ~TesselatedVertex(void) {
    delete (m_tag);
  }

  CVec coords(TriangleDm* domain) {
    return (m_tag->coords(domain));
  }
};

template <class TriangleIf, class VertexIf, class TriangleDm>
class TesselatedTriangle : public TriangleTp<TriangleIf, VertexIf> {
private:
  TriangleDm* m_domain;

public:
  TesselatedTriangle(TriangleDm* domain)
   : TriangleTp<TriangleIf, VertexIf>(), m_domain(domain) { }

  ~TesselatedTriangle(void) { }

  TriangleDm* domain(void) const {
    return (m_domain);
  }

  CVec coords(const CVec& coords) {
    assert(coords.x() >= 0.0f && coords.y() >= 0.0f && coords.z() >= 0.0f);
    assert(fabs(coords.x() + coords.y() + coords.z() - 1.0f) < 1.0e-7);

    return (coords.x() * vertex(V1)->coords(m_domain) +
	    coords.y() * vertex(V2)->coords(m_domain) +
	    coords.z() * vertex(V3)->coords(m_domain));
  }
};

/**
 * TesselatedMesh Class
 */
template <class TriangleIf, class VertexIf>
class TesselatedMesh : public MeshTp<TriangleIf, VertexIf> {
public:
  /**
   * Constructs an empty mesh.
   */
  TesselatedMesh(void) : MeshTp<TriangleIf, VertexIf>() { }

  /**
   * Destructor.
   */
  ~TesselatedMesh(void) { }

  TriangleIf* triangle(int index) {
    return (m_triangles[index]);
  }
};

#endif  /* __TESSELATEDMESH_H__ */
