// -*- Mode: c++ -*-
// $Id: EdgeTp.h,v 1.1.1.1 2000/01/06 20:46:10 eitan Exp $
// $Source: /ufs/multires/projects/Coke/Mesh/EdgeTp.h,v $

#ifndef __EDGETP_H__
#define __EDGETP_H__

#include <function.h>

template <class TriangleIf, class VertexIf>
class TriangleTp;

template <class TriangleIf, class Vertexif>
class VertexTp;

/**
 * EdgeTp Class
 */
template <class TriangleIf, class VertexIf>
class EdgeTp {
private:
  typedef TriangleTp<TriangleIf, VertexIf> TriangleTp;
  typedef VertexTp<TriangleIf, VertexIf> VertexTp;

  /** A triangle incident to this edge. */
  TriangleIf* m_triangle;

  /** The vertices incident to this edge. */
  VertexIf* m_vertices[2];

public:
  /**
   * Comparator in lexicographic order.
   */
  struct compare : public binary_function<EdgeTp*, EdgeTp*, bool> {
    bool operator()(const EdgeTp* ei, const EdgeTp* ej) const {
#ifdef _WIN32
      const VertexIf* mini = __min(ei->vertex(V1), ei->vertex(V2));
      const VertexIf* minj = __min(ej->vertex(V1), ej->vertex(V2));
#else
      const VertexIf* mini = min(ei->vertex(V1), ei->vertex(V2));
      const VertexIf* minj = min(ej->vertex(V1), ej->vertex(V2));
#endif
      if (mini < minj) {
	return (true);
      } else if (mini == minj) {
#ifdef _WIN32
	const VertexIf* maxi = __max(ei->vertex(V1), ei->vertex(V2));
	const VertexIf* maxj = __max(ej->vertex(V1), ej->vertex(V2));
#else
	const VertexIf* maxi = max(ei->vertex(V1), ei->vertex(V2));
	const VertexIf* maxj = max(ej->vertex(V1), ej->vertex(V2));
#endif
	return (maxi < maxj);
      }

      return (false);
    }
  };

public:
  /**
   * Constructs an uninitialized edge.
   */
  EdgeTp(void) : m_t(0) {
    m_vertices[V1] = 0;
    m_vertices[V2] = 0;
  }

  /**
   * Copy constructor.
   * @param instance  edge instance to copy
   */
  EdgeTp(const EdgeTp& instance) {
    operator=(instance);
  }

  /**
   * Constructs an edge incident to two specified vertices and a triangle.
   * @param vi  first vertex
   * @param vj  second vertex
   * @param triangle  incident triangle
   */
  EdgeTp(VertexIf* vi, VertexIf* vj, TriangleIf* triangle) {
    assert(triangle->hasIncidentVertex(vi) &&
	   triangle->hasIncidentVertex(vj));
    assert(vi != vj);

    m_triangle = triangle;
    m_vertices[V1] = vi;
    m_vertices[V2] = vj;
  }

  /**
   * Constructs an edge incident to two specified vertices.
   * @param vi  first vertex
   * @param vj  second vertex
   */
  EdgeTp(VertexIf* vi, VertexIf* vj) {
    m_vertices[V1] = vi;
    m_vertices[V2] = vj;

    VertexTp::TriangleIt ti = vi.triangleBegin();
    VertexTp::TriangleIt te = vi.triangleEnd();
    for (; ti != te; ++ti) {
      if ((*ti)->hasIncidentVertex(vj)) {
	m_triangle = *ti;
	return;
      }
    }
    die();
  }

  /**
   * Destructor.
   */
  ~EdgeTp(void) { }

  /**
   * Assignment operator.
   * @param instance  edge instance to copy
   */
  EdgeTp& operator=(const EdgeTp& instance) {
    m_triangle = instance.m_triangle;
    m_vertices[V1] = instance.m_vertices[V1];
    m_vertices[V2] = instance.m_vertices[V2];
  }

  /**
   * Returns the index of the specified vertex incident to this edge.
   * @param vertex  vertex incident to this edge
   * @return vertex index
   */
  int enumerateVertex(const VertexTp* vertex) const {
    if (vertex == (VertexTp*)m_vertices[V1]) {
      return (V1);
    } else {
      assert(vertex == (VertexTp*)m_vertices[V2]);
      return (V2);
    }
  }

  /**
   * Returns true if this edge lies on a boundary.
   * @return true if this edge lies on a boundary
   */
  bool isBoundaryEdge(void) const {
    return (oppositeTriangle() == 0);
  }


  /**
   * Returns the triangle across this edge opposite the reference triangle
   * returned by the triangle method.
   * @return triangle opposite the reference triangle
   */
  TriangleIf* oppositeTriangle(void) const {
    return (m_triangle->oppositeTriangle(m_vertices[V1], m_vertices[V2]));
  }

  /**
     * Returns a triangle incident to this edge.
     * @return triangle incident to this edge
     */
  TriangleIf* triangle(void) const {
    return (m_triangle);
  }

  /**
     * Returns the incident vertex associated with the specified index
     * relative to this edge.  Vertices are indexed in construction order.
     * @param index  vertex index
     * @return vertex associated with index
     */
  VertexIf* vertex(const int index) const {
    return (m_vertices[index]);
  }
};

#endif  /* __EDGETP_H__ */
