const float INF = 1e9;

class BoundingBox {
public:
  BoundingBox() : m_bmin(INF,INF,INF), m_bmax(-INF,-INF,-INF) {}

  void Clear() {
    m_bmin.x() = INF;
    m_bmin.y() = INF;
    m_bmin.z() = INF;
    m_bmax.x() = -INF;
    m_bmax.y() = -INF;
    m_bmax.z() = -INF;
  }

  void AddBox(BoundingBox &b) {
    assignMax(m_bmax, b.m_bmax);
    assignMin(m_bmin, b.m_bmin);
  }

  void AddPoint(CVec &p) {
    assignMax(m_bmax, p);
    assignMin(m_bmin, p);
  }

  bool Overlaps(BoundingBox &b) {
    if (m_bmax.x() < b.m_bmin.x()) return false;
    if (m_bmax.y() < b.m_bmin.y()) return false;
    if (m_bmax.z() < b.m_bmin.z()) return false;
    if (b.m_bmax.x() < m_bmin.x()) return false;
    if (b.m_bmax.y() < m_bmin.y()) return false;
    if (b.m_bmax.z() < m_bmin.z()) return false;
    return true;
  }

  bool Contains(BoundingBox &b) {
    if (m_bmax.x() < b.m_bmax.x()) return false;
    if (m_bmax.y() < b.m_bmax.y()) return false;
    if (m_bmax.z() < b.m_bmax.z()) return false;
    if (m_bmin.x() > b.m_bmin.x()) return false;
    if (m_bmin.y() > b.m_bmin.y()) return false;
    if (m_bmin.z() > b.m_bmin.z()) return false;
    return true;
  }

  CVec GetMaxExtent() {
    return m_bmax;
  }

  CVec GetMinExtent() {
    return m_bmin;
  }

private:
  CVec m_bmin, m_bmax;

  void assignMax(CVec &tgt, CVec &src) {
    tgt.x() = max(tgt.x(), src.x());
    tgt.y() = max(tgt.y(), src.y());
    tgt.z() = max(tgt.z(), src.z());
  }

  void assignMin(CVec &tgt, CVec &src) {
    tgt.x() = min(tgt.x(), src.x());
    tgt.y() = min(tgt.y(), src.y());
    tgt.z() = min(tgt.z(), src.z());
  }
};


static void WriteIvFile( SoSeparator* sep, const char* filename = NULL )
{
  SoWriteAction myAction;

  if( filename )
    myAction.getOutput()->openFile( filename );
  myAction.getOutput()->setBinary(FALSE);
  myAction.apply( sep );
  myAction.getOutput()->closeFile();
}

template <class T>
class hash<T *> {
public:
  size_t operator()(const T *v) const {
    return hash<char*>()((char*)v);
  }
};


void ProduceInventorFile(Mesh &mesh) {
  typedef Mesh::Vertex *VertexPtr_t;
  hash_map<Diamond<Mesh::Triangle,Mesh::Vertex> *, bool> diamondProcessed;
  Mesh::TriangleIt t, te;

  SoDB::init();

  SoSeparator *sep = new SoSeparator;
  sep->ref();

  SoComplexity *complexity = new SoComplexity;
  sep->addChild( complexity );
  complexity->type = SoComplexity::OBJECT_SPACE;
  complexity->value = 0;

  SoMaterial *surfaceMaterial = new SoMaterial;
  sep->addChild( surfaceMaterial );
  surfaceMaterial->ambientColor.setValue( 0.5, 0.5, 0.7 );
  surfaceMaterial->diffuseColor.setValue( 0.5, 0.5, 0.7 );
  surfaceMaterial->specularColor.setValue( 0.5, 0.5, 0.7 );

  SoCoordinate3* coord = new SoCoordinate3;
  coord->point.deleteValues( 0 );
  sep->addChild( coord );
  
  SoIndexedFaceSet* faceset = new SoIndexedFaceSet;
  faceset->materialIndex.deleteValues( 0 );
  faceset->coordIndex.deleteValues( 0 );
  sep->addChild( faceset );

  SoMaterial *sphereMaterial = new SoMaterial;
  sep->addChild( sphereMaterial );
  sphereMaterial->ambientColor.setValue( 1.0, 1.0, 0.1 );
  sphereMaterial->diffuseColor.setValue( 1.0, 1.0, 0.1 );
  sphereMaterial->specularColor.setValue( 1.0, 1.0, 0.1 );

  SoSeparator *sphereSep = new SoSeparator;
  sep->addChild( sphereSep );

  SoMaterial *vectorMaterial = new SoMaterial;
  sep->addChild( vectorMaterial );
  vectorMaterial->ambientColor.setValue( 0.8, 0.8, 1.0 );
  vectorMaterial->diffuseColor.setValue( 0.8, 0.8, 1.0 );
  vectorMaterial->specularColor.setValue( 0.8, 0.8, 1.0 );


  SoCoordinate3* linesetcoord = new SoCoordinate3;
  linesetcoord->point.deleteValues( 0 );
  sep->addChild( linesetcoord );

  SoLineSet* lineset = new SoLineSet;
  lineset->numVertices.deleteValues( 0 );
  lineset->vertexProperty = NULL;
  sep->addChild( lineset );

  BoundingBox box;

  for (t = mesh.triangleBegin(), te = mesh.triangleEnd();
       t != te; ++t) {
    for (int edgeNum=E1; edgeNum<=E3; ++edgeNum) {
      CVec pos = (*t)->getDiamond(EdgeLabel(edgeNum))->Evalp();
      box.AddPoint(pos);
    }
  }

  float ZExtent = box.GetMaxExtent().z() - box.GetMinExtent().z();
  float ZThreshold = box.GetMinExtent().z() + ZExtent/5.0;
  int idx = 0;

  for (t = mesh.triangleBegin(), te = mesh.triangleEnd();
       t != te; ++t) {
    for (int edgeNum=E1; edgeNum<=E3; ++edgeNum) {
      Diamond<Mesh::Triangle,Mesh::Vertex> *diamond = (*t)->getDiamond(EdgeLabel(edgeNum));

      if (diamondProcessed[diamond]) continue;
      diamondProcessed[diamond] = true;

      CVec pos = diamond->Evalp();

      //      if (pos.z() > ZThreshold) continue;

      SoTransformSeparator *sphereTSep = new SoTransformSeparator();
      SoTransform *sphereTrans = new SoTransform();
      SoSphere* sphere = new SoSphere();
      
      sphereTSep->addChild( sphereTrans );
      sphereTSep->addChild( sphere );
      sphereSep->addChild( sphereTSep );
      
      sphereTrans->translation.setValue( pos );
      sphere->radius = 0.05;

      CVec du = diamond->Evaldp1();
      CVec dv = diamond->Evaldp2();
      CVec n = du.cross(dv);
      n.normalize();

      lineset->numVertices.set1Value( idx/2, 2 );
      linesetcoord->point.set1Value( idx, pos - 0.1*n );
      ++idx;
      linesetcoord->point.set1Value( idx, pos + 0.1*n );
      ++idx;
    }
  }

  WriteIvFile( sep );
}


