#ifndef __SIMGEOMETRY_H
#define __SIMGEOMETRY_H

#include "Util/forall.h"
#include "Util/Real.h"
#include "SimTypes.h"

#include "fstream.h"

template <class T>
class SimGeometry {
public:
  DECLARE_TYPES(T)

  static void Initialize(Mesh &mesh) {
    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      SetxC (v, xR(v));
      Setx1C(v, Zero3Vec);
      Setx2C(v, Zero3Vec);

      SetxCprev (v, xR(v));
      Setx1Cprev(v, Zero3Vec);
      Setx2Cprev(v, Zero3Vec);
    endforall;
  }

  static void ScaleReference(Mesh &mesh, Real factor) {
    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      v->point() *= factor;
    endforall;
  }

  static void ScaleCurrent(Mesh &mesh, Real factor) {
    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      SetxCprev(v, xCprev(v)*factor);
      SetxC(    v, xCprev(v));
    endforall;
  }

  static void PullCurrent(Mesh &mesh, Real fraction) {
    Vertex *a, *b;
    Real dist = 0.0;
    a = b = NULL;
    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      if (a==NULL) a = v;
      Real d = (xR(v) - xR(a)).l2();
      if (d > dist) {
	dist = d;
	b = v;
      }
    endforall;
    assert(a && b && dist > 0.0);
    SetxCprev(a, xCprev(a) + fraction * (xR(a) - xR(b)));
    SetxCprev(b, xCprev(b) + fraction * (xR(b) - xR(a)));
    SetxC(a, xCprev(a));
    SetxC(b, xCprev(b));
  }

  static void Update(Mesh &mesh) {
  }


  // Current Coordinates

  static Real3Vec DispC(const SimVertex<T> *v) {
    return v->geometryData.m_dispC[0];
  }

  static void SetDispC(SimVertex<T> *v, Real3Vec dispC) {
    v->geometryData.m_dispC[0] = dispC;
  }


  static Real3Vec xC(const SimVertex<T> *v) {
    return xR(v) + v->geometryData.m_dispC[0];
  }

  static void SetxC(SimVertex<T> *v, Real3Vec xC) {
    v->geometryData.m_dispC[0] = xC - xR(v);
  }


  static Real3Vec x1C(const SimVertex<T> *v) {
    return v->geometryData.m_dispC[1];
  }

  static void Setx1C(SimVertex<T> *v, Real3Vec x1C) {
    v->geometryData.m_dispC[1] = x1C;
  }


  static Real3Vec x2C(const SimVertex<T> *v) {
    return v->geometryData.m_dispC[2];
  }

  static void Setx2C(SimVertex<T> *v, Real3Vec x2C) {
    v->geometryData.m_dispC[2] = x2C;
  }

  
  // Previous Coordinates (at last time step)
  
  static Real3Vec DispCprev(const SimVertex<T> *v) {
    return v->geometryData.m_dispCprev[0];
  }

  static void SetDispCprev(SimVertex<T> *v, Real3Vec dispCprev) {
    v->geometryData.m_dispCprev[0] = dispCprev;
  }


  static Real3Vec xCprev(const SimVertex<T> *v) {
    return xR(v) + v->geometryData.m_dispCprev[0];
  }

  static void SetxCprev(SimVertex<T> *v, Real3Vec xCprev) {
    v->geometryData.m_dispCprev[0] = xCprev - xR(v);
  }


  static Real3Vec x1Cprev(const SimVertex<T> *v) {
    return v->geometryData.m_dispCprev[1];
  }

  static void Setx1Cprev(SimVertex<T> *v, Real3Vec x1Cprev) {
    v->geometryData.m_dispCprev[1] = x1Cprev;
  }


  static Real3Vec x2Cprev(const SimVertex<T> *v) {
    return v->geometryData.m_dispCprev[2];
  }

  static void Setx2Cprev(SimVertex<T> *v, Real3Vec x2Cprev) {
    v->geometryData.m_dispCprev[2] = x2Cprev;
  }


  // Next Coordinates (proposed new coordinates)

  static Real3Vec DispCnext(const SimVertex<T> *v) {
    return v->geometryData.m_dispCnext;
  }

  static void SetDispCnext(SimVertex<T> *v, Real3Vec dispCnext) {
    v->geometryData.m_dispCnext = dispCnext;
  }


  static Real3Vec xCnext(const SimVertex<T> *v) {
    return xR(v) + v->geometryData.m_dispCnext;
  }

  static void SetxCnext(SimVertex<T> *v, Real3Vec xCnext) {
    v->geometryData.m_dispCnext = xCnext - xR(v);
  }



  // Temporary Coordinate Holders

  static Real3Vec DispCtemp(const SimVertex<T> *v, int tempIdx) {
    return v->geometryData.m_dispCtemp[tempIdx][0];
  }

  static void SetDispCtemp(SimVertex<T> *v, int tempIdx, Real3Vec dispCtemp) {
    v->geometryData.m_dispCtemp[tempIdx][0] = dispCtemp;
  }


  static Real3Vec xCtemp(const SimVertex<T> *v, int tempIdx) {
    return xR(v) + v->geometryData.m_dispCtemp[tempIdx][0];
  }

  static void SetxCtemp(SimVertex<T> *v, int tempIdx, Real3Vec xCtemp) {
    v->geometryData.m_dispCtemp[tempIdx][0] = xCtemp - xR(v);
  }


  static Real3Vec x1Ctemp(const SimVertex<T> *v, int tempIdx) {
    return v->geometryData.m_dispCtemp[tempIdx][1];
  }

  static void Setx1Ctemp(SimVertex<T> *v, int tempIdx, Real3Vec x1Ctemp) {
    v->geometryData.m_dispCtemp[tempIdx][1] = x1Ctemp;
  }


  static Real3Vec x2Ctemp(const SimVertex<T> *v, int tempIdx) {
    return v->geometryData.m_dispCtemp[tempIdx][2];
  }

  static void Setx2Ctemp(SimVertex<T> *v, int tempIdx, Real3Vec x2Ctemp) {
    v->geometryData.m_dispCtemp[tempIdx][2] = x2Ctemp;
  }



  // Reference Coordinates

  static Real3Vec xR(const SimVertex<T> *v) {
      return v->point();
  }

  static void SetxR(SimVertex<T> *v, Real3Vec xR) {
    xRref(v) = xR;
  }



  // Acceleration due to contact

  static Real3Vec x2Contact(const SimVertex<T> *v) {
    return v->geometryData.m_x2Contact;
  }

  static void Setx2Contact(SimVertex<T> *v, Real3Vec x2Contact) {
    v->geometryData.m_x2Contact = x2Contact;
  }


  // Load and save geometry state

  static void SaveState(SimVertex<T> *v, ofstream &fstate) {
    Real3Vec data;

    data = xR(v);
    fstate.write((char*)(&data), sizeof(data));

    data = DispC(v);
    fstate.write((char*)(&data), sizeof(data));
    
    data = x1C(v);
    fstate.write((char*)(&data), sizeof(data));
      
    data = x2C(v);
    fstate.write((char*)(&data), sizeof(data));

    data = DispCprev(v);
    fstate.write((char*)(&data), sizeof(data));
    
    data = x1Cprev(v);
    fstate.write((char*)(&data), sizeof(data));
      
    data = x2Cprev(v);
    fstate.write((char*)(&data), sizeof(data));
  }

  static void SaveStateASCII(SimVertex<T> *v, ostream &fstate) {
    fstate << xR(v) << endl;
    fstate << DispC(v) << endl;
    fstate << x1C(v) << endl;
    fstate << x2C(v) << endl;
    fstate << DispCprev(v) << endl;
    fstate << x1Cprev(v) << endl;
    fstate << x2Cprev(v) << endl;
    fstate << endl;
  }

  static void LoadState(SimVertex<T> *v, ifstream &fstate) {
    Real3Vec data;

    fstate.read((char*)(&data), sizeof(data));
    SetxR(v, data);

    fstate.read((char*)(&data), sizeof(data));
    SetDispC(v, data);

    fstate.read((char*)(&data), sizeof(data));
    Setx1C(v, data);

    fstate.read((char*)(&data), sizeof(data));
    Setx2C(v, data);

    fstate.read((char*)(&data), sizeof(data));
    SetDispCprev(v, data);

    fstate.read((char*)(&data), sizeof(data));
    Setx1Cprev(v, data);

    fstate.read((char*)(&data), sizeof(data));
    Setx2Cprev(v, data);
  }

private:

  Real3Vec m_dispC[3];

  Real3Vec m_dispCtemp[4][3];

  Real3Vec m_dispCprev[3];

  Real3Vec m_dispCnext;

  Real3Vec m_x2Contact;

  static Real3Vec &xRref(SimVertex<T> *v) {
      return v->point();
  }
};

#endif




