#ifndef __SIMTRIANGLE_H
#define __SIMTRIANGLE_H

#include "Util/Real.h"
#include "Quadrature/FEMTriangle.h"
#include "SimTypes.h"

#include "stdlib.h"

Real3Vec Intersect_NotFound = Real3Vec(RealLarge, RealLarge, RealLarge);

template <class T>
class SimTriangle : public FEMTriangle<typename T::Triangle, 
		                       typename T::Vertex, 
       		                       typename T::GaussPoint> {
public:
  DECLARE_TYPES(T)		 

  SimTriangle() : 
    FEMTriangle<Triangle, Vertex, GaussPoint>(),
    m_patch(NULL) {}

  LeafPatch *GetLeafPatch() {
    return m_patch;
  }

  void SetLeafPatch(LeafPatch &patch) {
    m_patch = &patch;
  }

  int Color() {
    return m_patch->Color();
  }

  Real3Vec NormalR() {
    Vertex *v1 = vertex(V1);
    Vertex *v2 = vertex(V2);
    Vertex *v3 = vertex(V3);

    Real3Vec e1 = v2->xR() - v1->xR();
    Real3Vec e2 = v3->xR() - v2->xR();
    Real3Vec e3 = v1->xR() - v3->xR();

    Real e1dot = e1.dot();
    Real e2dot = e2.dot();
    Real e3dot = e3.dot();

    if (e1dot < e2dot  &&  e1dot < e3dot)
      return e2.cross(e3);
    else if (e2dot < e3dot)
      return e3.cross(e1);

    return e1.cross(e2);
  }

  Real3Vec NormalC() {
    Vertex *v1 = vertex(V1);
    Vertex *v2 = vertex(V2);
    Vertex *v3 = vertex(V3);

    Real3Vec e1 = v2->xC() - v1->xC();
    Real3Vec e2 = v3->xC() - v2->xC();
    Real3Vec e3 = v1->xC() - v3->xC();

    Real e1dot = e1.dot();
    Real e2dot = e2.dot();
    Real e3dot =  e3.dot();

    if (e1dot < e2dot  &&  e1dot < e3dot)
      return e2.cross(e3);
    else if (e2dot < e3dot)
      return e3.cross(e1);

    return e1.cross(e2);
  }

  Real3Vec NormalCprev() {
    Vertex *v1 = vertex(V1);
    Vertex *v2 = vertex(V2);
    Vertex *v3 = vertex(V3);

    Real3Vec e1 = v2->xCprev() - v1->xCprev();
    Real3Vec e2 = v3->xCprev() - v2->xCprev();
    Real3Vec e3 = v1->xCprev() - v3->xCprev();

    Real e1dot = e1.dot();
    Real e2dot = e2.dot();
    Real e3dot = e3.dot();

    if (e1dot < e2dot  &&  e1dot < e3dot)
      return e2.cross(e3);
    else if (e2dot < e3dot)
      return e3.cross(e1);

    return e1.cross(e2);
  }

  Real3Vec IntersectEdgeC(Real3Vec p1, 
			  Real3Vec p2) {
    Real3Vec triNormal = NormalC();
    Real3Vec triPoint  = vertex(V1)->xC();
    Real3Vec rayDir    = p2 - p1;
    Real tNum = triNormal. dot (triPoint - p1);
    Real tDen = triNormal. dot (rayDir);
    
    // If edge is parallel to triangle, no intersection exists.
    // If there is an intersection along the line defined by the
    // edge, but NOT within the segment p1-p2, then no intersection
    // exists;
    if (tDen == 0.0 || abs(tNum) > abs(tDen) || sign(tNum) != sign(tDen))
      return Intersect_NotFound;
 
    return p1 + rayDir * tNum / tDen;
  }

  Real3Vec IntersectEdgeCprev(Real3Vec p1, 
			  Real3Vec p2) {
    Real3Vec triNormal = NormalCprev();
    Real3Vec triPoint  = vertex(V1)->xCprev();
    Real3Vec rayDir    = p2 - p1;
    Real tNum = triNormal. dot (triPoint - p1);
    Real tDen = triNormal. dot (rayDir);
    
    // If edge is parallel to triangle, no intersection exists.
    // If there is an intersection along the line defined by the
    // edge, but NOT within the segment p1-p2, then no intersection
    // exists;
    if (tDen == 0.0 || abs(tNum) > abs(tDen) || sign(tNum) != sign(tDen))
      return Intersect_NotFound;
 
    return p1 + rayDir * tNum / tDen;
  }

  friend ostream& operator<<( ostream&, const SimTriangle& );


  Real3Vec BarycentricC(Real3Vec ClosestPoint) {
    Real b1, b2, b3; 

    /*
      Basic idea is that bary c.'s are proportions of triangle hieghts. 
      Specifically, ||h1|| = ||hyp||sin(theta), where h1 = opposite
      and ||base cross hyp || = ||hyp||sin(theta)  */

    Real3Vec n1, n2, n3; 

    Real3Vec v1 = vertex(V1)->xC();
    Real3Vec v2 = vertex(V2)->xC();
    Real3Vec v3 = vertex(V3)->xC();

    n1 = v1 - v2; n1.normalize(); 
    n2 = v2 - v3; n2.normalize(); 
    n3 = v3 - v1; n3.normalize(); 

    b1 = (n2.cross(ClosestPoint - v2)).l2()/(n2.cross(v1 - v2)).l2(); 
    b2 = (n3.cross(ClosestPoint - v3)).l2()/(n3.cross(v2 - v3)).l2(); 
    b3 = (n1.cross(ClosestPoint - v1)).l2()/(n1.cross(v3 - v1)).l2(); 

    return Real3Vec(b1, b2, b3);
  }


  Real3Vec BarycentricCprev(Real3Vec ClosestPoint) {
    Real b1, b2, b3; 

    /*
      Basic idea is that bary c.'s are proportions of triangle hieghts. 
      Specifically, ||h1|| = ||hyp||sin(theta), where h1 = opposite
      and ||base cross hyp || = ||hyp||sin(theta)  */

    Real3Vec n1, n2, n3; 

    Real3Vec v1 = vertex(V1)->xCprev();
    Real3Vec v2 = vertex(V2)->xCprev();
    Real3Vec v3 = vertex(V3)->xCprev();

    n1 = v1 - v2; n1.normalize(); 
    n2 = v2 - v3; n2.normalize(); 
    n3 = v3 - v1; n3.normalize(); 

    b1 = (n2.cross(ClosestPoint - v2)).l2()/(n2.cross(v1 - v2)).l2(); 
    b2 = (n3.cross(ClosestPoint - v3)).l2()/(n3.cross(v2 - v3)).l2(); 
    b3 = (n1.cross(ClosestPoint - v1)).l2()/(n1.cross(v3 - v1)).l2(); 

    return Real3Vec(b1, b2, b3);
  }

protected:
  LeafPatch *m_patch;
};

template <class T>
ostream& operator<< (ostream &fout, const SimTriangle<T> &t) {
  fout << "\nTriangle {\n  " << t.v(V1)->xC()
       << "\n  " << t.v(V2)->xC()
       << "\n  " << t.v(V3)->xC()
       << ",\n}\n";
  return fout;
}

#endif
