#ifndef __SIMVERTEX_H
#define __SIMVERTEX_H

#include "Util/Real.h"
#include "Quadrature/ShellDOF.h"
#include "SimTypes.h"

template <class T>
class SimVertex : public ShellDOF<typename T::Triangle, typename T::Vertex> {
public:
  Real s, ds, sMax;

  DECLARE_TYPES(T)

  Geometry geometryData;
  Mass massData;
  Forces forcesData;
  Integrator integratorData;

  SimVertex( void )
    : ShellDOF<Triangle, Vertex>() {}
  SimVertex( const Real3Vec& p )
    : ShellDOF<Triangle, Vertex>(p) {}

  void SetColor(int col) {
    TriangleIt iter = triangleBegin(); 
    TriangleIt iterEnd = triangleEnd();
    for (; iter != iterEnd; ++iter) {
      Triangle *t = *iter;
      
      if (t->GetLeafPatch())
	t->GetLeafPatch()->SetColor( col );
      
    }
  }

  // Geometry
  ///////////////////////////
      
  // Current coordinates


  Real3Vec NormalC() const {
    Real3Vec normal = Zero3Vec;

    forall_const(TriangleIt, t, triangleBegin(), triangleEnd())
      normal += t->NormalC();
    endforall;

    normal.normalize();
    return normal;
  }


  Real3Vec DispC() const {
    return Geometry::DispC(this);
  }

  void SetDispC(Real3Vec dispC) {
    Geometry::SetDispC(this, dispC);
  }


  Real3Vec xC() const {
    return Geometry::xC(this);
  }

  void SetxC(Real3Vec xC) {
    Geometry::SetxC(this, xC);
  }


  Real3Vec x1C() const {
    return Geometry::x1C(this);
  }

  void Setx1C(Real3Vec x1C) {
    Geometry::Setx1C(this, x1C);
  }


  Real3Vec x2C() const {
    return Geometry::x2C(this);
  }

  void Setx2C(Real3Vec x2C) {
    Geometry::Setx2C(this, x2C);
  }


  // Previous coordinates (at previous time step)

  Real3Vec NormalCprev() const {
    Real3Vec normal = Zero3Vec;

    forall(TriangleIt, t, triangleBegin(), triangleEnd())
      normal += t->NormalCprev();
    endforall;

    normal.normalize();
    return normal;
  }


  Real3Vec DispCprev() const {
    return Geometry::DispCprev(this);
  }

  void SetDispCprev(Real3Vec dispCprev) {
    Geometry::SetDispCprev(this, dispCprev);
  }


  Real3Vec xCprev() const {
    return Geometry::xCprev(this);
  }

  void SetxCprev(Real3Vec xCprev) {
    Geometry::SetxCprev(this, xCprev);
  }


  Real3Vec x1Cprev() const {
    return Geometry::x1Cprev(this);
  }

  void Setx1Cprev(Real3Vec x1Cprev) {
    Geometry::Setx1Cprev(this, x1Cprev);
  }


  Real3Vec x2Cprev() const {
    return Geometry::x2Cprev(this);
  }

  void Setx2Cprev(Real3Vec x2Cprev) {
    Geometry::Setx2Cprev(this, x2Cprev);
  }


  // Next coordinates (proposed new coordinates)

  Real3Vec DispCnext() const {
    return Geometry::DispCnext(this);
  }

  void SetDispCnext(Real3Vec dispCnext) {
    Geometry::SetDispCnext(this, dispCnext);
  }


  Real3Vec xCnext() const {
    return Geometry::xCnext(this);
  }

  void SetxCnext(Real3Vec xCnext) {
    Geometry::SetxCnext(this, xCnext);
  }


  // Temporary coordinate placeholders

  Real3Vec DispCtemp(int tempIdx) const {
    return Geometry::DispCtemp(this, tempIdx);
  }

  void SetDispCtemp(Real3Vec dispCtemp, int tempIdx) {
    Geometry::SetDispCtemp(this, tempIdx, dispCtemp);
  }


  Real3Vec xCtemp(int tempIdx) const {
    return Geometry::xCtemp(this, tempIdx);
  }

  void SetxCtemp(Real3Vec xCtemp, int tempIdx) {
    Geometry::SetxCtemp(this, tempIdx, xCtemp);
  }


  Real3Vec x1Ctemp(int tempIdx) const {
    return Geometry::x1Ctemp(this, tempIdx);
  }

  void Setx1Ctemp(Real3Vec x1Ctemp, int tempIdx) {
    Geometry::Setx1Ctemp(this, tempIdx, x1Ctemp);
  }


  Real3Vec x2Ctemp(int tempIdx) const {
    return Geometry::x2Ctemp(this, tempIdx);
  }

  void Setx2Ctemp(Real3Vec x2Ctemp, int tempIdx) {
    Geometry::Setx2Ctemp(this, tempIdx, x2Ctemp);
  }



  // Reference coordinates

  Real3Vec NormalR() const {
    Real3Vec normal = Zero3Vec;

    forall(TriangleIt, t, triangleBegin(), triangleEnd())
      normal += t->NormalR();
    endforall;

    normal.normalize();
    return normal;
  }


  Real3Vec xR() const {
    return Geometry::xR(this);
  }

  void SetxR(Real3Vec xR) {
    Geometry::SetxR(this, xR);
  }


  // Acceleration due to contact

  Real3Vec x2Contact() const {
    return Geometry::x2Contact(this);
  }

  void Setx2Contact(Real3Vec x2Contact) {
    Geometry::Setx2Contact(this, x2Contact);
  }


  // Forces
  ////////////////

  Real3Vec Gstat() const {
    return Forces::Gstat(this);
  }

  Real3Vec Gext() const {
    return Forces::Gext(this);
  }

  // Mass
  //////////////////

  Real m() const {
    return Mass::m(this);
  }
};

#endif



#if 0
  Real3Vec &point() {
    return xR();
  }

  Real3Vec point() const {
    return xR();
  }

  Real3Vec x0C() const {
    return xC(0);
  }

  Real3Vec xC(int order) const {
    return Geometry::xC(this, order);
  }
#endif

