#ifndef __TENSOR3_H
#define __TENSOR3_H

#include "Vector3.h"

template <class Scalar>
class Tensor3 {
public:
  typedef Vector3<Scalar> Row;
  typedef Vector3<Scalar> Column;

  static const Tensor3 Zero;

  Tensor3(void) {
    m_r[0] = m_r[1] = m_r[2] = Row::Zero;
  }

  Tensor3(const Tensor3 &t) {
    m_r[0] = t.m_r[0];
    m_r[1] = t.m_r[1];
    m_r[2] = t.m_r[2];
  }
  
  Row &operator()(int i) {
    return row(i);
  }

  Row operator()(int i) const {
    return row(i);
  }

  Column operator *(const Column x) {
    return Column(m_r[0].dot(x),
		  m_r[1].dot(x),
		  m_r[2].dot(x));
  }

  Tensor3 operator *(const Tensor3 B) { 
    Tensor3 M;
    for (int i=0; i<3; ++i)
      for (int j=0; j<3; ++j)
	M(i)(j) = row(i). dot (B.col(j));
    return M;
  }

  Row &row(int idx) {
    assert(idx >= 0   &&  idx <= 2);
    return m_r[idx];
  }

  Row row(int idx) const {
    assert(idx >= 0   &&  idx <= 2);
    return m_r[idx];
  }

  Column col(int idx) const {
    assert(idx >= 0   &&  idx <= 2);
    return Column(row(0)(idx), row(1)(idx), row(2)(idx));
  }

  Tensor3 transpose() const {
    Tensor3 M;
    M.row(0) = Row(row(0).x(), row(1).x(), row(2).x());
    M.row(1) = Row(row(0).y(), row(1).y(), row(2).y());
    M.row(2) = Row(row(0).z(), row(1).z(), row(2).z());
    return M;
  }

  Scalar Trace() const {
    return row(0)(0) + row(1)(1) + row(2)(2);
  }

private:
  Row m_r[3];
};

template <class Scalar>
const Tensor3<Scalar> Tensor3<Scalar>::Zero = Tensor3<Scalar>();

template <class Scalar>
inline
Tensor3<Scalar>
TensorDiagonal(const Scalar d0,
	       const Scalar d1,
	       const Scalar d2) {
  Tensor3<Scalar> M;
  M(0)(0) = d0;
  M(1)(1) = d1;
  M(2)(2) = d2;
  return M;
}

template <class Scalar>
inline
Tensor3<Scalar>
TensorRows(const Vector3<Scalar> row0,
	   const Vector3<Scalar> row1,
	   const Vector3<Scalar> row2) {
  Tensor3<Scalar> M;
  M(0) = row0;
  M(1) = row1;
  M(2) = row2;
  return M;
}

template <class Scalar>
inline
Tensor3<Scalar>
TensorColumns(const Vector3<Scalar> col0,
	      const Vector3<Scalar> col1,
	      const Vector3<Scalar> col2) {
  Tensor3<Scalar> M;
  M(0) = col0;
  M(1) = col1;
  M(2) = col2;
  return M.transpose();
}

#endif
