// -*- Mode: c++ -*-
// $Id: LoopMesh.h,v 1.3 2000/09/29 21:00:16 eitan Exp $
// $Source: /cs/research/multires/cvsroot/Coke/Loop/LoopMesh.h,v $

#ifndef __LOOP_LOOPMESH_H__
#define __LOOP_LOOPMESH_H__

#include "Loop/LoopTesselator.h"
#include "Loop/Topology.h"
#include "Mesh/MeshTp.h"

/**
 * LoopMesh Class
 */
template <class MESHIF, class TRIANGLEIF, class VERTEXIF,
          class TRIANGLEMR, class VERTEXMR, class STENCIL>
class LoopMesh : public MeshTp<MESHIF, TRIANGLEIF, VERTEXIF> {
  template <class Mesh> friend class Accessor;

public:
  typedef TRIANGLEMR TriangleMr;
  typedef VERTEXMR VertexMr;
  typedef STENCIL Stencil;
  
protected:
  /** The subdivision stencil for this Loop subdivision mesh. */
  const Stencil& m_stencil;

protected:
  template <class Accessor>
  void even(TriangleMr* triangle, VertexLabel v, Accessor accessor) const {
    VertexMr* vertex = triangle->vertex(v);
    accessor.setValid(vertex, triangle->depth() + 1);
    accessor.clear(vertex, triangle->depth() + 1);
    m_stencil.even(triangle, v, accessor, accessor(vertex, triangle->depth() + 1));
  }

  template <class Accessor>
  void odd(TriangleMr* triangle, EdgeLabel e, Accessor accessor) const {
    VertexMr* vertex = triangle->child(T0)->vertex(OppositeVertex(e));
    accessor.setValid(vertex, triangle->depth() + 1);
    accessor.clear(vertex, triangle->depth() + 1);
    m_stencil.odd(triangle, e, accessor, accessor(vertex, triangle->depth() + 1));
  }

public:
  /**
   * Constructs an empty mesh with a specified subdivision stencil.
   * @param stencil  subdivision stencil
   */
  LoopMesh(const Stencil& stencil)
   : MeshTp<Mesh, Triangle, Vertex>(), m_stencil(stencil) { }
  
  /**
   * Copy constructor.
   */
  LoopMesh(const LoopMesh& instance)
   : MeshTp<Mesh, Triangle, Vertex>(instance),
     m_stencil(instance.m_stencil) { }
  
  /**
   * Destructor.
   */
  ~LoopMesh(void) { }

  /**
   * Evaluates the limit surface at a designated vertex in the mesh.
   * @param triangle  triangle containing vertex
   * @param v  label of vertex in triangle
   * @return limit position of vertex
   */
  CVec limit(TriangleMr* triangle, VertexLabel v) {
    VertexMr* vertex = triangle->vertex(v);
    VertexMr::Accessor accessor;

    if (vertex->isLimitValid() == false) {
      CVec limit = m_stencil.limit(triangle, v, accessor, m_stencil);
      accessor.setLimit(vertex, limit);
    }

    return (accessor.limit(vertex));
  }

  /**
   * Subdivides this Loop subdivision mesh uniformly to the specified depth.
   * @param depth  depth to which to uniformly subdivide this mesh
   */
  template <class Accessor>
  void subdivide(int depth, Accessor accessor) {
    TriangleIt ti = triangleBegin();
    TriangleIt te = triangleEnd();
    for (; ti != te; ++ti) {
      Triangle* triangle = *ti;
      ((Mesh*)this)->subdivide(triangle->root(), depth, accessor);
    }
  }

  class PointAccessor {
  public:
    CVec& operator()(MultiresVertex* vertex, int depth) const {
      assert(vertex->isValid(depth));
      return (vertex->point(depth));
    }

    void setValid(MultiresVertex* vertex, int depth) const {
      vertex->setValid(depth);
    }

    void clear(MultiresVertex* vertex, int depth) const {
      assert(vertex->isValid(depth));
      vertex->point(depth) = CVec(0,0,0);
    }
  };

  void subdivide(int depth) {
    subdivide(depth, PointAccessor());
  }

  /**
   * A convenience method to subdivide the specified triangle in the
   * subdivision hierarchy once.
   * @param triangle  triangle to subdivide
   */
  template <class Accessor>
  void subdivide(TriangleMr* triangle, Accessor accessor) {
    ((Mesh*)this)->subdivide(triangle, triangle->depth() + 1, accessor);
  }

  void subdivide(TriangleMr* triangle) {
    subdivide(triangle, PointAccessor());
  }

  /**
   * Subdivides the specified triangle in the subdivision hierarchy to the
   * specified depth.
   * @param triangle  triangle to subdivide
   * @param depth  depth to which to subdivide triangle
   */
  template <class Accessor>
  void subdivide(TriangleMr* triangle, int depth, Accessor accessor) {
    assert(depth <= __MAX_DEPTH);
    assert(triangle->depth() <= depth);

    if (triangle->depth() == depth)
      return;

    if (triangle->isLeaf()) {
      triangle->expand();
      validateOneRing(triangle);
    }

    int below = triangle->depth() + 1;
    for (int v = V1; v <= V3; v += 1) {
      VertexMr* vertex = triangle->vertex(VertexLabel(v));
      assert(vertex->isValid(triangle->depth()));
      if (!vertex->isValid(below)) {
	((Mesh*)this)->even(triangle, VertexLabel(v), accessor);
      }
    }

    TriangleMr* centre = triangle->child(T0);
    for (int e = E1; e <= E3; e += 1) {
      VertexMr* vertex = centre->vertex(OppositeVertex(EdgeLabel(e)));
      if (!vertex->isValid(below)) {
	((Mesh*)this)->odd(triangle, EdgeLabel(e), accessor);
      }
    }

    if (depth > below) {
      for (int t = T0; t <= T3; t += 1) {
	((Mesh*)this)->subdivide(triangle->child(TriangleLabel(t)), depth, accessor);
      }
    }
  }

  void subdivide(TriangleMr* triangle, int depth) {
    subdivide(triangle, depth, PointAccessor());
  }
  
  /**
   * Ensures that the one-ring neighbours of the specified subdivision triangle
   * exist and are valid.
   * @param triangle  triangle
   */
  void validateOneRing(TriangleMr* triangle) {
    if (triangle->isRoot())
      return;

    TriangleMr* parent = triangle->parent();
    int depth = triangle->depth();
    Permutation p;
    for (int v = V1; v <= V3; v += 1) {
      // In 3/4 of all cases, this point is incident to an edge of the
      // triangle's parent.
      if (VertexLabel(v) != ParentVertex(triangle->label())) {
	EdgeLabel e = OppositeEdge(VertexLabel(v));
	TriangleMr* adjacent = parent->adjacentTriangle(e, p);
	if (adjacent != 0)
	  ((Mesh*)this)->subdivide(adjacent, depth);
      } else {
	TriangleMr::TriangleIt ti = parent->triangleBegin(VertexLabel(v));
	TriangleMr::TriangleIt te = parent->triangleEnd(VertexLabel(v));
	while (ti != te) {
	  ((Mesh*)this)->subdivide(*ti, depth);
	  ++ti;
	}
      }
    }
  }
};

template <class MESHIF, class TRIANGLEIF, class VERTEXIF,
          class TRIANGLEMR, class VERTEXMR, class STENCIL>
ostream& operator<<(ostream& os, const LoopMesh<MESHIF, TRIANGLEIF, VERTEXIF,
		    TRIANGLEMR, VERTEXMR, STENCIL>& mesh)
{
  typedef MESHIF Mesh;
  typedef LoopTesselator<Mesh> Tesselator;

  Tesselator tesselator;
  Tesselator::Mesh* tesselation;

  tesselation = tesselator((Mesh&)mesh);
  os << *tesselation;
  delete (tesselation);
  return (os);
}

#define LOOPTYPE(LTYPE, LCLASS, TCLASS, VCLASS, MRTCLASS, MRVCLASS, SCLASS)   \
  class LTYPE##_MeshIf;                                                       \
  class LTYPE##_TriangleIf;                                                   \
  class LTYPE##_VertexIf;                                                     \
  class LTYPE##_TriangleMr;                                                   \
  typedef MRVCLASS LTYPE##_VertexMr;                                          \
  typedef SCLASS LTYPE##_Stencil;                                             \
                                                                              \
  class LTYPE##_MeshIf                                                        \
   : public LCLASS<LTYPE##_MeshIf, LTYPE##_TriangleIf, LTYPE##_VertexIf,      \
                   LTYPE##_TriangleMr, LTYPE##_VertexMr, LTYPE##_Stencil>     \
  {                                                                           \
   public:                                                                    \
    LTYPE##_MeshIf(const LTYPE##_Stencil& s)                                  \
     : LCLASS<LTYPE##_MeshIf, LTYPE##_TriangleIf, LTYPE##_VertexIf,           \
              LTYPE##_TriangleMr, LTYPE##_VertexMr, LTYPE##_Stencil>(s) { }   \
    LTYPE##_MeshIf(const LTYPE##_MeshIf& m)                                   \
     : LCLASS<LTYPE##_MeshIf, LTYPE##_TriangleIf, LTYPE##_VertexIf,           \
              LTYPE##_TriangleMr, LTYPE##_VertexMr, LTYPE##_Stencil>(m) { }   \
    ~LTYPE##_MeshIf(void) { }                                                 \
  };                                                                          \
                                                                              \
  class LTYPE##_TriangleIf                                                    \
   : public TCLASS<LTYPE##_TriangleIf, LTYPE##_VertexIf,                      \
                   LTYPE##_TriangleMr, LTYPE##_VertexMr>                      \
  {                                                                           \
   public:                                                                    \
    LTYPE##_TriangleIf(void)                                                  \
     : TCLASS<LTYPE##_TriangleIf, LTYPE##_VertexIf,                           \
              LTYPE##_TriangleMr, LTYPE##_VertexMr>() { }                     \
    LTYPE##_TriangleIf(const LTYPE##_TriangleIf& t)                           \
     : TCLASS<LTYPE##_TriangleIf, LTYPE##_VertexIf,                           \
              LTYPE##_TriangleMr, LTYPE##_VertexMr>(t) { }                    \
    ~LTYPE##_TriangleIf(void) { }                                             \
  };                                                                          \
                                                                              \
  class LTYPE##_VertexIf                                                      \
   : public VCLASS<LTYPE##_TriangleIf, LTYPE##_VertexIf, LTYPE##_VertexMr>    \
  {                                                                           \
   public:                                                                    \
    LTYPE##_VertexIf(void)                                                    \
     : VCLASS<LTYPE##_TriangleIf, LTYPE##_VertexIf, LTYPE##_VertexMr>() { }   \
    LTYPE##_VertexIf(const LTYPE##_VertexIf& v)                               \
     : VCLASS<LTYPE##_TriangleIf, LTYPE##_VertexIf, LTYPE##_VertexMr>(v) { }  \
    ~LTYPE##_VertexIf(void) { }                                               \
  };                                                                          \
                                                                              \
  class LTYPE##_TriangleMr                                                    \
   : public MRTCLASS<LTYPE##_TriangleIf, LTYPE##_VertexIf,                    \
                     LTYPE##_TriangleMr, LTYPE##_VertexMr>                    \
  {                                                                           \
   public:                                                                    \
    LTYPE##_TriangleMr(void)                                                  \
     : MRTCLASS<LTYPE##_TriangleIf, LTYPE##_VertexIf,                         \
                LTYPE##_TriangleMr, LTYPE##_VertexMr>() { }                   \
    ~LTYPE##_TriangleMr(void) { }                                             \
  };                                                                          \
                                                                              \
  typedef LCLASS<LTYPE##_MeshIf, LTYPE##_TriangleIf, LTYPE##_VertexIf,        \
                 LTYPE##_TriangleMr, LTYPE##_VertexMr, LTYPE##_Stencil> LTYPE;\

#endif  /* __LOOP_LOOPMESH_H__ */
