// -*- Mode: c++ -*-
// $Id: MeshTp.h,v 1.2 2000/09/16 04:35:29 eitan Exp $
// $Source: /cs/research/multires/cvsroot/Coke/Mesh/MeshTp.h,v $

#ifndef __MESH_MESHTP_H__
#define __MESH_MESHTP_H__

#include <algo.h>
#include <vector.h>
#include <iostream.h>
#include "Mesh/Accessor.h"
#include "Mesh/SMF.h"
#include "Util/util.h"

/**
 * MeshTp Class
 */
template <class MESHIF, class TRIANGLEIF, class VERTEXIF>
class MeshTp {
  friend class Accessor<MESHIF>;
  friend class Accessor<MeshTp>;

public:
  typedef MESHIF Mesh;
  typedef TRIANGLEIF Triangle;
  typedef VERTEXIF Vertex;

  typedef vector<Triangle*> TriangleCt;
  typedef TriangleCt::iterator TriangleIt;
  typedef TriangleCt::const_iterator const_TriangleIt;
  typedef vector<Vertex*> VertexCt;
  typedef VertexCt::iterator VertexIt;
  typedef VertexCt::const_iterator const_VertexIt;

protected:
  /** The set of triangles in this mesh. */
  TriangleCt m_triangles;
  
  /** The set of vertices in this mesh. */
  VertexCt m_vertices;

private:
  /**
   * Assignment operator: undefined operation.
   */
  MeshTp& operator=(const MeshTp&) {
    die();
    return (*this);
  }

public:
  /**
   * Constructs an empty mesh.
   */
  MeshTp(void) { }

  /**
   * Copy constructor.
   */
  MeshTp(const MeshTp&) { }

  /**
   * Destructor.
   */
  ~MeshTp(void) {
    for_each(triangleBegin(), triangleEnd(), Delete<Triangle>());
    for_each(vertexBegin(), vertexEnd(), Delete<Vertex>());
  }

  /**
   * Returns a constant iterator over the triangles in this mesh, positioned
   * at the start of the triangle list.
   * @return constant iterator over triangles in this mesh, positioned at start
   *	     of list
   */
  const_TriangleIt triangleBegin(void) const {
    return (m_triangles.begin());
  }

  /**
   * Returns an iterator over the triangles in this mesh, positioned at the
   * start of the triangle list.
   * @return iterator over triangles in this mesh, positioned at start of list
   */
  TriangleIt triangleBegin(void) {
    return (m_triangles.begin());
  }

  /**
   * Returns the number of triangles in this mesh.
   * @return number of triangles in this mesh
   */
  int triangleCount(void) const {
    return (m_triangles.size());
  }

  /**
   * Returns a constant iterator over the triangles in this mesh, positioned
   * at the end of the triangle list.
   * @return constant iterator over triangles in this mesh, positioned at end
   *	     of list
   */
  const_TriangleIt triangleEnd(void) const {
    return (m_triangles.end());
  }

  /**
   * Returns an iterator over the triangles in this mesh, positioned at the
   * end of the triangle list.
   * @return iterator over triangles in this mesh, positioned at end of list
   */
  TriangleIt triangleEnd(void) {
    return (m_triangles.end());
  }

  /**
   * Returns a constant iterator over the vertices in this mesh, positioned
   * at the start of the vertex list.
   * @return constant iterator over vertices in this mesh, positioned at start
   *	     of list
   */
  const_VertexIt vertexBegin(void) const {
    return (m_vertices.begin());
  }

  /**
   * Returns an iterator over the vertices in this mesh, positioned at the
   * start of the vertex list.
   * @return iterator over vertices in this mesh, positioned at start of list
   */
  VertexIt vertexBegin(void) {
    return (m_vertices.begin());
  }

  /**
   * Returns the number of vertices in this mesh.
   * @return number of vertices in this mesh
   */
  int vertexCount(void) const {
    return (m_vertices.size());
  }

  /**
   * Returns a constant iterator over the vertices in this mesh, positioned
   * at the end of the vertex list.
   * @return constant iterator over vertices in this mesh, positioned at end
   *         of list
   */
  const_VertexIt vertexEnd(void) const {
    return (m_vertices.end());
  }

  /**
   * Returns an iterator over the vertices in this mesh, positioned at the
   * end of the vertex list.
   * @return iterator over vertices in this mesh, positioned at end of list
   */
  VertexIt vertexEnd(void) {
    return (m_vertices.end());
  }

  /* Methods for backwards compatibility. */
  int triangleSize(void) const {
    return (m_triangles.size());
  }

  int vertexSize(void) const {
    return (m_vertices.size());
  }
};

template <class MESHIF, class TRIANGLEIF, class VERTEXIF>
istream& operator>>(istream& is, MeshTp<MESHIF, TRIANGLEIF, VERTEXIF>& mesh)
{
    SMF::readMesh(is, mesh);
    return (is);
}

template <class MESHIF, class TRIANGLEIF, class VERTEXIF>
ostream& operator<<(ostream& os,
		    const MeshTp<MESHIF, TRIANGLEIF, VERTEXIF>& mesh)
{
    SMF::writeMesh(os, mesh);
    return (os);
}

#define MESHTYPE(MTYPE, MCLASS, TCLASS, VCLASS)                               \
  class MTYPE##_MeshIf;                                                       \
  class MTYPE##_TriangleIf;                                                   \
  class MTYPE##_VertexIf;                                                     \
                                                                              \
  class MTYPE##_MeshIf                                                        \
   : public MCLASS<MTYPE##_MeshIf, MTYPE##_TriangleIf, MTYPE##_VertexIf>      \
  {                                                                           \
   public:                                                                    \
    MTYPE##_MeshIf(void)                                                      \
     : MCLASS<MTYPE##_MeshIf, MTYPE##_TriangleIf, MTYPE##_VertexIf>() { }     \
    MTYPE##_MeshIf(const MTYPE##_MeshIf& m)                                   \
     : MCLASS<MTYPE##_MeshIf, MTYPE##_TriangleIf, MTYPE##_VertexIf>(m) { }    \
    ~MTYPE##_MeshIf(void) { }                                                 \
  };                                                                          \
                                                                              \
  class MTYPE##_TriangleIf                                                    \
   : public TCLASS<MTYPE##_TriangleIf, MTYPE##_VertexIf>                      \
  {                                                                           \
   public:                                                                    \
    MTYPE##_TriangleIf(void)                                                  \
     : TCLASS<MTYPE##_TriangleIf, MTYPE##_VertexIf>() { }                     \
    MTYPE##_TriangleIf(const MTYPE##_TriangleIf& t)                           \
     : TCLASS<MTYPE##_TriangleIf, MTYPE##_VertexIf>(t) { }                    \
    ~MTYPE##_TriangleIf(void) { }                                             \
  };                                                                          \
                                                                              \
  class MTYPE##_VertexIf                                                      \
   : public VCLASS<MTYPE##_TriangleIf, MTYPE##_VertexIf>                      \
  {                                                                           \
   public:                                                                    \
    MTYPE##_VertexIf(void)                                                    \
     : VCLASS<MTYPE##_TriangleIf, MTYPE##_VertexIf>() { }                     \
    MTYPE##_VertexIf(const MTYPE##_VertexIf& v)                               \
     : VCLASS<MTYPE##_TriangleIf, MTYPE##_VertexIf>(v) { }                    \
    ~MTYPE##_VertexIf(void) { }                                               \
  };                                                                          \
                                                                              \
  typedef MCLASS<MTYPE##_MeshIf, MTYPE##_TriangleIf, MTYPE##_VertexIf> MTYPE; \

#endif  /* __MESH_MESHTP_H__ */
