#ifndef __SIMIMPLICITINTEGRATOR_H
#define __SIMIMPLICITINTEGRATOR_H

#undef PRODUCE_DEBUG_OUTPUT

#include "fstream.h"

#include "iomanip.h"

#include "SimTypes.h"

IOMANIPdeclare(Real3Vec)

#include "Util/forall.h"
#include "Util/random.h"

#include "nan.h"

template <class T>
class SimImplicitIntegrator {

public:
  DECLARE_TYPES(T)

  static void Initialize(Mesh &mesh) {
    Geometry       ::Initialize(mesh);
    Surface        ::Initialize(mesh);
    ShellIntegrator::Initialize(mesh);
    Material       ::Initialize(mesh);
    Mass           ::Initialize(mesh);
    Forces         ::Initialize(mesh);
  }

private:

  static Real m_time;

#ifdef PRODUCE_GNUPLOT_OUTPUT
  static ostream &cgnu;
#else
  static ofstream cgnu;
#endif

#ifdef PRODUCE_DEBUG_OUTPUT
  static ostream &cdebug;
#else
  static ofstream cdebug;
#endif

public:

  static Real Time() {
    return m_time;
  }

  static void InternalEnergy_AddIntegrationPoint(const GaussPoint *p, 
						 Surface::e2Iter e2Idx,
						 Real dV) {
  }


  static void TimeStep(Mesh &mesh) {

    // Copy current state over to prev. state. 

    cdebug.precision(5);

    SaveStateHistory(mesh); 

    Real dt = Parameters::dT;

    // Predict the new state. Assume constant acceleration.

    cdebug << "==================================" << endl << endl;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Mass::SetM(v, 1e-7);
  
      Real3Vec u0 = v->xCprev();
      Real3Vec v0 = v->x1Cprev();
      Real3Vec a0 = v->x2Cprev();
 
      Real3Vec u1 = u0 + dt * v0 + (dt*dt/2.0) * a0;
      
      Real3Vec v1 = v0 + dt * a0;

      v->SetxC(u1);
      v->Setx1C(v1);

    endforall;

    // Correct the prediction.

    MinimizeResidual(mesh, dt);

    // Compute velocity and acceleration for the corrected displacement.

    int idx = 0;
    Real Ek1 = 0.0;
    Real Ep1 = 0.0;
    Real Ek0 = 0.0;
    Real Ep0 = 0.0;

    cgnu << m_time << " ";

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Real gamma = Parameters::predcor_gamma;
      Real beta  = Parameters::predcor_beta;
  
      Real3Vec u0 = v->xCprev();
      Real3Vec v0 = v->x1Cprev();
      Real3Vec a0 = v->x2Cprev();

      Real3Vec u1 = v->xC();

      /* 
      Real3Vec v1 = (u1 - u0) * gamma/(beta*dt) 
                  + (1.0 - gamma/beta) * v0
                  + (1.0 - 1.0/(2*beta)) * dt * a0;
		  */

      Real3Vec a1 = (u1 - u0) / (beta*dt*dt)
                  - v0 / (beta*dt)
                  + a0 * (1.0 - 1.0/(2.0*beta));

      Real3Vec v1 = v0 + dt * ((1 - gamma) * a0  +  gamma * a1);

      v->Setx1C(v1);
      v->Setx2C(a1);

      if (!idx) {
	Real m = v->m();
	Ek0 += 0.5 * m * v0.x() * v0.x();
	Ek1 += 0.5 * m * v1.x() * v1.x();
	Ep0 += 0.5 * 1e-1 * u0.x() * u0.x();
	Ep1 += 0.5 * 1e-1 * u1.x() * u1.x();
	cdebug << "\tu0[" << idx << "] = " << u0;
	cdebug << "\tv0[" << idx << "] = " << v0;
	cdebug << "\ta0[" << idx << "] = " << a0 << endl;
	cdebug << "\tu1[" << idx << "] = " << u1;
	cdebug << "\tv1[" << idx << "] = " << v1;
	cdebug << "\ta1[" << idx << "] = " << a1 << endl;
	++idx;
	cgnu << " " << u1.x() << " " << v1.x() << " " << a1.x();
      }

    endforall;

    cgnu << " " << Ek1 << " " << Ep1 << " " << (Ek1 + Ep1) << endl;

    cdebug << "\tEk0 = " << Ek0 << "\t Ep0 = " << Ep0 << "\tE0 = " << (Ek0 + Ep0) << endl;
    cdebug << "\tEk1 = " << Ek1 << "\t Ep1 = " << Ep1 << "\tE1 = " << (Ek1 + Ep1) << endl;

    m_time += dt;

  }

private:

  Real3Vec P[2], Z[2], R[3], dRdP;

  static void MinimizeResidual(Mesh &mesh, Real dt) {

    // All computations in 
    // here assume that *current* is the guess state and *prev* is the 
    // state at the beginning of the timestep.

    int iter = 0;
    int k = 0;
    Real beta = 0.0;
    Real RR, RZ[2];

    cdebug << endl << "----------------------------" << endl << endl;

    cdebug << "t = " << m_time << endl;

    ComputeResidual(mesh, dt, k, RR);
    ConditionResidual(mesh, dt, k, RZ[k]);

    //cdebug << ClearLine() << m_time << "\t" << RR << "\t" << iter << endl;

    while (IsNANorINF(RR) || RR >= 1.0) {

      k = !k;
      ++iter;

      cdebug << "iter = " << iter << endl;

      ComputeDirectionP(mesh, k, beta);

      Real den = Compute_P_dot_dRdP(mesh, dt, k);

      Real alpha = - RZ[!k] / den;

      cdebug << "\talpha = " << alpha << "\tden = " << den << endl;

      if (den == 0.0)
	return;

      Real xC2 = 0.0;

      int idx = 0;
      forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

	Integrator &I = v->integratorData;

        Real3Vec xC =  v->xC() + alpha * I.P[k];
        v->SetxC( xC );

	xC2 += xC. dot (xC);

	//cdebug << "\tnew xC[" << idx << "] = " << xC << endl;
	++idx;
 	
      endforall;

      cdebug << "\tnew xC2 = " << xC2 << endl;

      ComputeResidual(mesh, dt, k, RR);
      ConditionResidual(mesh, dt, k, RZ[k]);

      beta = RZ[k] / RZ[!k];

      cdebug << ClearLine() << "\tt = " << m_time << "\tRR = " << RR 
	   << "\tRZ[" << iter << "] = " << RZ[k] 
	   << "\tRZ[" << iter-1 << "] = " << RZ[!k] 
	   << "\tbeta = " << beta
	   << endl;
    }

  }

  static char *ClearLine() {
    static char backspace[150];
    for (int i = 0; i < 150; ++i) 
      backspace[i] = '\b';
    return backspace;
  }

  static void SaveStateHistory(Mesh &mesh) {
    //Copy current state over previous state;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      
      v->SetDispCprev( v->DispC() );
      v->Setx1Cprev(   v->x1C()   );
      v->Setx2Cprev(   v->x2C()   );
      
    endforall;
  }

  static void RestoreStateHistory(Mesh &mesh) {
    //Copy history state over current state // CAREFULL WITH THIS!

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      
      v->SetDispC( v->DispCprev() );
      v->Setx1C(   v->x1Cprev()   );
      v->Setx2C(   v->x2Cprev()   );
      
    endforall;
  }

  static Real3Vec ComputeResidualComponent(Vertex *v, Real dt) {
    Real3Vec R = Zero3Vec;

    Real3Vec u0 = v->xCprev();
    Real3Vec v0 = v->x1Cprev();
    Real3Vec a0 = v->x2Cprev();
 
    Real3Vec u1 = v->xC();

    Real m = v->m();

    Real beta = Parameters::predcor_beta;
  
    Real3Vec Gdyn = m * ( (u1 - u0) / (beta * dt * dt)  
			  -  v0 / (beta * dt)  
			  +  a0 * (1.0 - 1.0/(2.0*beta)) );

    //cdebug << "\t\t\tGdyn = " << Gdyn << endl;
    //cdebug << "\t\t\tGext = " << v->Gext() << endl;
    //cdebug << "\t\t\tGstat = " << v->Gstat() << endl;

    R = Gdyn - v->Gext() + v->Gstat();

    return R;
  }

  static void ComputeResidual(Mesh &mesh, Real dt, int k, Real &RR) {

    int idx = 0;

    // Assume that *current* is the guess state (at time t + dt)
    // Assume that *prev* is the previous known state (at time t)

    Geometry       ::Update(mesh);
    Surface        ::Update(mesh);
    ShellIntegrator::Update(mesh);
    Material       ::Update(mesh);
    Mass           ::Update(mesh);
    Forces         ::Update(mesh);

    RR = 0.0;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Real3Vec R = ComputeResidualComponent(v, dt);

      Integrator &I = v->integratorData;
   
      I.R[k] = R;

      RR += R. dot (R);

      //cdebug << "\t\tR[" << idx << "] = " << I.R[k] << endl;
      ++idx;

    endforall;

    cdebug << "\t\t  RR = " << RR << endl;
  }

  static void ConditionResidual(Mesh &mesh, Real dt, int k, Real &RZ) {

    int idx = 0;

    RZ = 0.0;

    Real ZZ = 0.0;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Integrator &I = v->integratorData;
  
      Real3Vec R = I.R[k];

      Real m = v->m();

      Real beta = Parameters::predcor_beta;

      cdebug << (beta * dt * dt / m) << endl;

      Real3Vec Z = (beta * dt * dt / m) * R ; // Condition the system
      //Real3Vec Z = R;

      I.Z[k] = Z;

      RZ += R. dot (Z);

      ZZ += Z. dot (Z);

      //cdebug << "\t\tZ[" << idx << "] = " << I.Z[k] << "\tm = " << m << "\tbeta = " << beta << "\tdt = " << dt << endl;
      ++idx;

    endforall;

    cdebug << "\t\t  ZZ = " << ZZ << endl;
    cdebug << "\t\t  RZ = " << RZ << endl;
  }

  static void ComputeDirectionP(Mesh &mesh, int k, Real beta) {

    int idx = 0;

    Real PP = 0.0;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Integrator &I = v->integratorData;

      Real3Vec P = I.Z[!k] + beta * I.P[!k];

      I.P[k] = P;

      PP += P. dot (P);

      //cdebug << "\t\tP[" << idx << "] = " << P << endl;
      ++idx;

    endforall;

    cdebug << "\t\t  PP = " << PP << endl;
  }

  static void Compute_dRdP(Mesh &mesh, Real dt, int k) {

    Real eps = 1e-6;

    int idx = 0;

    Real maxP = 0;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Integrator &I = v->integratorData;

      Real3Vec P = I.P[k];

      maxP = max(maxP, abs(P.x()));
      maxP = max(maxP, abs(P.y()));
      maxP = max(maxP, abs(P.z()));

    endforall;

    eps /= maxP;

    // let u := u + eps * P
    cdebug << "\t\tComputing residual at x + " << (eps/maxP) << " * p: " << endl;
    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      Integrator &I = v->integratorData;

      Real3Vec xC = v->xC() + eps * I.P[k];
      v->SetxC( xC );

      //cdebug << "\t\t\txC[" << idx << "] = " << xC << endl;
      ++idx;
    endforall;

    Real RR;
    ComputeResidual(mesh, dt, 2, RR);

    Real delR2 = 0.0;
    Real dRdP2 = 0.0;

    // compute delta(R)/eps
    // and let u := u - eps * P (undo the previous modification)
    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Integrator &I = v->integratorData;

      Real3Vec delR = (I.R[2] - I.R[!k]);

      I.dRdP = delR / eps;

      //cdebug << "\t\t\tdRdP[" << idx << "] = " << I.dRdP << endl;

      delR2 = delR. dot (delR);
      dRdP2 = I.dRdP. dot (I.dRdP);

      v->SetxC( v->xC() - eps * I.P[k] );
 
    endforall;

    cdebug << "\t\t\tdelR2 = " << delR2 << endl;
    cdebug << "\t\t\tdRdP2 = " << dRdP2 << endl;
  }

  static Real Compute_P_dot_dRdP(Mesh &mesh, Real dt, int k) {

    Real result = 0.0;

    Compute_dRdP(mesh, dt, k);

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Integrator &I = v->integratorData;

      result += I.P[k]. dot (I.dRdP);
 
    endforall;

    return result;
  }

};

template <class T>
SimImplicitIntegrator<T>::Real SimImplicitIntegrator<T>::m_time;

#ifdef PRODUCE_GNUPLOT_OUTPUT
template <class T>
ostream &SimImplicitIntegrator<T>::cgnu = cout;
#else
template <class T>
ofstream SimImplicitIntegrator<T>::cgnu("/dev/null");
#endif

#ifdef PRODUCE_DEBUG_OUTPUT
template <class T>
ostream &SimImplicitIntegrator<T>::cdebug = cerr;
#else
template <class T>
ofstream SimImplicitIntegrator<T>::cdebug("/dev/null");
#endif

#endif



