#ifndef __SIMINTEGRATOR_H
#define __SIMINTEGRATOR_H

#include "fstream.h"

#include "iomanip.h"

#include "SimTypes.h"

IOMANIPdeclare(Real3Vec)

#include "Util/forall.h"
#include "Util/random.h"

bool collisionResponseFailure = false;

bool debugCollision = false;

int collisionCount = 0;

template <class T>
class SimIntegrator {
private:
  static ofstream fout;

public:
  DECLARE_TYPES(T)

  static void Initialize(Mesh &mesh) {
    Geometry       ::Initialize(mesh);
    Surface        ::Initialize(mesh);
    ShellIntegrator::Initialize(mesh);
    Material       ::Initialize(mesh);
    Mass           ::Initialize(mesh);
    Forces         ::Initialize(mesh);
    fout.open("isim.gnu");
    m_AdaptiveMode = true;
    m_useOlddT = 0;
  }

private:

  static Real m_time;

  static bool m_AdaptiveMode;

  static Real &time() {
    return m_time;
  }

  static int m_checkpointCount;

  static int m_useOlddT;
    
  static Real m_dT;

public:

  static Real Time() {
    return m_time;
  }

  static int CheckpointCount() {
    return m_checkpointCount;
  }

  static Real KineticEnergyDiscrete(Mesh &mesh) {

    Real Ekinetic = 0.0;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Ekinetic += 0.5 * v->m() * v->x1C().dot();

    endforall;

    return Ekinetic;
  }

  static Real m_Einternal;

  static void ClearInternalEnergy() {
    m_Einternal = 0.0;
  }

  static void InternalEnergy_AddIntegrationPoint(const GaussPoint *p, 
						 Surface::e2Iter e2Idx,
						 Real dV) {

//    Real Econtrib = 0.5 * (p->S(e2Idx) * p->Strain(e2Idx)).Trace();

#if 0
    CvBasis cvBasis = {p->gR(0,e2Idx), p->gR(1,e2Idx), p->gR(2,e2Idx)};

    CvCvTensor cvE = p->Strain(e2Idx);
    CnCnTensor cnE = RaiseIndices(cvE, cvBasis); 

    Real lambda, G;

    Real E  = Parameters::emod;
    Real nu = Parameters::nue;

    lambda = E * nu / (1 + nu) / (1 - 2 * nu);

    G      = E / (2 * (1 + nu));

    Real Econtrib = 0.5 * lambda * pow(Trace(cvE), 2) + G * ScalarProduct(cnE, cvE);
#endif

    Real Econtrib = 0.5 * ScalarProduct( p->S(e2Idx), p->Strain(e2Idx) );


    m_Einternal += Econtrib * dV;
  }

  static Real InternalEnergyContinuous(Mesh &mesh) {
    return m_Einternal;
  }

  static Real ComputeInternalEnergy(Mesh &mesh) {
    ClearInternalEnergy();

    Geometry       ::Update(mesh);
    Surface        ::Update(mesh);
    ShellIntegrator::Update(mesh);
    Material       ::Update(mesh);
    Mass           ::Update(mesh);
    Forces         ::Update(mesh);
    
    return InternalEnergyContinuous(mesh);
  }

  static Real TotalEnergy(Mesh &mesh) {
    return KineticEnergyDiscrete(mesh) + InternalEnergyContinuous(mesh);
  }
  
  static void TimeStep(Mesh &mesh) {

    if (m_useOlddT < 1)
      m_dT = Parameters::dT;
    else 
      --m_useOlddT;

    m_dT = min(m_dT, Parameters::dTmax);
    m_dT = max(m_dT, Parameters::dTmin);

    SaveStateHistory(mesh);

    TakeTimeStep(mesh, m_dT);

    int useOlddTmul = 1;

    while ( collisionResponseFailure ) {

      useOlddTmul *= 2;

      if (m_dT <= Parameters::dTmin) 
	break;
	
      RestoreStateHistory(mesh);

      m_dT = m_dT / 2.0;

      TakeTimeStep(mesh, m_dT);
	
      cerr.precision(30);
      cerr << ClearLine() 
	   << collisionCount << " t = " 
	   << time() << " dT = " << m_dT;
	   
    }
	
    if (useOlddTmul > 1)
      m_useOlddT = useOlddTmul;

    time() += m_dT;
 	
    cerr.precision(30);
    cerr << ClearLine() << collisionCount << " t = " << time() << " dT = " << m_dT;
    cerr << " * go gadget go! *       ";

    // Watchdog
    static Real prevInternalEnergy = 100;
    if (prevInternalEnergy > 100.0) {
      Real absChange = abs( InternalEnergyContinuous(mesh) - prevInternalEnergy );
      Real relChange = absChange / prevInternalEnergy;
      assert( relChange < 100 );
    }
    prevInternalEnergy = InternalEnergyContinuous(mesh);
	 

    // Output gnuplot data

    fout << setw(30) << setprecision(30) << time();
    int skip = 100;
    int plotCount = 1;
    fout << "\t"<< KineticEnergyDiscrete(mesh);
    fout << "\t"<< InternalEnergyContinuous(mesh);
    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      if (--skip > 0) continue;
      fout << "\t" << v->xC().l2();
      if (--plotCount < 1) break; 
    endforall;
    fout << endl;
  }

private:

  static char *ClearLine() {
    static char backspace[150];
    for (int i = 0; i < 150; ++i) 
      backspace[i] = '\b';
    return backspace;
  }

  static void SaveStateHistory(Mesh &mesh) {
    //Copy current state over previous state;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      
      v->SetDispCprev( v->DispC() );
      v->Setx1Cprev(   v->x1C()   );
      v->Setx2Cprev(   v->x2C()   );
      
    endforall;
  }

  static void RestoreStateHistory(Mesh &mesh) {
    //Copy history state over current state // CAREFULL WITH THIS!

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      
      v->SetDispC( v->DispCprev() );
      v->Setx1C(   v->x1Cprev()   );
      v->Setx2C(   v->x2Cprev()   );
      
    endforall;
  }

  static void TakeTimeStep(Mesh &mesh, Real dT) {

    Real gamma   = Parameters::predcor_gamma;
    Real damping = Parameters::damping;

    //Predictor

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      Real3Vec disp = v->DispC(),
	       x1   = v->x1C(),
	       x2   = v->x2C(),
	       xPre1;

      assert(disp == v->DispCprev());
      assert(x1 == v->x1Cprev());
      assert(x2 == v->x2Cprev());

      // eqn.34
      disp += x1 * dT  +  x2 * dT*dT / 2.0;

      // eqn.35
      xPre1 = x1  +  (1.0 - gamma) * dT * x2;

      v->SetxCtemp(xPre1, 0);
      v->SetDispC(disp);

      if ( (v->xC() - v->xCprev()).l2() >= 100.0 ) {
	cerr << "\nxCprev: " << v->xCprev() << endl;
	cerr << "xC: " << v->xC() << endl;
	cerr << "x1: " << x1 << endl;
	cerr << "x2: " << x2 << endl;
	cerr << "dT: " << dT << endl;
      }

      assert( (v->xC() - v->xCprev()).l2() < 100.0 );
    endforall;

    ClearInternalEnergy();

    Geometry       ::Update(mesh);
    Surface        ::Update(mesh);
    ShellIntegrator::Update(mesh);
    Material       ::Update(mesh);
    Mass           ::Update(mesh);
    Forces         ::Update(mesh);



    // Corrector

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Real3Vec xPre1 = v->xCtemp(0);
      Real3Vec disp = v->DispC(), 
	       x1   = v->x1C(),
	       x2   = v->x2C();

      // eqn.36
      x2  = (v->Gext() - v->Gstat()) / v->m();
      x2 -= damping * x1;
      Real3Vec x1delta = gamma * dT * x2;
      x1  = xPre1  +  x1delta;

      v->Setx1C(x1);
      v->Setx2C(x2);

      v->SetDispCnext(disp);
      v->Setx2Contact(Zero3Vec);

    endforall;

    Real olddT = Parameters::dT;
    Parameters::dT = dT;
    ResolveCollisions(mesh);
    Parameters::dT = olddT;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Real3Vec x1   = v->x1C(),
	       x2   = v->x2C(),
               x2Contact = v->x2Contact();

      if (v->DispC() != v->DispCnext()) {
	Real3Vec normal = v->NormalC();
	Real a = x1. dot (normal);

	assert(Parameters::elasticCollision <= 1.0);
	assert(Parameters::elasticCollision >= 0.0);
	Real b = 1.0 + Parameters::elasticCollision;

	x1 -= a * b * normal;
      }

      x1 += gamma * dT * x2Contact;
      x2 += x2Contact;

      v->Setx1C(x1);
      v->Setx2C(x2);

    endforall;

    // BUG BUG  BUG BUG  BUG BUG  BUG BUG  BUG BUG  BUG BUG  BUG BUG  BUG BUG 
    Geometry       ::Update(mesh); // WARNING: ERROR: BUG BUG BUG
  }

  static int sign(Real r) {
    if (r > 0) return 1;
    if (r < 0) return -1;
    return 0;
  }

  static void ResolveCollisions(Mesh &mesh) {
    if (mesh.rootPatch == NULL) return;

    int attempts = 8;

    while (1) {
      if (! --attempts)
	return;

      mesh.rootPatch->Update();
      collisionResponseFailure = false;
      pair<int,int> result;
      CollisionClear(mesh);
      result = mesh.rootPatch->FindSelfIntersections();
      CollisionUpdate(mesh);
      int count = result.first;
      int fixed = result.second;
      if (count) {
	cerr << ClearLine() << count << " collision(s) found (" << fixed << " fixed)";
      }
      collisionCount += count;
      if (!count || !fixed) 
	return;
    }
  }


  static void OldResolveCollisions(Mesh &mesh) {
    if (mesh.rootPatch == NULL) return;

    int repeatCollision = 4;
    while (--repeatCollision > 0) {
#if 0
      if (repeatCollision < 4) {
	cerr << "\n\nResolving collisions (" << repeatCollision << ")\n\n\n";
	debugCollision = true;
      }
#endif
      mesh.rootPatch->Update();
      collisionResponseFailure = false;
      pair<int,int> result;
      CollisionClear(mesh);
      result = mesh.rootPatch->FindSelfIntersections();
      int count = result.first;
      int fixed = result.second;
      if (collisionResponseFailure) {
	cerr << "\n\nCollision response failure! Try smaller timestep?\n\n" << endl;
	return;
      }
      if (count) {
	cerr << ClearLine() << count << " collision(s) found (" << fixed << " fixed) <"
	     << repeatCollision << ">         ...";
      }
      if (!count || !fixed) 
	return;
    }
    collisionResponseFailure = true;
    return;
#if 0
    cerr << "\n\nCould not resolve all collisions! Giving up!\n\n";
    assert(0);
#endif
  }

public:

#define CHECKPOINT_VERSION (4)

  static void SaveCheckpoint(Mesh &mesh, char *filename) {
    char str[100];
    ++m_checkpointCount;
    sprintf(str, "%s.state.%d", filename, m_checkpointCount);
    SaveState(mesh, str);
    sprintf(str, "%s.iv.%d", filename, m_checkpointCount);
    WriteIvFile(mesh, str);
  }

  static void SaveState(Mesh &mesh, char *filename) {
    ofstream fstate(filename);

    int version = CHECKPOINT_VERSION;
    fstate.write((char*)(&version), sizeof(version));

    fstate.write((char*)(&m_time), sizeof(m_time));    

    fstate.write((char*)(&m_checkpointCount), sizeof(m_checkpointCount));

    fstate.write((char*)(&m_dT), sizeof(m_dT));

    fstate.write((char*)(&m_useOlddT), sizeof(m_useOlddT));

    Random::SaveState(fstate);

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      
      Geometry::SaveState(v, fstate);

    endforall;
  }

  static void SaveStateASCII(Mesh &mesh, ostream &fstate) {

    int version = CHECKPOINT_VERSION;
    fstate << version << endl;
    fstate << m_time << endl;
    fstate << m_checkpointCount << endl;
    fstate << m_dT << endl;
    fstate << m_useOlddT << endl;

    Random::SaveStateASCII(fstate);

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      
      Geometry::SaveStateASCII(v, fstate);

    endforall;
  }

  static void LoadState(Mesh &mesh, char *filename) {
    ifstream fstate(filename);

    int version;
    fstate.read((char*)(&version), sizeof(version));
    if (version < 3 || version > 100) {
      cerr << "Checkpoint file: Incompatible version " 
	   << "(expecting v." << CHECKPOINT_VERSION
	   << ", read v." << version << ")\n";
      exit(1);
    }

    fstate.read((char*)(&m_time), sizeof(m_time));

    fstate.read((char*)(&m_checkpointCount), sizeof(m_checkpointCount));

    fstate.read((char*)(&m_dT), sizeof(m_dT));

    fstate.read((char*)(&m_useOlddT), sizeof(m_useOlddT));

    if (version == 4) {
      Random::LoadState(fstate);
    }
    else {
      cerr << "Warning! Checkpoint file does not contain Random Number Generator state\n";
    }

    cerr << "Loading checkpoint file, version " << version
	 << ", time = " << Time() << endl;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())

      Geometry::LoadState(v, fstate);

    endforall;
  }
};

template <class T>
ofstream SimIntegrator<T>::fout;

template <class T>
SimIntegrator<T>::Real SimIntegrator<T>::m_Einternal;


template <class T>
SimIntegrator<T>::Real SimIntegrator<T>::m_time;

template <class T>
bool SimIntegrator<T>::m_AdaptiveMode;

template <class T>
int SimIntegrator<T>::m_checkpointCount;

template <class T>
int SimIntegrator<T>::m_useOlddT;
    
template <class T>
SimIntegrator<T>::Real SimIntegrator<T>::m_dT;


#endif





#if PRINT
    
    cerr << "=================================================" << endl << endl;
    cerr << "=================================================" << endl << endl;
    cerr << "=================================================" << endl << endl;
    cerr << "Positions" << endl;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      cerr << v->xC() << endl;
    endforall;

    cerr << "Gauss points" << endl;

    forall(Mesh::GaussPointIt, p, mesh.gaussPointBegin(), mesh.gaussPointEnd())
      cerr << p->gC(0) << endl;
    endforall;

    cerr << "Velocities" << endl;
    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      cerr << v->x1C() << endl;
    endforall;

    cerr << "Accelerations" << endl;
    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      cerr << v->x2C() << endl;
    endforall;

#endif




#if PRINT

    cerr << "=================================================" << endl << endl;

    cerr << "Positions" << endl;

    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      cerr << v->xC() << endl;
    endforall;

    cerr << "Gauss points" << endl;

    forall(Mesh::GaussPointIt, p, mesh.gaussPointBegin(), mesh.gaussPointEnd())
      cerr << p->gC(0) << endl;
    endforall;


    cerr << "Velocities" << endl;
    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      cerr << v->x1C() << endl;
    endforall;

    cerr << "Accelerations" << endl;
    forall(Mesh::VertexIt, v, mesh.vertexBegin(), mesh.vertexEnd())
      cerr << v->x2C() << endl;
    endforall;

    cerr << "Forces and mass" << endl;
#endif
