#ifndef __SIMPARAMETERS_H
#define __SIMPARAMETERS_H

#include "SimTypes.h"

#include "Util/util.h"

template <class T>
class SimParameters {

public:

  DECLARE_TYPES(T)

  // Surface properties
  static Real rho; // Shell density
  static Real h;   // Uniform shell thickness
  static int thicknessSamples; // Number of points for thickness integral

  // Material properties
  static Real emod; 
  static Real nue;

  // Integration parameters
  static Real predcor_gamma; // predictor-corrector gamma
  static Real predcor_beta;  // predictor-corrector beta
  static Real damping;
  static Real dT;
  static Real dTmin;
  static Real dTmax;
  static Real relErrorBound;

  // Collision Response Properties
  static Real elasticCollision;

  static void Initialize() {
    rho              = 2710;
    h                = 0.00026;
    thicknessSamples = 3;

    emod             = 70e9;
    nue              = 0.33;

    predcor_gamma    = 0.5;
    predcor_beta     = 0.0;
    damping          = 0.0;
    dT               = 1e-6;
    dTmin            = 1e-7;
    dTmax            = 1e-1;

    elasticCollision = 0.8;
  }

  static bool ParseCommand(char *command, istream &fin) {
    if (!strcasecmp(command,"rho")) {
      fin >> rho;
    }
    else if (!strcasecmp(command,"h")) {
      fin >> h;
    }
    else if (!strcasecmp(command,"thickness-samples")) {
      fin >> thicknessSamples;
    }
    else if (!strcasecmp(command,"emod")) {
      fin >> emod;
    }
    else if (!strcasecmp(command,"poisson")) {
      fin >> nue;
    }
    else if (!strcasecmp(command,"gamma")) {
      fin >> predcor_gamma;
    }
    else if (!strcasecmp(command,"beta")) {
      fin >> predcor_beta;
    }
    else if (!strcasecmp(command,"damping")) {
      fin >> damping;
    }
    else if (!strcasecmp(command,"dt")) {
      fin >> dT;
    }
    else if (!strcasecmp(command,"dt-max")) {
      fin >> dTmax;
    }
    else if (!strcasecmp(command,"dt-min")) {
      fin >> dTmin;
    }
    else if (!strcasecmp(command,"relative-error-bound")) {
      fin >> relErrorBound;
    }
    else if (!strcasecmp(command,"elastic-collision")) {
      fin >> elasticCollision;
    }
    else return false;
    return true;
  }

  static void Load(char *fname) {
    ifstream fin(fname);
    Load(fin);
  }

  static void Load(istream &fin) {
    char command[100];
    while (!fin.eof()) {
      fin >> command;
      if (!Parameters::ParseCommand(command, fin)) {
	cerr << "Unknown parameter/command '" << command << "'\n";
	die();
      }
    }
  }

  static void Save(ostream &fout) {
    fout.precision(30);
    fout << "rho " << rho << endl;
    fout << "h " << h << endl;
    fout << "thickness-samples " << thicknessSamples << endl;
    fout << "emod " << emod << endl;
    fout << "poisson " << nue << endl;
    fout << "gamma " << predcor_gamma << endl;
    fout << "beta " << predcor_beta << endl;
    fout << "damping " << damping << endl;
    fout << "dt " << dT << endl;
    fout << "dt-max " << dTmax << endl;
    fout << "dt-min " << dTmin << endl;
    fout << "relative-error-bound " << relErrorBound << endl;
    fout << "elastic-collision " << elasticCollision << endl;
  }
};



template <class T>
SimParameters<T>::Real SimParameters<T>::rho;

template <class T>
SimParameters<T>::Real SimParameters<T>::h;

template <class T>
int  SimParameters<T>::thicknessSamples;

template <class T>
SimParameters<T>::Real SimParameters<T>::emod;

template <class T>
SimParameters<T>::Real SimParameters<T>::nue;

template <class T>
SimParameters<T>::Real SimParameters<T>::predcor_gamma;

template <class T>
SimParameters<T>::Real SimParameters<T>::predcor_beta;

template <class T>
SimParameters<T>::Real SimParameters<T>::damping;

template <class T>
SimParameters<T>::Real SimParameters<T>::dT;

template <class T>
SimParameters<T>::Real SimParameters<T>::dTmin;

template <class T>
SimParameters<T>::Real SimParameters<T>::dTmax;

template <class T>
SimParameters<T>::Real SimParameters<T>::relErrorBound;

template <class T>
SimParameters<T>::Real SimParameters<T>::elasticCollision;

#endif





