#ifndef __TENSOR3_H
#define __TENSOR3_H

#include <Util/Tensor.h>
#include <Util/Real.h>
#include <Util/Invert.h>

template <class VectorClass>
struct CovBasis {

public:

  CovBasis(const VectorClass &b0, const VectorClass &b1, const VectorClass &b2) {
    m_b[0] = b0;
    m_b[1] = b1;
    m_b[2] = b2;
  }

  const VectorClass &operator[](unsigned int i) const {
    assert(i < 3);
    return m_b[i];
  }

private:

  VectorClass m_b[3];
};

template <class VectorClass>
struct ConBasis {

public:

  ConBasis(const VectorClass &b0, const VectorClass &b1, const VectorClass &b2) {
    m_b[0] = b0;
    m_b[1] = b1;
    m_b[2] = b2;
  }

  const VectorClass &operator()(unsigned int i) const {
    assert(i < 3);
    return m_b[i];
  }

private:

  VectorClass m_b[3];
};

typedef CovBasis<Real3Vec> CvBasis;

template <class VectorClass, class Scalar>
class CovCovMetric3Tensor : public Tensor<Tensor<Scalar, VectorClass, Scalar, 3, COVARIANT>, 
			                 VectorClass, Scalar, 3, COVARIANT> {
public:
  CovCovMetric3Tensor(const CovBasis<VectorClass> &b) {
    CovCovMetric3Tensor &g = *this;
    int i, j;

    for (i = 0; i < 3; ++i)   // OPTIMIZE: use symmetry
       for (j = 0; j < 3; ++j)
          g[i][j] = b[i]. dot (b[j]);    
  }

  CovCovMetric3Tensor(Scalar gCoeffs[3][3]) {
    CovCovMetric3Tensor &g = *this;
    int i, j;

    for (i = 0; i <3; ++i)
      for (j = 0; j < 3; ++j)
	g(i)(j) = gCoeffs[i][j];
  }
};

template <class VectorClass, class Scalar>
class ConConMetric3Tensor : public Tensor<Tensor<Scalar, VectorClass, Scalar, 3, CONTRAVARIANT>, 
                                         VectorClass, Scalar, 3, CONTRAVARIANT> {
public:

  typedef Tensor<Tensor<Scalar, VectorClass, Scalar, 3, CONTRAVARIANT>, 
                 VectorClass, Scalar, 3, CONTRAVARIANT> SuperClass;

  ConConMetric3Tensor(const ConBasis<VectorClass> &b) {
    ConConMetric3Tensor &g = *this;
    int i, j;

    for (i = 0; i < 3; ++i)   // OPTIMIZE: use symmetry
       for (j = 0; j < 3; ++)
          g(i)(j) = b(i). dot (b(j));    
  }

  ConConMetric3Tensor(Scalar gCoeffs[3][3]) {
    ConConMetric3Tensor &g = *this;
    int i, j;

    for (i = 0; i < 3; ++i)
      for (j = 0; j < 3; ++j)
	g(i)(j) = gCoeffs[i][j];
  }

  VectorClass operator *(const VectorClass &v) const {
    const ConConMetric3Tensor &g = *this;
    return VectorClass( g(0)(0)*v(0) + g(0)(1)*v(1) + g(0)(2)*v(2),
			g(1)(0)*v(0) + g(1)(1)*v(1) + g(1)(2)*v(2),
			g(2)(0)*v(0) + g(2)(1)*v(1) + g(2)(2)*v(2) );
  }
};


template <class VectorClass, class Scalar>
class CovCon3Tensor : public Tensor<Tensor<Scalar, VectorClass, Scalar, 3, CONTRAVARIANT>, 
                                    VectorClass, Scalar, 3, COVARIANT> {
public:

  typedef Tensor<Tensor<Scalar, VectorClass, Scalar, 3, CONTRAVARIANT>, 
                 VectorClass, Scalar, 3, COVARIANT> SuperClass;

  CovCon3Tensor(const CovBasis<VectorClass> &newBasis,
		const ConBasis<VectorClass> &oldBasis) {

    CovCon3Tensor &T = *this;
    int i, j;

    for (i = 0; i < 3; ++i)   
       for (j = 0; j < 3; ++j)
          T[i](j) = newBasis[i]. dot (oldBasis(j));    
  }

  VectorClass operator *(const VectorClass &v) const {
    const CovCon3Tensor &T = *this;
    return VectorClass( T[0](0)*v(0) + T[0](1)*v(1) + T[0](2)*v(2),
			T[1](0)*v(0) + T[1](1)*v(1) + T[1](2)*v(2),
			T[2](0)*v(0) + T[2](1)*v(1) + T[2](2)*v(2) );
  }
};


template <class VectorClass, class Scalar>
class ConCov3Tensor : public Tensor<Tensor<Scalar, VectorClass, Scalar, 3, COVARIANT>, 
                                    VectorClass, Scalar, 3, CONTRAVARIANT> {
public:

  typedef Tensor<Tensor<Scalar, VectorClass, Scalar, 3, COVARIANT>, 
                 VectorClass, Scalar, 3, CONTRAVARIANT> SuperClass;

  ConCov3Tensor(const ConBasis<VectorClass> &newBasis,
		const CovBasis<VectorClass> &oldBasis) {
    ConCov3Tensor &T = *this;
    int i, j;

    for (i = 0; i < 3; ++i)   
       for (j = 0; j < 3; ++j)
          T(i)[j] = newBasis(i). dot (oldBasis[j]);    
  }

  VectorClass operator *(const VectorClass &v) const {
    const ConCov3Tensor &T = *this;
    return VectorClass( T(0)[0]*v(0) + T(0)[1]*v(1) + T(0)[2]*v(2),
			T(1)[0]*v(0) + T(1)[1]*v(1) + T(1)[2]*v(2),
			T(2)[0]*v(0) + T(2)[1]*v(1) + T(2)[2]*v(2) );
  }
};





typedef Tensor<Real, Real3Vec, Real, 3, COVARIANT    > Cv3Tensor;
typedef Tensor<Real, Real3Vec, Real, 3, CONTRAVARIANT> Cn3Tensor;

typedef Tensor<Cv3Tensor, Real3Vec, Real, 3, COVARIANT    > CvCv3Tensor;
typedef Tensor<Cn3Tensor, Real3Vec, Real, 3, CONTRAVARIANT> CnCn3Tensor;

typedef CovCovMetric3Tensor<Real3Vec, Real> CvCvMetric3Tensor;
typedef ConConMetric3Tensor<Real3Vec, Real> CnCnMetric3Tensor;

typedef ConCov3Tensor<Real3Vec, Real> CnCv3Tensor;
typedef CovCon3Tensor<Real3Vec, Real> CvCn3Tensor;

typedef CovBasis<Real3Vec> CvBasis;
typedef ConBasis<Real3Vec> CnBasis;

CnCnMetric3Tensor RaiseIndices(CvCvMetric3Tensor cov1) {
  int i, j;
  Real gm[3][3];

  for (i = 0; i < 3; ++i)
    for (j = 0; j < 3; ++j)
      gm[i][j] = cov1[i][j];

  MapleInvert3by3(gm);

  CnCnMetric3Tensor con1(gm);

  return con1;
}

CnCn3Tensor RaiseIndices(CvCv3Tensor covT, CvBasis b) {
  int i, j, k, l;
  CvCvMetric3Tensor cov1(b);
  CnCnMetric3Tensor con1 = RaiseIndices(cov1);

  CnCn3Tensor conT;

  for (i = 0; i < 3; ++i)
    for (j = 0; j < 3; ++j)
      for (k = 0; k < 3; ++k)
	for (l = 0; l < 3; ++l)
	  conT(i)(j) += con1(i)(k) * con1(j)(l) * covT[k][l];

  return conT;
}

CnBasis RaiseIndices(const CvBasis &covb) {
  int i, j;
  Real b[3][3];
  
  for (i = 0; i < 3; ++i)
    for (j = 0; j < 3; ++j)
      b[i][j] = covb[i][j];

  MapleInvert3by3(b);

  return CnBasis( Real3Vec(b[0][0], b[1][0], b[2][0]),
		  Real3Vec(b[0][1], b[1][1], b[2][1]),
		  Real3Vec(b[0][2], b[1][2], b[2][2]));
}

Real ScalarProduct(CnCn3Tensor conT, CvCv3Tensor covS) {
  Real sum = 0;
  int i, j;

  for (i = 0; i < 3; ++i)
    for (j = 0; j < 3; ++j)
      sum += conT(i)(j) * covS[i][j];

  return sum;
}

Real Trace(CvCv3Tensor covT) {
  Real sum = 0;
  int i;

  for (i = 0; i < 3; ++i)
    sum += covT[i][i];

  return sum;
}

#endif



