#include "Wconfig.h"

#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wpss.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wcontface.h"
#include "Wtmem3.h"
#include "Whistry.h"
#include "Wnewmarki.h"

#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)

typedef struct priv_elem_t {
  W_node_t         *node[3];
  W_float_t         thickness;
  W_index_t         material;
  W_matstate_t      matstate;
  double            mass;
}              priv_elem_t;

/*
  C
  C  PROCEDURE CALCULATES ELASTIC COEFFICIENTS OF STIFFNESS MATRIX
  C  OF FACET TRIANGULAR SHELL ELEMENT BASED ON
  C  KENNEDY,BELYTSCHKO,LIN (1986),BATHE,HO (1981)
  C
  C
  C     call FACTES (SM, E, POIS, SCF, THS, X2, X3, Y3)
  C  Input args:
  C
  C  E           Young's modulus
  C  POIS        Poisson coefficient
  C  SCF         shear correction factor, accounts for shear deformation due to
  C              Mindlin theory
  C  THS         thickness of the shell (uniform over the area)
  C  X2, X3, Y3  local coordinates of the node number 2 (X-coord) and node
  C              number 3 (X and Y-coord).
  C
  C                    Y ^
  C                      |   + 3
  C                      |
  C                      +-------+-> X
  C                     1        2
  C
  C  Output args:
  C
  C  SM          stiffness square matrix 18 x 18, double precision
  C
*/
static void
factes (double SM[9][9],        /* IN/OUT: stiffness matrix
                                 in the local coordinates
                                */
        double E,               /* Young's modulus */
        double POIS,            /* Poisson coefficient */
        double THS,             /* thickness */
        double X2,              /*
                                  X2, X3, Y3  local coordinates
                                  of the node number 2 (X-coord) and node
                                  number 3 (X and Y-coord).
                                    
                                                        Y ^
                                                          |   + 3
                                                          |
                                                          +-------+-> X
                                                          1        2
                                */
        double X3,
        double Y3
        )
{
  int I, J, M, N;
  double G, D, AGHPSI, AUX1, aux2;
  double x2inv, y3inv;

  x2inv = 1./X2;
  y3inv = 1./Y3;
  
  D      = E/(1.-POIS*POIS);
  G      = E/(1.0+POIS)*0.5;

  /*  MEMBRANE PART */
  
  SM[0][0] = SM[0][0] + (D*Y3*x2inv + G*(X3-X2)*(X3-X2)*x2inv*y3inv)*THS*0.50;
  SM[0][1] = SM[0][1] + (-D*POIS - G)*(X3-X2)*x2inv*THS*0.50;
  SM[1][0] = SM[0][1];
  SM[0][3] = SM[0][3] + (-D*Y3*x2inv - G*(X3-X2)*X3*x2inv*y3inv)*THS*0.50;
  SM[3][0] = SM[0][3];
  SM[0][4] = SM[0][4] + (D*POIS*X3*x2inv + G*(X3-X2)*x2inv)*THS*0.50;
  SM[4][0] = SM[0][4];
  SM[0][6] = SM[0][6] + G*(X3-X2)*y3inv*THS*0.50;
  SM[6][0] = SM[0][6];
  SM[0][7] = SM[0][7] + (-D*POIS)*THS*0.50;
  SM[7][0] = SM[0][7];
  SM[1][1] = SM[1][1] + (D*(X3-X2)*(X3-X2)*x2inv*y3inv + G*Y3*x2inv)*THS*0.50;
  SM[1][3] = SM[1][3] + (D*POIS*(X3-X2)*x2inv + G*X3*x2inv)*THS*0.50;
  SM[3][1] = SM[1][3];
  SM[1][4] = SM[1][4] + (-D*(X3-X2)*X3*x2inv*y3inv - G*Y3*x2inv)*THS*0.50;
  SM[4][1] = SM[1][4];
  SM[1][6] = SM[1][6] - G*THS*0.50;
  SM[6][1] = SM[1][6];
  SM[1][7] = SM[1][7] + D*(X3-X2)*y3inv*THS*0.50;
  SM[7][1] = SM[1][7];
  SM[3][3] = SM[3][3] + (D*Y3*x2inv + G*X3*X3*x2inv*y3inv)*THS*0.50;
  SM[3][4] = SM[3][4] + (-D*POIS - G)*X3*x2inv*THS*0.50;
  SM[4][3] = SM[3][4];
  SM[3][6] = SM[3][6] - G*X3*y3inv*THS*0.50;
  SM[6][3] = SM[3][6];
  SM[3][7] = SM[3][7] + D*POIS*THS*0.50;
  SM[7][3] = SM[3][7];
  SM[4][4] = SM[4][4] + (D*X3*X3*x2inv*y3inv + G*Y3*x2inv)*THS*0.50;
  SM[4][6] = SM[4][6] + G*THS*0.50;
  SM[6][4] = SM[4][6];
  SM[4][7] = SM[4][7] - D*X3*y3inv*THS*0.50;
  SM[7][4] = SM[4][7];
  SM[6][6] = SM[6][6] + G*X2*y3inv*THS*0.50;
  SM[7][7] = SM[7][7] + D*X2*y3inv*THS*0.50;
}

static int
elem_elcode (W_elem_t *el)
{
  return TMEM3_ELCODE;
}


static char *
type_name(W_elem_t *el)
{
  static char buf[] = "tmem3";
  return buf;
}

static int 
num_of_faces (W_elem_t *el)
{
  return 1;
}

static int 
get_face_vertex_nodes (W_elem_t *el, int nface, W_node_t *n[3])
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  int i;
  FOR (i, 3) {
    n[i] = h->node[i];
  }
  return 3;
}

static int
get_contact_faces (W_elem_t *el, int nface, WCF_contact_face_t *cf[], int maxcf)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  int i;
  
  if (maxcf < 1) {
    FOR (i, maxcf) { cf[i] = NULL; }
    return 1;
  }
  
  cf[0] = WCF_new_contact_face (h->node[0], h->node[1], h->node[2], NULL);
  
  return 1;
}

static void
compute_netdef (double x[3], double y[3], double xt[3], double yt[3],
                double netdef[3])
{
  netdef[0] = xt[1] - x[1];
  netdef[1] = xt[2] - x[2];
  netdef[2] = yt[2] - y[2];
}

/*
  COMPUTE INTERNAL FORCES OF FACET TRIANGLE FROM NET DEFORMATIONS
    NX,NY,NXY...MEMBRANE FORCES            ALL IN LOCAL ELEMENT COORDS
    F...VECTOR OF FORCES ACTING ON NODES IN ELEMENT COORDINATE SYSTEM
*/
static void
intfof (double X2, double X3, double Y3, double EY, double POIS, double THS,
        double NETDEF[3], double *NX, double *NY, double *NXY, double F[9])
{
  /*  ASSEMBLE INTERNAL FORCES */
  *NX =EY*THS/(1.0-POIS*POIS)* (+1.0/X2 * NETDEF[0] + POIS/Y3 * NETDEF[2]);
  *NY =EY*THS/(1.0-POIS*POIS)* (+POIS/X2 * NETDEF[0] + 1.0/Y3 * NETDEF[2]);
  *NXY=EY*THS/2.0/(1.0+POIS)* (-X3/X2/Y3 * NETDEF[0] +1.0/Y3 * NETDEF[1]);
  
  /*
    ASSEMBLE FORCE-VECTOR IN ELEMENT COORDINATE SYSTEM ACTING ON NODES
    (NOT ON ELEMENT VERTICES !)
    MEMBRANE EFFECT
  */
  F[0] = -(-Y3/2*(*NX) + (X3-X2)/2*(*NXY));
  F[1] = -((X3-X2)/2*(*NY) - Y3/2*(*NXY));
  F[3] = -(Y3/2*(*NX) - X3/2*(*NXY));
  F[4] = -(-X3/2*(*NY) + Y3/2*(*NXY));
  F[6] = -(X2/2*(*NXY));
  F[7] = -(X2/2*(*NY));
}

static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  W_coords_3d_t c[3], refc[3], e0, e1, e2;
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_material_t *m;
  double x[3], y[3], xt[3], yt[3];
  double netdef[3];
  W_vect_t f;
  double E, nu, nx, ny, nxy, fl[9], fg[9];

  m = WD_get_material_by_idx(el->domain, t->material);
  if (t->matstate == NULL) t->matstate = WM_get_mat_state(m);
  WM_get_uniax_lin_el_consts (m, &E, &nu);
  
  /* kinematic quantities */
#define GET_FROM_NODE_ALL(i)                                    \
  {                                                             \
    GET_FROM_NODE (t->node[i], WN_get_coords, c[i]);            \
    GET_FROM_NODE (t->node[i], WN_get_ref_coords, refc[i]);     \
  }
  GET_FROM_NODE_ALL(0);
  GET_FROM_NODE_ALL(1);
  GET_FROM_NODE_ALL(2);
#undef GET_FROM_NODE_ALL

  G3D_tri_to_tri2d (&refc[0], &refc[1], &refc[2], x, y);
  G3D_tri_to_tri2d (&c[0], &c[1], &c[2], xt, yt);
  compute_netdef (x, y, xt, yt, netdef);

  intfof (xt[1], xt[2], yt[2], E, nu, t->thickness, netdef, &nx, &ny, &nxy, fl);
  
  G3D_tri_to_tri2d_triad (&c[0], &c[1], &c[2], &e0, &e1, &e2);

  VECMAC_3_VEC (f,  =, fl[0] * e0, +, fl[1] * e1);
  WN_assemble_force (t->node[0], &f);
  VECMAC_3_VEC (f,  =, fl[3] * e0, +, fl[4] * e1);
  WN_assemble_force (t->node[1], &f);
  VECMAC_3_VEC (f,  =, fl[6] * e0, +, fl[7] * e1);
  WN_assemble_force (t->node[2], &f);
  
  /* deformation energy */
  el->deformation_energy = 0;   /* RAW */
}

static double 
compute_total_mass (priv_elem_t *t, W_domain_t *d)
{
  W_float_t a;
  W_float_t volume_mass;
  W_coords_3d_t c[3];
  W_material_t *m;
  int i;
  double x[3], y[3];

  FOR (i, 3) {
    GET_FROM_NODE (t->node[i], WN_get_coords, c[i]);
  }

  G3D_tri_to_tri2d (&c[0], &c[1], &c[2], x, y);
  a = G3D_tri2d_area (x[0], y[0], x[1], y[1], x[2], y[2]);
  
  m = WD_get_material_by_idx (d, t->material);
  if (!m) {
    sprintf (W_buf(), "Material %d not found", t->material);
    REPORT_FATAL_ERROR (d->errh, W_buf());
  }
  volume_mass = WM_get_volume_mass (m);

  if (a <= 0) {
    REPORT_FATAL_ERROR (d->errh, "Non-positive area");
  }

  return (a * volume_mass * t->thickness);
}

static W_float_t
total_mass (W_elem_t *el)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  return t->mass;
}

static void
assemble_mass (W_elem_t *el)
{
  W_float_t m;
  priv_elem_t *t = (priv_elem_t *)el->data;
  
  m = total_mass(el) / 3;

  WN_assemble_mass (t->node[0], m);
  WN_assemble_mass (t->node[1], m);
  WN_assemble_mass (t->node[2], m);
}


static W_float_t 
suggested_time_step (W_elem_t *el)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_coords_3d_t c[3];
  W_float_t E, nu, G, A;
  W_float_t lambda, rho;
  W_material_t *m;
  int i;
  double x[3], y[3];

  FOR (i, 3) {
    GET_FROM_NODE (t->node[i], WN_get_coords, c[i]);
  }
  G3D_tri_to_tri2d (&c[0], &c[1], &c[2], x, y);

  m = WD_get_material_by_idx(el->domain, t->material);
  if (!WM_get_uniax_lin_el_consts(m, &E, &nu))
    REPORT_FATAL_ERROR(el->domain->errh, "No uniax_lin_el_consts");
  rho      = WM_get_volume_mass(m);
  A        = G3D_tri2d_area (x[0], y[0], x[1], y[1], x[2], y[2]);
  lambda   = nu * E / ((1 + nu)*(1 - 2*nu));
  G        = E / (2 * (1 + nu));

  /* Formula identical to: Flanagan, Belytschko 1981 */
  return sqrt(A) * sqrt(rho / (lambda + 2*G));
}

static void *
new_history_client_data (W_elem_t *e,
                         W_index_t idx, /* unique integer  */
                         W_history_writer_t *r, /* history writer */
                         char *input_line /* input line (to allow
                                             for additional parsing of
                                             the line contents by this
                                             function) */
                         )
{
 /* RAW */
  return NULL;
}

static void
update_history (W_elem_t *e, /* generic element */
                W_solution_history_t *h, /* solution history */
                W_float_t time
                )
{
 /* RAW */
}

static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB,
                     int type, char *mode, void *optdata)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_vect_t refc[3], u[3];
  W_material_t *m;
  int i, ival, pt, reqlen;
  G3D_loc_t *pts[3];
  static int buflen        = 0;
  static double *values[3] = {NULL,NULL,NULL};
  int output_spec
    = WOPT_OPT (el->domain->options, postfile_output_spec);
  
  if (type == HEXVUE_POSTPROC_FILE) {
    double nx = 0, ny = 0, nxy = 0;
    double equiv_pl_def = 0, von_mises = 0;

    reqlen = 0;
    if (output_spec & W_POSTFILE_DISPL_MASK) 
      reqlen += 3;
    if (output_spec & W_POSTFILE_STRESS_MASK)
      reqlen += 6;
    if (output_spec & W_POSTFILE_EQPLS_MASK)
      reqlen++;
    if (output_spec & W_POSTFILE_VON_MISES_MASK)
      reqlen++;
    if (output_spec & W_POSTFILE_MATERIAL_MASK)
      reqlen++;
    
    if (buflen < reqlen) {
      FOR (i, 3) {
        values[i] = NEWP_REALLOC_OBJ_ARRAY (values[i], reqlen, double);
      }
      buflen = reqlen;
    }
  
    FOR (i, 3) {
      GET_FROM_NODE (h->node[i], WN_get_ref_coords, refc[i]);
      GET_FROM_NODE (h->node[i], WN_get_displ, u[i]);
    }

    m = WD_get_material_by_idx (el->domain, h->material);
    
    pt = 0;
    FOR(i, 3) {
      pts[pt]   = &refc[pt];
      ival = 0;
      if (output_spec & W_POSTFILE_DISPL_MASK) {
        values[pt][ival] = u[pt].x; ival++;
        values[pt][ival] = u[pt].y; ival++;
        values[pt][ival] = u[pt].z; ival++;
      }
      
      /* RAW evaluate stress etc */
      
      if (output_spec & W_POSTFILE_STRESS_MASK) {
        values[pt][ival] = nx; ival++;
        values[pt][ival] = ny; ival++;
        values[pt][ival] = nxy; ival++;
        values[pt][ival] = 0; ival++;
        values[pt][ival] = 0; ival++;
        values[pt][ival] = 0; ival++;
      }
      if (output_spec & W_POSTFILE_EQPLS_MASK) {
        values[pt][ival] = equiv_pl_def; ival++;
      }
      if (output_spec & W_POSTFILE_VON_MISES_MASK) {
        values[pt][ival] = von_mises; ival++;
      }
      if (output_spec & W_POSTFILE_MATERIAL_MASK) {
        values[pt][ival] = WM_get_id (m); ival++;
      }
      if (ival > reqlen) REPORT_FATAL_ERROR (el->domain->errh, "wrong");
      pt++;
    }
    W_write_hexvue_tri3 (IOB, pts, reqlen, values, mode);
  } /* type */
}

#ifdef X_XFACE
static void
modify_elem_grep(W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  int i;
  W_vect_t refc, u;
  WCRec pnts[3];
  W_material_t *m;
 
  for (i = 0; i < 3; i++) {
    GET_FROM_NODE (h->node[i], WN_get_ref_coords, refc);
    GET_FROM_NODE (h->node[i], WN_get_displ, u);
    pnts[i].x = refc.x + displ_scale * u.x;
    pnts[i].y = refc.y + displ_scale * u.y;
    pnts[i].z = refc.z + displ_scale * u.z;
  }

  if (el->grep == NULL) {
    el->grep = CreateTriangle3D (pnts);
    m = WD_get_material_by_idx(el->domain, h->material);
    EASValsSetColor (WM_get_color(m));
    EASValsSetLayer (WX_ELEM_LAYER);
    EASValsSetFillStyle (FILL_SOLID);
    EGWithMaskChangeAttributes(COLOR_MASK | FILL_MASK | LAYER_MASK, el->grep);
    EMNoDrawAddGraphicsToModel(ESIModel (), el->grep);
  } else {
    EGModifyGraphicsGeometry(el->grep, pnts);
  }
}
#endif

static void
assemble_inertial_forces_diag (W_elem_t *el)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_vect_t f, a;
  int i;
  double m = h->mass / 3;

  FOR (i, 3) {
    GET_FROM_NODE (h->node[i], WN_get_acceleration, a);
    VECMAC_2_VEC (f, =, -m * a);
    WN_assemble_force (h->node[i], &f);
  }
}

static void
assemble_residual_forces (W_elem_t *el,
                          W_mat_state_update_form_enum update_form,
                          W_mass_type_enum mass_type)
{
  assemble_restoring_forces (el, update_form);
  assemble_inertial_forces_diag (el); /* supports only diagonal mass */
}


static void
sub_T (W_coords_3d_t *e0, W_coords_3d_t *e1, W_coords_3d_t *e2,
        double T[3][3])
{
  T[0][0] = e0->x;
  T[1][0] = e0->y;
  T[2][0] = e0->z;

  T[0][1] = e1->x;
  T[1][1] = e1->y;
  T[2][1] = e1->z;

  T[0][2] = e2->x;
  T[1][2] = e2->y;
  T[2][2] = e2->z;
}

static void
full_T (W_coords_3d_t *e0, W_coords_3d_t *e1, W_coords_3d_t *e2,
        double T[9][9])
{
  int i, j;
  FOR (i, 9) FOR (j, 9) T[i][j] = 0;
#define SI 0
  T[SI+0][SI+0] = e0->x;
  T[SI+1][SI+0] = e0->y;
  T[SI+2][SI+0] = e0->z;

  T[SI+0][SI+1] = e1->x;
  T[SI+1][SI+1] = e1->y;
  T[SI+2][SI+1] = e1->z;

  T[SI+0][SI+2] = e2->x;
  T[SI+1][SI+2] = e2->y;
  T[SI+2][SI+2] = e2->z;
#undef SI 
#define SI 3
  T[SI+0][SI+0] = e0->x;
  T[SI+1][SI+0] = e0->y;
  T[SI+2][SI+0] = e0->z;

  T[SI+0][SI+1] = e1->x;
  T[SI+1][SI+1] = e1->y;
  T[SI+2][SI+1] = e1->z;

  T[SI+0][SI+2] = e2->x;
  T[SI+1][SI+2] = e2->y;
  T[SI+2][SI+2] = e2->z;
#undef SI 
#define SI 6
  T[SI+0][SI+0] = e0->x;
  T[SI+1][SI+0] = e0->y;
  T[SI+2][SI+0] = e0->z;

  T[SI+0][SI+1] = e1->x;
  T[SI+1][SI+1] = e1->y;
  T[SI+2][SI+1] = e1->z;

  T[SI+0][SI+2] = e2->x;
  T[SI+1][SI+2] = e2->y;
  T[SI+2][SI+2] = e2->z;
#undef SI 
}

static void
globalize_full (double K[9][9],
                double stiff_mult, double locK[9][9],
                W_coords_3d_t *e0, W_coords_3d_t *e1, W_coords_3d_t *e2)
{
  int i, j, k, m;
  double r;
  double T[9][9];
  
  full_T (e0, e1, e2, T);

  FOR (i, 9) {
    FOR (j, 9) {
      r = 0;
      FOR (k, 9) {
        FOR (m, 9) {
          r += locK[k][m] * T[i][k] * T[j][m];
        }
      }
      K[i][j] = stiff_mult * r;
    }
  }
}

static void
scale_locK (double locK[9][9], double stiff_mult)
{
  int i, j;
  FOR (i, 9) {
    FOR (j, 9) {
      if (locK[i][j] != 0) locK[i][j] *= stiff_mult;
    }
  }
}

#define QDIM  9
#define NTSUB 3
#include "transf.c"

static void
globalize_blocked (double K[9][9],
                   double stiff_mult, double locK[9][9],
                   W_coords_3d_t *e0, W_coords_3d_t *e1, W_coords_3d_t *e2)
{
  double T[3][3];
  double tmp[9][9];
  
  sub_T (e0, e1, e2, T);

  if (stiff_mult != 1) {
    scale_locK (locK, stiff_mult);
  }
  
  R_is_block_diag_T_times_Q        (tmp, T, locK);
  R_is_Q_times_transp_block_diag_T (K, tmp, T);
}

static void
global_stiff (W_elem_t *el,
              W_float_t stiff_mult,
              double K[9][9]
              )
{
  double locK[9][9]={{0,0,0,0,0,0,0,0,0},
                     {0,0,0,0,0,0,0,0,0},
                     {0,0,0,0,0,0,0,0,0},
                     {0,0,0,0,0,0,0,0,0},
                     {0,0,0,0,0,0,0,0,0},
                     {0,0,0,0,0,0,0,0,0},
                     {0,0,0,0,0,0,0,0,0},
                     {0,0,0,0,0,0,0,0,0},
                     {0,0,0,0,0,0,0,0,0}};
  double E, nu;
  W_coords_3d_t c[3], e0, e1, e2;
  double x[3], y[3], xt[3], yt[3];
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_material_t *m;

  m = WD_get_material_by_idx(el->domain, t->material);
  if (t->matstate == NULL) t->matstate = WM_get_mat_state(m);
  WM_get_uniax_lin_el_consts (m, &E, &nu);

#define GET_FROM_NODE_ALL(i)                                    \
  {                                                             \
    GET_FROM_NODE (t->node[i], WN_get_coords, c[i]);            \
  }
  GET_FROM_NODE_ALL(0);
  GET_FROM_NODE_ALL(1);
  GET_FROM_NODE_ALL(2);
#undef GET_FROM_NODE_ALL

  G3D_tri_to_tri2d (&c[0], &c[1], &c[2], xt, yt);
  
  G3D_tri_to_tri2d_triad (&c[0], &c[1], &c[2], &e0, &e1, &e2);

  factes (locK, E, nu, t->thickness, xt[1], xt[2], yt[2]);
  globalize_blocked (K, stiff_mult, locK, &e0, &e1, &e2);
}

static void
assemble_eff_stiffness (W_elem_t *el,
                        W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                        W_float_t stiff_mult,
                        W_float_t mass_mult,
                        W_mass_type_enum mass_type
                        )
{
  int i, j;
  priv_elem_t *t = (priv_elem_t *)el->data;
  double K[9][9];
  W_index_t idcs[9];
  W_bool_t constrained[9];
  
  FOR (i, 9) {FOR (j, 9) { K[i][j] = 0; }}

  /* RAW elastic stiffness */
  global_stiff (el, stiff_mult, K);

  { /* add mass matrix; supports only diagonal mass */
    double m = mass_mult * t->mass / 3;
    FOR (i, 9) { K[i][i] += m; }
  }
  
  WN_eqn_numbers_and_constraints (t->node[0], &idcs[0], &constrained[0]);
  WN_eqn_numbers_and_constraints (t->node[1], &idcs[3], &constrained[3]);
  WN_eqn_numbers_and_constraints (t->node[2], &idcs[6], &constrained[6]);
  (*assemble_eff_stiff_callback)  (el->domain->newmark_integrator, FLES_ADD,
                                   9, idcs, (double *) K, constrained);
}

static W_float_t
kinetic_energy (W_elem_t *el, W_mass_type_enum mass_type)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  W_vect_t v;
  int i;
  double ke = 0;
  double m = h->mass / 3;

  FOR (i, 3) {
    GET_FROM_NODE (h->node[i], WN_get_velocity, v);
    ke += VECMAC_DOT (v, v);
  }
  return (m * ke);
}

static W_elem_methods_t methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  new_history_client_data,
  update_history,
  NULL,
  write_postproc_file,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  num_of_faces,
  get_face_vertex_nodes,
  get_contact_faces,
  assemble_residual_forces,
  assemble_eff_stiffness,
  kinetic_energy
};

/* Reads element data from stream as long as there is no */
/* END_OF_BLOCK tag in the input line. Creates elements. */

W_bool_t 
tmem3_reader (W_domain_t *d)
{
  const char *input_line;
  long elem_num;
  priv_elem_t *t;
  int i, j, nodeid;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      t = w_new (priv_elem_t);
      CHECK_MALLOC(t);
      TOKENS_PARSE(input_line);
      elem_num             = TOKEN_INT(1);
      FOR (i, 3) {
        nodeid = TOKEN_INT(2+i);
        t->node[i] = WD_get_node_by_idx (d, nodeid);
        if (t->node[i] == NULL) {
          sprintf (W_buf (), "No such node: %d", (int)nodeid);
          REPORT_FATAL_ERROR (d->errh, W_buf ());
        }
      }
      t->material  = TOKEN_INT(5);
      t->thickness = TOKEN_DBL(6);
      t->matstate  = NULL;
      t->mass      = compute_total_mass (t, d);
#ifdef PVM_NEED_GHOST_ELEMENTS
      is_ghost = FALSE;
      if (TOKENS_HAVE_KEYWORD("GHOST"))
        is_ghost = TRUE;
#endif      
      WD_add_element(d, elem_num,
                     WE_new_element(d, &methods, t
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                    , is_ghost
#                                   endif
                                    ));
    }
  }
  return FALSE;
}

